package nand.apps.chat.camera

import android.Manifest
import android.content.Context
import android.content.Intent
import android.content.pm.PackageManager
import android.net.Uri
import androidx.activity.result.ActivityResultLauncher
import androidx.activity.result.contract.ActivityResultContracts.StartActivityForResult
import androidx.activity.result.contract.ActivityResultContracts.TakePicture
import androidx.core.content.FileProvider
import nand.apps.chat.MainActivity
import nand.apps.chat.model.event.*
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.log.getLogger
import nand.apps.chat.permission.PermissionManager
import nand.apps.chat.res.getStringBlocking
import okio.Path
import org.koin.core.Koin
import seers.composeapp.generated.resources.*

const val INTENT_ACTION_QR_CODE_CAMERA_RESULT = "QR_CODE_CAMERA_RESULT"

class AndroidCameraLauncher(
    private val appContext: Context,
    koin: Koin,
) : CameraLauncher {

    override val isSupported = appContext.packageManager.hasSystemFeature(PackageManager.FEATURE_CAMERA_ANY)
    
    private var activity: MainActivity? = null
    private var activityLauncher: ActivityResultLauncher<Intent>? = null
    private var photoLauncher: ActivityResultLauncher<Uri>? = null
    private var photoPath: Path? = null
    private var permissionManager: PermissionManager? = null

    private val eventHandler: ChatEventHandler by koin.inject()
    private val logger = getLogger("AndroidCameraLauncher")
    
    fun start(activity: MainActivity, permissionManager: PermissionManager) {
        this.activity = activity
        this.permissionManager = permissionManager
        activityLauncher = activity.registerForActivityResult(StartActivityForResult()) { result ->
            when (result.data?.action) {
                INTENT_ACTION_QR_CODE_CAMERA_RESULT -> {
                    val text = result.data?.data ?: return@registerForActivityResult
                    eventHandler += TakePhotoQrCodeEvent(text.toString())
                }
            }
        }
        photoLauncher = activity.registerForActivityResult(TakePicture()) { result ->
            if (result) {
                photoPath?.let { eventHandler += TakePhotoEvent(it) }
            }
            photoPath = null
        }
    }
    
    fun stop() {
        activityLauncher?.unregister()
        photoLauncher?.unregister()
        activity = null
        permissionManager = null
    }

    fun resumeLaunch() {
        photoPath?.let { launchForPhoto(it) } ?: launchForQrCode()
    }

    fun cancelLaunch() {
        photoPath = null
    }

    override fun launchForPhoto(path: Path) {
        try {
            photoPath = path
            permissionManager?.requestPermissions(Manifest.permission.CAMERA) {
                val uri = FileProvider.getUriForFile(appContext, "${appContext.packageName}.provider", path.toFile())
                photoLauncher?.launch(uri)
            }
        } catch (e: Exception) {
            photoPath = null
            logger.error("Failed to launch camera app", e)
            eventHandler += ErrorDialogEvent(getStringBlocking(Res.string.camera_launcher_failed))
        }
    }

    override fun launchForQrCode() {
        activity?.let { activity ->
            try {
                permissionManager?.requestPermissions(Manifest.permission.CAMERA) {
                    activityLauncher?.launch(Intent(activity, QrCodeCameraActivity::class.java))
                }
            } catch (e: Exception) {
                logger.error("Failed to launch QR code activity", e)
                eventHandler += ErrorDialogEvent(getStringBlocking(Res.string.camera_launcher_failed))
            }
        }
    }
}