package nand.apps.chat.notification

import android.Manifest
import android.app.Notification
import android.app.NotificationChannel
import android.app.NotificationManager
import android.content.Context
import android.content.pm.PackageManager
import android.os.Build
import android.os.VibrationEffect
import androidx.core.app.ActivityCompat
import androidx.core.app.NotificationManagerCompat
import androidx.core.content.ContextCompat.getSystemService
import chat.seers.app.R
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import nand.apps.chat.model.notification.*
import nand.apps.chat.repo.NotificationRepo
import nand.apps.chat.ui.util.collect
import org.koin.core.Koin

class AndroidNotificationController(
    private val context: Context,
    koin: Koin,
) {
    private val messageChannel = NotificationChannel(NOTIFICATION_CHANNEL_MESSAGES, "Messages", NotificationManager.IMPORTANCE_DEFAULT)
    private val callChannel = NotificationChannel(NOTIFICATION_CHANNEL_CALLS, "Calls", NotificationManager.IMPORTANCE_HIGH)
    private val manager = NotificationManagerCompat.from(context)
    private val notificationRepo: NotificationRepo by koin.inject()
    
    private val messageController by lazy { AndroidMessageNotificationReceiver(context, this, koin) }
    private val callController by lazy { AndroidCallNotificationReceiver(context, this, koin) }
    private val controllers = listOf(messageController, callController)
    
    private var collectorJob: Job? = null
    private val notificationGroups = mutableSetOf<ChatNotificationGroup>()
    private var isActive = false
    
    fun start(coroutineScope: CoroutineScope) {
        if (!isActive) {
            with(messageChannel) {
                name = context.getString(R.string.notification_channel_messages)
                enableVibration(true)
                enableLights(true)
                vibrationPattern = longArrayOf(0, 200, 100, 100, 100, 200, 100, 200)
            }
            with(callChannel) {
                name = context.getString(R.string.notification_channel_calls)
                enableVibration(true)
                enableLights(true)
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.VANILLA_ICE_CREAM) {
                    vibrationEffect = VibrationEffect.createWaveform(longArrayOf(0, 2000, 3000), 2)
                } else {
                    vibrationPattern = longArrayOf(0, 2000, 3000, 2000, 3000, 2000)
                }
            }
            getSystemService(context, NotificationManager::class.java)?.createNotificationChannels(
                listOf(messageChannel, callChannel)
            )
            controllers.forEach { it.start(coroutineScope) }
            isActive = true
            collectorJob = coroutineScope.collect(notificationRepo.observer) { notifications ->
                notifications.forEach { (_, notification) -> show(notification) }
                notificationGroups.forEach { group ->
                    if (group !in notifications) dismiss(group)
                }
                notificationGroups.clear()
                notificationGroups.addAll(notifications.keys)
            }
        }
    }
    
    fun stop() {
        if (isActive) {
            getSystemService(context, NotificationManager::class.java)?.run {
                deleteNotificationChannel(NOTIFICATION_CHANNEL_MESSAGES)
                deleteNotificationChannel(NOTIFICATION_CHANNEL_CALLS)
            }
            controllers.forEach { it.stop() }
            collectorJob?.cancel()
            collectorJob = null
            isActive = false
        }
    }

    fun dismiss(group: ChatNotificationGroup) {
        manager.cancel(group.notificationId)
        notificationRepo.dismiss(group)
    }
    
    fun getActiveNotification(group: ChatNotificationGroup): Notification? {
        return manager.activeNotifications.find { it.id == group.notificationId }?.notification
    }

    fun notify(group: ChatNotificationGroup, notification: Notification) {
        if (ActivityCompat.checkSelfPermission(context, Manifest.permission.POST_NOTIFICATIONS) == PackageManager.PERMISSION_GRANTED) {
            manager.notify(group.notificationId, notification)
        }
    }

    private fun show(notification: ChatNotification) {
        val appNotification = when (notification) {
            is MessageNotification -> messageController.getAppNotification(notification)
            is CallNotification -> callController.getAppNotification(notification)
            is AppUpdateNotification -> return // Not supported
        } ?: return
        notify(notification.group, appNotification)
    }
}