package nand.apps.chat.di

import io.ktor.client.*
import nand.apps.chat.audio.AudioFileWriterFactory
import nand.apps.chat.audio.AudioMetadataReader
import nand.apps.chat.audio.SoundEffectCache
import nand.apps.chat.audio.id3.Id3TagReader
import nand.apps.chat.audio.opus.OpusFileWriterFactory
import nand.apps.chat.dev.DevCommandDefaultHandlers
import nand.apps.chat.embed.EmbedPreviewFactory
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.engine.ChatEngineCommandHandler
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.engine.DefaultChatEngineCallback
import nand.apps.chat.engine.mux.ChatEngineEncryptionMux
import nand.apps.chat.engine.mux.ChatEngineMux
import nand.apps.chat.engine.mux.ChatEngineCommandHandlerMux
import nand.apps.chat.engine.mux.ChatEngineProfileAuthenticator
import nand.apps.chat.net.TempFileCache
import nand.apps.chat.monitor.AutoAwayMonitor
import nand.apps.chat.net.FileDownloader
import nand.apps.chat.net.FileDownloaderImpl
import nand.apps.chat.net.FileUploader
import nand.apps.chat.net.FileUploaderImpl
import nand.apps.chat.net.packet.ChatPacketConverter
import nand.apps.chat.net.packet.ChatPacketRepo
import nand.apps.chat.repo.NotificationRepo
import nand.apps.chat.repo.*
import nand.apps.chat.emoji.EmojiRepo
import nand.apps.chat.emoji.EmojiRepoImpl
import nand.apps.chat.time.DateTimeFormatter
import nand.apps.chat.ui.app.AppViewModel
import nand.apps.chat.ui.chat.screen.ChatScreenViewModel
import nand.apps.chat.ui.contacts.ContactListViewModel
import nand.apps.chat.ui.dialog.image.ImageCropperViewModel
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.event.ChatEventHandlerImpl
import nand.apps.chat.event.VisibleMessageTracker
import nand.apps.chat.io.*
import nand.apps.chat.net.packet.ChatPacketRepoImpl
import nand.apps.chat.net.packet.MessagePacketHandler
import nand.apps.chat.net.remote.ChatRemoteServer
import nand.apps.chat.net.remote.command.ChatRemoteCommandHandler
import nand.apps.chat.profile.ProfileAuthenticator
import nand.apps.chat.profile.ProfileBrowser
import nand.apps.chat.search.ChatSearch
import nand.apps.chat.ui.chat.message.input.MessageInputViewModel
import nand.apps.chat.ui.dialog.audio.AudioRecorderDialogViewModel
import nand.apps.chat.ui.embed.audio.AudioEmbedViewModel
import nand.apps.chat.ui.emoji.EmojiPopupViewModel
import nand.apps.chat.ui.file.ChatFilePickerViewModel
import nand.apps.chat.ui.file.sender.ChatFileSenderViewModel
import nand.apps.chat.ui.friends.AddFriendDialogViewModel
import nand.apps.chat.ui.groups.CreateGroupDialogViewModel
import nand.apps.chat.ui.groups.JoinGroupDialogViewModel
import nand.apps.chat.ui.groups.settings.GroupSettingsDialogViewModel
import nand.apps.chat.ui.login.LoginScreenViewModel
import nand.apps.chat.ui.profile.SelfProfileDialogViewModel
import nand.apps.chat.ui.profile.SelfProfileRowViewModel
import nand.apps.chat.ui.profile.auth.ProfileAuthenticationDialogViewModel
import nand.apps.chat.ui.profile.export.ProfileExportDialogViewModel
import nand.apps.chat.ui.profile.password.LoginPasswordDialogViewModel
import nand.apps.chat.ui.profile.selector.ProfileSelectorViewModel
import nand.apps.chat.ui.profile.username.LoginUsernameDialogViewModel
import nand.apps.chat.ui.settings.SettingsViewModel
import nand.apps.chat.ui.settings.net.tox.ToxNodeSettingsViewModel
import nand.apps.chat.ui.tutorial.TutorialViewModel
import nand.apps.chat.ui.update.AppUpdateDialogViewModel
import nand.apps.chat.update.AppUpdater
import nand.apps.chat.update.AppUpdaterImpl
import nand.apps.chat.util.createDefaultHttpClient
import okio.FileSystem
import okio.SYSTEM
import org.koin.core.Koin
import org.koin.core.KoinApplication
import org.koin.core.context.startKoin
import org.koin.core.module.dsl.factoryOf
import org.koin.core.module.dsl.singleOf
import org.koin.dsl.KoinAppDeclaration
import org.koin.dsl.module

fun initKoin(
    appDeclaration: KoinAppDeclaration = {},
): KoinApplication {
    val registeredModules = module {
        // Global
        single<Koin> { getKoin() }
        factory<HttpClient> { createDefaultHttpClient() }
        single<FileSystem> { FileSystem.SYSTEM }
        single<ChatBackgroundWorker> { DefaultBackgroundWorker() }
        single<CoroutineDelayScope> { DefaultCoroutineDelayScope() }
        singleOf(::ChatFileSystem)
        singleOf(::TempFileCache)
        singleOf(::SoundEffectCache)
        single<FileUploader> { FileUploaderImpl(get(), get(), get()) }
        single<FileDownloader> { FileDownloaderImpl(get(), get(), get()) }
        single<AudioFileWriterFactory> { OpusFileWriterFactory() }
        single<AppUpdater> { AppUpdaterImpl(get(), get(), get(), get(), get(), get()) }
        singleOf(::AutoAwayMonitor)
        singleOf(::ChatPacketConverter)
        singleOf(::DateTimeFormatter)
        singleOf(::ProfileBrowser)
        singleOf(::Id3TagReader)
        singleOf(::AudioMetadataReader)
        
        // Core chat engine
        single<ChatEngine> { ChatEngineMux(get()) }
        single<ChatEngineEncryption> { ChatEngineEncryptionMux(get()) }
        single<ChatEngineCommandHandler> { ChatEngineCommandHandlerMux(get()) }
        single<ChatEngineConstraints> {
            val engine: ChatEngine by getKoin().inject()
            engine.constraints
        }
        single<ChatEngineCallback> { DefaultChatEngineCallback(get()) }
        single<ProfileAuthenticator> { ChatEngineProfileAuthenticator(get()) }
        singleOf(::ChatRemoteServer)
        singleOf(::ChatRemoteCommandHandler)
        
        // Repos
        single<SettingsRepo> { SettingsRepoImpl(get(), get()) }
        single<FileTransferRepo> { FileTransferRepoImpl(get()) }
        single<ChatAvatarRepo> { ChatAvatarRepoImpl(get()) }
        single<SelfProfileRepo> { SelfProfileRepoImpl(get()) }
        single<CallRepo> { CallRepoImpl(get()) }
        single<VoicemailRepo> { VoicemailRepoImpl(get()) }
        single<ChatMessageRepo> { ChatMessageRepoImpl(get()) }
        single<ContactRepo> { ContactRepoImpl(get()) }
        single<ChatPacketRepo> { ChatPacketRepoImpl(get()) }
        single<NotificationRepo> { NotificationRepoImpl(get()) }
        single<EmojiRepo> { EmojiRepoImpl(get()) }
        
        // Event handlers
        single<ChatEventHandler> { ChatEventHandlerImpl(get()) }
        singleOf(::VisibleMessageTracker)
        singleOf(::MessagePacketHandler)
        
        // UI + view models
        singleOf(::EmbedPreviewFactory)
        singleOf(::ChatSearch)
        factoryOf(::AppViewModel)
        factoryOf(::LoginScreenViewModel)
        factoryOf(::ChatScreenViewModel)
        factoryOf(::ContactListViewModel)
        factoryOf(::MessageInputViewModel)
        factoryOf(::SelfProfileRowViewModel)
        factoryOf(::SelfProfileDialogViewModel)
        factoryOf(::ProfileExportDialogViewModel)
        factoryOf(::ProfileAuthenticationDialogViewModel)
        factoryOf(::ProfileSelectorViewModel)
        factoryOf(::LoginUsernameDialogViewModel)
        factoryOf(::LoginPasswordDialogViewModel)
        factoryOf(::ImageCropperViewModel)
        factoryOf(::AddFriendDialogViewModel)
        factoryOf(::CreateGroupDialogViewModel)
        factoryOf(::JoinGroupDialogViewModel)
        factoryOf(::ChatFilePickerViewModel)
        factoryOf(::ChatFileSenderViewModel)
        factoryOf(::GroupSettingsDialogViewModel)
        factoryOf(::SettingsViewModel)
        factoryOf(::TutorialViewModel)
        factoryOf(::AudioEmbedViewModel)
        factoryOf(::AudioRecorderDialogViewModel)
        factoryOf(::AppUpdateDialogViewModel)
        factoryOf(::EmojiPopupViewModel)
        factoryOf(::ToxNodeSettingsViewModel)
        
        // Developer tools
        singleOf(::DevCommandDefaultHandlers)
    }
    return startKoin {
        modules(registeredModules)
        modules(getChatEngineKoinModules())
        appDeclaration()
    }
}