package nand.apps.chat.embed

import com.fleeksoft.ksoup.nodes.Document
import io.ktor.client.*
import io.ktor.client.request.*
import io.ktor.http.*
import nand.apps.chat.embed.file.AudioEmbedData
import nand.apps.chat.embed.file.ImageEmbedData
import nand.apps.chat.embed.web.WebPageEmbedData
import nand.apps.chat.model.file.contentTypeFromFileName
import nand.apps.chat.model.file.isSupportedAudioFormat
import nand.apps.chat.util.getHostName
import nand.apps.chat.util.getHtmlDocument
import nand.apps.chat.util.getUrlFileName
import nand.apps.chat.util.takeIfNotBlank

class DefaultUrlParser : EmbedUrlParser {

    override fun matches(url: String) = true

    override suspend fun parse(url: String, client: HttpClient): EmbedPreviewData? {
        val response = client.get(url)
        val fileName = url.getUrlFileName()
        val type = response.contentType()?.let { type ->
            if (type == ContentType.Application.OctetStream) {
                contentTypeFromFileName(fileName)
            } else {
                type
            }
        }
        if (type == null || response.status.value >= 400) return null
        return when {
            type == ContentType.Text.Html -> {
                parseHtml(url, client)
            }
            type.contentType == "image" -> {
                ImageEmbedData(url, type.contentSubtype)
            }
            type.isSupportedAudioFormat() -> {
                AudioEmbedData(url, type)
            }
            else -> null
        }
    }
    
    private suspend fun parseHtml(url: String, client: HttpClient): WebPageEmbedData? {
        val doc = client.getHtmlDocument(url) ?: return null
        
        val domain = doc.getTwitterMeta("domain") ?: url.getHostName() ?: return null
        
        val title = doc.getTwitterMeta("title")
            ?: doc.getOpenGraphMeta("title")
            ?: doc.title().takeIfNotBlank()
            ?: return null
        
        val desc = doc.getTwitterMeta("description")
            ?: doc.getOpenGraphMeta("description")
            ?: return null
        
        val imageUrl = doc.getTwitterMeta("image")
            ?: doc.getOpenGraphMeta("image")
            ?: doc.selectFirst("link[rel*=\"icon\"]")?.attr("href")
            ?: getFavIconUrl(url)
        
        return WebPageEmbedData(
            url = url,
            domain = domain,
            title = title,
            description = desc,
            imageUrl = imageUrl,
        )
    }
    
    private fun Document.getTwitterMeta(name: String) =
        selectFirst("meta[name=\"twitter:$name\"]")?.attr("content")

    private fun Document.getOpenGraphMeta(property: String) =
        selectFirst("meta[property=\"og:$property\"]")?.attr("content")
    
    private fun getFavIconUrl(url: String): String {
        val protocol = url.substringBefore(":")
        val host = url.getHostName()
        return "$protocol://$host/favicon.ico"
    }
}