package nand.apps.chat.engine.mux

import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.BuildConfig.MOCK_ENGINE
import nand.apps.audio.AudioFrame
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.engine.ChatEngineType
import nand.apps.chat.engine.matrix.MatrixChatEngine
import nand.apps.chat.engine.mock.MockChatEngine
import nand.apps.chat.engine.tox.ToxChatEngine
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.file.ChatFileControl
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.repo.SettingsRepo
import org.koin.core.Koin

class ChatEngineMux(koin: Koin) : ChatEngine {

    private val settingsRepo: SettingsRepo by koin.inject()
    private val toxChatEngine: ToxChatEngine by koin.inject()
    private val matrixChatEngine: MatrixChatEngine by koin.inject()
    private val mockChatEngine: MockChatEngine by koin.inject()

    val engine: ChatEngine
        get() = when {
        MOCK_ENGINE -> mockChatEngine
        else -> when (settingsRepo.chatEngine) {
            ChatEngineType.TOX -> toxChatEngine
            ChatEngineType.MATRIX -> matrixChatEngine
            else -> mockChatEngine
        }
    }

    override val version get() = engine.version
    override val hostInfo get() = engine.hostInfo
    override val constraints get() = engine.constraints
    override val encryption get() = engine.encryption
    override val authenticator get() = engine.authenticator
    override val commandHandler get() = engine.commandHandler

    override suspend fun initialize(profile: ChatProfileData) = engine.initialize(profile)

    override fun start(coroutineScope: CoroutineScope) = engine.start(coroutineScope)

    override fun stop() = engine.stop()

    override fun getSelfUserData() = engine.getSelfUserData()

    override fun setSelfUserData(name: String, statusMessage: String) =
        engine.setSelfUserData(name, statusMessage)

    override fun setSelfStatus(status: UserStatus) =
        engine.setSelfStatus(status)

    override fun getSelfAddress() = engine.getSelfAddress()

    override fun resetSelfAddress() = engine.resetSelfAddress()

    override suspend fun getFriendList() = engine.getFriendList()

    override suspend fun getFriendRequestList() = engine.getFriendRequestList()

    override fun sendFriendRequest(address: String, message: String) =
        engine.sendFriendRequest(address, message)

    override fun acceptFriendRequest(requestUid: FriendRequestUid) =
        engine.acceptFriendRequest(requestUid)

    override fun declineFriendRequest(requestUid: FriendRequestUid) =
        engine.declineFriendRequest(requestUid)

    override fun removeFriend(friendUid: UserUid) =
        engine.removeFriend(friendUid)

    override fun setFriendAlias(friendUid: UserUid, alias: String?) =
        engine.setFriendAlias(friendUid, alias)

    override fun sendMessage(message: ChatMessage) =
        engine.sendMessage(message)

    override fun editMessage(old: ChatMessage, new: ChatMessage) =
        engine.editMessage(old, new)

    override fun setMessageReactions(message: ChatMessage, reactions: Set<String>) =
        engine.setMessageReactions(message, reactions)

    override fun removeMessage(message: ChatMessage) =
        engine.removeMessage(message)

    override fun sendPacket(uid: ContactUid, packet: ByteArray) =
        engine.sendPacket(uid, packet)

    override fun sendPeerPacket(groupUid: GroupUid, peerUid: UserUid, packet: ByteArray) =
        engine.sendPeerPacket(groupUid, peerUid, packet)

    override suspend fun getGroupList() = engine.getGroupList()

    override fun getGroupPeerList(groupUid: GroupUid) =
        engine.getGroupPeerList(groupUid)

    override fun createGroup(title: String, type: ChatGroupType, visibility: ChatGroupVisibility) =
        engine.createGroup(title, type, visibility)

    override fun setGroupTitle(groupUid: GroupUid, title: String) =
        engine.setGroupTitle(groupUid, title)

    override fun setGroupDescription(groupUid: GroupUid, description: String) =
        engine.setGroupDescription(groupUid, description)

    override fun setGroupDescriptionLocked(groupUid: GroupUid, isLocked: Boolean) =
        engine.setGroupDescriptionLocked(groupUid, isLocked)

    override fun setGroupPassword(groupUid: GroupUid, password: String?) =
        engine.setGroupPassword(groupUid, password)

    override fun setGroupNickname(groupUid: GroupUid, name: String) =
        engine.setGroupNickname(groupUid, name)

    override fun setGroupAvatar(groupUid: GroupUid, avatarHash: String) =
        engine.setGroupAvatar(groupUid, avatarHash)

    override fun setGroupAudioEnabled(groupUid: GroupUid, isEnabled: Boolean) =
        engine.setGroupAudioEnabled(groupUid, isEnabled)

    override fun setGroupPeerLimit(groupUid: GroupUid, peerLimit: Int) =
        engine.setGroupPeerLimit(groupUid, peerLimit)

    override fun setGroupTalkState(groupUid: GroupUid, talkState: ChatGroupTalkState) =
        engine.setGroupTalkState(groupUid, talkState)

    override fun setGroupVisibility(groupUid: GroupUid, visibility: ChatGroupVisibility) =
        engine.setGroupVisibility(groupUid, visibility)

    override fun sendGroupInvite(friendUid: UserUid, groupUid: GroupUid) =
        engine.sendGroupInvite(friendUid, groupUid)

    override fun acceptGroupInvite(invite: ChatGroupInvite) =
        engine.acceptGroupInvite(invite)

    override fun setGroupRole(groupUid: GroupUid, peerUid: UserUid, role: ChatGroupRole) =
        engine.setGroupRole(groupUid, peerUid, role)

    override fun kickGroupPeer(groupUid: GroupUid, peerUid: UserUid) =
        engine.kickGroupPeer(groupUid, peerUid)

    override fun setGroupPeerBlocked(groupUid: GroupUid, peerUid: UserUid, block: Boolean) =
        engine.setGroupPeerBlocked(groupUid, peerUid, block)

    override fun joinGroup(groupUid: GroupUid, selfNickname: String, password: String?) =
        engine.joinGroup(groupUid, selfNickname, password)

    override fun removeGroup(groupUid: GroupUid, partingMessage: String?) =
        engine.removeGroup(groupUid, partingMessage)

    override fun setContactMuted(uid: ContactUid, isMuted: Boolean) =
        engine.setContactMuted(uid, isMuted)

    override fun sendFile(friendUid: UserUid, kind: ChatFileType, fileSize: Long, fileName: String, fileUid: FileUid) =
        engine.sendFile(friendUid, kind, fileSize, fileName, fileUid)

    override fun sendFileChunk(friendUid: UserUid, fileUid: FileUid, position: Long, data: ByteArray) =
        engine.sendFileChunk(friendUid, fileUid, position, data)

    override fun sendFileControl(friendUid: UserUid, fileUid: FileUid, control: ChatFileControl) =
        engine.sendFileControl(friendUid, fileUid, control)

    override fun startCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean) =
        engine.startCall(uid, isAudioEnabled, isVideoEnabled)

    override fun answerCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean) =
        engine.answerCall(uid, isAudioEnabled, isVideoEnabled)

    override fun sendCallControl(uid: ContactUid, control: CallControl) =
        engine.sendCallControl(uid, control)

    override fun sendAudioFrame(uid: ContactUid, frame: AudioFrame) =
        engine.sendAudioFrame(uid, frame)

    override fun sendAudioBitRate(uid: ContactUid, bitRate: Int) =
        engine.sendAudioBitRate(uid, bitRate)

    override fun generateHash(data: ByteArray) = engine.generateHash(data)
}