package nand.apps.chat.net.remote

import com.ionspin.kotlin.crypto.LibsodiumInitializer
import io.ktor.http.*
import io.ktor.network.selector.*
import io.ktor.network.sockets.*
import io.ktor.utils.io.*
import kotlinx.coroutines.*
import kotlinx.serialization.ExperimentalSerializationApi
import kotlinx.serialization.SerializationException
import kotlinx.serialization.json.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.settings.net.ChatRemoteSettingsData
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.net.remote.command.*
import nand.apps.chat.net.remote.event.*
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.util.collect
import nand.apps.chat.util.fifoSharedFlow

private const val WAIT_DELAY_MS = 50L
private const val RESPONSE_TIMEOUT_MS = 30_000L

@OptIn(ExperimentalSerializationApi::class)
class ChatRemoteClient {

    private val json = Json {
        ignoreUnknownKeys = true
        classDiscriminatorMode = ClassDiscriminatorMode.NONE
    }
    private val logger = getLogger("ChatRemoteClient")

    private lateinit var socket: Socket
    private lateinit var encryption: ChatRemoteEncryption
    private lateinit var receiveChannel: ByteReadChannel
    private lateinit var sendChannel: ByteWriteChannel
    private lateinit var coroutineScope: CoroutineScope

    private val isStarted get() = this::socket.isInitialized && socket.isActive
    private var isReady = false

    val onEvent = fifoSharedFlow<ChatRemoteEvent>()
    private val onCommandResponse = fifoSharedFlow<CommandResponseEvent>(replay = 1)

    @OptIn(ExperimentalUnsignedTypes::class)
    fun init(
        profile: ChatProfileData,
        settings: ChatRemoteSettingsData,
        coroutineScope: CoroutineScope,
    ) {
        this.coroutineScope = coroutineScope
        coroutineScope.launch {
            val selectorManager = SelectorManager(Dispatchers.IO)
            socket = aSocket(selectorManager).tcp().connect(
                hostname = settings.host,
                port = settings.port,
            )
            receiveChannel = socket.openReadChannel()
            sendChannel = socket.openWriteChannel(autoFlush = true)

            if (!LibsodiumInitializer.isInitialized()) {
                LibsodiumInitializer.initialize()
            }

            coroutineScope.launch(Dispatchers.IO) {
                val nonce = receiveChannel.readByteArray(NONCE_BYTE_COUNT)
                logger.debug("Received nonce from server")
                encryption = ChatRemoteEncryption(
                    profile = profile,
                    nonce = nonce.toUByteArray(),
                )
                isReady = true
                try {
                    while (true) {
                        val numBytes = receiveChannel.readInt()
                        val bytes = receiveChannel.readByteArray(numBytes).let {
                            if (encryption.isDecryptionSupported()) encryption.decryptData(it) else it
                        } ?: continue

                        try {
                            val packet = json.decodeFromString<ChatRemoteEventPacket>(bytes.decodeToString())
                            logger.debug("Received ${packet.type} event from server")
                            val event = when (packet.type) {
                                ChatRemoteEventType.COMMAND_RESPONSE -> json.decodeFromJsonElement<CommandResponseEvent>(packet.data)
                                ChatRemoteEventType.CONTACT_LIST -> json.decodeFromJsonElement<ContactListEvent>(packet.data)
                                ChatRemoteEventType.NEW_MESSAGE -> json.decodeFromJsonElement<NewMessageEvent>(packet.data)
                                ChatRemoteEventType.FRIEND_REQUEST -> json.decodeFromJsonElement<FriendRequestEvent>(packet.data)
                                ChatRemoteEventType.GROUP_INVITE -> json.decodeFromJsonElement<GroupInviteEvent>(packet.data)
                            }
                            if (event is CommandResponseEvent) {
                                onCommandResponse.emit(event)
                            } else {
                                onEvent.emit(event)
                            }
                        } catch (e: SerializationException) {
                            logger.error("Failed to deserialize JSON packet", e)
                        } catch (e: IllegalArgumentException) {
                            logger.error("Failed to deserialize command packet", e)
                        }
                    }
                } catch (e: Exception) {
                    logger.error("Failed to read packet", e)
                    socket.close()
                }
            }
        }
    }

    fun stop() {
        if (isStarted) socket.close()
    }

    suspend fun sendCommand(command: ChatRemoteCommand): HttpStatusCode {
        while (!isReady) delay(WAIT_DELAY_MS)

        val packet = ChatRemoteCommandPacket(
            type = when (command) {
                is HandshakeCommand -> ChatRemoteCommandType.HANDSHAKE
                is SetNameCommand -> ChatRemoteCommandType.SET_NAME
                is SetStatusCommand -> ChatRemoteCommandType.SET_STATUS
                is SendMessageCommand -> ChatRemoteCommandType.SEND_MESSAGE
                is SendFriendRequestCommand -> ChatRemoteCommandType.SEND_FRIEND_REQUEST
                is AcceptFriendRequestCommand -> ChatRemoteCommandType.ACCEPT_FRIEND_REQUEST
                is DeclineFriendRequestCommand -> ChatRemoteCommandType.DECLINE_FRIEND_REQUEST
                is RemoveFriendCommand -> ChatRemoteCommandType.REMOVE_FRIEND
                is SendGroupInviteCommand -> ChatRemoteCommandType.SEND_GROUP_INVITE
                is AcceptGroupInviteCommand -> ChatRemoteCommandType.ACCEPT_GROUP_INVITE
                is DeclineGroupInviteCommand -> ChatRemoteCommandType.DECLINE_GROUP_INVITE
            },
            data = json.encodeToJsonElement(command).jsonObject,
        )
        val packetJson = json.encodeToString(packet)
        return encryption.encryptData(packetJson.encodeToByteArray())?.let { bytes ->
            logger.debug("Sending ${packet.type} command to server")
            sendChannel.writeInt(bytes.size)
            sendChannel.writeByteArray(bytes)
            var responseCode = HttpStatusCode.RequestTimeout
            val start = currentTimeMillis()
            val job = coroutineScope.collect(onCommandResponse) { event ->
                if (event.type == packet.type && event.uid == packet.uid) {
                    responseCode = HttpStatusCode.fromValue(event.code)
                }
            }
            while (responseCode == HttpStatusCode.RequestTimeout && currentTimeMillis() - start < RESPONSE_TIMEOUT_MS) {
                delay(WAIT_DELAY_MS)
            }
            job.cancel()
            responseCode
        } ?: run {
            error("Failed to encrypt packet")
        }
    }
}