package nand.apps.chat.net.remote

import com.ionspin.kotlin.crypto.LibsodiumInitializer
import io.ktor.network.selector.*
import io.ktor.network.sockets.*
import kotlinx.coroutines.*
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.settings.net.ChatRemoteSettingsData
import nand.apps.chat.model.settings.net.DEFAULT_REMOTE_SERVER_HOST
import nand.apps.chat.model.settings.net.DEFAULT_REMOTE_SERVER_PORT
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.util.collect
import org.koin.core.Koin

@OptIn(ExperimentalUnsignedTypes::class)
class ChatRemoteServer(
    private val koin: Koin,
): ChatEngineCallback {

    private val settingsRepo: SettingsRepo by koin.inject()
    private var profile: ChatProfileData? = null

    private lateinit var coroutineScope: CoroutineScope
    private var serverSocket: ServerSocket? = null
    private var socketJob: Job? = null
    private var host: String = DEFAULT_REMOTE_SERVER_HOST
    private var port: Int = DEFAULT_REMOTE_SERVER_PORT

    private val logger = getLogger("ChatRemoteServer")

    override suspend fun onEngineInit(profile: ChatProfileData) {
        this.profile = profile
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = CoroutineScope(Dispatchers.Main)
        this.coroutineScope.collect(settingsRepo.observer) { settings ->
            val remoteSettings = settings.networkSettings.remoteSettings
            if (remoteSettings.isEnabled) {
                startServer(remoteSettings)
            } else {
                stopServer()
            }
        }
    }

    override fun onEngineStop() {
        stopServer()
        profile = null
        coroutineScope.cancel()
    }

    private fun startServer(settings: ChatRemoteSettingsData) {
        if (socketJob != null && (settings.host != host || settings.port != port)) {
            stopServer()
            startServer(settings)
            return
        }
        if (socketJob != null) return
        val profile = this.profile ?: run {
            logger.error("Remote server failed to start - Encryption profile not set")
            return
        }
        val selectorManager = SelectorManager(Dispatchers.IO)
        socketJob = coroutineScope.launch {
            if (!LibsodiumInitializer.isInitialized()) {
                LibsodiumInitializer.initialize()
            }
            host = settings.host
            port = settings.port
            serverSocket = aSocket(selectorManager).tcp().bind(
                hostname = host,
                port = port,
            )
            logger.debug("Started remote server $host:$port")
            while (socketJob != null) {
                try {
                    val socket = serverSocket?.accept() ?: break
                    logger.debug("Accepted socket ${socket.localAddress}")
                    val encryption = ChatRemoteEncryption(profile)
                    val socketHandler = ChatRemoteSocketHandler(
                        socket = socket,
                        encryption = encryption,
                        koin = koin,
                    )
                    socketHandler.open()
                } catch (ignore: Exception) {
                }
            }
        }
    }

    private fun stopServer() {
        serverSocket?.run {
            logger.debug("Stopped remote server $host:$port")
            close()
        }
        serverSocket = null
        socketJob?.cancel()
        socketJob = null
    }
}