package nand.apps.chat.net.remote

import io.ktor.http.*
import io.ktor.network.sockets.*
import io.ktor.utils.io.*
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.Flow
import kotlinx.serialization.ExperimentalSerializationApi
import kotlinx.serialization.SerializationException
import kotlinx.serialization.json.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.group.toPersistedGroupData
import nand.apps.chat.model.group.toPersistedGroupInvite
import nand.apps.chat.model.user.toPersistedUserData
import nand.apps.chat.net.remote.command.*
import nand.apps.chat.net.remote.event.*
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.ui.util.collect
import okio.EOFException
import org.koin.core.Koin

@OptIn(ExperimentalSerializationApi::class, ExperimentalUnsignedTypes::class)
class ChatRemoteSocketHandler(
    private val socket: Socket,
    private val encryption: ChatRemoteEncryption,
    koin: Koin,
) {

    private val contactRepo: ContactRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val commandHandler: ChatRemoteCommandHandler by koin.inject()

    private val coroutineScope = CoroutineScope(Dispatchers.Main)
    private val json = Json {
        ignoreUnknownKeys = true
        classDiscriminatorMode = ClassDiscriminatorMode.NONE
    }

    private lateinit var receiveChannel: ByteReadChannel
    private lateinit var sendChannel: ByteWriteChannel

    private val logger = getLogger("ChatRemoteSocketHandler")

    fun open() {
        coroutineScope.launch {
            try {
                receiveChannel = socket.openReadChannel()
                sendChannel = socket.openWriteChannel(autoFlush = true)
                startEventListeners()
                sendChannel.writeByteArray(encryption.nonce.toByteArray())
                while (true) {
                    readPacket()
                }
            } catch (e: EOFException) {
                logger.debug("Disconnected from remote client")
                close()
            } catch (e: Exception) {
                logger.error("Socket connection error", e)
                close()
            }
        }
    }

    fun close() {
        coroutineScope.cancel()
        socket.close()
    }

    private suspend fun readPacket() {
        val numBytes = receiveChannel.readInt()
        val bytes = receiveChannel.readByteArray(numBytes).let {
            if (encryption.isDecryptionSupported()) encryption.decryptData(it) else it
        } ?: return
        try {
            val packet = json.decodeFromString<ChatRemoteCommandPacket>(bytes.decodeToString())
            val command = when (packet.type) {
                ChatRemoteCommandType.HANDSHAKE -> packet.decode<HandshakeCommand>()
                ChatRemoteCommandType.SET_NAME -> packet.decode<SetNameCommand>()
                ChatRemoteCommandType.SET_STATUS -> packet.decode<SetStatusCommand>()
                ChatRemoteCommandType.SEND_MESSAGE -> packet.decode<SendMessageCommand>()
                ChatRemoteCommandType.SEND_FRIEND_REQUEST -> packet.decode<SendFriendRequestCommand>()
                ChatRemoteCommandType.ACCEPT_FRIEND_REQUEST -> packet.decode<AcceptFriendRequestCommand>()
                ChatRemoteCommandType.DECLINE_FRIEND_REQUEST -> packet.decode<DeclineFriendRequestCommand>()
                ChatRemoteCommandType.REMOVE_FRIEND -> packet.decode<RemoveFriendCommand>()
                ChatRemoteCommandType.SEND_GROUP_INVITE -> packet.decode<SendGroupInviteCommand>()
                ChatRemoteCommandType.ACCEPT_GROUP_INVITE -> packet.decode<AcceptGroupInviteCommand>()
                ChatRemoteCommandType.DECLINE_GROUP_INVITE -> packet.decode<DeclineGroupInviteCommand>()
            }
            val isSuccessful = commandHandler.onCommand(command)
            val code = if (isSuccessful) HttpStatusCode.OK else HttpStatusCode.BadRequest
            sendEvent(
                CommandResponseEvent(
                    type = packet.type,
                    code = code.value,
                    uid = packet.uid,
                )
            )
            if (command is HandshakeCommand) {
                sendEvent(
                    ContactListEvent(
                        friends = contactRepo.getFriends().map { it.toPersistedUserData() },
                        groups = contactRepo.getGroups().map { it.toPersistedGroupData() },
                    )
                )
            }
        } catch (e: SerializationException) {
            logger.error("Failed to deserialize JSON packet", e)
        } catch (e: IllegalArgumentException) {
            logger.error("Failed to deserialize command packet", e)
        }
    }

    private fun startEventListeners() {
        collect(messageRepo.onNewMessage) {
            NewMessageEvent(
                conversationUid = conversationUid.value,
                senderUid = senderUid.value,
                message = text,
                type = type,
            )
        }
        collect(contactRepo.onFriendRequest) {
            FriendRequestEvent(
                friendUid = uid.value,
                message = message,
            )
        }
        collect(contactRepo.onGroupInvite) {
            GroupInviteEvent(toPersistedGroupInvite())
        }
    }

    private suspend fun sendEvent(event: ChatRemoteEvent) {
        try {
            val packet = ChatRemoteEventPacket(
                type = when (event) {
                    is CommandResponseEvent -> ChatRemoteEventType.COMMAND_RESPONSE
                    is ContactListEvent -> ChatRemoteEventType.CONTACT_LIST
                    is NewMessageEvent -> ChatRemoteEventType.NEW_MESSAGE
                    is FriendRequestEvent -> ChatRemoteEventType.FRIEND_REQUEST
                    is GroupInviteEvent -> ChatRemoteEventType.GROUP_INVITE
                },
                data = json.encodeToJsonElement(event).jsonObject,
            )
            encryption.encryptData(json.encodeToString(packet).encodeToByteArray())?.let { bytes ->
                sendChannel.writeInt(bytes.size)
                sendChannel.writeByteArray(bytes)
            } ?: run {
                error("Failed to encrypt packet")
            }
        } catch (e: Exception) {
            logger.error("Failed to send event packet", e)
        }
    }

    private inline fun <reified T> ChatRemoteCommandPacket.decode() =
        json.decodeFromJsonElement<T>(data)

    private fun <T> collect(flow: Flow<T>, onEvent: T.() -> ChatRemoteEvent) {
        coroutineScope.collect(flow) { sendEvent(it.onEvent()) }
    }
}