package nand.apps.chat.repo

import io.ktor.util.*
import kotlinx.coroutines.*
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import kotlinx.serialization.json.Json
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.event.VisibleMessageTracker
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.io.ChatDirectory
import nand.apps.chat.io.ChatEngineEncryption
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.event.ErrorDialogEvent
import nand.apps.chat.model.file.ChatFileTransfer
import nand.apps.chat.model.file.ChatFileTransferDirection
import nand.apps.chat.model.file.ChatFileTransferState
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.*
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserData
import nand.apps.chat.res.StringResourceProvider
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.groups.getLabelResource
import nand.apps.chat.util.*
import org.koin.core.Koin
import seers.composeapp.generated.resources.*
import kotlin.math.abs

private const val FIND_MESSAGE_TIMESTAMP_THRESHOLD = 5 * 1000L // 5 seconds

class ChatMessageRepoImpl(koin: Koin) : ChatMessageRepo {

    private val contactRepo: ContactRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val fileTransferRepo: FileTransferRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private val visibleMessageTracker: VisibleMessageTracker by koin.inject()
    private val eventHandler: ChatEventHandler by koin.inject()
    private val engine: ChatEngine by koin.inject()
    private val encryption: ChatEngineEncryption by koin.inject()
    private val fileSystem: ChatFileSystem by koin.inject()
    private val stringProvider: StringResourceProvider by koin.inject()
    private val backgroundWorker: ChatBackgroundWorker by koin.inject()

    private val messages = mutableMapOf<ContactUid, MutableMap<MessageUid, ChatMessage>>()
    private val unreadMessages = MutableStateFlowMap<ContactUid, UnreadMessageCounts>()
    private val json = Json { ignoreUnknownKeys = true }
    private val persistedMutex = Mutex()
    private val logger = getLogger("ChatMessageRepo")

    private lateinit var messagesDirectory: ChatDirectory
    private lateinit var coroutineScope: CoroutineScope

    override val onNewMessage = fifoSharedFlow<ChatMessage>()
    override val onMessagesChanged = fifoSharedFlow<ContactUid>()
    override val unreadObserver = unreadMessages
    
    private val settings get() = settingsRepo.settings.chatSettings
    private val maxMessages get() = settings.maxMessages
    
    override fun createMessage(text: String, conversationUid: ContactUid): ChatMessage {
        val messageType = if (text.startsWith("/me ")) ChatMessageType.ACTION else ChatMessageType.NORMAL
        val formattedMessage = if (messageType == ChatMessageType.ACTION) text.substringAfter(' ') else text
        return ChatMessage(
            conversationUid = conversationUid,
            senderUid = selfProfileRepo.getSelfUid(conversationUid),
            text = formattedMessage,
            type = messageType,
        )
    }

    override fun getMessages(conversationUid: ContactUid): List<ChatMessage> =
        messages[conversationUid]?.values?.toList().orEmpty()

    override fun getMessage(conversationUid: ContactUid, messageUid: MessageUid) =
        messages[conversationUid]?.get(messageUid)

    override fun getUnreadMessageCounts(conversationUid: ContactUid): UnreadMessageCounts {
        var directMessages = 0
        var groupMessages = 0
        val contact = contactRepo.getContact(conversationUid) ?: return UnreadMessageCounts()
        messages[conversationUid]?.forEach { (_, message) ->
            if (message.isUnread) {
                when (contact) {
                    is ChatGroupData -> {
                        if (message.containsSelfMention(contact)) directMessages++ else groupMessages++
                    }
                    else -> directMessages++
                }
            }
        }
        return UnreadMessageCounts(
            directMessages = directMessages,
            groupMessages = groupMessages,
        )
    }

    override fun sendMessage(message: ChatMessage): ChatMessage? {
        return engine.sendMessage(message)?.also { addMessage(it) }
    }

    override fun addMessage(message: ChatMessage) {
        val conversation = contactRepo.getContact(message.conversationUid) ?: return
        val selfUid = selfProfileRepo.getSelfUid(message.conversationUid)
        val isUnread = message.senderUid != selfUid && !conversation.isMuted &&
                (visibleMessageTracker.isScrolledUp || !visibleMessageTracker.isConversationVisible(message.conversationUid))
        val modified = if (isUnread) {
            message.copy(isUnread = true)
        } else {
            message
        }
        val messages = messages.getOrPut(message.conversationUid) { mutableMapOf() }
        messages[message.uid] = modified
        if (messages.size > maxMessages) messages.removeFirst()
        onNewMessage.tryEmit(modified)
        if (shouldPersistMessage(modified)) {
            persistedMessageLock { persistMessage(modified) }
        }
        if (isUnread) {
            val unreadMessages = unreadMessages[message.conversationUid] ?: UnreadMessageCounts()
            this.unreadMessages[message.conversationUid] = unreadMessages.run {
                if (conversation is UserData || conversation is ChatGroupData && modified.containsSelfMention(conversation)) {
                    copy(directMessages = directMessages + 1)
                } else {
                    copy(groupMessages = groupMessages + 1)
                }
            }
        }
    }
    
    override fun updateMessage(message: ChatMessage, localOnly: Boolean, oldUid: MessageUid) {
        val messages = messages[message.conversationUid] ?: return
        val existing = messages[oldUid]
        if (oldUid != message.uid) {
            messages.remove(oldUid)
        }
        messages[message.uid] = message
        onMessagesChanged.tryEmit(message.conversationUid)
        existing?.let {
            if (!localOnly) engine.editMessage(old = it, new = message)
        }
        if (shouldPersistMessage(message)) {
            persistedMessageLock { editPersistedMessage(message, oldUid) }
        }
    }

    override fun markMessagesRead(conversationUid: ContactUid, messages: List<ChatMessage>) {
        val messageMap = this.messages[conversationUid] ?: return
        val unreadMessages = mutableMapOf<String, ChatMessage>()
        messages.forEach { message ->
            val existing = messageMap[message.uid] ?: return@forEach
            if (existing.isUnread) {
                val updated = existing.copy(isUnread = false)
                messageMap[message.uid] = updated
                unreadMessages[message.uid.value] = updated
            }
        }
        if (unreadMessages.isNotEmpty()) {
            this.unreadMessages[conversationUid] = getUnreadMessageCounts(conversationUid)
            markPersistedMessagesRead(
                conversationUid = conversationUid,
                unreadMessages = unreadMessages,
            )
        }
    }
    
    override fun addReaction(message: ChatMessage, reaction: String) {
        val selfUid = selfProfileRepo.getSelfUid(message.conversationUid)
        val reactions = message.reactions[selfUid]?.toMutableSet() ?: mutableSetOf()
        if (reaction in reactions) {
            reactions -= reaction
        } else {
            reactions += reaction
        }
        if (reactions.size <= MAX_REACTIONS_PER_USER) {
            setReactions(
                senderUid = selfUid,
                message = message,
                reactions = reactions,
                localOnly = false,
            )
        } else {
            eventHandler += ErrorDialogEvent(stringProvider[Res.string.message_reaction_limit_message])
        }
    }
    
    override fun setReactions(
        senderUid: UserUid,
        message: ChatMessage,
        reactions: Set<String>,
        localOnly: Boolean,
    ) {
        if (reactions.size > MAX_REACTIONS_PER_USER) return
        updateMessage(
            message = message.setReactions(
                senderUid = senderUid,
                reactions = reactions,
            ),
            localOnly = true,
        )
        if (!localOnly) {
            engine.setMessageReactions(
                message = message,
                reactions = reactions,
            )
        }
    }

    override fun removeMessage(message: ChatMessage, localOnly: Boolean) {
        messages[message.conversationUid]?.remove(message.uid)
        onMessagesChanged.tryEmit(message.conversationUid)
        if (!localOnly) engine.removeMessage(message)
        if (shouldPersistMessage(message)) {
            persistedMessageLock { removePersistedMessage(message) }
        }
    }

    override fun removeMessages(conversationUid: ContactUid, isRemovingContact: Boolean) {
        messages.remove(conversationUid)
        unreadMessages.remove(conversationUid)
        deletePersistedMessages(conversationUid)
        if (!isRemovingContact) {
            onMessagesChanged.tryEmit(conversationUid)
        }
    }

    override fun removeGroupInvite(groupInvite: ChatGroupInvite) {
        val messages = messages[groupInvite.friendUid] ?: return
        val message = messages.values.find { message ->
            groupInvite.cookie == message.groupInvite?.cookie
        } ?: return
        messages[message.uid] = message.copy(groupInvite = null)
        onMessagesChanged.tryEmit(groupInvite.friendUid)
    }

    override fun findMessage(
        conversationUid: ContactUid,
        senderUid: ContactUid,
        textHash: String,
        timestamp: Long,
    ) = messages[conversationUid]?.values?.find {
        it.senderUid == senderUid &&
                abs(it.timestamp - timestamp) <= FIND_MESSAGE_TIMESTAMP_THRESHOLD &&
                it.text.sha256() == textHash
    }

    override suspend fun onEngineInit(profile: ChatProfileData) {
        messagesDirectory = fileSystem.getProfileDirectory(profile, "messages")
        loadPersistedMessages()
        loadPersistedFriendRequests()
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        this.coroutineScope = coroutineScope
    }

    override fun onEngineStop() {
        messages.clear()
        visibleMessageTracker.reset()
    }

    override fun onFriendRequest(request: FriendRequest) {
        logger.debug("onFriendRequest($request)")
        addMessage(
            ChatMessage(
                conversationUid = request.uid,
                senderUid = request.uid,
                text = request.message,
                friendRequest = request,
            )
        )
    }

    override fun onFriendRequestAccepted(request: FriendRequest, newFriend: UserData) {
        messages.remove(request.uid)?.let { existingMessages ->
            messages[newFriend.uid] = existingMessages.map { (uid, message) ->
                uid to message.copy(
                    conversationUid = request.uid,
                    senderUid = newFriend.uid,
                    friendRequest = null,
                )
            }.toMutableMap()
        }
        addMessage(
            ChatMessage(
                conversationUid = newFriend.uid,
                senderUid = newFriend.uid,
                text = stringProvider[Res.string.user_added_message],
                type = ChatMessageType.ACTION,
            )
        )
    }

    override fun onFriendMessage(friendUid: UserUid, messageType: ChatMessageType, message: String) {
        logger.debug("onFriendMessage($friendUid, $messageType, $message)")
        addMessage(
            ChatMessage(
                conversationUid = friendUid,
                senderUid = friendUid,
                text = message,
                type = messageType,
            )
        )
    }

    override fun onFriendReadReceipt(friendUid: UserUid, messageUid: MessageUid) {
        logger.debug("onFriendReadReceipt($friendUid, $messageUid)")
        val messages = messages[friendUid] ?: return
        val message = messages[messageUid] ?: return
        messages[messageUid] = message.copy(
            timestamp = currentTimeMillis(),
            isDelivered = true,
        )
        onMessagesChanged.tryEmit(friendUid)
    }

    override fun onGroupInvite(invite: ChatGroupInvite) {
        logger.debug("onGroupInvite($invite)")
        addMessage(
            ChatMessage(
                conversationUid = invite.friendUid,
                senderUid = invite.friendUid,
                text = stringProvider.getString(Res.string.group_invite_peer_message, "\n\$tiny\$${invite.cookie}"),
                groupInvite = invite,
                type = ChatMessageType.ACTION,
            )
        )
    }

    override fun onGroupMessage(groupUid: GroupUid, peerUid: UserUid, messageType: ChatMessageType, message: String) {
        logger.debug("onGroupMessage($groupUid, $peerUid, $messageType, $message)")
        addMessage(
            ChatMessage(
                conversationUid = groupUid,
                senderUid = peerUid,
                text = message,
                type = messageType,
            )
        )
    }

    override fun onGroupPeerMessage(groupUid: GroupUid, peerUid: UserUid, messageType: ChatMessageType, message: String) {
        logger.debug("onGroupPeerMessage($groupUid, $peerUid, $messageType, $message)")
        addMessage(
            ChatMessage(
                conversationUid = GroupPeerUid(groupUid, peerUid),
                senderUid = peerUid,
                text = message,
                type = messageType,
            )
        )
    }

    override fun onGroupUserName(groupUid: GroupUid, peerUid: UserUid, oldName: String, newName: String) {
        logger.debug("onGroupUserName($groupUid, $peerUid, $oldName, $newName)")
        if (settings.isPeerNameMessagesEnabled) {
            addMessage(
                ChatMessage(
                    conversationUid = groupUid,
                    senderUid = groupUid,
                    text = if (oldName.isBlank() || oldName == newName) {
                        stringProvider.getString(Res.string.peer_joined_group_message, newName)
                    } else {
                        stringProvider.getString(Res.string.peer_name_changed_message, oldName, newName)
                    },
                    type = ChatMessageType.GROUP,
                )
            )
        }
    }

    override fun onGroupPeerConnection(groupUid: GroupUid, peerUid: UserUid, isOnline: Boolean) {
        logger.debug("onGroupPeerConnection($groupUid, $peerUid, $isOnline)")
        if (settings.isPeerConnectMessagesEnabled) {
            addMessage(
                ChatMessage(
                    conversationUid = groupUid,
                    senderUid = peerUid,
                    text = "",
                    type = if (isOnline) ChatMessageType.PEER_ONLINE else ChatMessageType.PEER_OFFLINE,
                )
            )
        }
    }

    override fun onGroupPeerKicked(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid) {
        logger.debug("onGroupPeerKicked($groupUid, $sourceUid, $targetUid)")
        val group = contactRepo.getGroup(groupUid) ?: return
        val source = group.peers[sourceUid] ?: return
        val target = group.peers[targetUid] ?: return
        addMessage(
            ChatMessage(
                conversationUid = groupUid,
                senderUid = groupUid,
                text = stringProvider.getString(Res.string.group_event_peer_kicked_message, source.displayName, target.displayName),
                type = ChatMessageType.GROUP,
            )
        )
    }

    override fun onGroupPeerRoleChanged(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid, role: ChatGroupRole) {
        logger.debug("onGroupPeerRoleChanged($groupUid, $sourceUid, $targetUid, $role)")
        val group = contactRepo.getGroup(groupUid) ?: return
        val source = group.peers[sourceUid] ?: return
        val target = group.peers[targetUid] ?: return
        addMessage(
            ChatMessage(
                conversationUid = groupUid,
                senderUid = groupUid,
                text = stringProvider.getString(Res.string.group_event_peer_role_changed_message, source.displayName, target.displayName, role.getLabelResource()),
                type = ChatMessageType.GROUP,
            )
        )
    }

    override fun onGroupTalkStateChanged(groupUid: GroupUid, talkState: ChatGroupTalkState) {
        logger.debug("onGroupTalkStateChanged($groupUid, $talkState)")
        addMessage(
            ChatMessage(
                conversationUid = groupUid,
                senderUid = groupUid,
                text = stringProvider.getString(Res.string.group_event_talk_state_changed_message, talkState.getLabelResource()),
                type = ChatMessageType.GROUP,
            )
        )
    }
    
    override fun addFileTransferMessage(
        transfer: ChatFileTransfer,
        isAutoAccepted: Boolean,
    ) = with(transfer) {
        if (transfer.type == ChatFileType.AVATAR) return
        val senderUid = when (transfer.direction) {
            ChatFileTransferDirection.OUTGOING -> selfProfileRepo.getSelfUid()
            ChatFileTransferDirection.INCOMING -> friendUid
        }
        val text = when {
            isAutoAccepted || transfer.direction == ChatFileTransferDirection.OUTGOING -> ""
            else -> stringProvider[Res.string.user_sent_file_message]
        }
        addMessage(
            ChatMessage(
                conversationUid = friendUid,
                senderUid = senderUid,
                text = text,
                type = if (isAutoAccepted) ChatMessageType.NORMAL else ChatMessageType.ACTION,
                fileTransferUid = uid,
            )
        )
    }

    override fun markConversationRead(conversationUid: ContactUid) {
        unreadMessages.remove(conversationUid)
        val messages = messages[conversationUid] ?: return
        val unreadMessages = messages.filterValues { it.isUnread }
        if (unreadMessages.isEmpty()) return
        unreadMessages.forEach { (uid, message) ->
            messages[uid] = message.copy(isUnread = false)
        }
        markPersistedMessagesRead(
            conversationUid = conversationUid,
            unreadMessages = unreadMessages.mapKeys { it.key.value },
        )
    }

    private fun markPersistedMessagesRead(
        conversationUid: ContactUid,
        unreadMessages: Map<String, ChatMessage>,
    ) = persistedMessageLock {
        if (conversationUid.value in messagesDirectory) {
            messagesDirectory.editLines(conversationUid.value) { line ->
                val uid = line.substringBefore(':')
                val message = unreadMessages[uid]
                if (message != null) {
                    message.copy(isUnread = false).toPersistedMessageLine()
                } else {
                    line
                }
            }
        }
    }

    private fun ChatMessage.containsSelfMention(group: ChatGroupData) = containsSelfMention(
        group = group,
        mentionsRegex = engine.constraints.mentionRegex,
    )
    
    private fun shouldPersistMessage(message: ChatMessage): Boolean {
        return this::messagesDirectory.isInitialized &&
                settings.isMessageHistoryEnabled &&
                message.conversationUid !is FriendRequestUid
    }
    
    private fun persistedMessageLock(block: () -> Unit) {
        if (this::coroutineScope.isInitialized) {
            coroutineScope.launch(backgroundWorker.dispatcher) { persistedMutex.withLock(action = block) }
        } else {
            logger.warn("Failed to get persisted message lock - coroutine scope is uninitialized", Exception())
        }
    }
    
    private fun persistMessage(message: ChatMessage) {
        val line = message.toPersistedMessageLine()
        messagesDirectory.appendToFile(
            fileName = message.conversationUid.value,
            data = "$line\n".encodeToByteArray(),
        )
    }

    private fun editPersistedMessage(message: ChatMessage, oldUid: MessageUid = message.uid) {
        messagesDirectory.editLines(message.conversationUid.value) { line ->
            val uid = line.substringBefore(':')
            if (uid == oldUid.value) {
                message.toPersistedMessageLine()
            } else {
                line
            }
        }
    }
    
    private fun removePersistedMessage(message: ChatMessage) {
        messagesDirectory.editLines(message.conversationUid.value) { line ->
            val uid = line.substringBefore(':')
            if (uid == message.uid.value) null else line
        }
    }
    
    private fun ChatMessage.toPersistedMessageLine(): String? {
        val fileTransfer = fileTransferUid?.let { fileTransferRepo[it] }
        val persistedMessage = toPersistedMessage(fileTransfer)
        val messageJson = json.encodeToString(persistedMessage)
        val bytes = messageJson.encodeToByteArray().let {
            if (encryption.isEncryptionSupported()) encryption.encryptData(it) else it
        } ?: run {
            logger.error("Failed to encrypt persisted message: $uid")
            return null
        }
        return uid.value + ":" + bytes.encodeBase64()
    }
    
    private suspend fun loadPersistedMessages() {
        if (!this::messagesDirectory.isInitialized || !settings.isMessageHistoryEnabled) return
        
        val start = currentTimeMillis()
        val messageLimit = maxMessages
        withContext(backgroundWorker.dispatcher) {
            messagesDirectory.listFiles().forEach { path ->
                var directMessages = 0
                var groupMessages = 0
                val conversationUid = when {
                    path.name.contains("-") -> {
                        val (groupUid, peerUid) = path.name.split("-")
                        GroupPeerUid(
                            groupUid = GroupUid(groupUid),
                            peerUid = UserUid(peerUid),
                        )
                    }
                    else -> contactRepo.getContact(path.name)?.uid
                }
                val conversation = conversationUid?.let { contactRepo.getContact(it) } ?: run {
                    messagesDirectory.removeFile(path.name)
                    return@forEach
                }
                var lineCount = 0
                val loadedMessages = mutableMapOf<MessageUid, ChatMessage>()
                val rawLines = mutableListOf<String>()
                messagesDirectory.readLines(path.name) { line ->
                    lineCount++
                    decodePersistedMessage(line, conversationUid)?.let { message ->
                        loadedMessages[message.uid] = message
                        rawLines += line
                        if (message.isUnread) {
                            when (conversation) {
                                is ChatGroupData -> {
                                    if (message.containsSelfMention(conversation)) directMessages++ else groupMessages++
                                }
                                else -> directMessages++
                            }
                        }
                    }
                    if (loadedMessages.size > messageLimit) {
                        loadedMessages.removeFirst()
                        rawLines.removeAt(0)
                    }
                    true
                }
                if (rawLines.size < lineCount) {
                    // Time to trim the fat
                    messagesDirectory.writeLines(
                        fileName = path.name,
                        lines = rawLines,
                    )
                }
                messages[conversationUid] = loadedMessages
                if (directMessages > 0 || groupMessages > 0) {
                    unreadMessages[conversationUid] = UnreadMessageCounts(
                        directMessages = directMessages,
                        groupMessages = groupMessages,
                    )
                }
            }
        }
        val loadTime = currentTimeMillis() - start
        if (loadTime > 1000L) {
            logger.warn("Loading persisted messages took $loadTime ms!")
        } else {
            logger.verbose("Loading persisted messages took $loadTime ms")
        }
    }

    private suspend fun loadPersistedFriendRequests() {
        engine.getFriendRequestList().fastForEach { request ->
            val message = ChatMessage(
                conversationUid = request.uid,
                senderUid = request.uid,
                text = request.message,
                friendRequest = request,
            )
            messages[request.uid] = mutableMapOf(message.uid to message)
        }
    }
    
    private fun decodePersistedMessage(line: String, conversationUid: ContactUid): ChatMessage? {
        val uid = line.substringBefore(':', randomUid())
        val message = line.substringAfter(':').decodeBase64Bytes().let {
            if (encryption.isEncryptionSupported() && encryption.isEncrypted(it)) encryption.decryptData(it) else it
        }?.let {
            try {
                json.decodeFromString<PersistedChatMessage>(it.decodeToString())
            } catch (e: Exception) {
                null
            }
        } ?: return null
        val sender = contactRepo.getContact(
            uid = message.senderUid,
            userGroup = when (conversationUid) {
                is GroupUid -> contactRepo.getGroup(conversationUid)
                is GroupPeerUid -> contactRepo.getGroup(conversationUid.groupUid)
                else -> null
            },
        ) ?: return null
        val fileTransfer = message.fileTransfer?.run {
            ChatFileTransfer(
                fileUid = FileUid(fileUid),
                friendUid = UserUid(message.senderUid),
                name = name,
                path = path,
                size = size,
                direction = direction,
                state = when {
                    fileSystem.exists(path) && fileSystem.getSize(path) == size -> ChatFileTransferState.COMPLETE
                    else -> ChatFileTransferState.REMOVED
                },
            ).also {
                fileTransferRepo.addHistoricalTransfer(it)
            }
        }
        return ChatMessage(
            conversationUid = conversationUid,
            senderUid = sender.uid,
            text = message.text,
            uid = MessageUid(uid),
            timestamp = message.timestamp,
            type = message.type,
            isDelivered = message.isDelivered,
            isEdited = message.isEdited,
            isUnread = message.isUnread,
            reactions = message.reactions.orEmpty(),
            fileTransferUid = fileTransfer?.uid,
        )
    }
    
    private fun deletePersistedMessages(conversationUid: ContactUid) {
        if (this::messagesDirectory.isInitialized && conversationUid.value in messagesDirectory) {
            messagesDirectory.removeFile(conversationUid.value)
        }
    }
}