package nand.apps.chat.repo

import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.group.*
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.user.*
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.model.event.*
import nand.apps.chat.model.uid.*
import nand.apps.chat.res.StringResourceProvider
import nand.apps.chat.util.LOCAL_CLIENT_VERSION
import nand.apps.chat.util.fifoSharedFlow
import nand.apps.chat.util.MutableStateFlowMap
import okio.Path.Companion.toPath
import org.koin.core.Koin
import seers.composeapp.generated.resources.*

class ContactRepoImpl(koin: Koin) : ContactRepo {

    private val engine: ChatEngine by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val eventHandler: ChatEventHandler by koin.inject()
    private val stringProvider: StringResourceProvider by koin.inject()
    
    private val friends = MutableStateFlowMap<UserUid, UserData>()
    private val groups = MutableStateFlowMap<GroupUid, ChatGroupData>()
    private val friendRequests = MutableStateFlowMap<FriendRequestUid, FriendRequest>()
    private val groupInvites = MutableStateFlowMap<String, ChatGroupInvite>()
    
    private val logger = getLogger("ContactRepo")

    override val friendObserver = friends
    override val groupObserver = groups
    override val friendRequestObserver = friendRequests

    override val onFriendRequest = fifoSharedFlow<FriendRequest>()
    override val onGroupInvite = fifoSharedFlow<ChatGroupInvite>()

    override val self get() = selfProfileRepo.self

    override fun getFriends() = friends.values.toList()

    override fun getGroups() = groups.values.toList()

    override fun getOnlineFriends() = friends.values.filter { it.isOnline }

    override fun getOnlineGroups() = groups.values.filter { it.isOnline }

    override fun getOfflineGroups() = groups.values.filter { !it.isOnline }

    override fun getFriend(uid: UserUid) = friends[uid]

    override fun getFriendRequest(uid: FriendRequestUid) = friendRequests[uid]

    override fun getGroup(uid: GroupUid) = groups[uid]
    
    override fun getContact(uid: ContactUid, userGroup: ChatGroupData?) = when (uid) {
        is UserUid -> when {
            uid == self.uid -> self
            userGroup?.peers?.containsKey(uid) == true -> userGroup.peers[uid]
            else -> friends[uid]
        }
        is FriendRequestUid -> friendRequests[uid]
        is GroupUid -> groups[uid]
        is GroupPeerUid -> groups[uid.groupUid]?.peers?.get(uid.peerUid)
    }
    
    override fun getContact(uid: String, userGroup: ChatGroupData?): ChatContact? {
        return getContact(UserUid(uid), userGroup) ?: groups[GroupUid(uid)]
    }

    override fun getRandomStatusMessage(): String {
        return stringProvider.getStringArray(Res.array.default_status_messages).random()
    }
    
    override fun isPeerBlocked(groupUid: GroupUid, peerUid: UserUid) =
        getGroup(groupUid)?.isPeerBlocked(peerUid) == true

    override suspend fun onEngineInit(profile: ChatProfileData) {
        friends.putAll(engine.getFriendList().associateBy { it.uid })
        friendRequests.putAll(engine.getFriendRequestList().associateBy { it.uid })
        groups.putAll(engine.getGroupList().associateBy { it.uid })
    }

    override fun onEngineStart(coroutineScope: CoroutineScope) {
        // Nothing to do here
    }

    override fun onEngineStop() {
        friends.clear()
        groups.clear()
        friendRequests.clear()
        groupInvites.clear()
    }

    override fun onFriendName(friendUid: UserUid, name: String): Boolean {
        logger.debug("onFriendName($friendUid, $name)")
        return updateFriend(friendUid) { it.copy(name = name) }
    }

    override fun onFriendStatus(friendUid: UserUid, status: UserStatus): Boolean {
        logger.debug("onFriendStatus($friendUid, $status)")
        return updateFriend(friendUid) { it.copy(status = status) }
    }

    override fun onFriendStatusMessage(friendUid: UserUid, message: String): Boolean {
        logger.debug("onFriendStatusMessage($friendUid, $message)")
        return updateFriend(friendUid) { it.copy(statusMessage = message) }
    }

    override fun onFriendClientVersion(friendUid: UserUid, version: ChatClientVersion): Boolean {
        logger.debug("onFriendClientVersion($friendUid, $version)")
        return updateFriend(friendUid) { it.copy(clientVersion = version) }
    }

    override fun onFriendConnectionProtocol(friendUid: UserUid, protocol: ConnectionProtocol) {
        logger.debug("onFriendConnectionProtocol($friendUid, $protocol)")
        val friend = friends[friendUid] ?: return
        val wasOnline = friend.isOnline
        val isOnline = protocol != ConnectionProtocol.NONE
        val lastOnline = if (wasOnline && !isOnline) currentTimeMillis() else friend.lastOnline
        updateFriend(friendUid) {
            it.copy(
                protocol = protocol,
                lastOnline = lastOnline,
            )
        }
        if (!wasOnline && isOnline) {
            avatarRepo.checkSendSelfAvatar(self.uid, friendUid)
        }
    }

    override fun onUserAvatarChanged(uid: UserUid, groupUid: GroupUid?, avatarPath: String?) {
        logger.debug("onUserAvatarChanged($uid, $groupUid, $avatarPath)")
        when {
            self.uid == uid -> selfProfileRepo.setCurrentAvatarUri(avatarPath)
            groupUid != null -> groups[groupUid]?.let { group ->
                group.peers[uid]?.let { peer ->
                    updateGroupPeer(groupUid, peer.uid) { it.copy(avatarUri = avatarPath) }
                }
            }
            else -> friends[uid]?.let {
                updateFriend(uid) { it.copy(avatarUri = avatarPath) }
            }
        }
    }

    override fun onFriendRequest(request: FriendRequest) {
        logger.debug("onFriendRequest($request)")
        friendRequests[request.uid] = request
        onFriendRequest.tryEmit(request)
    }

    override fun onGroupInvite(invite: ChatGroupInvite) {
        logger.debug("onGroupInvite($invite)")
        groupInvites[invite.cookie] = invite
        onGroupInvite.tryEmit(invite)
    }

    override fun onGroupAdded(group: ChatGroupData) {
        logger.debug("onGroupAdded($group)")
        if (group.uid !in groups) {
            groups[group.uid] = group
        } else {
            logger.warn("Attempted to add group that already exists: $group")
        }
    }
    
    override fun onGroupConnected(
        groupUid: GroupUid,
        groupData: ChatGroupData?,
        fetchPeers: Boolean,
    ) {
        logger.debug("onGroupConnected($groupUid)")
        updateGroup(groupUid) {
            (groupData ?: it).copy(
                isOnline = true,
                peers = if (fetchPeers) engine.getGroupPeerList(groupUid) else it.peers,
            )
        }
    }

    override fun onGroupErrorState(groupUid: GroupUid, error: ChatGroupErrorType) {
        logger.debug("onGroupErrorState($groupUid, $error)")
        updateGroup(groupUid) { it.copy(error = error, isOnline = false) }
    }

    override fun onGroupTitle(groupUid: GroupUid, title: String): Boolean {
        logger.debug("onGroupTitle($groupUid, $title)")
        return updateGroup(groupUid) { it.copy(title = title) }
    }

    override fun onGroupDescription(groupUid: GroupUid, description: String): Boolean {
        logger.debug("onGroupDescription($groupUid, $description)")
        return updateGroup(groupUid) { it.copy(description = description) }
    }

    override fun onGroupDescriptionLocked(groupUid: GroupUid, isLocked: Boolean): Boolean {
        logger.debug("onGroupDescriptionLocked($groupUid, $isLocked)")
        return updateGroup(groupUid) { it.copy(isDescriptionLocked = isLocked) }
    }

    override fun onGroupPassword(groupUid: GroupUid, password: String?): Boolean {
        logger.debug("onGroupPassword($groupUid, *****)")
        return updateGroup(groupUid) { it.copy(password = password) }
    }

    override fun onGroupVisibility(groupUid: GroupUid, visibility: ChatGroupVisibility): Boolean {
        logger.debug("onGroupVisibility($groupUid, $visibility)")
        return updateGroup(groupUid) { it.copy(visibility = visibility) }
    }

    override fun onGroupPeerLimit(groupUid: GroupUid, peerLimit: Int): Boolean {
        logger.debug("onGroupPeerLimit($groupUid, $peerLimit)")
        return updateGroup(groupUid) { it.copy(peerLimit = peerLimit) }
    }

    override fun onGroupTalkState(groupUid: GroupUid, talkState: ChatGroupTalkState): Boolean {
        logger.debug("onGroupTalkState($groupUid, $talkState)")
        messageRepo.onGroupTalkStateChanged(groupUid, talkState)
        return updateGroup(groupUid) { it.copy(talkState = talkState) }
    }

    override fun onGroupAvatar(groupUid: GroupUid, avatarUri: String?): Boolean {
        logger.debug("onGroupAvatar($groupUid, $avatarUri)")
        return updateGroup(groupUid) { it.copy(avatarUri = avatarUri) }
    }

    override fun onGroupUsersChanged(groupUid: GroupUid): Boolean {
        val group = groups[groupUid]
        if (group?.isOnline != true) return false
        logger.debug("onGroupUsersChanged($groupUid)")
        
        // Track which users have just connected or disconnected from the group
        val newPeers = engine.getGroupPeerList(groupUid)
        group.peers.forEach { (uid, peer) ->
            val newPeer = newPeers[uid] ?: return@forEach
            if (peer.isOnline != newPeer.isOnline && !group.isPeerBlocked(peer.uid)) {
                messageRepo.onGroupPeerConnection(groupUid, peer.uid, peer.isOnline)
            }
        }
        
        return updateGroup(groupUid) { it.copy(peers = newPeers) }
    }

    override fun onGroupPeerJoin(groupUid: GroupUid, peer: UserData, isSelf: Boolean): Boolean {
        val group = groups[groupUid] ?: return false
        logger.debug("onGroupPeerJoin($groupUid, ${peer.uid})")
        if (group.peers[peer.uid]?.isOnline == false && !group.isPeerBlocked(peer.uid)) {
            messageRepo.onGroupPeerConnection(groupUid, peer.uid, isOnline = true)
        }
        return updateGroup(groupUid) {
            val peers = it.peers.toMutableMap()
            peers[peer.uid] = initGroupPeer(it, peer, isSelf)
            it.copy(peers = peers)
        }
    }

    override fun onGroupPeerExit(groupUid: GroupUid, peerUid: UserUid): Boolean {
        val group = groups[groupUid] ?: return false
        logger.debug("onGroupPeerExit($groupUid, $peerUid)")
        if (group.peers[peerUid]?.isOnline == true && !group.isPeerBlocked(peerUid)) {
            messageRepo.onGroupPeerConnection(groupUid, peerUid, isOnline = false)
        }
        return updateGroupPeer(groupUid, peerUid) {
            it.copy(
                protocol = ConnectionProtocol.NONE,
                lastOnline = currentTimeMillis(),
            )
        }
    }

    override fun onGroupPeerName(groupUid: GroupUid, peerUid: UserUid, name: String): Boolean {
        if (groups[groupUid]?.isOnline != true) return false
        logger.debug("onGroupUserName($groupUid, $peerUid, $name)")
        return updateGroupPeer(groupUid, peerUid) { peer ->
            if (peer.role != ChatGroupRole.OBSERVER && !isPeerBlocked(groupUid, peerUid)) {
                messageRepo.onGroupUserName(groupUid, peerUid, peer.name, name)
            }
            peer.copy(name = name)
        }
    }

    override fun onGroupPeerStatus(groupUid: GroupUid, peerUid: UserUid, status: UserStatus): Boolean {
        logger.debug("onGroupPeerStatus($groupUid, $peerUid, $status)")
        return updateGroupPeer(groupUid, peerUid) { it.copy(status = status) }
    }

    override fun onGroupPeerClientVersion(groupUid: GroupUid, peerUid: UserUid, version: ChatClientVersion): Boolean {
        logger.debug("onGroupPeerClientVersion($groupUid, $peerUid, $version)")
        return updateGroupPeer(groupUid, peerUid) { it.copy(clientVersion = version) }
    }

    override fun onGroupPeerKicked(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid) {
        logger.debug("onGroupPeerKicked($groupUid, $sourceUid, $targetUid)")
        if (targetUid == selfProfileRepo.getSelfUid(groupUid)) {
            onGroupErrorState(
                groupUid = groupUid,
                error = ChatGroupErrorType.KICKED,
            )
        }
        updateGroupPeer(groupUid, targetUid) { it.copy(protocol = ConnectionProtocol.NONE) }
    }

    override fun onGroupPeerRoleChanged(groupUid: GroupUid, sourceUid: UserUid, targetUid: UserUid, role: ChatGroupRole) {
        logger.debug("onGroupPeerRoleChanged($groupUid, $sourceUid, $targetUid, $role)")
        updateGroupPeer(groupUid, targetUid) { it.copy(role = role) }
    }
    
    override fun sendFriendRequest(name: String, address: String, message: String): UserData? {
        logger.debug("sendFriendRequest($name, $address, $message)")
        val friendUid = engine.sendFriendRequest(address, message) ?: return null
        val newFriend = UserData(
            uid = friendUid,
            name = name,
        )
        friends[friendUid] = newFriend
        return newFriend
    }

    override fun acceptFriendRequest(friendRequest: FriendRequest): UserData? {
        logger.debug("acceptFriendRequest($friendRequest)")
        friendRequests.remove(friendRequest.uid)
        val friendUid = engine.acceptFriendRequest(friendRequest.uid) ?: return null
        val newFriend = UserData(
            uid = friendUid,
            name = "(Pending Friend)",
        )
        friends[friendUid] = newFriend
        return newFriend
    }
    
    override fun declineFriendRequest(friendRequest: FriendRequest): Boolean {
        logger.debug("declineFriendRequest($friendRequest)")
        friendRequests.remove(friendRequest.uid)
        engine.declineFriendRequest(friendRequest.uid)
        return true
    }
    
    override fun removeFriend(friendUid: UserUid): Boolean {
        logger.debug("removeFriend($friendUid)")
        engine.removeFriend(friendUid)
        friends.remove(friendUid)
        return true
    }

    override fun setFriendAlias(friendUid: UserUid, alias: String?) {
        logger.debug("setFriendAlias($friendUid, $alias)")
        if (engine.setFriendAlias(friendUid, alias)) {
            updateFriend(friendUid) { it.copy(alias = alias) }
        }
    }
    
    override fun createGroup(title: String, type: ChatGroupType, visibility: ChatGroupVisibility): ChatGroupData? {
        logger.debug("createGroup($title, $type)")
        return engine.createGroup(title, type, visibility)?.also { group -> groups[group.uid] = group }
    }
    
    override fun setGroupTitle(groupUid: GroupUid, title: String) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupTitle($groupUid, $title)")
        if (title != group.title && engine.setGroupTitle(groupUid, title)) {
            groups[groupUid] = group.copy(title = title)
        }
    }
    
    override fun setGroupDescription(groupUid: GroupUid, description: String) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupDescription($groupUid, $description)")
        if (description != group.description && engine.setGroupDescription(groupUid, description)) {
            groups[groupUid] = group.copy(description = description)
        }
    }

    override fun setGroupDescriptionLocked(groupUid: GroupUid, isLocked: Boolean) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupDescriptionLocked($groupUid, $isLocked)")
        if (isLocked != group.isDescriptionLocked && engine.setGroupDescriptionLocked(groupUid, isLocked)) {
            groups[groupUid] = group.copy(isDescriptionLocked = isLocked)
        }
    }
    
    override fun setGroupPassword(groupUid: GroupUid, password: String?) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupPassword($groupUid, *****)")
        if (password != group.password && engine.setGroupPassword(groupUid, password)) {
            groups[groupUid] = group.copy(password = password)
        }
    }
    
    override fun setGroupNickname(groupUid: GroupUid, name: String) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupNickname($groupUid, $name)")
        val self = group.findSelf(self)
        if (name != self.name && engine.setGroupNickname(groupUid, name)) {
            updateGroupPeer(groupUid, self.uid) { it.copy(name = name) }
        }
    }

    override fun setGroupPeerLimit(groupUid: GroupUid, peerLimit: Int) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupPeerLimit($groupUid, $peerLimit)")
        if (peerLimit != group.peerLimit && engine.setGroupPeerLimit(groupUid, peerLimit)) {
            updateGroup(groupUid) { it.copy(peerLimit = peerLimit) }
        }
    }

    override fun setGroupTalkState(groupUid: GroupUid, talkState: ChatGroupTalkState) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupTalkState($groupUid, $talkState)")
        if (talkState != group.talkState && engine.setGroupTalkState(groupUid, talkState)) {
            messageRepo.onGroupTalkStateChanged(groupUid, talkState)
            updateGroup(groupUid) { it.copy(talkState = talkState) }
        }
    }

    override fun setGroupVisibility(groupUid: GroupUid, visibility: ChatGroupVisibility) {
        val group = groups[groupUid] ?: return
        logger.debug("setGroupVisibility($groupUid, $visibility)")
        if (visibility != group.visibility && engine.setGroupVisibility(groupUid, visibility)) {
            updateGroup(groupUid) { it.copy(visibility = visibility) }
        }
    }
    
    override fun setGroupRole(groupUid: GroupUid, peerUid: UserUid, role: ChatGroupRole) {
        val group = groups[groupUid] ?: return
        val peer = group.peers[peerUid] ?: return
        if (peer.role == role) return
        logger.debug("setGroupRole($groupUid, $peerUid, $role)")
        val self = group.findSelf(self)
        if (self.hasAuthorityOver(peer) && engine.setGroupRole(groupUid, peerUid, role)) {
            updateGroupPeer(groupUid, peerUid) { it.copy(role = role) }
        } else {
            eventHandler += ErrorDialogEvent(
                message = stringProvider[Res.string.group_event_peer_role_change_failed],
            )
        }
    }
    
    override fun kickGroupPeer(groupUid: GroupUid, peerUid: UserUid) {
        val group = groups[groupUid] ?: return
        val peer = group.peers[peerUid] ?: return
        logger.debug("kickGroupPeer($groupUid, $peerUid)")
        val self = group.findSelf(self)
        if (self.hasAuthorityOver(peer) && engine.kickGroupPeer(groupUid, peerUid)) {
            onGroupPeerExit(groupUid, peerUid)
        } else {
            eventHandler += ErrorDialogEvent(
                message = stringProvider[Res.string.group_event_peer_kick_failed],
            )
        }
    }

    override fun setGroupPeerBlocked(groupUid: GroupUid, peerUid: UserUid, block: Boolean) {
        val group = groups[groupUid] ?: return
        val peer = group.peers[peerUid] ?: return
        logger.debug("setGroupPeerBlocked($groupUid, $peerUid, $block)")
        if (group.isPeerBlocked(peerUid) != block && engine.setGroupPeerBlocked(groupUid, peerUid, block)) {
            updateGroup(groupUid) { it.setPeerBlock(peer.uid, block) }
        }
    }
    
    override fun joinGroup(groupName: String?, groupUid: GroupUid, selfNickname: String?, password: String?): ChatGroupData? {
        val success = engine.joinGroup(
            groupUid = groupUid,
            selfNickname = selfNickname ?: self.name,
            password = password,
        )
        return if (success) {
            groups[groupUid] ?: ChatGroupData(
                uid = groupUid,
                selfUid = self.uid,
                title = groupName ?: stringProvider[Res.string.group_untitled],
                type = ChatGroupType.ADVANCED,
                password = password,
                visibility = ChatGroupVisibility.PUBLIC,
            ).also { groups[groupUid] = it }
        } else {
            null
        }
    }
    
    override fun sendGroupInvite(friendUid: UserUid, groupUid: GroupUid): Boolean {
        return engine.sendGroupInvite(friendUid, groupUid).also { success ->
            if (!success) logger.error("Failed to send group invite for $groupUid to $friendUid")
        }
    }
    
    override fun acceptGroupInvite(invite: ChatGroupInvite): ChatGroupData? {
        removeGroupInvite(invite)
        val newGroup = engine.acceptGroupInvite(invite) ?: return null
        groups[newGroup.uid] = newGroup
        return newGroup
    }

    override fun removeGroupInvite(invite: ChatGroupInvite): Boolean {
        groupInvites.remove(invite.cookie)
        return true
    }
    
    override fun removeGroup(groupUid: GroupUid, partingMessage: String?) {
        logger.debug("removeGroup($groupUid)")
        engine.removeGroup(groupUid, partingMessage)
        groups.remove(groupUid)
    }

    override fun setContactMuted(uid: ContactUid, isMuted: Boolean) {
        logger.debug("setContactMuted($uid, $isMuted)")
        when (uid) {
            is UserUid -> if (engine.setContactMuted(uid, isMuted)) {
                updateFriend(uid) { it.copy(isMuted = isMuted) }
            }
            is GroupUid -> if (engine.setContactMuted(uid, isMuted)) {
                updateGroup(uid) { it.copy(isMuted = isMuted) }
            }
            else -> {} // Not applicable
        }
    }
    
    private fun updateFriend(uid: UserUid, block: (UserData) -> UserData): Boolean {
        val oldData = friends[uid]
        val newData = block(friends[uid] ?: UserData(uid))
        friends[uid] = newData
        return oldData != newData
    }

    private fun updateGroup(uid: GroupUid, block: (ChatGroupData) -> ChatGroupData): Boolean {
        val oldData = groups[uid]
        val newData = block(groups[uid] ?: ChatGroupData(
            uid = uid,
            selfUid = self.uid,
        ).also { logger.warn("Failed to find existing group $uid") })
        groups[uid] = newData
        return oldData != newData
    }
    
    private fun updateGroupPeer(
        groupUid: GroupUid,
        peerUid: UserUid,
        block: (UserData) -> UserData,
    ) = updateGroup(groupUid) {
        val peer = it.peers[peerUid] ?: run {
            logger.warn("Failed to find existing peer $peerUid in $groupUid")
            return@updateGroup it.copy(peers = engine.getGroupPeerList(groupUid))
        }
        val peers = it.peers.toMutableMap()
        peers[peerUid] = block(peer)
        it.copy(peers = peers)
    }
    
    private fun initGroupPeer(group: ChatGroupData, peer: UserData, isSelf: Boolean): UserData {        
        return if (isSelf) {
            val avatarPath = avatarRepo.getAvatarPath(self.uid)
            avatarPath?.let {
                avatarRepo.onGroupPeerAvatarReceived(
                    group = group,
                    peer = peer,
                    hash = it.toPath().name,
                )
            }
            peer.copy(
                avatarUri = avatarPath,
                clientVersion = LOCAL_CLIENT_VERSION,
            )
        } else {
            peer.copy(avatarUri = avatarRepo.getAvatarPath(peer.uid))
        }
    }
}