package nand.apps.chat.repo

import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.io.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.file.*
import nand.apps.chat.model.uid.FileTransferUid
import nand.apps.chat.model.uid.FileUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.util.MutableStateFlowMap
import nand.apps.chat.util.sha256
import okio.Path.Companion.toPath
import org.koin.core.Koin

class FileTransferRepoImpl(koin: Koin) : FileTransferRepo {

    private val engine: ChatEngine by koin.inject()
    private val fileSystem: ChatFileSystem by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val avatarRepo: ChatAvatarRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    
    private val fileTransfers = MutableStateFlowMap<FileTransferUid, ChatFileTransfer>()

    private val downloadsDirectory get() = fileSystem.getDirectory(settings.downloadDirectory ?: fileSystem.paths.downloadPath.toString())
    private val settings get() = settingsRepo.settings.fileSettings

    override val observer = fileTransfers

    private val logger = getLogger("FileTransfers")

    override suspend fun onEngineInit(profile: ChatProfileData) {
        // Nothing to do here
    }
    
    override fun onEngineStart(coroutineScope: CoroutineScope) {
        // Nothing to do here
    }

    override fun onEngineStop() {
        fileTransfers.clear()
    }

    override operator fun get(uid: FileTransferUid) = fileTransfers[uid]

    override fun addHistoricalTransfer(fileTransfer: ChatFileTransfer) {
        if (fileTransfer.uid !in fileTransfers) {
            fileTransfers[fileTransfer.uid] = fileTransfer
        }
    }

    override fun sendFile(transfer: ChatFileTransfer): Boolean {
        logger.debug("sendFile(${transfer.friendUid}, ${transfer.name})")
        val success = engine.sendFile(
            friendUid = transfer.friendUid,
            kind = transfer.type,
            fileSize = transfer.size,
            fileName = transfer.name,
            fileUid = transfer.fileUid,
        )
        if (success) {
            updateTransfer(transfer.copy(state = ChatFileTransferState.ACTIVE))
            messageRepo.addFileTransferMessage(
                transfer = transfer,
                isAutoAccepted = true,
            )
        }
        return success
    }

    override fun sendFile(path: String, friendUid: UserUid): Boolean {
        val directory = fileSystem.getDirectoryForFile(path) ?: return false
        val fileName = path.toPath().name
        return sendFile(
            ChatFileTransfer(
                fileUid = FileUid((path + currentTimeMillis()).sha256()),
                friendUid = friendUid,
                name = fileName,
                path = path,
                size = directory.getFileSize(fileName),
                direction = ChatFileTransferDirection.OUTGOING,
            )
        )
    }
    
    override fun sendFileControl(transfer: ChatFileTransfer, control: ChatFileControl) {
        val res = engine.sendFileControl(
            friendUid = transfer.friendUid,
            fileUid = transfer.fileUid,
            control = control,
        )
        when {
            control == ChatFileControl.CANCEL || res == ChatFileControlResponse.OK -> {
                updateTransfer(transfer.copy(state = control.toState()))
                if (control == ChatFileControl.CANCEL) transfer.removeIncomingFile()
            }
            else -> {
                logger.error("Got bad control response for file ${transfer.uid} (${transfer.name}): $res")
            }
        }
    }

    override fun onFileReceived(friendUid: UserUid, fileUid: FileUid, fileSize: Long, fileName: String) {
        logger.debug("onFileReceived($friendUid, $fileUid, $fileSize, $fileName)")
        var transferName = if (!fileName.isValidFileName()) {
            fileName.sanitizeFileName().also {
                logger.debug("Sanitized file name: $fileName -> $it")
            }
        } else {
            fileName
        }
        transferName = downloadsDirectory.resolveDuplicateFileName(transferName)
        val filePath = downloadsDirectory.getFilePath(transferName).toString()
        val transfer = ChatFileTransfer(
            fileUid = fileUid,
            friendUid = friendUid,
            name = transferName,
            path = filePath,
            size = fileSize,
            direction = ChatFileTransferDirection.INCOMING,
        )
        val isAutoAccepted = isFileAutoAcceptable(transferName)
        if (isAutoAccepted) {
            sendFileControl(transfer, ChatFileControl.RESUME)
        } else {
            updateTransfer(transfer)
        }
        messageRepo.addFileTransferMessage(
            transfer = transfer,
            isAutoAccepted = isAutoAccepted,
        )
    }

    override fun onFileChunkReceived(friendUid: UserUid, fileUid: FileUid, position: Long, data: ByteArray) {
        if (position == 0L) {
            logger.debug("onFileChunkReceived($friendUid, $fileUid, $position, B[${data.size}])")
        } else {
            logger.verbose("onFileChunkReceived($friendUid, $fileUid, $position, B[${data.size}])")
        }
        val transfer = getTransfer(friendUid, fileUid, isError = data.isNotEmpty()) ?: return
        val directory = transfer.getIncomingDirectory()
        updateTransfer(transfer.copy(position = position))
        if (data.isNotEmpty() && !directory.appendToFile(transfer.name, data)) {
            sendFileControl(transfer, ChatFileControl.CANCEL)
            return
        }
        if (data.isEmpty() || directory.getFileSize(transfer.name) == transfer.size) {
            logger.debug("File transfer complete: ${transfer.name}")
            if (transfer.type == ChatFileType.AVATAR) {
                removeTransfer(transfer)
                avatarRepo.updateAvatar(
                    hash = transfer.name,
                    uid = friendUid,
                )
            } else {
                updateTransfer(transfer.copy(state = ChatFileTransferState.COMPLETE))
            }
        }
    }
    
    override fun onFileChunkRequest(friendUid: UserUid, fileUid: FileUid, position: Long, length: Long) {
        if (position == 0L) {
            logger.debug("onFileChunkRequest($friendUid, $fileUid, $position, $length)")
        } else {
            logger.verbose("onFileChunkRequest($friendUid, $fileUid, $position, $length)")
        }
        val transfer = getTransfer(friendUid, fileUid, isError = true) ?: return
        if (length == 0L) {
            if (transfer.type == ChatFileType.AVATAR) {
                avatarRepo.onAvatarSent(transfer)
            } else {
                updateTransfer(
                    transfer.copy(
                        position = transfer.size,
                        state = ChatFileTransferState.COMPLETE,
                    )
                )
            }
        } else {
            val data = fileSystem.getDirectoryForFile(transfer.path)?.readFromFile(
                fileName = transfer.name,
                position = position,
                length = length,
            )
            if (data != null && data.isNotEmpty()) {
                engine.sendFileChunk(friendUid, fileUid, position, data)
            } else {
                engine.sendFileChunk(friendUid, fileUid, 0L, ByteArray(0))
            }
            updateTransfer(transfer.copy(position = position))
        }
    }
    
    override fun onFileControl(friendUid: UserUid, fileUid: FileUid, control: ChatFileControl) {
        logger.debug("onFileControl($friendUid, $fileUid, $control)")
        val transfer = getTransfer(friendUid, fileUid, isError = false) ?: return
        updateTransfer(transfer.copy(state = control.toState()))
        if (control == ChatFileControl.CANCEL) {
            if (transfer.direction == ChatFileTransferDirection.OUTGOING && transfer.type == ChatFileType.AVATAR) {
                avatarRepo.onAvatarSent(transfer)
            } else {
                transfer.removeIncomingFile()
            }
        }
    }

    override fun updateTransfer(transfer: ChatFileTransfer) {
        fileTransfers[transfer.uid] = transfer
    }

    override fun removeTransfer(transfer: ChatFileTransfer) {
        fileTransfers.remove(transfer.uid)
    }

    override fun rejectFileTransfer(
        friendUid: UserUid,
        fileUid: FileUid,
    ) = engine.sendFileControl(
        friendUid = friendUid,
        fileUid = fileUid,
        control = ChatFileControl.CANCEL,
    )
    
    private fun isFileAutoAcceptable(fileName: String): Boolean {
        val extension = fileName.substringAfterLast('.').lowercase()
        return settings.autoAcceptFileTypes.contains(extension)
    }

    private fun getTransfer(
        friendUid: UserUid,
        fileUid: FileUid,
        isError: Boolean = true,
    ): ChatFileTransfer? {
        val uid = FileTransferUid(fileUid, friendUid)
        return fileTransfers[uid].takeIf { it?.friendUid == friendUid } ?: run {
            if (isError) {
                logger.warn("Received command for unknown file $fileUid from friend $friendUid")
                rejectFileTransfer(friendUid, fileUid)
            }
            null
        }
    }
    
    private fun ChatFileTransfer.removeIncomingFile() {
        if (direction == ChatFileTransferDirection.INCOMING) {
            getIncomingDirectory().removeFile(name)
        }
    }
    
    private fun ChatFileTransfer.getIncomingDirectory() = when (type) {
        ChatFileType.GENERIC -> downloadsDirectory
        ChatFileType.AVATAR -> avatarRepo.incomingDirectory
    }
    
    private fun ChatFileControl.toState() = when (this) {
        ChatFileControl.CANCEL -> ChatFileTransferState.CANCELLED
        ChatFileControl.RESUME -> ChatFileTransferState.ACTIVE
        ChatFileControl.PAUSE -> ChatFileTransferState.IDLE
    }
}