package nand.apps.chat.repo

import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.audio.AudioPlayer
import nand.apps.chat.audio.queueSound
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.event.VisibleMessageTracker
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.event.AppUpdateReadyEvent
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.notification.*
import nand.apps.chat.model.uid.GroupPeerUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserData
import nand.apps.chat.platform.AppState
import nand.apps.chat.platform.AppStateProvider
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.res.StringResourceProvider
import nand.apps.chat.ui.util.collect
import nand.apps.chat.util.MutableStateFlowMap
import org.koin.core.Koin
import seers.composeapp.generated.resources.*

class NotificationRepoImpl(koin: Koin) : NotificationRepo {

    private val chatEngine: ChatEngine by koin.inject()
    private val contactRepo: ContactRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()
    private val callRepo: CallRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private val audioPlayer: AudioPlayer by koin.inject()
    private val appStateProvider: AppStateProvider by koin.inject()
    private val stringProvider: StringResourceProvider by koin.inject()
    private val visibleMessageTracker: VisibleMessageTracker by koin.inject()
    private val eventHandler: ChatEventHandler by koin.inject()

    private val notifications = MutableStateFlowMap<ChatNotificationGroup, ChatNotification>()
    private val settings get() = settingsRepo.settings.notificationSettings

    override val observer = notifications

    override fun start(coroutineScope: CoroutineScope) {
        coroutineScope.collect(messageRepo.onNewMessage) { message ->
            val isSelfMessage = message.senderUid == selfProfileRepo.getSelfUid(message.conversationUid)
            if (settings.isMessageNotificationsEnabled && !isSelfMessage) {
                val soundSettings = settings.soundSettings
                when {
                    message.shouldShowNotification() -> {
                        showMessageNotification(message)
                    }
                    soundSettings.isEnabled && soundSettings.isNewMessageTriggerEnabled -> {
                        audioPlayer.queueSound(soundSettings.soundType)
                    }
                }
            }
        }
        coroutineScope.collect(callRepo.onCallReceived) { call ->
            if (settings.isCallNotificationsEnabled &&
                (getPlatform().isMobile || !visibleMessageTracker.isConversationVisible(call.contactUid))) {
                showCallNotification(call)
            }
        }
        coroutineScope.collect(messageRepo.unreadObserver) { unreadStates ->
            val latestMessageNotification = notifications[ChatNotificationGroup.MESSAGES] as? MessageNotification ?: return@collect
            val unreadCount = unreadStates[latestMessageNotification.message.conversationUid] ?: 0
            if (unreadCount == 0) dismiss(ChatNotificationGroup.MESSAGES)
        }
        coroutineScope.collect(eventHandler.onEvent) { event ->
            if (event is AppUpdateReadyEvent && appStateProvider.currentState == AppState.BACKGROUND) {
                showNotification(AppUpdateNotification(event.release))
            }
        }
    }

    override fun dismiss(group: ChatNotificationGroup) {
        notifications.remove(group)
    }
    
    override fun dismissAll() {
        notifications.clear()
    }
    
    private fun showNotification(notification: ChatNotification) {
        notifications[notification.group] = notification
    }
    
    private fun showMessageNotification(message: ChatMessage) {
        val conversation = message.getConversationContact()
        val sender = message.getSenderContact().takeIf { it != contactRepo.self } ?: return
        showNotification(
            MessageNotification(
                title = when (conversation) {
                    is UserData -> stringProvider.getString(Res.string.notification_new_message_from_user, conversation.name)
                    is ChatGroupData -> stringProvider.getString(Res.string.notification_new_message_in_group, conversation.title)
                    is FriendRequest -> stringProvider.getString(Res.string.notification_new_friend_request)
                    else -> return
                },
                sender = sender.getName() ?: return,
                avatar = sender.getAvatar() ?: return,
                message = message,
            )
        )
    }
    
    private fun showCallNotification(call: CallData) {
        val sender = contactRepo.getContact(call.contactUid) ?: return
        val senderName = sender.getName() ?: return
        showNotification(
            CallNotification(
                sender = senderName,
                avatar = sender.getAvatar() ?: return,
                contactUid = call.contactUid,
                callUid = call.uid,
            )
        )
    }
    
    private fun ChatContact.getName() = when (this) {
        is UserData -> displayName
        is ChatGroupData -> title
        is FriendRequest -> stringProvider[Res.string.friend_request]
        else -> null
    }
    
    private fun ChatContact.getAvatar() = when (this) {
        is UserData -> avatar
        is ChatGroupData -> avatar
        is FriendRequest -> avatar
        else -> null
    }

    private fun ChatMessage.containsSelfMention(): Boolean {
        return when (conversationUid) {
            is GroupUid -> {
                val group = getGroupContact() ?: return false
                return containsSelfMention(
                    group = group,
                    mentionsRegex = chatEngine.constraints.mentionRegex,
                )
            }
            else -> true
        }
    }

    private fun ChatMessage.getConversationContact() = contactRepo.getContact(conversationUid)

    private fun ChatMessage.getGroupContact() = when (val conversationUid = conversationUid) {
        is GroupUid -> contactRepo.getGroup(conversationUid)
        is GroupPeerUid -> contactRepo.getGroup(conversationUid.groupUid)
        else -> null
    }

    private fun ChatMessage.getSenderContact() = contactRepo.getContact(
        uid = senderUid,
        userGroup = getGroupContact(),
    )

    private fun ChatMessage.shouldShowNotification() = !isGroupConnectionEvent &&
            !visibleMessageTracker.isConversationVisible(conversationUid) &&
            getConversationContact()?.isMuted != true &&
            containsSelfMention()
}