package nand.apps.chat.repo

import kotlinx.coroutines.withContext
import kotlinx.serialization.json.Json
import nand.apps.chat.BuildConfig.MATRIX_ENGINE
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.model.engine.ChatEngineType
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.settings.*
import nand.apps.chat.model.settings.filetransfer.FileTransferSettingsData
import nand.apps.chat.model.settings.net.NetworkSettingsData
import nand.apps.chat.model.settings.notifications.NotificationSettingsData
import nand.apps.chat.model.settings.theme.ThemeSettingsData
import nand.apps.chat.util.MutableStateFlowDelegate

private const val SETTINGS_FILE = "settings.json"

private const val VERSION_0 = 0
private const val VERSION_1 = 1
const val SETTINGS_VERSION = VERSION_1

class SettingsRepoImpl(
    fileSystem: ChatFileSystem,
    private val backgroundWorker: ChatBackgroundWorker,
) : SettingsRepo {

    private val settingsState = MutableStateFlowDelegate(SettingsData())
    override var settings by settingsState
    override var isLoaded = false
        private set

    override val observer = settingsState

    override val chatEngine get() = settings.run {
        when (val engine = generalSettings.chatEngine) {
            ChatEngineType.MATRIX -> when {
                MATRIX_ENGINE && devSettings.isMatrixChatEnabled -> ChatEngineType.MATRIX
                else -> ChatEngineType.TOX
            }
            else -> engine
        }
    }

    private val json = Json { ignoreUnknownKeys = true }
    private val configDir = fileSystem.getConfigDirectory()
    private val logger = getLogger("SettingsRepo")

    override suspend fun save() {
        logger.debug("Saving to $SETTINGS_FILE")
        val settings = settings.copy(version = SETTINGS_VERSION)
        withContext(backgroundWorker.dispatcher) {
            configDir.writeToFile(
                fileName = SETTINGS_FILE,
                data = json.encodeToString(settings).encodeToByteArray(),
            )
        }
    }
    
    override suspend fun load() {
        if (SETTINGS_FILE in configDir) {
            logger.verbose("Loading from $SETTINGS_FILE")
            val settings: SettingsData? = try {
                withContext(backgroundWorker.dispatcher) {
                    configDir.readFromFile(SETTINGS_FILE)?.let {
                        json.decodeFromString(it.decodeToString())
                    }
                }
            } catch (e: Exception) {
                logger.error("Failed to parse settings file", e)
                null
            }
            settings?.let {
                this.settings = it.upgrade()
                logger.verbose("Successfully loaded from $SETTINGS_FILE")
            }
        }
        isLoaded = true
    }
    
    override fun update(block: SettingsData.() -> Any) {
        settings = settings.run {
            when (val data = block()) {
                is SettingsData -> data
                is GeneralSettingsData -> copy(generalSettings = data)
                is ChatSettingsData -> copy(chatSettings = data)
                is NotificationSettingsData -> copy(notificationSettings = data)
                is EmbedSettingsData -> copy(embedSettings = data)
                is FileTransferSettingsData -> copy(fileSettings = data)
                is NetworkSettingsData -> copy(networkSettings = data)
                is ThemeSettingsData -> copy(themeSettings = data)
                is DevSettingsData -> copy(devSettings = data)
                else -> return
            }
        }
    }

    @Suppress("Deprecation")
    private fun SettingsData.upgrade(): SettingsData {
        if (version >= SETTINGS_VERSION) return this
        return when (version) {
            VERSION_0 -> copy(
                themeSettings = ThemeSettingsData(
                    theme = generalSettings.appTheme,
                    baseHue = generalSettings.appThemeHue,
                ),
                appUpdateSettings = appUpdateSettingsData,
            )
            else -> this
        }.copy(version = version + 1).upgrade()
    }
}