package nand.apps.chat.search

import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.user.UserData
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.chat.message.MessageTextEffect
import nand.apps.chat.util.findMatchRanges

class ChatSearch(
    private val contactRepo: ContactRepo,
    private val settingsRepo: SettingsRepo,
) {

    fun filterMessages(
        messages: List<ChatMessage>,
        terms: String,
        params: ChatSearchParams,
    ): List<ChatMessage> {
        if (terms.isBlank()) return messages

        val regex = try {
            Regex(
                pattern = when {
                    params.matchRegex -> terms
                    params.wholeWords -> "\\b($terms)\\b"
                    else -> Regex.escape(terms)
                },
                options = when {
                    params.matchCase -> emptySet()
                    else -> setOf(RegexOption.IGNORE_CASE)
                }
            )
        } catch (ignore: Exception) {
            return emptyList()
        }
        val request = SearchRequest(
            terms = terms,
            regex = regex,
            params = params,
        )

        val maxTextEffects = settingsRepo.settings.chatSettings.maxTextEffects
        val nameMap = mutableMapOf<ContactUid, String>()
        return messages.mapNotNull { message ->
            val displayName = nameMap[message.senderUid] ?: run {
                message.getDisplayName().also { nameMap[message.senderUid] = it }
            }
            var messageText = message.text
            if (messageText.startsWith("/") && maxTextEffects > 0) {
                messageText = MessageTextEffect.parseEffects(
                    text = messageText,
                    maxEffects = maxTextEffects,
                ).first
            }
            if (message.isAction) {
                messageText = "$displayName $messageText"
            }
            // TODO - Parse mentions and font tags
            val nameMatches = if (params.matchNames) displayName.findSearchMatches(request) else emptyList()
            val textMatches = messageText.findSearchMatches(request)
            if (nameMatches.isNotEmpty() || textMatches.isNotEmpty()) {
                message.copy(
                    nameMatches = nameMatches,
                    textMatches = textMatches,
                )
            } else {
                null
            }
        }
    }

    private fun ChatMessage.getDisplayName(): String {
        val user = contactRepo.getContact(
            uid = senderUid,
            userGroup = (conversationUid as? GroupUid)?.let { contactRepo.getGroup(it) },
        )
        return when (user) {
            is UserData -> user.displayName
            is ChatGroupData -> user.title
            else -> ""
        }
    }

    private fun String.findSearchMatches(request: SearchRequest): List<IntRange> {
        val ignoreCase = !request.params.matchCase
        val matchStart = request.terms.length == 1 && !request.params.wholeWords
        return when {
            request.params.matchRegex || !matchStart && contains(request.terms, ignoreCase = ignoreCase) -> {
                findMatchRanges(request.regex)
            }
            matchStart && startsWith(request.terms, ignoreCase = ignoreCase) -> {
                listOf(0..<1)
            }
            else -> emptyList()
        }
    }
}

private data class SearchRequest(
    val terms: String,
    val regex: Regex,
    val params: ChatSearchParams,
)