package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.material.LinearProgressIndicator
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Cancel
import androidx.compose.material.icons.filled.OpenInBrowser
import androidx.compose.material.icons.filled.Pause
import androidx.compose.material.icons.filled.PlayArrow
import androidx.compose.runtime.Composable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.text.font.FontStyle
import nand.apps.chat.model.file.*
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.embed.audio.AudioEmbedRenderer
import nand.apps.chat.ui.embed.image.ImageEmbedRenderer
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.formatFileSize
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*
import kotlin.math.roundToInt
import kotlin.math.roundToLong

@Composable
fun FileTransferRow(
    fileTransfer: ChatFileTransfer,
    onFileTransferControl: (ChatFileControl) -> Unit,
    onOpenFile: () -> Unit,
    onDeleteFile: () -> Unit,
    modifier: Modifier = Modifier,
) = with(fileTransfer) {
    // TODO - Use embed framework for this
    when (state) {
        ChatFileTransferState.COMPLETE -> {
            val embedSettings = LocalSettingsProvider.current.embedSettings
            val contentType = contentType
            when {
                embedSettings.isEnabled && embedSettings.isImagesEnabled && contentType?.contentType == "image" -> {
                    ImageEmbedRenderer(
                        uri = path,
                        type = contentType.contentSubtype,
                        modifier = modifier,
                    )
                    return
                }
                contentType?.isSupportedAudioFormat() == true -> {
                    AudioEmbedRenderer(
                        uri = path,
                        type = contentType,
                        onDeleteFile = onDeleteFile,
                        modifier = modifier,
                    )
                    return
                }
            }
        }
        ChatFileTransferState.REMOVED -> {
            TextLine(
                text = stringResource(Res.string.past_file_transfer_removed, name),
                style = ChatTheme.typography.body2.copy(fontStyle = FontStyle.Italic),
                color = ChatTheme.colors.onBackground2,
            )
            return
        }
        else -> {}
    }
    FileTransferCard(
        fileTransfer = fileTransfer,
        onFileTransferControl = onFileTransferControl,
        onOpenFile = onOpenFile,
        modifier = modifier,
    )
}

@Composable
private fun FileTransferCard(
    fileTransfer: ChatFileTransfer,
    onFileTransferControl: (ChatFileControl) -> Unit,
    onOpenFile: () -> Unit,
    modifier: Modifier = Modifier,
) = with(fileTransfer) {
    val background = when (state) {
        ChatFileTransferState.CANCELLED -> ChatTheme.colors.error
        else -> ChatTheme.colors.surface2
    }
    Column(
        modifier = modifier
            .fillMaxWidth()
            .clip(ChatTheme.shapes.medium)
            .background(background)
            .padding(start = ChatTheme.dimens.spacingMedium)
            .padding(vertical = ChatTheme.dimens.spacingMedium)
    ) {
        Row(
            verticalAlignment = Alignment.CenterVertically,
            modifier = Modifier.fillMaxWidth()
        ) {
            Column(
                modifier = Modifier.weight(1f),
            ) {
                val transferDuration = (currentTimeMillis() - startTime) / 1000.0
                val transferSpeed = when {
                    position > 0 && transferDuration >= 1 ->"(${formatFileSize((position / transferDuration).roundToLong())}/s)"
                    else -> ""
                }
                TextLine(
                    text = "$name (${stringResource(fileTransfer.getStateString())})",
                    style = ChatTheme.typography.body2,
                )
                TextLine(
                    text = when {
                        isInProgress -> "${formatFileSize(position)} / ${formatFileSize(size)} $transferSpeed"
                        else -> formatFileSize(size)
                    },
                    style = ChatTheme.typography.caption,
                    color = ChatTheme.colors.onBackground,
                )
            }
            if (isInProgress) {
                IconButton(
                    icon = when (state) {
                        ChatFileTransferState.ACTIVE -> Icons.Default.Pause
                        else -> Icons.Default.PlayArrow
                    },
                    contentDescription = when (state) {
                        ChatFileTransferState.ACTIVE -> stringResource(Res.string.pause_transfer)
                        else -> stringResource(Res.string.resume_transfer)
                    },
                    onClick = {
                        when (state) {
                            ChatFileTransferState.ACTIVE -> onFileTransferControl(ChatFileControl.PAUSE)
                            else -> onFileTransferControl(ChatFileControl.RESUME)
                        }
                    }
                )
                IconButton(
                    icon = Icons.Default.Cancel,
                    contentDescription = stringResource(Res.string.stop_transfer),
                    onClick = { onFileTransferControl(ChatFileControl.CANCEL) }
                )
            } else if (state == ChatFileTransferState.COMPLETE) {
                IconButton(
                    icon = Icons.Default.OpenInBrowser,
                    contentDescription = stringResource(Res.string.open_file),
                    onClick = onOpenFile,
                )
            }
        }
        if (isInProgress) {
            Row(
                verticalAlignment = Alignment.CenterVertically,
                horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                modifier = Modifier
                    .fillMaxWidth()
                    .padding(
                        top = ChatTheme.dimens.spacingMedium,
                        end = ChatTheme.dimens.spacingMedium,
                    ),
            ) {
                val percentage = position.toFloat() / size
                LinearProgressIndicator(
                    progress = percentage,
                    modifier = Modifier.weight(1f)
                )
                TextLine(
                    text = "${(percentage * 100).roundToInt()}%",
                    style = ChatTheme.typography.caption,
                )
            }
        }
    }
}

private fun ChatFileTransfer.getStateString() = when (state) {
    ChatFileTransferState.IDLE -> Res.string.file_state_idle
    ChatFileTransferState.ACTIVE -> {
        if (direction == ChatFileTransferDirection.INCOMING) {
            Res.string.file_state_receiving
        } else {
            Res.string.file_state_sending
        }
    }
    ChatFileTransferState.CANCELLED,
    ChatFileTransferState.REMOVED -> Res.string.file_state_cancelled
    ChatFileTransferState.COMPLETE -> Res.string.file_state_completed
}