package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.layout.padding
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalUriHandler
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupType
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.net.openFile
import nand.apps.chat.ui.chat.screen.LocalChatScreenState
import nand.apps.chat.ui.dialog.PasswordDialog
import nand.apps.chat.ui.embed.EmbedContainer
import nand.apps.chat.ui.embed.qrcode.QrCodeEmbedRenderer
import nand.apps.chat.ui.emoji.EmojiRenderer
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.Res
import seers.composeapp.generated.resources.password
import seers.composeapp.generated.resources.set_group_password

@Composable
fun MessageBlockExtra(
    message: ChatMessage,
    stylized: StylizedMessage,
    modifier: Modifier = Modifier,
    isQrCode: Boolean,
) {
    val screenState = LocalChatScreenState.current
    val eventHandler = LocalEventHandler.current
    val engineConstraints = LocalChatEngineConstraints.current
    val uriHandler = LocalUriHandler.current
    val fileSystem: ChatFileSystem by getKoin().inject()
    val extraModifier = modifier.padding(top = ChatTheme.dimens.spacingSmall)
    val fileTransfer = screenState.fileTransfers[message.fileTransferUid]
    var isPasswordDialogVisible by rememberSaveable { mutableStateOf(false) }
    when {
        fileTransfer != null -> FileTransferRow(
            fileTransfer = fileTransfer,
            modifier = extraModifier,
            onFileTransferControl = { eventHandler += FileTransferControlEvent(fileTransfer, it) },
            onOpenFile = {
                if (fileSystem.exists(fileTransfer.path)) {
                    uriHandler.openFile(fileTransfer.path)
                }
            },
            onDeleteFile = { eventHandler += UpdateMessageEvent(message.copy(fileTransferUid = null)) }
        )
        message.friendRequest != null -> message.friendRequest?.let {
            AcceptDeclineButtons(
                modifier = extraModifier,
                onAccept = { eventHandler += AcceptFriendRequestEvent(it) },
                onDecline = { eventHandler += DeclineFriendRequestEvent(it) },
            )
        }
        message.groupInvite != null -> message.groupInvite?.let { invite ->
            AcceptDeclineButtons(
                modifier = extraModifier,
                onAccept = {
                    if (invite.groupType == ChatGroupType.ADVANCED) {
                        isPasswordDialogVisible = true
                    } else {
                        eventHandler += AcceptGroupInviteEvent(invite)
                    }
                },
                onDecline = { eventHandler += DeclineGroupInviteEvent(invite) },
            )
        }
        isQrCode -> QrCodeEmbedRenderer(
            text = stylized.text.text,
            modifier = extraModifier,
        )
        stylized.isEmojiMessage -> EmojiRenderer(
            text = stylized.text.text,
            modifier = extraModifier,
        )
        message.urls.size == 1 && MessageTextEffect.NOEMBED !in stylized.effects -> {
            EmbedContainer(
                url = message.urls.first(),
                message = message,
                modifier = extraModifier,
            )
        }
    }
    if (isPasswordDialogVisible) {
        PasswordDialog(
            title = Res.string.set_group_password,
            label = Res.string.password,
            maxLength = engineConstraints.maxGroupPasswordLength,
            isOptional = true,
            onConfirm = { password ->
                isPasswordDialogVisible = false
                message.groupInvite?.let { invite ->
                    eventHandler += AcceptGroupInviteEvent(
                        invite = invite.copy(password = password)
                    )
                }
            },
            onCancel = { isPasswordDialogVisible = false },
        )
    }
}