package nand.apps.chat.ui.chat.message

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.unit.dp
import nand.apps.chat.model.avatar.ChatGroupAvatar
import nand.apps.chat.model.avatar.ChatUserAvatar
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.user.UserData
import nand.apps.chat.ui.avatar.ChatAvatarBox
import nand.apps.chat.ui.avatar.ChatAvatarSize
import nand.apps.chat.ui.chat.message.input.LocalChatMessageInputState
import nand.apps.chat.ui.chat.message.reaction.MessageReactionsRow
import nand.apps.chat.ui.chat.screen.LocalChatScreenState
import nand.apps.chat.ui.provider.LocalDateTimeFormatProvider
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.res.getStringBlocking
import seers.composeapp.generated.resources.*

@Composable
fun MessageRow(
    message: ChatMessage,
    user: UserData?,
    group: ChatGroupData?,
    isHeaderVisible: Boolean,
    isDateSeparatorVisible: Boolean,
    modifier: Modifier = Modifier,
) {
    val screenState = LocalChatScreenState.current
    val settings = LocalSettingsProvider.current.chatSettings
    val displayName = remember(message, user, group) {
        when {
            message.isGroupMessage -> group?.title ?: getStringBlocking(Res.string.group_message)
            message.friendRequest != null -> getStringBlocking(Res.string.friend_request)
            user != null -> user.displayName
            else -> getStringBlocking(Res.string.unknown_sender)
        }
    }
    Column(modifier = modifier) {
        if (isDateSeparatorVisible) {
            MessageDateSeparator(
                timestamp = message.timestamp,
                modifier = Modifier.fillMaxWidth(),
            )
        }
        MessageAligner(
            alignment = settings.messageAlignment,
            isSelf = message.senderUid == screenState.selfUid,
        ) {
            when {
                group != null && message.isGroupConnectionEvent -> {
                    GroupEventMessageRow(
                        message = message,
                        displayName = displayName,
                        modifier = Modifier.fillMaxWidth(),
                    )
                }
                else -> MessageRow(
                    message = message,
                    user = user,
                    group = group,
                    displayName = displayName,
                    isHeaderVisible = isHeaderVisible,
                )
            }
        }
    }
}

@Composable
private fun MessageRow(
    message: ChatMessage,
    user: UserData?,
    group: ChatGroupData?,
    displayName: String,
    isHeaderVisible: Boolean,
) = Row {
    val screenState = LocalChatScreenState.current
    val inputState = LocalChatMessageInputState.current
    val settings = LocalSettingsProvider.current.chatSettings
    val dateTimeFormatter = LocalDateTimeFormatProvider.current
    val avatarSize = if (isPortraitMode()) ChatAvatarSize.MEDIUM else ChatAvatarSize.SMALL
    
    if (isHeaderVisible && settings.isAvatarsEnabled) {
        val friendRequest = message.friendRequest
        val avatar = remember(user, group) {
            when {
                message.isGroupMessage -> group?.avatar ?: ChatGroupAvatar("")
                friendRequest != null -> friendRequest.avatar
                else -> user?.avatar ?: ChatUserAvatar(displayName)
            }
        }
        ChatAvatarBox(
            avatar = avatar,
            size = avatarSize,
        )
    }
    val rowBackground = when {
        inputState.editMessage?.uid == message.uid -> ChatTheme.colors.primaryVariant
        else -> Color.Unspecified
    }
    MessageActionMenu(
        message = message,
        user = user,
        group = group,
        displayName = displayName,
        modifier = Modifier
            .align(Alignment.Top)
            .padding(start = ChatTheme.dimens.spacingMedium + if (isHeaderVisible || !settings.isAvatarsEnabled) 0.dp else avatarSize.getSizeDp())
            .background(rowBackground)
    ) {
        Column {
            if (isHeaderVisible) {
                Box(modifier = Modifier.fillMaxWidth()) {
                    val timestampFormatted = remember(message) {
                        buildString {
                            if (message.isDelivered) append("✓ ")
                            append(dateTimeFormatter.formatTime(message.timestamp))
                        }
                    }
                    val textStyle = if (isPortraitMode()) ChatTheme.typography.body2 else ChatTheme.typography.caption
                    MessageAligner {
                        TextLine(
                            text = when {
                                message.nameMatches.isNotEmpty() -> {
                                    displayName.addMatchStyles(
                                        matches = message.nameMatches,
                                        style = SpanStyle(
                                            color = ChatTheme.colors.onMention,
                                            background = ChatTheme.colors.mention,
                                        ),
                                    )
                                }
                                else -> AnnotatedString(displayName)
                            },
                            style = textStyle.copy(fontWeight = FontWeight.Bold),
                            color = when {
                                message.isGroupMessage -> ChatTheme.colors.statusOnline
                                else -> ChatTheme.colors.onSurface
                            },
                        )
                        TextLine(
                            text = timestampFormatted,
                            style = ChatTheme.typography.caption,
                            color = ChatTheme.colors.onBackground2,
                            modifier = Modifier.align(Alignment.CenterEnd),
                        )
                    }
                }
            }
            MessageBlock(
                message = message,
                displayName = displayName,
            )
            if (settings.isMessageReactionsEnabled) {
                MessageReactionsRow(
                    message = message,
                    selfUid = screenState.selfUid,
                    group = screenState.conversationGroup,
                    modifier = Modifier.padding(top = ChatTheme.dimens.spacingSmall),
                )
            }
        }
    }
}