package nand.apps.chat.ui.chat.message.input

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.animation.core.*
import androidx.compose.foundation.ExperimentalFoundationApi
import androidx.compose.foundation.background
import androidx.compose.foundation.draganddrop.dragAndDropTarget
import androidx.compose.foundation.layout.*
import androidx.compose.material.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.Send
import androidx.compose.material.icons.filled.Warning
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.layout.*
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.input.VisualTransformation
import androidx.compose.ui.unit.*
import kotlinx.coroutines.delay
import nand.apps.chat.model.event.*
import nand.apps.chat.platform.AppState
import nand.apps.chat.platform.AppStateProvider
import nand.apps.chat.res.stringResource
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.button.StyledIconButton
import nand.apps.chat.ui.dialog.FileTransferConfirmationDialog
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.emoji.EmojiPopup
import nand.apps.chat.ui.file.sender.ChatFileSenderDialog
import nand.apps.chat.ui.modifier.*
import nand.apps.chat.ui.provider.*
import nand.apps.chat.ui.settings.FileTransferUploadConfigRoute
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.shortcut.toChatKeyEvent
import nand.apps.chat.ui.text.*
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.LaunchedCollect
import nand.apps.chat.ui.util.getClipboardFilePath
import nand.apps.chat.ui.util.getFilePath
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import kotlin.math.roundToInt
import seers.composeapp.generated.resources.*

@OptIn(ExperimentalFoundationApi::class)
@Composable
fun MessageInputField(
    textStyle: TextStyle,
    modifier: Modifier = Modifier,
    viewModel: MessageInputViewModel,
) {
    val state = LocalChatMessageInputState.current
    val settings = LocalSettingsProvider.current
    val navController = LocalNavController.current
    val focusManager = LocalFocusManager.current
    val density = LocalDensity.current
    val eventHandler = LocalEventHandler.current
    val appStateProvider: AppStateProvider by getKoin().inject()
    val isPortrait = isPortraitMode()
    val chatSettings = settings.chatSettings
    val fileSettings = settings.fileSettings

    var composablePosition by remember { mutableStateOf(IntOffset.Zero) }
    var textInputPosition by remember { mutableStateOf(IntOffset.Zero) }
    var textInputSize by remember { mutableStateOf(IntSize.Zero) }

    var mentionPopupOffset by remember { mutableStateOf(IntOffset.Zero) }
    var mentionPopupSize by remember { mutableStateOf(DpSize.Zero) }

    var emojiPopupSize by remember { mutableStateOf(DpSize.Zero) }
    
    val focusRequester = remember { FocusRequester() }

    var message by rememberTextFieldValue()
    var visualTransformation by remember { mutableStateOf(VisualTransformation.None) }

    // View model actions
    LaunchedCollect(viewModel.actionData) { action ->
        when (action) {
            is UpdateTextAction -> {
                message = action.textField
                visualTransformation = StylizedTextVisualTransformation(action.stylizedText)
            }
            is ClearFocusAction -> focusManager.clearFocus()
        }
    }

    // Update text style and color theme for styling
    val colors = ChatTheme.colors
    LaunchedEffect(textStyle, colors) {
        viewModel.onTextStyleChanged(textStyle, colors)
    }

    // Focus input field after re-focusing the app on desktop
    LaunchedEffect(Unit) {
        if (state.isAutoFocus) focusRequester.requestFocus()
        eventHandler.onEvent.collect { event ->
            if (state.isAutoFocus) {
                when {
                    event is AppFocusChangedEvent && event.isFocused -> {
                        delay(50) // Need to wait a few frames or the focus doesn't work
                        if (appStateProvider.currentState == AppState.FOREGROUND) {
                            focusRequester.requestFocus()
                        }
                    }
                    event is OpenConversationEvent -> {
                        focusRequester.requestFocus()
                    }
                }
            }
        }
    }
    
    // Shortcuts
    ShortcutHandler(onShortcut = viewModel::onShortcutAction)
    
    Column(
        horizontalAlignment = Alignment.CenterHorizontally,
        modifier = Modifier.onGloballyPositioned {
            composablePosition = it.boundsInRoot().topLeft.round()
        }
    ) {
        AnimatedVisibility(
            visible = state.isEmojiPopupVisible,
        ) {
            EmojiPopup(
                size = emojiPopupSize,
                inputFocusRequester = focusRequester,
                onEmoji = viewModel::onEmojiSelected,
                onDismiss = { viewModel.toggleEmojiPopup(false) },
                modifier = Modifier.padding(bottom = ChatTheme.dimens.spacingSmall),
            )
        }
        Row(
            verticalAlignment = Alignment.Bottom,
            modifier = modifier.dragAndDropTarget(
                shouldStartDragAndDrop = { event ->
                    fileSettings.isDragDropEnabled && state.isButtonsEnabled && event.getFilePath() != null
                },
                target = viewModel,
            ).onGloballyPositioned {
                val root = it.findRootCoordinates()
                val maxWidthDp = (it.size.width / density.density).dp
                val maxHeightDp = (root.size.height / density.density).dp
                emojiPopupSize = DpSize(
                    width = maxWidthDp.coerceAtMost(512.dp),
                    height = (maxHeightDp / 2).coerceAtMost(256.dp),
                )
                mentionPopupSize = DpSize(
                    width = maxWidthDp / if (isPortrait) 1 else 3,
                    height = maxHeightDp / 2
                )
            },
        ) {
            val infiniteTransition = rememberInfiniteTransition(label = "File Upload Spinner")
            val iconRotation = if (state.isBusy) {
                infiniteTransition.animateFloat(
                    initialValue = 0f,
                    targetValue = 360f,
                    animationSpec = infiniteRepeatable(tween(1000, easing = LinearEasing), RepeatMode.Restart),
                    label = "Rotate"
                )
            } else {
                null
            }
            StyledIconButton(
                icon = if (state.isBusy) Icons.Outlined.Sync else Icons.Outlined.AddPhotoAlternate,
                contentDescription = Res.string.chat_screen_send_file,
                isEnabled = state.isButtonsEnabled && state.isConversationOnline,
                onClick = {
                    if (state.isFileTransferEnabled) {
                        viewModel.toggleSendSelector(true)
                    } else {
                        viewModel.toggleGroupUploadDialog(true)
                    }
                },
                rotation = iconRotation?.value ?: 0f,
                modifier = Modifier
                    .padding(end = ChatTheme.dimens.spacingSmall)
                    .size(ChatTheme.dimens.buttonXLarge),
            )
            SimpleTextField(
                value = message,
                onValueChange = viewModel::onTextFieldChanged,
                hintText = state.hintText?.let { stringResource(it) },
                isEnabled = state.isButtonsEnabled,
                isError = state.isUploadError,
                isSingleLine = false,
                textStyle = textStyle,
                visualTransformation = when {
                    chatSettings.isInputTextStylesEnabled -> visualTransformation
                    else -> VisualTransformation.None
                },
                leadingIcon = if (state.isEditable && !state.isConversationOnline) {
                    {
                        val contactName = state.conversationTitle ?: stringResource(Res.string.contact_is_offline)
                        Icon(
                            imageVector = Icons.Default.Warning,
                            contentDescription = stringResource(Res.string.contact_is_offline, contactName),
                            tint = ChatTheme.colors.warning,
                            modifier = Modifier.size(ChatTheme.dimens.iconSmall),
                        )
                    }
                } else null,
                trailingIcon = {
                    Row(
                        modifier = Modifier.padding(end = ChatTheme.dimens.spacingSmall),
                    ) {
                        IconButton(
                            icon = Icons.Outlined.AddReaction,
                            contentDescription = stringResource(Res.string.chat_message_add_emojis),
                            isEnabled = state.isButtonsEnabled || state.reactMessage != null,
                            color = when {
                                state.isEmojiPopupVisible || state.reactMessage != null -> ChatTheme.colors.primary
                                else -> ChatTheme.colors.onBackground2
                            },
                            onClick = {
                                state.reactMessage?.let {
                                    viewModel.addReaction(null)
                                } ?: run {
                                    viewModel.toggleEmojiPopup()
                                    focusRequester.requestFocus()
                                }
                            },
                        )
                        state.editMessage?.let {
                            IconButton(
                                icon = Icons.Outlined.Cancel,
                                contentDescription = stringResource(Res.string.cancel),
                                color = ChatTheme.colors.onBackground2,
                                onClick = viewModel::cancelMessageEdit,
                            )
                        }
                    }
                },
                onTextLayout = { layout ->
                    val visualText = (visualTransformation as? StylizedTextVisualTransformation)?.text ?: "".toTransformedText()
                    val index = when {
                        state.mentionIndex in 0..visualText.text.length -> state.mentionIndex
                        else -> visualText.offsetMapping.originalToTransformed(message.selection.end)
                    }
                    mentionPopupOffset = IntOffset(
                        x = layout.getHorizontalPosition(
                            offset = index,
                            usePrimaryDirection = true,
                        ).toInt(),
                        y = layout.getLineTop(layout.getLineForOffset(index)).roundToInt(),
                    )
                },
                modifier = Modifier
                    .background(if (state.dragFilePath != null) ChatTheme.colors.surface3 else Color.Unspecified)
                    .weight(1f)
                    .focusRequester(focusRequester)
                    .onGloballyPositioned { coords ->
                        textInputPosition = coords.boundsInRoot().topLeft.round()
                        textInputSize = coords.size
                    }
                    .onKeyDown { viewModel.onKeyDown(it.toChatKeyEvent()) }
                    .onPasteKey { viewModel.onPaste(getClipboardFilePath()) },
            )
            AnimatedVisibility(
                visible = message.text.isNotEmpty(),
            ) {
                StyledIconButton(
                    icon = when {
                        state.editMessage != null -> Icons.Outlined.DoneOutline
                        else -> Icons.AutoMirrored.Default.Send
                    },
                    contentDescription = Res.string.chat_screen_send_message,
                    isEnabled = state.isButtonsEnabled,
                    onClick = { viewModel.submitMessage() },
                    modifier = Modifier
                        .padding(start = ChatTheme.dimens.spacingSmall)
                        .size(ChatTheme.dimens.buttonXLarge),
                )
            }
        }
    }
    when {
        state.isGroupUploadDialogVisible -> SimpleDialog(
            title = Res.string.group_file_upload_config_dialog_title,
            confirmLabel = Res.string.open_settings,
            onConfirm = {
                viewModel.toggleGroupUploadDialog(false)
                navController.navigate(FileTransferUploadConfigRoute)
            },
            onCancel = { viewModel.toggleGroupUploadDialog(false) },
        ) {
            Text(
                text = stringResource(Res.string.group_file_upload_config_dialog_message),
                style = ChatTheme.typography.body2,
            )
        }
        state.fileConfirmPath != null -> FileTransferConfirmationDialog(
            path = state.fileConfirmPath.toPath(),
            onConfirm = {
                viewModel.onSendFile(state.fileConfirmPath)
                viewModel.setFileConfirmationPath(null)
            },
            onCancel = { viewModel.setFileConfirmationPath(null) },
        )
        state.isSendSelectorVisible -> ChatFileSenderDialog(
            onSelectFile = { path ->
                viewModel.toggleSendSelector(false)
                viewModel.onSendFile(path)
            },
            onDismiss = { viewModel.toggleSendSelector(false) },
        )
        state.isMentionPopupVisible -> state.conversationGroup?.let { group ->
            MentionPopup(
                startIndex = state.mentionIndex,
                message = message,
                group = group,
                size = mentionPopupSize,
                offset = when {
                    isPortrait -> IntOffset(0, -textInputSize.height)
                    else -> (textInputPosition - composablePosition) + mentionPopupOffset + IntOffset(0, -textInputSize.height)
                },
                onUpdateMessage = { newMessage ->
                    viewModel.dismissMentionPopup()
                    viewModel.onTextFieldChanged(newMessage)
                },
                onDismiss = { viewModel.dismissMentionPopup() },
                keyEvent = state.mentionKeyEvent,
            )
        }
    }
}