package nand.apps.chat.ui.chat.message.input

import androidx.compose.ui.draganddrop.DragAndDropEvent
import androidx.compose.ui.draganddrop.DragAndDropTarget
import androidx.compose.ui.input.key.Key
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.TextRange
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.input.TextFieldValue
import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.io.isFilePath
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupRole
import nand.apps.chat.model.group.ChatGroupTalkState
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.settings.ChatSettingsData
import nand.apps.chat.model.settings.SettingsData
import nand.apps.chat.model.settings.filetransfer.FileTransferSettingsData
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.UserData
import nand.apps.chat.net.FileUploader
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.repo.FileTransferRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.res.FormattedStringResource
import nand.apps.chat.ui.chat.message.toStylizedInputText
import nand.apps.chat.ui.contacts.ContactsState
import nand.apps.chat.ui.shortcut.ChatKeyEvent
import nand.apps.chat.ui.text.*
import nand.apps.chat.ui.theme.ChatThemeColors
import nand.apps.chat.ui.util.collect
import nand.apps.chat.ui.util.getFilePath
import nand.apps.chat.util.fifoSharedFlow
import nand.apps.chat.util.takeIfNotBlank
import okio.Path
import seers.composeapp.generated.resources.*

class MessageInputViewModel(
    contactRepo: ContactRepo,
    private val messageRepo: ChatMessageRepo,
    private val fileTransferRepo: FileTransferRepo,
    private val settingsRepo: SettingsRepo,
    private val eventHandler: ChatEventHandler,
    private val fileUploader: FileUploader,
    private val fileSystem: ChatFileSystem,
    private val constraints: ChatEngineConstraints,
) : ViewModel(), DragAndDropTarget {

    val stateData = MutableStateFlow(MessageInputState())
    val actionData = fifoSharedFlow<MessageInputAction>()

    private val state get() = stateData.value
    private val conversationUid get() = state.conversationUid
    private val conversationContact get() = state.conversationContact
    private val editMessage get() = state.editMessage
    private val reactMessage get() = state.reactMessage

    private var contactsState = ContactsState()
    private var textField = TextFieldValue()
    private var stylizedText = "".toTransformedText()
    private var textStyleParams: StylizedTextParameters? = null
    private val messageHistory = mutableListOf<String>()
    private var messageHistoryIndex = 0

    private var chatSettings = ChatSettingsData()
    private var fileSettings = FileTransferSettingsData()

    init {
        collect(contactRepo.friendObserver) { friends ->
            onFriendsChanged(friends)
        }
        collect(contactRepo.groupObserver) { groups ->
            onGroupsChanged(groups)
        }
        collect(settingsRepo.observer) { settings ->
            onSettingsChanged(settings)
        }
        collect(eventHandler.onEvent, ::onEvent)
    }

    private fun onEvent(event: ChatEvent) = when (event) {
        is OpenConversationEvent -> {
            onOpenConversationEvent(event.contactUid)
        }
        is ExitConversationEvent -> {
            onExitConversationEvent()
        }
        is PromptEditMessageEvent -> {
            editMessage(event.message)
        }
        is PromptReactMessageEvent -> {
            reactMessage(event.message)
        }
        else -> Unit
    }

    private fun updateState(block: (MessageInputState) -> MessageInputState) {
        stateData.tryEmit(block(state))
    }

    fun onTextFieldChanged(field: TextFieldValue) {
        val existingText = textField.text
        textField = field
            .takeIfUnderLength(constraints.maxMessageLength, textField)
            .checkMentions(
                original = textField,
                originalTransformed = stylizedText,
            )
        if (existingText != textField.text) {
            stylizedText = getStylizedText(textField)
        }
        actionData.tryEmit(UpdateTextAction(textField, stylizedText))

        with(textField) {
            if (text.isNotEmpty() && selection.length == 0 && selection.min > 0 && text[selection.min - 1] == '@') {
                updateState { it.copy(mentionIndex = selection.min) }
            } else {
                checkDismissMentionPopup()
            }
        }
    }

    fun onTextStyleChanged(textStyle: TextStyle?, colors: ChatThemeColors) {
        textStyleParams = StylizedTextParameters(
            settings = chatSettings,
            colors = colors,
            tagPreviewStyle = if (chatSettings.isInputTagPreviewEnabled) textStyle else null,
            mentionsRegex = constraints.mentionRegex,
            contactsState = contactsState,
        )
        stylizedText = getStylizedText(textField)
        actionData.tryEmit(UpdateTextAction(textField, stylizedText))
    }

    fun submitMessage() {
        val conversationUid = conversationUid ?: return
        val text = textField.text
        if (state.isBusy || reactMessage != null || text.isEmpty()) {
            return
        }
        editMessage?.let { message ->
            val edited = messageRepo.createMessage(text, conversationUid)
            eventHandler += UpdateMessageEvent(
                message = message.copy(
                    text = edited.text,
                    type = edited.type,
                    isEdited = true,
                ),
                localOnly = false,
            )
            editMessage(null)
        } ?: run {
            eventHandler += SendMessageEvent(
                conversationUid = conversationUid,
                text = text,
            )
            if (text.isNotBlank()) {
                messageHistory += text
                messageHistoryIndex = messageHistory.size
            }
            updateInputText("")
        }
    }

    fun cancelMessageEdit() {
        editMessage?.let { editMessage(null) }
        reactMessage?.let { reactMessage(null) }
    }

    fun onSendFile(path: String): Boolean {
        if (conversationUid == null || !path.isFilePath() || !fileSystem.exists(path)) return false
        sendFile(path)
        return true
    }

    fun onEmojiSelected(emoji: String) {
        reactMessage?.let { addReaction(emoji) } ?: onTextFieldChanged(textField.insert(emoji))
    }

    fun addReaction(reaction: String?) {
        val message = reactMessage ?: return
        reaction?.let {
            eventHandler += ReactMessageEvent(
                message = message,
                reaction = reaction,
            )
        }
        reactMessage(null)
    }

    fun toggleEmojiPopup(
        isVisible: Boolean = !state.isEmojiPopupVisible,
    ) = updateState { state ->
        state.copy(
            reactMessage = if (isVisible) state.reactMessage else null,
            isEmojiPopupVisible = isVisible,
        )
    }

    fun toggleGroupUploadDialog(isVisible: Boolean) = updateState {
        it.copy(isGroupUploadDialogVisible = isVisible)
    }

    fun toggleSendSelector(isVisible: Boolean) = updateState {
        it.copy(isSendSelectorVisible = isVisible)
    }

    fun dismissMentionPopup() = updateState {
        it.copy(
            mentionIndex = -1,
            mentionKeyEvent = null,
        )
    }

    fun setFileConfirmationPath(path: String?) = updateState {
        it.copy(fileConfirmPath = path)
    }

    override fun onEntered(event: DragAndDropEvent) {
        setDragPath(event.getFilePath())
    }

    override fun onExited(event: DragAndDropEvent) {
        setDragPath(null)
    }

    override fun onEnded(event: DragAndDropEvent) {
        setDragPath(null)
    }

    override fun onDrop(event: DragAndDropEvent): Boolean {
        state.dragFilePath?.let { filePath ->
            when {
                !state.isFileTransferEnabled -> toggleGroupUploadDialog(true)
                fileSettings.showFileTransferConfirmation -> setFileConfirmationPath(filePath)
                else -> onSendFile(filePath)
            }
        }
        setDragPath(null)
        return true
    }

    fun onKeyDown(event: ChatKeyEvent): Boolean {
        return when (event.key) {
            Key.Enter, Key.NumPadEnter -> {
                when {
                    state.isMentionPopupVisible -> setMentionKeyEvent(event)
                    event.isShiftPressed -> onTextFieldChanged(textField.insertNewline())
                    else -> submitMessage()
                }
                true
            }
            Key.DirectionUp -> {
                when {
                    state.isMentionPopupVisible -> setMentionKeyEvent(event)
                    else -> return cycleMessage(MessageCycleDirection.UP)
                }
                true
            }
            Key.DirectionDown -> {
                when {
                    state.isMentionPopupVisible -> setMentionKeyEvent(event)
                    else -> return cycleMessage(MessageCycleDirection.DOWN)
                }
                true
            }
            else -> false
        }
    }

    fun onShortcutAction(action: ShortcutAction) = when (action) {
        ShortcutAction.TOGGLE_EMOJI_POPUP -> {
            toggleEmojiPopup()
            true
        }
        ShortcutAction.SEND_FILE -> {
            if (state.isButtonsEnabled) {
                if (state.isFileTransferEnabled) {
                    toggleSendSelector(true)
                } else {
                    toggleGroupUploadDialog(true)
                }
                true
            } else {
                false
            }
        }
        else -> false
    }

    fun onPaste(path: Path?) = when {
        path == null || !fileSettings.isCopyPasteEnabled -> false
        !state.isFileTransferEnabled -> {
            toggleGroupUploadDialog(true)
            true
        }
        fileSettings.showFileTransferConfirmation -> {
            setFileConfirmationPath(path.toString())
            true
        }
        else -> onSendFile(path.toString())
    }

    private fun cycleMessage(direction: MessageCycleDirection): Boolean {
        if (messageHistory.isEmpty() || messageHistoryIndex >= messageHistory.size && textField.text.isNotBlank()) {
            return false
        }

        val selection = textField.selection
        if (selection.length > 0) {
            return false
        }

        val isEndSelected = selection.max == textField.text.length
        val canCycle = when (direction) {
            MessageCycleDirection.UP -> selection.min == 0 || isEndSelected && messageHistoryIndex < messageHistory.size
            MessageCycleDirection.DOWN -> isEndSelected
        }

        val newIndex = (messageHistoryIndex + direction.value).coerceAtLeast(0).coerceAtMost(messageHistory.size)
        if (!canCycle || newIndex == messageHistoryIndex) {
            return false
        }

        messageHistoryIndex = newIndex
        val inputText = if (messageHistoryIndex in messageHistory.indices) messageHistory[messageHistoryIndex] else ""
        updateInputText(inputText)
        return true
    }

    private fun checkDismissMentionPopup() {
        if (state.isDirectMessage || textField.selection.min < state.mentionIndex || !chatSettings.isMentionsEnabled) {
            dismissMentionPopup()
        }
    }

    private fun setMentionKeyEvent(event: ChatKeyEvent) = updateState {
        it.copy(mentionKeyEvent = event)
    }

    private fun setDragPath(path: String?) = updateState {
        it.copy(dragFilePath = path)
    }

    private fun onOpenConversationEvent(conversationUid: ContactUid) {
        if (this.conversationUid != conversationUid) {
            cancelMessageEdit()
            setConversation(conversationUid)
        }
    }

    private fun onExitConversationEvent() {
        cancelMessageEdit()
        setConversation(null)
    }

    private fun onFriendsChanged(friends: Map<UserUid, UserData>) {
        contactsState = contactsState.copy(friends = friends)
        setConversation(conversationUid)
        updateTextStyle()
    }

    private fun onGroupsChanged(groups: Map<GroupUid, ChatGroupData>) {
        contactsState = contactsState.copy(groups = groups)
        setConversation(conversationUid)
        updateTextStyle()
    }

    private fun onSettingsChanged(settings: SettingsData) {
        chatSettings = settings.chatSettings
        fileSettings = settings.fileSettings
        updateState { it.copy(isFileTransferEnabled = isFileTransferEnabled(conversationUid)) }
        updateTextStyle()
        checkDismissMentionPopup()
    }

    private fun updateTextStyle() {
        textStyleParams?.let {
            onTextStyleChanged(
                textStyle = it.tagPreviewStyle,
                colors = it.colors,
            )
        }
    }

    private fun getStylizedText(field: TextFieldValue) = textStyleParams?.let {
        AnnotatedString(field.text).toStylizedInputText(it)
    } ?: field.text.toTransformedText()

    private fun setConversation(conversationUid: ContactUid?) {
        val conversationContact = when (conversationUid) {
            is UserUid -> contactsState.friends[conversationUid]
            is GroupUid -> contactsState.groups[conversationUid]
            is GroupPeerUid -> contactsState.groups[conversationUid.groupUid]?.peers?.get(conversationUid.peerUid)
            else -> null
        }
        contactsState = contactsState.copy(conversationUid = conversationUid)
        if (this.conversationUid != conversationUid || this.conversationContact != conversationContact) {
            val user = conversationContact as? UserData
            val group = conversationContact as? ChatGroupData
            val talkStateWarning = group?.run {
                when {
                    talkState == ChatGroupTalkState.FOUNDER_ONLY && selfRole != ChatGroupRole.FOUNDER -> {
                        ChatGroupTalkState.FOUNDER_ONLY
                    }
                    talkState == ChatGroupTalkState.STAFF_ONLY && !selfRole.hasAuthorityOver(ChatGroupRole.USER) -> {
                        ChatGroupTalkState.STAFF_ONLY
                    }
                    else -> null
                }
            }
            updateState {
                it.copy(
                    isFileTransferEnabled = isFileTransferEnabled(conversationUid),
                    conversationUid = conversationUid,
                    conversationContact = conversationContact,
                    conversationTitle = when (conversationUid) {
                        is UserUid,
                        is GroupPeerUid -> user?.displayName
                        is GroupUid -> group?.title?.takeIfNotBlank()
                        else -> null
                    },
                    talkStateWarning = talkStateWarning,
                )
            }
            updateTextStyle()
            checkDismissMentionPopup()
        }
    }

    private fun updateInputText(text: String) = onTextFieldChanged(
        TextFieldValue(
            text = text,
            selection = TextRange(text.length),
        )
    )

    private fun editMessage(message: ChatMessage?) {
        updateState { state ->
            state.copy(
                reactMessage = null,
                editMessage = message?.takeIf { it.text.isNotEmpty() },
                isEmojiPopupVisible = false,
            )
        }
        updateInputText(
            text = message?.text?.let { text ->
                when (message.type) {
                    ChatMessageType.ACTION -> "/me $text"
                    else -> text
                }
            }.orEmpty()
        )
    }

    private fun reactMessage(message: ChatMessage?) = updateState {
        it.copy(
            editMessage = null,
            reactMessage = message,
            isEmojiPopupVisible = message != null,
        )
    }

    private fun sendFile(filePath: String) {
        val conversationUid = this.conversationUid ?: return
        when {
            conversationUid is GroupUid && fileSettings.isGroupUploadEnabled -> {
                updateState {
                    it.copy(
                        isUploadError = false,
                        isBusy = true,
                    )
                }
                viewModelScope.launch {
                    val url = fileUploader.upload(filePath).orEmpty()
                    updateState {
                        it.copy(
                            isUploadError = url.isBlank(),
                            isBusy = false,
                        )
                    }
                    if (url.isNotBlank()) {
                        updateInputText("${textField.text}\n".trimStart('\n') + url)
                    }
                }
            }
            conversationUid is UserUid -> {
                if (!fileTransferRepo.sendFile(filePath, conversationUid)) {
                    eventHandler += ErrorSnackbarEvent(FormattedStringResource(Res.string.chat_screen_failed_to_send_file))
                }
            }
        }
    }

    private fun isFileTransferEnabled(conversationUid: ContactUid?) = conversationUid is UserUid ||
            fileSettings.isGroupUploadEnabled && fileSettings.uploadConfig.url.isNotEmpty()
}