package nand.apps.chat.ui.chat.screen

import androidx.compose.foundation.background
import androidx.compose.foundation.basicMarquee
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.shape.CircleShape
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.outlined.ArrowForwardIos
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import nand.apps.chat.model.group.ChatGroupFeature
import nand.apps.chat.model.group.ChatGroupRole
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.uid.FriendRequestUid
import nand.apps.chat.model.uid.GroupPeerUid
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.ui.call.CallButton
import nand.apps.chat.ui.avatar.ChatAvatarBox
import nand.apps.chat.ui.avatar.ChatAvatarSize
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.chat.message.toStylizedMessage
import nand.apps.chat.ui.chat.search.ChatSearchState
import nand.apps.chat.ui.groups.settings.GroupSettingsDialog
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.text.StylizedTextParameters
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.res.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun ChatHeader(
    onBackPressed: () -> Unit,
    onTogglePeers: () -> Unit,
    onToggleSearch: () -> Unit,
    modifier: Modifier = Modifier,
) {
    val state = LocalChatScreenState.current
    val isPortrait = isPortraitMode()
    val padding = if (isPortrait) ChatTheme.dimens.spacingLarge else ChatTheme.dimens.spacingMedium
    val isBackButtonVisible = isPortrait || state.conversationUid is GroupPeerUid
    val settings = LocalSettingsProvider.current
    val isGroupPeersCollapsed = settings.generalSettings.isGroupPeersCollapsed
    val isCallsEnabled = settings.callSettings.isEnabled
    ShortcutHandler(
        action = ShortcutAction.TOGGLE_PEERS,
        isEnabled = !isPortrait,
        onShortcut = onTogglePeers,
    )
    Row(
        verticalAlignment = Alignment.CenterVertically,
        modifier = modifier.padding(
            start = if (isBackButtonVisible) 0.dp else padding,
            top = padding,
            bottom = padding,
            end = padding,
        ),
    ) {
        if (isBackButtonVisible) {
            IconButton(
                icon = Icons.Outlined.ArrowBackIosNew,
                contentDescription = stringResource(Res.string.back),
                size = ChatTheme.dimens.iconMedium,
                onClick = onBackPressed,
                modifier = Modifier.padding(horizontal = ChatTheme.dimens.spacingSmall),
            )
        }
        val avatarSize = if (isPortrait) ChatAvatarSize.LARGE else ChatAvatarSize.MEDIUM
        when (state.conversationUid) {
            is UserUid,
            is GroupPeerUid -> state.conversationUser?.let {
                ChatAvatarBox(
                    avatar = it.avatar,
                    status = it.status,
                    isOnline = it.isOnline,
                    size = avatarSize,
                )
            }
            is GroupUid -> state.conversationGroup?.let {
                ChatAvatarBox(
                    avatar = it.avatar,
                    status = UserStatus.DEFAULT,
                    isOnline = it.isOnline,
                    size = avatarSize,
                )
            }
            else -> Unit
        }
        val colors = ChatTheme.colors
        val conversationDesc = remember(state, colors, settings.chatSettings) {
            when (val uid = state.conversationUid) {
                is UserUid -> state.conversationUser?.statusMessage
                is GroupUid -> state.conversationGroup?.run {
                    description.ifBlank { getStringBlocking(Res.string.group_chat_peers_online, numPeersOnline, peers.size) }
                }
                is GroupPeerUid -> state.groups[uid.groupUid]?.title
                is FriendRequestUid -> uid.value
                else -> null
            }?.toStylizedMessage(
                params = StylizedTextParameters(
                    settings = settings.chatSettings,
                    colors = colors,
                ),
                removeEffectTags = true,
            )
        }
        Column(
            modifier = Modifier
                .padding(start = padding)
                .weight(1f),
        ) {
            TextLine(
                text = state.conversationTitle.orEmpty(),
                style = ChatTheme.typography.h6,
                modifier = Modifier.basicMarquee(),
            )
            conversationDesc?.text?.run {
                Text(
                    text = text,
                    style = if (isPortrait) ChatTheme.typography.body1 else ChatTheme.typography.body2,
                    color = ChatTheme.colors.onBackground,
                    modifier = Modifier.basicMarquee(),
                )
            }
        }
        state.conversationUser?.takeIf { state.conversationUid is UserUid }?.run {
            CallButton(
                contactUid = uid,
                call = state.conversationCall,
                isEnabled = isOnline && isCallsEnabled,
            )
            SearchButton(
                state = state.searchState,
                onToggleSearch = onToggleSearch,
            )
        }
        state.conversationGroup?.run {
            var isOptionsDialogVisible by remember(state.conversationUid) { mutableStateOf(false) }
            var isSettingsDialogVisible by remember(state.conversationUid) { mutableStateOf(false) }
            val selfPeer = remember(state) { findSelf(state.self) }
            if (ChatGroupFeature.AUDIO_CALLS in features) {
                CallButton(
                    contactUid = uid,
                    call = state.conversationCall,
                    isEnabled = isOnline && isCallsEnabled,
                )
            }
            val showViewPeersButton = isGroupPeersCollapsed || isPortrait
            if (selfPeer.role == ChatGroupRole.OBSERVER) {
                SearchButton(
                    state = state.searchState,
                    onToggleSearch = onToggleSearch,
                )
            } else {
                IconButton(
                    icon = Icons.Outlined.MoreVert,
                    contentDescription = stringResource(Res.string.group_settings),
                    onClick = { isOptionsDialogVisible = true },
                    size = ChatTheme.dimens.iconMedium,
                )
            }
            Box {
                IconButton(
                    icon = when {
                        showViewPeersButton -> Icons.Outlined.Groups
                        else -> Icons.AutoMirrored.Outlined.ArrowForwardIos
                    },
                    contentDescription = when {
                        showViewPeersButton -> stringResource(Res.string.group_chat_view_peers)
                        else -> stringResource(Res.string.group_chat_hide_peers)
                    },
                    onClick = onTogglePeers,
                    size = ChatTheme.dimens.iconMedium,
                )
                if (showViewPeersButton) {
                    val unreadCount = remember(state.conversationUid, state.unreadCounts) {
                        state.getPeerUnreadCount(uid)
                    }
                    if (unreadCount > 0) {
                        TextLine(
                            text = unreadCount.toString(),
                            color = Color.White,
                            style = ChatTheme.typography.overline.copy(lineHeight = 12.sp),
                            modifier = Modifier
                                .clip(CircleShape)
                                .background(Color.Red)
                                .padding(start = ChatTheme.dimens.spacingXSmall)
                                .align(Alignment.TopEnd),
                        )
                    }
                }
            }
            if (isOptionsDialogVisible) {
                ChatScreenOptionsDialog(
                    onSearch = {
                        isOptionsDialogVisible = false
                        onToggleSearch()
                    },
                    onSettings = {
                        isOptionsDialogVisible = false
                        isSettingsDialogVisible = true
                    },
                    onDismiss = { isOptionsDialogVisible = false },
                )
            }
            if (isSettingsDialogVisible) {
                GroupSettingsDialog(
                    group = this,
                    selfPeer = selfPeer,
                    onDismiss = { isSettingsDialogVisible = false },
                )
            }
        }
    }
}

@Composable
@NonRestartableComposable
private fun SearchButton(
    state: ChatSearchState,
    onToggleSearch: () -> Unit,
) = IconButton(
    icon = when {
        state.isActive -> Icons.Outlined.SearchOff
        else -> Icons.Outlined.Search
    },
    contentDescription = when {
        state.isActive -> stringResource(Res.string.chat_screen_search_cancel)
        else -> stringResource(Res.string.chat_screen_search_messages)
    },
    size = ChatTheme.dimens.iconMedium,
    onClick = onToggleSearch,
)