package nand.apps.chat.ui.chat.screen

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import nand.apps.chat.model.event.*
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.ui.call.CallHeader
import nand.apps.chat.ui.chat.message.MessagesColumn
import nand.apps.chat.ui.chat.message.input.MessageInputField
import nand.apps.chat.ui.chat.message.input.MessageInputViewModel
import nand.apps.chat.ui.chat.search.ChatSearchHeader
import nand.apps.chat.ui.groups.GroupErrorScreen
import nand.apps.chat.ui.groups.GroupPeersColumn
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.navigation.BackHandler
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.theme.ChatTheme

@Composable
fun ChatScreen(
    viewModel: ChatScreenViewModel,
    inputViewModel: MessageInputViewModel,
    modifier: Modifier = Modifier,
) = Box(modifier = modifier) {

    val state = LocalChatScreenState.current
    val eventHandler = LocalEventHandler.current
    val settings = LocalSettingsProvider.current
    val isPortrait = isPortraitMode()
    var isGroupPeersSelected by remember { mutableStateOf(false) }
    val isGroupPeersCollapsed = settings.generalSettings.isGroupPeersCollapsed
    
    BackHandler(
        isEnabled = isPortrait && isGroupPeersSelected,
    ) {
        isGroupPeersSelected = false
    }
    ShortcutHandler(
        action = ShortcutAction.TOGGLE_SEARCH,
        isEnabled = state.conversationUid != null,
        onShortcut = viewModel::toggleSearch,
    )

    Row(
        modifier = Modifier.fillMaxSize(),
    ) {
        Column(
            modifier = Modifier
                .weight(1f)
                .fillMaxHeight(),
        ) {
            ChatHeader(
                onBackPressed = { eventHandler += ExitConversationEvent },
                onTogglePeers = {
                    if (isPortrait) {
                        isGroupPeersSelected = true
                    } else {
                        eventHandler += UpdateSettingsEvent(
                            settings = settings.copy(
                                generalSettings = settings.generalSettings.copy(
                                    isGroupPeersCollapsed = !isGroupPeersCollapsed,
                                )
                            )
                        )
                    }
                },
                onToggleSearch = viewModel::toggleSearch,
                modifier = Modifier.fillMaxWidth(),
            )
            CallHeader(
                call = state.conversationCall,
                contactName = state.conversationTitle.orEmpty(),
                modifier = Modifier.fillMaxWidth(),
            )
            ChatSearchHeader(
                state = state.searchState,
                modifier = Modifier.fillMaxWidth(),
                onValueChange = viewModel::search,
                onCancel = viewModel::cancelSearch,
            )
            Box(
                modifier = Modifier
                    .background(ChatTheme.colors.onBackground2)
                    .fillMaxWidth()
                    .height(2.dp)
            )
            MessagesColumn(
                viewModel = viewModel,
                modifier = Modifier
                    .fillMaxWidth()
                    .weight(1f),
            )
            MessageInputField(
                textStyle = if (isPortrait) ChatTheme.typography.body1 else ChatTheme.typography.body2,
                viewModel = inputViewModel,
                modifier = Modifier
                    .fillMaxWidth()
                    .padding(horizontal = ChatTheme.dimens.spacingLarge)
                    .padding(bottom = ChatTheme.dimens.spacingLarge),
            )
        }
        state.conversationGroup?.let { group ->
            AnimatedVisibility(
                visible = isPortrait && isGroupPeersSelected || !isPortrait && !isGroupPeersCollapsed,
            ) {
                GroupPeersColumn(
                    group = group,
                    sidebarState = if (isPortrait) SidebarState.FULL else SidebarState.DEFAULT,
                    onExit = { isGroupPeersSelected = false },
                    modifier = Modifier
                        .background(ChatTheme.colors.background)
                        .fillMaxHeight(),
                )
            }
        }
    }
    state.conversationGroup?.let { group ->
        group.error?.let { error ->
            GroupErrorScreen(
                group = group,
                error = error,
                modifier = Modifier
                    .background(ChatTheme.colors.backgroundFloating)
                    .matchParentSize(),
            )
        }
    }
}