package nand.apps.chat.ui.chat.screen

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.Job
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.event.VisibleMessageTracker
import nand.apps.chat.model.event.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.GroupPeerUid
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.repo.*
import nand.apps.chat.search.ChatSearch
import nand.apps.chat.ui.chat.search.ChatSearchState
import nand.apps.chat.ui.util.collect
import nand.apps.chat.util.clamp

class ChatScreenViewModel(
    private val messageRepo: ChatMessageRepo,
    private val contactRepo: ContactRepo,
    selfProfileRepo: SelfProfileRepo,
    private val settingsRepo: SettingsRepo,
    fileTransferRepo: FileTransferRepo,
    callRepo: CallRepo,
    private val eventHandler: ChatEventHandler,
    private val searchHandler: ChatSearch,
    private val visibleMessageTracker: VisibleMessageTracker,
) : ViewModel() {

    val stateData = MutableStateFlow(ChatScreenState())
    
    private val conversationUid get() = stateData.value.conversationUid

    private val isSearchActive get() = stateData.value.searchState.isActive
    private val searchTerms get() = stateData.value.searchState.terms
    private val searchParams get() = stateData.value.searchState.params
    private val searchTextEffects get() = stateData.value.searchState.maxTextEffects

    private var firstVisibleMessageIndex = 0
    private var searchJob: Job? = null
    
    init {
        collect(messageRepo.onNewMessage) { message ->
            if (message.conversationUid == conversationUid) {
                addToCurrentConversation(message)
            }
        }
        collect(messageRepo.onMessagesChanged) { conversationUid ->
            if (conversationUid == this.conversationUid) {
                if (isSearchActive) {
                    updateMessageSearch()
                } else {
                    refreshMessages(conversationUid)
                }
            }
        }
        collect(messageRepo.unreadObserver) { unreadCounts ->
            stateData.emit(stateData.value.copy(unreadCounts = unreadCounts))
        }
        collect(selfProfileRepo.selfObserver) { self ->
            stateData.emit(stateData.value.copy(self = self))
        }
        collect(contactRepo.friendObserver) { friends ->
            stateData.emit(stateData.value.copy(friends = friends))
        }
        collect(contactRepo.groupObserver) { groups ->
            stateData.emit(stateData.value.copy(groups = groups))
        }
        collect(fileTransferRepo.observer) { fileTransfers ->
            stateData.emit(stateData.value.copy(fileTransfers = fileTransfers))
        }
        collect(callRepo.callObserver) { calls ->
            stateData.emit(stateData.value.copy(calls = calls))
        }
        collect(callRepo.callActivityObserver) { callActivity ->
            stateData.emit(stateData.value.copy(callActivity = callActivity))
        }
        collect(settingsRepo.observer) { newSettings ->
            val settings = newSettings.chatSettings
            if (isSearchActive && (searchParams != settings.searchParams || searchTextEffects != settings.maxTextEffects)) {
                updateSearchState {
                    it.copy(
                        params = settings.searchParams,
                        maxTextEffects = settings.maxTextEffects,
                    )
                }
                updateMessageSearch()
            }
        }
        collect(eventHandler.onEvent, ::onEvent)
    }
    
    private fun onEvent(event: ChatEvent) = when (event) {
        is OpenConversationEvent -> {
            onOpenConversationEvent(event.contactUid)
        }
        is ExitConversationEvent -> {
            onExitConversationEvent()
        }
        is AcceptFriendRequestEvent -> {
            acceptFriendRequest(event.request)
        }
        is DeclineFriendRequestEvent -> {
            checkExitConversation(event.request.uid)
        }
        is RemoveFriendEvent -> {
            checkExitConversation(event.friend.uid)
        }
        is RemoveGroupEvent -> {
            checkExitConversation(event.group.uid)
        }
        else -> Unit
    }

    private fun onOpenConversationEvent(conversationUid: ContactUid) {
        if (conversationUid != this.conversationUid) {
            cancelSearch(updateMessages = false)
            refreshMessages(conversationUid)
        }
    }

    private fun onExitConversationEvent() {
        val conversationUid = conversationUid
        if (conversationUid is GroupPeerUid) {
            eventHandler += OpenConversationEvent(conversationUid.groupUid)
        } else {
            stateData.tryEmit(stateData.value.copy(conversationUid = null))
            visibleMessageTracker.onConversationChanged(null)
        }
    }

    private suspend fun addToCurrentConversation(message: ChatMessage) {
        if (isSearchActive && searchTerms.isNotBlank()) {
            val filtered = searchHandler.filterMessages(
                messages = listOf(message),
                terms = searchTerms,
                params = searchParams,
            )
            if (filtered.isNotEmpty()) {
                stateData.emit(stateData.value.copy(messages = stateData.value.messages + filtered))
                return
            }
        }
        stateData.emit(stateData.value.copy(messages = stateData.value.messages + message))
    }
    
    fun setFirstVisibleMessageIndex(index: Int) {
        val oldIndex = firstVisibleMessageIndex
        firstVisibleMessageIndex = index
        if (oldIndex <= index) {
            // Only should call this when scrolling down (for now)
            return
        }

        val conversationUid = conversationUid ?: return
        val messages = stateData.value.messages
        if (messages.isEmpty()) return

        // TODO - Stop at top visible index and show a "mark messages read" top banner
        val actualIndex = (messages.lastIndex - index).clamp(0, messages.lastIndex)
        val unreadMessages = messages.slice(0..actualIndex).filter { it.isUnread }
        messageRepo.markMessagesRead(conversationUid, unreadMessages)
    }

    fun onScrolledUp(isScrolledUp: Boolean) {
        visibleMessageTracker.onScrolledUp(isScrolledUp)
    }

    fun search(terms: String) {
        if (terms != searchTerms) {
            updateSearchState {
                it.copy(
                    terms = terms,
                    maxTextEffects = settingsRepo.settings.chatSettings.maxTextEffects,
                    params = settingsRepo.settings.chatSettings.searchParams,
                )
            }
            updateMessageSearch()
        }
    }

    fun cancelSearch(updateMessages: Boolean = true) {
        if (isSearchActive) {
            stateData.tryEmit(stateData.value.copy(searchState = ChatSearchState()))
            if (updateMessages) {
                updateMessageSearch()
            } else {
                searchJob?.cancel()
                searchJob = null
            }
        }
    }

    fun toggleSearch() {
        if (isSearchActive) {
            cancelSearch()
        } else {
            updateSearchState { it.copy(isActive = true) }
        }
    }
    
    private fun refreshMessages(conversationUid: ContactUid) {
        val messages = messageRepo.getMessages(conversationUid)
        stateData.tryEmit(
            stateData.value.copy(
                conversationUid = conversationUid,
                messages = messages,
            )
        )
        visibleMessageTracker.onConversationChanged(conversationUid)
    }

    private fun checkExitConversation(conversationUid: ContactUid?) {
        if (this.conversationUid == conversationUid) {
            eventHandler += ExitConversationEvent
        }
    }

    private fun updateSearchState(block: (ChatSearchState) -> ChatSearchState) {
        stateData.tryEmit(
            stateData.value.copy(
                searchState = block(stateData.value.searchState),
            )
        )
    }
    
    private fun acceptFriendRequest(request: FriendRequest) {
        val newFriend = contactRepo.getFriend(UserUid(request.uid.value))
        if (newFriend != null) {
            eventHandler += OpenConversationEvent(newFriend.uid)
        } else {
            eventHandler += ExitConversationEvent
        }
    }

    private fun updateMessageSearch() {
        val conversationUid = conversationUid ?: return
        searchJob?.cancel()
        searchJob = viewModelScope.launch {
            var messages = messageRepo.getMessages(conversationUid)
            withContext(Dispatchers.IO) {
                messages = messages.filterMessageSearch()
            }
            if (stateData.value.conversationUid == conversationUid) {
                stateData.emit(stateData.value.copy(messages = messages))
            }
        }
    }

    private fun List<ChatMessage>.filterMessageSearch(): List<ChatMessage> {
        return if (isSearchActive) {
            searchHandler.filterMessages(
                messages = this,
                terms = searchTerms,
                params = searchParams,
            )
        } else {
            this
        }
    }
}