package nand.apps.chat.ui.contacts

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.outlined.ArrowBackIos
import androidx.compose.material.icons.automirrored.outlined.ArrowForwardIos
import androidx.compose.material.icons.filled.GroupAdd
import androidx.compose.material.icons.filled.PersonAddAlt1
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.alpha
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.TextFieldValue
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.message.UnreadMessageCounts
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserData
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.chat.screen.LocalChatScreenState
import nand.apps.chat.ui.friends.AddFriendDialog
import nand.apps.chat.ui.friends.FriendContactRow
import nand.apps.chat.ui.friends.FriendRequestRow
import nand.apps.chat.ui.groups.AddGroupDialog
import nand.apps.chat.ui.groups.CreateGroupDialog
import nand.apps.chat.ui.groups.GroupContactRow
import nand.apps.chat.ui.groups.JoinGroupDialog
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.provider.isPortraitMode
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun ContactListColumn(
    sidebarState: SidebarState,
    modifier: Modifier = Modifier,
    viewModel: ContactListViewModel = koinKmpViewModel(),
    onToggleCollapsedState: () -> Unit,
) = Column(
    horizontalAlignment = Alignment.CenterHorizontally,
    modifier = modifier,
) {
    val chatState = LocalChatScreenState.current
    val isCollapsed = sidebarState == SidebarState.COLLAPSED
    val conversationUid = chatState.conversationUid
    val unreadCounts = chatState.unreadCounts
    val state by viewModel.stateData.collectAsState()
    var searchInput by rememberTextFieldValue()
    
    LaunchedEffect(isCollapsed) {
        searchInput = TextFieldValue()
        viewModel.search("")
    }
    
    ShortcutHandler(
        action = ShortcutAction.TOGGLE_CONTACTS,
        isEnabled = !isPortraitMode(),
        onShortcut = onToggleCollapsedState,
    )
    
    if (isCollapsed) {
        IconButton(
            icon = Icons.AutoMirrored.Outlined.ArrowForwardIos,
            contentDescription = stringResource(Res.string.expand_contacts),
            onClick = onToggleCollapsedState,
            modifier = Modifier.padding(vertical = ChatTheme.dimens.spacingSmall)
        )
    } else {
        Row(
            modifier = Modifier.fillMaxWidth(),
            verticalAlignment = Alignment.CenterVertically,
        ) {
            val focusManager = LocalFocusManager.current
            SimpleTextField(
                value = searchInput,
                onValueChange = {
                    searchInput = it
                    viewModel.search(it.text)
                },
                leadingIcon = if (!isPortraitMode()) {
                    {
                        IconButton(
                            icon = Icons.AutoMirrored.Outlined.ArrowBackIos,
                            contentDescription = stringResource(Res.string.collapse_contacts),
                            onClick = onToggleCollapsedState,
                        )
                    }
                } else null,
                hintText = stringResource(Res.string.search_contacts),
                keyboardOptions = KeyboardOptions(
                    autoCorrectEnabled = false,
                    imeAction = ImeAction.Search
                ),
                onSubmit = { focusManager.clearFocus() },
                modifier = Modifier.fillMaxWidth(),
            )
        }
    }
    
    LazyColumn(
        contentPadding = PaddingValues(vertical = ChatTheme.dimens.spacingSmall),
        modifier = Modifier.fillMaxHeight(),
    ) {
        items(
            items = state.contacts,
            key = { it.uid.value },
        ) { contact ->
            val rowModifiers = Modifier
                .run { if (isCollapsed) wrapContentWidth() else fillMaxWidth() }
                .animateItem()
            when (contact) {
                is FriendRequest -> {
                    FriendRequestRow(
                        request = contact,
                        sidebarState = sidebarState,
                        modifier = Modifier
                            .background(if (conversationUid == contact.uid) ChatTheme.colors.primaryVariant else ChatTheme.colors.secondaryVariant)
                            .then(rowModifiers)
                    )
                }
                is ChatGroupData -> {
                    val groupUnreadCount = remember(contact.uid, unreadCounts) {
                        val unreadCount = unreadCounts[contact.uid]
                        chatState.getPeerUnreadCount(contact.uid).takeIf { it > 0 }?.let {
                            UnreadMessageCounts(directMessages = it + (unreadCount?.directMessages ?: 0))
                        } ?: unreadCount
                    }
                    GroupContactRow(
                        group = contact,
                        selfPeer = contact.findSelf(state.self),
                        sidebarState = sidebarState,
                        unreadCount = groupUnreadCount,
                        call = chatState.calls[contact.uid],
                        modifier = Modifier
                            .background(if (conversationUid == contact.uid) ChatTheme.colors.primaryVariant else Color.Unspecified)
                            .alpha(if (contact.isOnline && !contact.isMuted) 1f else 0.5f)
                            .then(rowModifiers)
                    )
                }
                is UserData -> {
                    FriendContactRow(
                        friend = contact,
                        unreadCount = unreadCounts[contact.uid],
                        sidebarState = sidebarState,
                        modifier = Modifier
                            .background(if (conversationUid == contact.uid) ChatTheme.colors.primaryVariant else Color.Unspecified)
                            .alpha(if (contact.isOnline && !contact.isMuted) 1f else 0.5f)
                            .then(rowModifiers)
                    )
                }
            }
        }
        if (isCollapsed) {
            item(
                key = "Add Friend",
            ) {
                IconButton(
                    icon = Icons.Default.PersonAddAlt1,
                    contentDescription = stringResource(Res.string.add_friend_dialog_title),
                    size = ChatTheme.dimens.iconMedium,
                    onClick = { viewModel.openDialog(ContactDialogType.ADD_FRIEND) },
                    modifier = Modifier.padding(horizontal = ChatTheme.dimens.spacingSmall),
                )
            }
            item(
                key = "Add Group",
            ) {
                IconButton(
                    icon = Icons.Default.GroupAdd,
                    contentDescription = stringResource(Res.string.create_group_dialog_title),
                    size = ChatTheme.dimens.iconMedium,
                    onClick = { viewModel.openDialog(ContactDialogType.ADD_GROUP) },
                    modifier = Modifier.padding(horizontal = ChatTheme.dimens.spacingSmall),
                )
            }
        }
    }
    when (state.dialog) {
        ContactDialogType.ADD_FRIEND -> AddFriendDialog(
            onDismiss = viewModel::dismissDialog,
        )
        ContactDialogType.ADD_GROUP -> AddGroupDialog(
            onSelect = viewModel::openDialog,
            onDismiss = viewModel::dismissDialog,
        )
        ContactDialogType.CREATE_GROUP -> CreateGroupDialog(
            onDismiss = viewModel::dismissDialog,
        )
        ContactDialogType.JOIN_GROUP -> JoinGroupDialog(
            onDismiss = viewModel::dismissDialog,
        )
        else -> Unit
    }
}
