package nand.apps.chat.ui.dialog

import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import kotlinx.coroutines.launch
import nand.apps.chat.io.IMAGE_EXTENSIONS
import nand.apps.chat.io.isValidFileName
import nand.apps.chat.io.sanitizeFileName
import nand.apps.chat.model.event.ErrorSnackbarEvent
import nand.apps.chat.model.event.SuccessSnackbarEvent
import nand.apps.chat.net.FileDownloader
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.res.FormattedStringResource
import nand.apps.chat.ui.file.ChatFilePickerDialog
import nand.apps.chat.ui.file.ChatFilePickerMode
import nand.apps.chat.ui.file.ChatFilePickerOptions
import nand.apps.chat.ui.provider.LocalCoroutineScope
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.util.getUrlFileName
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun FileDownloadDialog(
    url: String,
    onConfirm: (String) -> Unit,
    onCancel: () -> Unit,
) {
    val coroutineScope = LocalCoroutineScope.current
    val eventHandler = LocalEventHandler.current
    val downloader: FileDownloader by getKoin().inject()
    val downloadFile: (path: String) -> Unit = { path ->
        coroutineScope.launch {
            val fileName = path.toPath().name
            if (downloader.download(url = url, filePath = path)) {
                eventHandler += SuccessSnackbarEvent(
                    FormattedStringResource(Res.string.file_download_complete, fileName)
                )
            } else {
                eventHandler += ErrorSnackbarEvent(
                    FormattedStringResource(Res.string.file_download_failed, fileName)
                )
            }
        }
        onConfirm(path)
    }
    if (getPlatform().isDesktop) {
        ChatFilePickerDialog(
            options = ChatFilePickerOptions(
                fileName = url.getUrlFileName().sanitizeFileName(),
                directory = downloader.directory,
                extensions = IMAGE_EXTENSIONS,
                mode = ChatFilePickerMode.FILES_AND_DIRECTORIES,
                isOpenDialog = false,
            ),
            onConfirm = downloadFile,
            onCancel = onCancel,
        )
    } else {
        var fileName by rememberTextFieldValue(url.getUrlFileName().sanitizeFileName())
        SimpleDialog(
            title = Res.string.download,
            isConfirmEnabled = fileName.text.isValidFileName(),
            onConfirm = {
                downloadFile(downloader.directory.toPath().resolve(fileName.text).toString())
            },
            onCancel = onCancel,
        ) {
            SimpleTextField(
                value = fileName,
                onValueChange = { fileName = it },
                labelText = stringResource(Res.string.file_name),
                modifier = Modifier.fillMaxWidth(),
            )
        }
    }
}