package nand.apps.chat.ui.dialog.audio

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.layout.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Stop
import androidx.compose.material.icons.outlined.Mic
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import io.ktor.http.*
import nand.apps.chat.model.audio.AudioMetadata
import nand.apps.chat.model.file.OPUS
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.embed.audio.AudioEmbedRenderer
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.util.toDurationString
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.stringResource
import org.koin.core.parameter.parametersOf
import seers.composeapp.generated.resources.*

@Composable
fun AudioRecorderDialog(
    title: StringResource,
    filePath: String,
    recordBitRate: Int,
    onConfirm: () -> Unit,
    onDismiss: () -> Unit,
    metadata: AudioMetadata? = null,
    shouldEncrypt: Boolean = false,
    viewModel: AudioRecorderDialogViewModel = composableViewModel(
        parameters = {
            parametersOf(
                AudioRecorderDialogState(
                    filePath = filePath,
                    bitRate = recordBitRate,
                    metadata = metadata,
                    shouldEncrypt = shouldEncrypt,
                )
            )
        },
    ),
) {
    val state by viewModel.stateData.collectAsState()

    SimpleDialog(
        title = title,
        titleStyle = ChatTheme.typography.body1,
        confirmLabel = when {
            state.isActive -> Res.string.stop_recording
            !state.isPathValid -> Res.string.start_recording
            else -> Res.string.confirm
        },
        onConfirm = {
            when {
                !state.isPathValid || state.isActive -> viewModel.toggleRecording()
                else -> onConfirm()
            }
        },
        onCancel = {
            viewModel.deleteTempFile()
            onDismiss()
        },
    ) {
        Column {
            Row(
                horizontalArrangement = Arrangement.Center,
                verticalAlignment = Alignment.CenterVertically,
                modifier = Modifier.fillMaxWidth()
            ) {
                IconButton(
                    icon = when {
                        state.isActive -> Icons.Filled.Stop
                        else -> Icons.Outlined.Mic
                    },
                    contentDescription = stringResource(
                        when {
                            state.isActive -> Res.string.stop_recording
                            else -> Res.string.start_recording
                        }
                    ),
                    onClick = {
                        viewModel.toggleRecording()
                    },
                )
                TextLine(
                    text = when {
                        state.isActive -> state.recordDurationSecs.toInt().toDurationString()
                        else -> "--:--"
                    },
                    style = ChatTheme.typography.body1,
                    modifier = Modifier.padding(start = ChatTheme.dimens.spacingMedium),
                )
            }
            AnimatedVisibility(
                visible = state.isPathValid,
            ) {
                AudioEmbedRenderer(
                    uri = filePath,
                    type = ContentType.Audio.OPUS,
                    reuseCache = false,
                    onDeleteFile = viewModel::onDeleteFile,
                )
            }
        }
    }
}