package nand.apps.chat.ui.dialog.audio

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableStateFlow
import nand.apps.audio.AudioFileWriter
import nand.apps.chat.audio.AudioFileWriterFactory
import nand.apps.chat.audio.AudioRecorder
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.io.ChatEngineEncryption
import nand.apps.chat.io.ChatFileSystem
import okio.Path.Companion.toPath

class AudioRecorderDialogViewModel(
    private val audioRecorder: AudioRecorder,
    private val fileSystem: ChatFileSystem,
    private val encryption: ChatEngineEncryption,
    private val fileWriterFactory: AudioFileWriterFactory,
    private val backgroundWorker: ChatBackgroundWorker,
    initialState: AudioRecorderDialogState,
) : ViewModel() {

    val stateData = MutableStateFlow(initialState)
    private val state get() = stateData.value

    private var fileWriter: AudioFileWriter? = null
    private var recordJob: Job? = null

    init {
        if (fileSystem.exists(initialState.filePath)) {
            updateState { it.copy(isPathValid = true) }
        }
    }

    fun toggleRecording() {
        recordJob?.cancel()
        if (state.isActive) {
            updateState { it.copy(isActive = false) }
            if (!state.isSharedRecorderActive) audioRecorder.stop()
            stopRecording()
        } else {
            val isSharedRecorderActive = audioRecorder.isActive
            if (!isSharedRecorderActive) audioRecorder.start()
            updateState {
                it.copy(
                    isActive = audioRecorder.isActive,
                    isSharedRecorderActive = isSharedRecorderActive,
                    recordDurationSecs = 0.0,
                )
            }
            startRecording()
        }
    }

    fun deleteTempFile() {
        fileWriter?.run {
            close()
            fileSystem.deleteFile(state.filePath)
        }
        fileWriter = null
    }

    fun onDeleteFile() = updateState {
        it.copy(isPathValid = false)
    }

    private fun startRecording() {
        updateState { it.copy(isPathValid = false) }
        recordJob = viewModelScope.launch {
            audioRecorder.onFrame.collect { frame ->
                if (fileWriter == null) {
                    val filePath = state.filePath
                    fileSystem.makeDirectoriesForFile(filePath)
                    fileSystem.deleteFile(filePath)
                    fileWriter = fileWriterFactory.getFileWriter(
                        path = filePath,
                        sampleRate = frame.sampleRate,
                        channels = frame.channels,
                        bitRate = state.bitRate,
                        metadata = state.metadata,
                    )
                }
                fileWriter?.write(frame.samples)
                updateState {
                    it.copy(
                        recordDurationSecs = it.recordDurationSecs + frame.duration,
                    )
                }
            }
        }
    }

    private fun stopRecording() {
        fileWriter?.run {
            close()
            viewModelScope.launch {
                val filePath = state.filePath
                if (state.shouldEncrypt && encryption.isEncryptionSupported()) {
                    withContext(backgroundWorker.dispatcher) {
                        val dir = fileSystem.getDirectoryForFile(filePath)
                        val fileName = filePath.toPath().name
                        val bytes = dir?.readFromFile(fileName) ?: return@withContext
                        dir.writeToFile(
                            fileName = fileName,
                            data = bytes,
                            encryption = encryption,
                        )
                    }
                }
                updateState { it.copy(isPathValid = fileSystem.exists(filePath)) }
            }
        }
        fileWriter = null
    }

    private fun updateState(block: (AudioRecorderDialogState) -> AudioRecorderDialogState) {
        stateData.tryEmit(block(state))
    }
}