package nand.apps.chat.ui.embed.audio

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.background
import androidx.compose.foundation.gestures.detectTapGestures
import androidx.compose.foundation.layout.*
import androidx.compose.material.CircularProgressIndicator
import androidx.compose.material.LinearProgressIndicator
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.input.pointer.pointerInput
import androidx.compose.ui.layout.onSizeChanged
import io.ktor.http.*
import nand.apps.chat.embed.file.AudioEmbedData
import nand.apps.chat.model.audio.AudioPlaybackState
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.dialog.PermanentActionDialog
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.LaunchedCollect
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.util.toDurationString
import org.jetbrains.compose.resources.stringResource
import org.koin.core.parameter.parametersOf
import seers.composeapp.generated.resources.*

@Composable
fun AudioEmbedRenderer(
    uri: String,
    type: ContentType,
    modifier: Modifier = Modifier,
    reuseCache: Boolean = true,
    onDeleteFile: (() -> Unit)? = null,
    viewModel: AudioEmbedViewModel = composableViewModel(
        parameters = {
            parametersOf(
                AudioEmbedState(
                    uri = uri,
                    type = type,
                    reuseCache = reuseCache,
                )
            )
        },
    ),
) {
    val state by viewModel.stateData.collectAsState()
    var seekWidth by remember { mutableStateOf(0.0) }
    
    // Stop the sound when the player is removed
    DisposableEffect(Unit) {
        onDispose { viewModel.pausePlayback() }
    }

    LaunchedCollect(viewModel.actionData) { action ->
        when (action) {
            is DeleteFileAction -> onDeleteFile?.invoke()
        }
    }
    
    Column(
        modifier = modifier
            .fillMaxWidth()
            .clip(ChatTheme.shapes.medium)
            .background(ChatTheme.colors.surface2)
    ) {
        Row(
            verticalAlignment = Alignment.CenterVertically,
            modifier = Modifier.fillMaxWidth()
        ) {
            Column(
                modifier = Modifier
                    .weight(1f)
                    .padding(ChatTheme.dimens.spacingMedium),
            ) {
                val title = state.metadata?.title
                val artist = state.metadata?.artist
                TextLine(
                    text = title ?: state.fileName,
                    style = ChatTheme.typography.body1,
                )
                AnimatedVisibility(
                    visible = artist != null,
                ) {
                    TextLine(
                        text = artist.orEmpty(),
                        style = ChatTheme.typography.body2,
                    )
                }
            }
            if (state.playbackState == AudioPlaybackState.LOADING) {
                CircularProgressIndicator(
                    modifier = Modifier.size(ChatTheme.dimens.iconMedium),
                )
            } else {
                IconButton(
                    icon = when (state.playbackState) {
                        AudioPlaybackState.PLAYING -> Icons.Default.Pause
                        else -> Icons.Default.PlayArrow
                    },
                    contentDescription = when (state.playbackState) {
                        AudioPlaybackState.PLAYING -> stringResource(Res.string.pause)
                        else -> stringResource(Res.string.play)
                    },
                    onClick = viewModel::togglePlayback,
                )
            }
            IconButton(
                icon = Icons.Default.Stop,
                contentDescription = stringResource(Res.string.stop),
                onClick = viewModel::stopPlayback,
            )
            if (onDeleteFile != null) {
                IconButton(
                    icon = Icons.Default.Delete,
                    contentDescription = stringResource(Res.string.delete),
                    onClick = {
                        viewModel.stopPlayback()
                        viewModel.showDeleteDialog()
                    },
                )
            }
        }
        AnimatedVisibility(
            visible = state.playablePath != null,
        ) {
            Row(
                verticalAlignment = Alignment.CenterVertically,
                horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                modifier = Modifier
                    .fillMaxWidth()
                    .padding(ChatTheme.dimens.spacingMedium),
            ) {
                TextLine(
                    text = if (state.duration > 0) state.timestamp.toInt().toDurationString() else "--:--",
                    style = ChatTheme.typography.button,
                )
                LinearProgressIndicator(
                    progress = when {
                        state.duration > 0 -> (state.timestamp / state.duration).toFloat()
                        else -> 0f
                    },
                    modifier = Modifier
                        .clip(ChatTheme.shapes.small)
                        .weight(1f)
                        .height(ChatTheme.dimens.spacingMedium)
                        .onSizeChanged { seekWidth = it.width.toDouble() }
                        .pointerInput(state.duration, seekWidth) {
                            detectTapGestures(
                                onPress = { offset ->
                                    val seekTime = (offset.x / seekWidth) * state.duration
                                    viewModel.seekTo(seekTime)
                                }
                            )
                        },
                )
                TextLine(
                    text = if (state.duration > 0) state.duration.toInt().toDurationString() else "--:--",
                    style = ChatTheme.typography.button,
                )
            }
        }
    }
    
    if (state.isDeleteDialogVisible) {
        PermanentActionDialog(
            title = Res.string.delete_audio,
            message = stringResource(Res.string.delete_audio_confirmation),
            onConfirm = viewModel::confirmDeletion,
            onCancel = viewModel::dismissDeleteDialog,
        )
    }
}

@Composable
@NonRestartableComposable
fun AudioEmbedData.Renderer(modifier: Modifier) = AudioEmbedRenderer(
    uri = uri,
    type = type,
    modifier = modifier,
)