package nand.apps.chat.ui.embed.audio

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.*
import kotlinx.coroutines.flow.MutableStateFlow
import nand.apps.audio.AudioHandle
import nand.apps.chat.audio.AudioMetadataReader
import nand.apps.chat.audio.AudioPlayer
import nand.apps.chat.io.*
import nand.apps.chat.model.audio.AudioPlaybackState
import nand.apps.chat.model.file.isSupportedAudioFormat
import nand.apps.chat.net.TempFileCache
import nand.apps.chat.util.fifoSharedFlow
import nand.apps.chat.util.sha256
import okio.Path.Companion.toPath

class AudioEmbedViewModel(
    private val metadataReader: AudioMetadataReader,
    private val tempCache: TempFileCache,
    private val fileSystem: ChatFileSystem,
    private val encryption: ChatEngineEncryption,
    private val audioPlayer: AudioPlayer,
    private val backgroundWorker: ChatBackgroundWorker,
    private val delayScope: CoroutineDelayScope,
    initialState: AudioEmbedState,
) : ViewModel() {

    val stateData = MutableStateFlow(initialState)
    val actionData = fifoSharedFlow<AudioEmbedAction>()
    private val state get() = stateData.value

    private var playbackJob: Job? = null
    private var audioHandle: AudioHandle? = null

    fun togglePlayback() {
        val newState = when (state.playbackState) {
            AudioPlaybackState.PLAYING -> AudioPlaybackState.PAUSED
            AudioPlaybackState.PAUSED -> AudioPlaybackState.PLAYING
            AudioPlaybackState.STOPPED -> {
                if (state.playablePath != null) AudioPlaybackState.PLAYING else AudioPlaybackState.LOADING
            }
            else -> return
        }
        setPlaybackState(newState)
    }

    fun stopPlayback() {
        setPlaybackState(AudioPlaybackState.STOPPED)
    }

    fun seekTo(timestamp: Double) {
        if (state.duration > 0) {
            audioHandle?.seekTo(timestamp)
            updateState { it.copy(timestamp = timestamp) }
        }
    }

    fun showDeleteDialog() = updateState {
        it.copy(isDeleteDialogVisible = true)
    }

    fun confirmDeletion() {
        fileSystem.deleteFile(state.uri)
        state.playablePath?.let { fileSystem.deleteFile(it) }
        updateState {
            it.copy(
                playablePath = null,
                isDeleteDialogVisible = false,
            )
        }
        actionData.tryEmit(DeleteFileAction)
    }

    fun dismissDeleteDialog() = updateState {
        it.copy(isDeleteDialogVisible = false)
    }

    fun pausePlayback() {
        audioHandle?.let { audioPlayer.stopSound(it) }
    }

    private fun setPlaybackState(playbackState: AudioPlaybackState) {
        if (!state.type.isSupportedAudioFormat()) {
            updateState {
                it.copy(
                    playbackState = AudioPlaybackState.STOPPED,
                    duration = 0.0,
                )
            }
            return
        }
        updateState { it.copy(playbackState = playbackState) }
        playbackJob?.cancel()
        playbackJob = viewModelScope.launch {
            when (playbackState) {
                AudioPlaybackState.LOADING -> {
                    loadFile()
                }
                AudioPlaybackState.PLAYING -> {
                    playbackLoop()
                }
                AudioPlaybackState.PAUSED -> {
                    pausePlayback()
                }
                AudioPlaybackState.STOPPED -> {
                    onStopPlaybackState()
                }
            }
        }
    }

    private suspend fun loadFile() {
        val playablePath = if (state.isUrl) {
            withContext(backgroundWorker.dispatcher) {
                tempCache.readFromUrl(state.uri)?.path?.toString()
            }
        } else {
            copyToCache()
        }
        updateState { state ->
            state.copy(
                playablePath = playablePath,
                metadata = playablePath?.let { metadataReader.readMetadata(it) },
            )
        }
        setPlaybackState(playablePath?.let { AudioPlaybackState.PLAYING } ?: AudioPlaybackState.STOPPED)
    }

    private suspend fun playbackLoop() {
        audioHandle = state.playablePath?.let {
            audioPlayer.queueFile(
                path = it.toPath(),
                startTime = state.timestamp,
            )
        } ?: run {
            setPlaybackState(AudioPlaybackState.STOPPED)
            return
        }
        val duration = audioHandle?.duration ?: 0.0
        updateState { it.copy(duration = duration) }
        with(delayScope) {
            while (state.playbackState == AudioPlaybackState.PLAYING && state.timestamp < duration && audioHandle != null) {
                val timestamp = audioHandle?.timestamp ?: 0.0
                updateState { it.copy(timestamp = timestamp) }
                delay(20)
            }
        }
        if (state.playbackState != AudioPlaybackState.STOPPED && (state.timestamp >= duration || audioHandle == null)) {
            setPlaybackState(AudioPlaybackState.STOPPED)
        }
    }

    private fun onStopPlaybackState() {
        var timestamp = state.timestamp
        audioHandle?.let {
            audioPlayer.stopSound(it)
            timestamp = 0.0
        }
        audioHandle = null
        updateState { it.copy(timestamp = timestamp) }
    }

    private fun updateState(block: (AudioEmbedState) -> AudioEmbedState) {
        stateData.tryEmit(block(state))
    }

    private suspend fun copyToCache(): String? {
        val properPath = state.uri.toPath()
        val fileName = properPath.name
        val tmpDir = fileSystem.getTempDirectory("audio")
        val outName = "${state.uri.sha256()}.${properPath.extension}"
        val tmpFilePath = tmpDir.getFilePath(outName).toString()
        if (state.reuseCache && outName in tmpDir) return tmpFilePath

        val audioDir = fileSystem.getDirectoryForFile(state.uri) ?: return null

        val success = withContext(backgroundWorker.dispatcher) {
            val audioData = audioDir.readFromFile(
                fileName = fileName,
                encryption = encryption,
            ) ?: return@withContext false

            tmpDir.writeToFile(
                fileName = outName,
                data = audioData,
                safe = false,
            )
        }
        return if (success) return tmpFilePath else null
    }
}