package nand.apps.chat.ui.file

import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.layout.width
import androidx.compose.material.Icon
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.ColorFilter
import androidx.compose.ui.graphics.vector.rememberVectorPainter
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.unit.dp
import coil3.compose.AsyncImage
import nand.apps.chat.io.*
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.icons.fileExtensionIcon
import nand.apps.chat.ui.provider.LocalDateTimeFormatProvider
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.formatFileSize
import nand.apps.chat.util.formatPlural
import okio.Path.Companion.toPath
import org.koin.compose.getKoin

@Composable
fun ChatFileRow(
    file: ChatFileMetadata,
    modifier: Modifier = Modifier,
    color: Color = ChatTheme.colors.onBackground,
    iconSize: Dp = ChatTheme.dimens.iconMedium,
    textStyle: TextStyle = ChatTheme.typography.body1,
    isTextVisible: Boolean = true,
    fileSizeWidth: Dp = 0.dp,
    fileDateWidth: Dp = 0.dp,
) = Row(
    verticalAlignment = Alignment.CenterVertically,
    horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
    modifier = modifier,
) {
    val fileName = file.name.takeIf { it.isNotEmpty() } ?: run {
        if (file.isDirectory) file.path else ""
    }
    if (file.extension in IMAGE_EXTENSIONS) {
        var isLoading by remember { mutableStateOf(true) }
        AsyncImage(
            model = file.path,
            placeholder = rememberVectorPainter(Icons.Outlined.Image),
            colorFilter = if (isLoading) ColorFilter.tint(color) else null,
            contentDescription = fileName,
            modifier = Modifier.size(iconSize),
            onLoading = { isLoading = true },
            onSuccess = { isLoading = false },
            onError = { isLoading = false },
        )
    } else {
        Icon(
            imageVector = when {
                file.isDirectory -> run {
                    val filePaths: ChatFilePaths by getKoin().inject()
                    when (file.path.toPath()) {
                        in filePaths.roots -> if (getPlatform().isMobile) Icons.Outlined.Smartphone else Icons.Outlined.Computer
                        filePaths.homePath -> Icons.Outlined.Home
                        filePaths.picturesPath -> Icons.Outlined.PhotoLibrary
                        filePaths.videosPath -> Icons.Outlined.VideoLibrary
                        filePaths.musicPath -> Icons.Outlined.LibraryMusic
                        filePaths.downloadPath -> Icons.Outlined.Download
                        else -> Icons.Outlined.Folder
                    }
                }
                else -> fileExtensionIcon(file.extension)
            },
            contentDescription = fileName,
            tint = color,
            modifier = Modifier.size(iconSize),
        )
    }
    if (isTextVisible) {
        TextLine(
            text = fileName,
            style = textStyle,
            color = color,
            modifier = Modifier.weight(1f)
        )
        if (fileSizeWidth > 0.dp) {
            TextLine(
                text = when {
                    file.isDirectory && file.fileCount > 0 -> file.fileCount.toString() + formatPlural(" file", file.fileCount)
                    file.isDirectory && file.fileCount == 0 -> "Empty"
                    else -> formatFileSize(file.size)
                },
                style = ChatTheme.typography.body2,
                color = if (file.isDirectory && file.fileCount == 0) ChatTheme.colors.onBackground2 else color,
                textAlign = TextAlign.End,
                modifier = Modifier.width(fileSizeWidth),
            )
        }
        if (fileDateWidth > 0.dp) {
            TextLine(
                text = LocalDateTimeFormatProvider.current.formatDate(file.lastModified),
                style = ChatTheme.typography.body2,
                color = color,
                textAlign = TextAlign.End,
                modifier = Modifier.width(fileDateWidth),
            )
        }
    }
}