package nand.apps.chat.ui.file.sender

import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.AttachFile
import androidx.compose.material.icons.outlined.Image
import androidx.compose.material.icons.outlined.PhotoCamera
import androidx.compose.material.icons.outlined.RecordVoiceOver
import androidx.compose.runtime.*
import nand.apps.chat.model.audio.AudioMetadata
import nand.apps.chat.image.ImageCompressionParams
import nand.apps.chat.image.ImageFormat
import nand.apps.chat.ui.dialog.audio.AudioRecorderDialog
import nand.apps.chat.ui.dialog.image.ImageCropperDialog
import nand.apps.chat.ui.file.ChatFilePickerDialog
import nand.apps.chat.ui.file.ChatFilePickerOptions
import nand.apps.chat.ui.menu.ActionMenuDialog
import nand.apps.chat.ui.menu.ActionMenuRow
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.util.LaunchedCollect
import nand.apps.chat.ui.viewmodel.composableViewModel
import okio.Path.Companion.toPath
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

private const val MAX_PHOTO_SIZE = 1024

@Composable
fun ChatFileSenderDialog(
    onSelectFile: (String) -> Unit,
    onDismiss: () -> Unit,
    viewModel: ChatFileSenderViewModel = composableViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    val settings = LocalSettingsProvider.current

    LaunchedCollect(viewModel.actionData) { action ->
        when (action) {
            is ConfirmFileAction -> onSelectFile(action.path)
            is DismissAction -> onDismiss()
        }
    }
    
    when {
        state.isFilePickerVisible -> {
            ChatFilePickerDialog(
                options = ChatFilePickerOptions(directory = state.defaultFilePickerDirectory),
                onConfirm = { viewModel.confirmFilePickerSelection(it.toPath()) },
                onCancel = onDismiss,
            )
        }
        state.isRecordAudioDialogVisible -> {
            AudioRecorderDialog(
                title = Res.string.chat_screen_record_audio,
                filePath = state.audioFilePath,
                recordBitRate = settings.callSettings.voicemailAudioBitRate,
                metadata = AudioMetadata(
                    title = stringResource(Res.string.chat_screen_record_audio_voice_message),
                    artist = state.selfName,
                ),
                onConfirm = viewModel::onAudioRecorded,
                onDismiss = viewModel::dismissAudioRecorder,
            )
        }
        state.cropImagePath.isNotBlank() -> {
            ImageCropperDialog(
                title = Res.string.crop_image,
                imagePath = state.cropImagePath,
                maxSize = MAX_PHOTO_SIZE,
                forceSquare = false,
                compressionParams = ImageCompressionParams(
                    outPath = state.outputImagePath,
                    format = ImageFormat.JPEG,
                    quality = 0.9f,
                ),
                onConfirm = { viewModel.cropFile() },
                onCancel = onDismiss,
            )
        }
        else -> ActionMenuDialog(
            onDismiss = onDismiss,
        ) {
            ActionMenuRow(
                icon = Icons.Outlined.AttachFile,
                label = stringResource(Res.string.chat_screen_send_file),
                onClick = viewModel::attachFile,
            )
            if (state.isGallerySupported) {
                ActionMenuRow(
                    icon = Icons.Outlined.Image,
                    label = stringResource(Res.string.chat_screen_send_image),
                    onClick = viewModel::launchGallery,
                )
            }
            if (state.isCameraSupported) {
                ActionMenuRow(
                    icon = Icons.Outlined.PhotoCamera,
                    label = stringResource(Res.string.chat_screen_take_photo),
                    onClick = viewModel::launchCamera,
                )
            }
            ActionMenuRow(
                icon = Icons.Outlined.RecordVoiceOver,
                label = stringResource(Res.string.chat_screen_record_audio),
                onClick = viewModel::launchAudioRecorder,
            )
        }
    }
}