package nand.apps.chat.ui.file.sender

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import nand.apps.chat.camera.CameraLauncher
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.ChatFileBrowser
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.io.nameWithoutExtension
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.event.ChatEvent
import nand.apps.chat.model.event.TakePhotoEvent
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.time.DateTimeFormatter
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.ui.util.collect
import nand.apps.chat.util.fifoSharedFlow
import okio.FileSystem
import okio.Path
import okio.Path.Companion.toPath

class ChatFileSenderViewModel(
    selfProfileRepo: SelfProfileRepo,
    private val settingsRepo: SettingsRepo,
    eventHandler: ChatEventHandler,
    private val cameraLauncher: CameraLauncher,
    private val fileBrowser: ChatFileBrowser,
    private val chatFileSystem: ChatFileSystem,
    private val fileSystem: FileSystem,
    private val dateTimeFormatter: DateTimeFormatter,
) : ViewModel() {

    val stateData = MutableStateFlow(
        ChatFileSenderState(
            isGallerySupported = fileBrowser.isGallerySupported(),
            isCameraSupported = cameraLauncher.isSupported,
        )
    )
    val actionData = fifoSharedFlow<ChatFileSenderAction>()

    private val state get() = stateData.value
    private val logger = getLogger("ChatFileSenderViewModel")

    init {
        collect(selfProfileRepo.selfObserver) { self ->
            updateState { it.copy(selfName = self.displayName) }
        }
        if (cameraLauncher.isSupported) {
            collect(eventHandler.onEvent) { onEvent(it) }
        }
    }

    fun confirmFilePickerSelection(path: Path) {
        settingsRepo.update { chatSettings.copy(defaultSendDirectory = path.parent?.toString()) }
        finish(path.toString())
    }

    fun cropFile() {
        try {
            if (state.cropImagePath != state.outputImagePath) {
                fileSystem.delete(state.cropImagePath.toPath())
            }
            if (fileSystem.exists(state.outputImagePath.toPath())) {
                finish(state.outputImagePath)
            } else {
                dismiss()
            }
        } catch (e: Exception) {
            logger.error("Failed to crop file", e)
        }
    }

    fun attachFile() {
        if (fileBrowser.isBrowserSupported()) {
            fileBrowser.launchBrowser { path -> path?.let { confirmFilePickerSelection(it) } }
        } else {
            val defaultPath = settingsRepo.settings.chatSettings.defaultSendDirectory ?: chatFileSystem.paths.picturesPath.toString()
            updateState {
                it.copy(
                    isFilePickerVisible = true,
                    defaultFilePickerDirectory = defaultPath,
                )
            }
        }
    }

    fun launchGallery() {
        fileBrowser.launchGallery(::setCropImagePath)
    }

    fun launchCamera() {
        val dir = chatFileSystem.paths.picturesPath.resolve(".tmp")
        fileSystem.createDirectories(dir)
        cameraLauncher.launchForPhoto(dir.resolve("${currentTimeMillis()}.jpg"))
    }

    fun launchAudioRecorder() {
        val dateTime = dateTimeFormatter.formatDateTime(currentTimeMillis()).replace(':', '_')
        updateState {
            it.copy(
                isRecordAudioDialogVisible = true,
                audioFilePath = chatFileSystem.getTempDirectory("audio").getFilePath("$dateTime.opus").toString(),
            )
        }
    }

    fun onAudioRecorded() {
        updateState { it.copy(isRecordAudioDialogVisible = false) }
        if (state.audioFilePath.isNotBlank()) {
            finish(state.audioFilePath)
        }
    }

    fun dismissAudioRecorder() {
        updateState { it.copy(isRecordAudioDialogVisible = false) }
        dismiss()
    }

    private fun setCropImagePath(path: Path?) {
        val pathString = path?.toString().orEmpty()
        updateState {
            it.copy(
                cropImagePath = pathString,
                outputImagePath = path?.run {
                    parent?.resolve("$nameWithoutExtension-cropped.jpg")?.toString()
                } ?: pathString,
            )
        }
    }

    private fun onEvent(event: ChatEvent) {
        when (event) {
            is TakePhotoEvent -> setCropImagePath(event.path)
            else -> return
        }
    }

    private fun finish(path: String) = actionData.tryEmit(ConfirmFileAction(path))

    private fun dismiss() = actionData.tryEmit(DismissAction)

    private fun updateState(block: (ChatFileSenderState) -> ChatFileSenderState) {
        stateData.tryEmit(block(state))
    }
}