package nand.apps.chat.ui.groups

import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.Logout
import androidx.compose.material.icons.filled.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import nand.apps.chat.model.call.CallData
import nand.apps.chat.model.call.getActions
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupRole
import nand.apps.chat.model.group.ChatGroupVisibility
import nand.apps.chat.model.message.UnreadMessageCounts
import nand.apps.chat.model.user.UserData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.ui.call.icon
import nand.apps.chat.ui.call.label
import nand.apps.chat.ui.contacts.ContactRow
import nand.apps.chat.ui.dialog.ContactAddressDialog
import nand.apps.chat.ui.dialog.PermanentActionDialog
import nand.apps.chat.ui.groups.settings.GroupSettingsDialog
import nand.apps.chat.ui.main.SidebarState
import nand.apps.chat.ui.menu.ActionMenuItem
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.res.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
@NonRestartableComposable
fun GroupContactRow(
    group: ChatGroupData,
    selfPeer: UserData,
    sidebarState: SidebarState,
    modifier: Modifier = Modifier,
    unreadCount: UnreadMessageCounts? = null,
    call: CallData? = null,
) {
    val eventHandler = LocalEventHandler.current
    var dialogType by rememberSaveable { mutableStateOf<GroupRowDialog?>(null) }
    ContactRow(
        heading = group.title.ifBlank { stringResource(Res.string.group_untitled) },
        subHeading = group.description.ifBlank {
            stringResource(Res.string.group_chat_peers_online, group.numPeersOnline, group.peers.size)
        },
        status = UserStatus.DEFAULT,
        isOnline = group.isOnline,
        iconOverlay = when {
            group.error != null -> Icons.Default.ErrorOutline
            group.isMuted -> Icons.Default.Snooze
            else -> null
        },
        iconOverlayColor = when {
            group.error != null -> ChatTheme.colors.warning
            else -> ChatTheme.colors.icon
        },
        avatar = group.avatar,
        unreadCount = unreadCount,
        sidebarState = sidebarState,
        onClick = { eventHandler += OpenConversationEvent(group.uid) },
        modifier = modifier,
        call = call,
        isActionMenuEnabled = true,
        actionMenuItems = {
            mutableListOf<ActionMenuItem>().apply {
                call?.takeIf { it.isInProgress }?.let { call ->
                    call.status.getActions().forEach { action ->
                        this += ActionMenuItem(
                            icon = action.icon,
                            label = getStringBlocking(action.label),
                            onClick = { eventHandler += action.getEvent(call.contactUid) },
                        )
                    }
                }
                if (selfPeer.role != ChatGroupRole.OBSERVER) {
                    this += ActionMenuItem(
                        icon = Icons.Default.Settings,
                        label = getStringBlocking(Res.string.group_settings),
                        onClick = { dialogType = GroupRowDialog.SETTINGS },
                    )
                }
                if (group.visibility == ChatGroupVisibility.PUBLIC) {
                    this += ActionMenuItem(
                        icon = Icons.Default.QrCode2,
                        label = getStringBlocking(Res.string.share),
                        onClick = { dialogType = GroupRowDialog.SHARE },
                    )
                }
                if (group.isMuted) {
                    this += ActionMenuItem(
                        icon = Icons.Default.NotificationsActive,
                        label = getStringBlocking(Res.string.unmute_conversation),
                        onClick = {
                            eventHandler += MuteContactEvent(group.uid, isMuted = false)
                        },
                    )
                } else {
                    this += ActionMenuItem(
                        icon = Icons.Default.NotificationsPaused,
                        label = getStringBlocking(Res.string.mute_conversation),
                        onClick = {
                            eventHandler += MuteContactEvent(group.uid, isMuted = true)
                        },
                    )
                }
                this += ActionMenuItem(
                    icon = Icons.Default.ClearAll,
                    label = getStringBlocking(Res.string.clear_messages),
                    onClick = { dialogType = GroupRowDialog.CLEAR_MESSAGES },
                )
                this += ActionMenuItem(
                    icon = Icons.AutoMirrored.Default.Logout,
                    label = getStringBlocking(Res.string.leave_group),
                    onClick = { dialogType = GroupRowDialog.REMOVE },
                )
            }
        }
    )
    when (dialogType) {
        GroupRowDialog.SETTINGS -> GroupSettingsDialog(
            group = group,
            selfPeer = selfPeer,
            onDismiss = { dialogType = null },
        )
        GroupRowDialog.SHARE -> ContactAddressDialog(
            address = group.uid.value,
            description = stringResource(Res.string.group_address_dialog_title),
            onClose = { dialogType = null },
        )
        GroupRowDialog.CLEAR_MESSAGES -> PermanentActionDialog(
            title = Res.string.clear_messages,
            message = stringResource(Res.string.clear_messages_confirmation),
            onConfirm = {
                dialogType = null
                eventHandler += ClearMessagesEvent(group.uid)
            },
            onCancel = { dialogType = null },
        )
        GroupRowDialog.REMOVE -> PermanentActionDialog(
            title = Res.string.leave_group,
            message = stringResource(Res.string.leave_group_confirmation, group.title),
            onConfirm = {
                dialogType = null
                // TODO - Add optional parting message input to dialog
                eventHandler += RemoveGroupEvent(
                    group = group,
                )
            },
            onCancel = { dialogType = null },
        )
        else -> {}
    }
}

private enum class GroupRowDialog {
    SETTINGS,
    SHARE,
    CLEAR_MESSAGES,
    REMOVE,
}