package nand.apps.chat.ui.groups

import androidx.compose.foundation.layout.*
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.SnackbarHostState
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.input.ImeAction
import nand.apps.chat.ui.contacts.ContactAddressInputField
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.groups.JoinGroupDialogAction.*
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.text.SimplePasswordField
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.text.takeIfUnderLength
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.LaunchedCollect
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.res.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun JoinGroupDialog(
    onDismiss: () -> Unit,
    viewModel: JoinGroupDialogViewModel = composableViewModel(),
) {
    var groupName by rememberTextFieldValue()
    var nickName by rememberTextFieldValue(viewModel.selfName)
    var address by rememberTextFieldValue()
    var password by rememberTextFieldValue()
    val errorSnackbars = remember { SnackbarHostState() }
    val successSnackbars = remember { SnackbarHostState() }
    val constraints = LocalChatEngineConstraints.current

    val addressRegex = remember(constraints) {
        Regex("[${constraints.uidCharacterRange}]+")
    }

    // View model action handler
    LaunchedCollect(viewModel.actionData) { action ->
        when (action) {
            is ExitDialog -> onDismiss()
            is ShowError -> {
                errorSnackbars.showSnackbar(getStringBlocking(Res.string.join_group_dialog_error_message))
            }
        }
    }

    SimpleDialog(
        title = Res.string.add_friend_dialog_title,
        isConfirmEnabled = addressRegex.matches(address.text) && nickName.text.isNotBlank(),
        onConfirm = {
            viewModel.joinGroup(
                groupName = groupName.text,
                address = address.text,
                nickName = nickName.text,
                password = password.text,
            )
        },
        onCancel = onDismiss,
    ) {
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
        ) {
            SimpleTextField(
                value = groupName,
                onValueChange = { groupName = it.takeIfUnderLength(constraints.maxGroupNameLength, groupName) },
                labelText = stringResource(Res.string.join_group_dialog_name_label),
                keyboardOptions = KeyboardOptions(
                    autoCorrectEnabled = false,
                    imeAction = ImeAction.Next
                ),
                modifier = Modifier.fillMaxWidth(),
            )
            ContactAddressInputField(
                address = address,
                onValueChange = { address = it },
                isPublicAddress = true,
                successSnackbar = successSnackbars,
                errorSnackbar = errorSnackbars,
            )
            SimplePasswordField(
                password = password,
                onValueChange = { password = it },
                modifier = Modifier.fillMaxWidth(),
            )
            SimpleTextField(
                value = nickName,
                onValueChange = { nickName = it.takeIfUnderLength(constraints.maxNameLength, nickName) },
                labelText = stringResource(Res.string.join_group_dialog_nickname_label),
                keyboardOptions = KeyboardOptions(
                    autoCorrectEnabled = false,
                    imeAction = ImeAction.Next
                ),
                modifier = Modifier.fillMaxWidth(),
            )
        }
        SimpleSnackbarHost(
            state = errorSnackbars,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier.align(Alignment.BottomCenter),
        )
        SimpleSnackbarHost(
            state = successSnackbars,
            backgroundColor = ChatTheme.colors.secondary,
            modifier = Modifier.align(Alignment.BottomCenter),
        )
    }
}