package nand.apps.chat.ui.login

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.rememberScrollState
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.foundation.verticalScroll
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.FolderOpen
import androidx.compose.material.icons.filled.Settings
import androidx.compose.material.icons.outlined.CrueltyFree
import androidx.compose.material.icons.outlined.DataArray
import androidx.compose.material.icons.outlined.Info
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusDirection
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.graphics.Brush
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.KeyboardType
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import nand.apps.chat.BuildConfig.MATRIX_ENGINE
import nand.apps.chat.model.engine.ChatEngineType
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.icons.Tox
import nand.apps.chat.ui.logo.SeersLogo
import nand.apps.chat.ui.logo.SeersLogoText
import nand.apps.chat.ui.modifier.appendIf
import nand.apps.chat.ui.modifier.bringIntoView
import nand.apps.chat.ui.navigation.MainRoute
import nand.apps.chat.ui.navigation.SettingsRoute
import nand.apps.chat.ui.profile.selector.ProfileSelectorDialog
import nand.apps.chat.ui.provider.LocalChatEngineConstraints
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.settings.AboutScreenRoute
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.snackbar.show
import nand.apps.chat.ui.text.*
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.LaunchedCollect
import nand.apps.chat.ui.util.rememberGridBrush
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun LoginScreen(
    onLoginComplete: () -> Unit,
    modifier: Modifier = Modifier,
    viewModel: LoginScreenViewModel = koinKmpViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    var server by rememberTextFieldValue(
        key = state.chatEngineType,
        initialValue = state.host.orEmpty(),
    )
    var name by rememberTextFieldValue(state.username)
    var password by rememberTextFieldValue()
    val focusManager = LocalFocusManager.current
    val engineConstraints = LocalChatEngineConstraints.current
    val navController = LocalNavController.current
    val settings = LocalSettingsProvider.current
    val hostFocus = remember { FocusRequester() }
    val nameFocus = remember { FocusRequester() }
    val passwordFocus = remember { FocusRequester() }
    val snackbarHostState = remember { SnackbarHostState() }
    val filePickerSnackbar = remember { SnackbarHostState() }

    val performLogin = {
        viewModel.performLogin(
            username = name.text,
            password = password.text,
            server = server.text,
        )
    }

    LaunchedCollect(viewModel.actionData) { action ->
        when (action) {
            is UpdateName -> {
                name = name.copy(text = action.name)
            }
            is ShowLoginError -> {
                snackbarHostState.showLoginError(action.error)
            }
            is ShowInvalidFileError -> {
                filePickerSnackbar.show(Res.string.login_error_please_select_valid_profile)
            }
            is ShowInvalidNameError -> {
                snackbarHostState.show(Res.string.login_error_invalid_profile_name)
            }
            is ShowImportFailedError -> {
                snackbarHostState.show(Res.string.login_error_import_failed)
            }
            is FinishLogin -> {
                onLoginComplete()
                navController.navigate(MainRoute)
            }
        }
    }
    LaunchedEffect(Unit) {
        viewModel.init()
    }
    Box(
        contentAlignment = Alignment.Center,
        modifier = modifier
            .background(
                brush = rememberGridBrush(
                    size = ChatTheme.dimens.iconXLarge,
                    backgroundColor = ChatTheme.colors.background,
                    foregroundColor = ChatTheme.colors.surface,
                )
            )
            .appendIf(!ChatTheme.colors.isLight) {
                background(
                    brush = Brush.radialGradient(
                        0.75f to Color.Transparent,
                        1f to Color.Black.copy(alpha = 0.25f),
                    ),
                )
            },
    ) {
        val scrollState = rememberScrollState()
        Box(
            contentAlignment = Alignment.TopCenter,
            modifier = Modifier
                .verticalScroll(scrollState)
                .windowInsetsPadding(WindowInsets.ime)
        ) {
            val logoSize = ChatTheme.dimens.imageThumbnail
            // Logo shadow
            Box(
                modifier = Modifier
                    .background(
                        brush = Brush.radialGradient(
                            0.9f to Color.Black.copy(alpha = 0.5f),
                            1f to Color.Transparent,
                        ),
                    )
                    .size(logoSize + ChatTheme.dimens.spacingMedium),
            )
            // Logo + fields
            Column(
                horizontalAlignment = Alignment.CenterHorizontally,
                verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                modifier = Modifier.padding(top = ChatTheme.dimens.spacingSmall),
            ) {
                if (getPlatform().isDesktop) {
                    LaunchedEffect(Unit) {
                        when {
                            engineConstraints.isHostRequired && server.text.isEmpty() -> hostFocus.requestFocus()
                            name.text.isEmpty() -> nameFocus.requestFocus()
                            password.text.isEmpty() -> passwordFocus.requestFocus()
                        }
                    }
                }
                SeersLogo(
                    modifier = Modifier.size(logoSize),
                )
                Spacer(
                    modifier = Modifier.height(ChatTheme.dimens.buttonMedium),
                )
                Column(
                    verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
                    modifier = Modifier
                        .background(
                            color = ChatTheme.colors.background,
                            shape = ChatTheme.shapes.small,
                        )
                        .padding(bottom = ChatTheme.dimens.spacingMedium)
                        .padding(horizontal = ChatTheme.dimens.spacingMedium)
                ) {
                    AnimatedVisibility(
                        visible = engineConstraints.isHostRequired,
                    ) {
                        SimpleTextField(
                            value = server,
                            onValueChange = { server = it },
                            labelText = stringResource(Res.string.server),
                            onSubmit = { focusManager.moveFocus(FocusDirection.Down) },
                            keyboardOptions = KeyboardOptions(
                                keyboardType = KeyboardType.Uri,
                                autoCorrectEnabled = false,
                                imeAction = ImeAction.Next
                            ),
                            modifier = Modifier
                                .focusRequester(hostFocus)
                                .bringIntoView(scrollState),
                        )
                    }
                    SimpleTextField(
                        value = name,
                        onValueChange = { name = it.takeIfUnderLength(engineConstraints.maxNameLength, name) },
                        labelText = stringResource(Res.string.username),
                        trailingIcon = {
                            IconButton(
                                icon = Icons.Filled.FolderOpen,
                                contentDescription = stringResource(Res.string.open_profile),
                                isEnabled = !state.isLoginBusy,
                                onClick = { viewModel.openProfileSelector() },
                            )
                        },
                        onSubmit = { focusManager.moveFocus(FocusDirection.Down) },
                        keyboardOptions = KeyboardOptions(
                            autoCorrectEnabled = false,
                            imeAction = ImeAction.Next
                        ),
                        isEnabled = !state.isLoginBusy,
                        modifier = Modifier
                            .focusRequester(nameFocus)
                            .bringIntoView(scrollState),
                    )
                    SimplePasswordField(
                        password = password,
                        onValueChange = { password = it },
                        onToggleVisibility = { passwordFocus.requestFocus() },
                        onSubmit = {
                            focusManager.clearFocus()
                            performLogin()
                        },
                        isEnabled = !state.isLoginBusy,
                        modifier = Modifier
                            .focusRequester(passwordFocus)
                            .bringIntoView(scrollState),
                    )
                }
                SimpleTextButton(
                    text = stringResource(Res.string.login),
                    isEnabled = !state.isLoginBusy && name.text.isNotBlank() &&
                            (!engineConstraints.isPasswordRequired || password.text.isNotBlank()) &&
                            (!engineConstraints.isHostRequired || server.text.isNotBlank()),
                    isLoading = state.isLoginBusy,
                    onClick = performLogin,
                )
            }
            // Text part of the logo
            SeersLogoText(
                modifier = Modifier.offset(y = logoSize - ChatTheme.dimens.spacingSmall),
            )
        }
        Box(
            modifier = Modifier
                .fillMaxSize()
                .padding(ChatTheme.dimens.spacingMedium)
                .windowInsetsPadding(WindowInsets.safeDrawing),
        ) {
            if (MATRIX_ENGINE && settings.devSettings.isMatrixChatEnabled) {
                IconButton(
                    icon = when (state.chatEngineType) {
                        ChatEngineType.TOX -> Icons.Outlined.Tox
                        ChatEngineType.MATRIX -> Icons.Outlined.DataArray
                        ChatEngineType.MOCK -> Icons.Outlined.CrueltyFree
                    },
                    contentDescription = when (state.chatEngineType) {
                        ChatEngineType.TOX -> stringResource(Res.string.login_chat_engine_tox)
                        ChatEngineType.MATRIX -> stringResource(Res.string.login_chat_engine_matrix)
                        ChatEngineType.MOCK -> "MOCKED"
                    },
                    size = ChatTheme.dimens.iconMedium,
                    onClick = { viewModel.toggleChatEngine() },
                    modifier = Modifier.align(Alignment.TopStart),
                )
            }
            IconButton(
                icon = Icons.Outlined.Info,
                contentDescription = stringResource(Res.string.about_screen_title),
                size = ChatTheme.dimens.iconMedium,
                isEnabled = !state.isLoginBusy,
                onClick = { navController.navigate(AboutScreenRoute) },
                modifier = Modifier.align(Alignment.BottomStart),
            )
            IconButton(
                icon = Icons.Default.Settings,
                contentDescription = stringResource(Res.string.settings),
                size = ChatTheme.dimens.iconMedium,
                isEnabled = !state.isLoginBusy,
                onClick = { navController.navigate(SettingsRoute) },
                modifier = Modifier.align(Alignment.BottomEnd),
            )
            SimpleSnackbarHost(
                state = snackbarHostState,
                backgroundColor = ChatTheme.colors.error,
                modifier = Modifier.align(Alignment.BottomCenter),
            )
        }
    }
    
    when {
        state.isProfileSelectorVisible -> ProfileSelectorDialog(
            onProfileSelected = viewModel::selectProfile,
            onDismiss = viewModel::dismissProfileSelector,
        )
        state.isRenameRequired -> SimpleDialog(
            title = Res.string.login_profile_rename_required_title,
            onConfirm = { viewModel.confirmRename(name.text) },
            onCancel = viewModel::cancelProfileImport,
        ) {
            Column(
                verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
            ) {
                Text(
                    text = stringResource(Res.string.login_profile_rename_required_message, state.username),
                    color = ChatTheme.colors.onSurface,
                )
                SimpleTextField(
                    value = name,
                    onValueChange = { name = it },
                    labelText = stringResource(Res.string.username),
                    onSubmit = { viewModel.confirmRename(it.text) },
                    keyboardOptions = KeyboardOptions(
                        autoCorrectEnabled = false,
                        imeAction = ImeAction.Done,
                    ),
                )
            }
        }
        state.isImportRequired -> SimpleDialog(
            title = Res.string.login_profile_import_required_title,
            onConfirm = viewModel::confirmProfileImport,
            onCancel = viewModel::cancelProfileImport,
        ) {
            Text(
                text = stringResource(Res.string.login_profile_import_required_message),
                color = ChatTheme.colors.onSurface,
            )
        }
        state.isPasswordWarningVisible -> {
            SimpleDialog(
                title = Res.string.login_new_profile_password_empty_title,
                onConfirm = performLogin,
                onCancel = viewModel::dismissNewProfileDialog,
            ) {
                Text(
                    text = stringResource(Res.string.login_new_profile_password_empty_message),
                    color = ChatTheme.colors.onSurface,
                )
            }
        }
        state.isProfileNew -> {
            var confirmedPassword by rememberTextFieldValue()
            val isConfirmEnabled = password.text.isEmpty() || password.text == confirmedPassword.text
            SimpleDialog(
                title = Res.string.login_new_profile_title,
                isConfirmEnabled = isConfirmEnabled,
                onConfirm = performLogin,
                onCancel = viewModel::dismissNewProfileDialog,
            ) {
                Column(
                    verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                ) {
                    Text(
                        text = stringResource(Res.string.login_new_profile_message, name.text),
                        color = ChatTheme.colors.onSurface,
                    )
                    if (password.text.isNotEmpty()) {
                        SimplePasswordField(
                            password = confirmedPassword,
                            label = Res.string.login_new_profile_confirm_password,
                            onValueChange = { confirmedPassword = it },
                            onSubmit = {
                                if (isConfirmEnabled) performLogin()
                            },
                        )
                    }
                }
            }
        }
    }
}