package nand.apps.chat.ui.login

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import nand.apps.chat.BuildConfig.MOCK_ENGINE
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.engine.ChatEngineType
import nand.apps.chat.io.*
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.profile.*
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.util.fifoSharedFlow
import nand.apps.chat.util.getHostName
import nand.apps.chat.util.takeIfNotBlank

private const val MOCK_PROFILE_NAME = "MOCK"

class LoginScreenViewModel(
    private val fileSystem: ChatFileSystem,
    private val engine: ChatEngine,
    private val settingsRepo: SettingsRepo,
    private val credentialsStore: ProfileCredentialsStore,
) : ViewModel() {

    private val profileDirectory = fileSystem.getProfileDirectory()
    private val defaultUsername = settingsRepo.settings.generalSettings.loginUsername
    private val defaultHost = settingsRepo.settings.generalSettings.loginHost
    
    val stateData = MutableStateFlow(
        LoginScreenState(
            username = defaultUsername,
            host = defaultHost,
            chatEngineType = settingsRepo.chatEngine,
        )
    )
    val actionData = fifoSharedFlow<LoginScreenAction>()
    
    suspend fun init() {
        when {
            MOCK_ENGINE || settingsRepo.chatEngine == ChatEngineType.MOCK -> {
                stateData.emit(stateData.value.copy(isLoginBusy = true))
                engine.initialize(
                    ChatProfileData(
                        filePath = fileSystem
                            .getProfileDirectory(MOCK_PROFILE_NAME)
                            .getDefaultProfilePath(ChatEngineType.MOCK)
                            .toString(),
                        username = MOCK_PROFILE_NAME,
                    )
                )
                stateData.emit(stateData.value.copy(isLoginBusy = false))
                actionData.emit(FinishLogin)
            }
            defaultUsername.isNotBlank() -> {
                val cachedPassword = credentialsStore.loadPassword()
                if (!cachedPassword.isNullOrBlank()) {
                    performLogin(
                        username = defaultUsername,
                        password = cachedPassword,
                        server = defaultHost.takeIf { engine.constraints.isHostRequired },
                        isAutomatic = true,
                    )
                }
            }
        }
    }
    
    fun performLogin(
        username: String,
        password: String,
        server: String?,
        isAutomatic: Boolean = false,
    ) {
        val host = server?.getHostName() ?: server
        val profilePath = if (engine.constraints.isHostRequired) {
            if (host.isNullOrBlank()) return
            profileDirectory.getSubDirectory("$username@$host".sanitizeFileName()).getFilePath(".profile")
        } else {
            if (!username.isValidFileName()) {
                if (isAutomatic) {
                    stateData.tryEmit(stateData.value.copy(username = ""))
                } else {
                    actionData.tryEmit(ShowInvalidNameError)
                }
                return
            }
            val dir = profileDirectory.getSubDirectory(username)
            val profile = dir.getPrimaryProfile() ?: dir.getDefaultProfilePath(stateData.value.chatEngineType)
            if (profile.name !in dir) {
                if (isAutomatic) return
                when {
                    !stateData.value.isProfileNew -> {
                        stateData.tryEmit(stateData.value.copy(isProfileNew = true))
                        return
                    }
                    password.isEmpty() && !stateData.value.isPasswordWarningVisible -> {
                        stateData.tryEmit(stateData.value.copy(isPasswordWarningVisible = true))
                        return
                    }
                }
            }
            profile
        }
        viewModelScope.launch {
            stateData.emit(
                stateData.value.copy(
                    isLoginBusy = true,
                    isProfileNew = false,
                    isPasswordWarningVisible = false,
                )
            )
            val result = engine.initialize(
                ChatProfileData(
                    username = username,
                    password = password,
                    host = host,
                    filePath = profilePath.toString(),
                    networkSettings = settingsRepo.settings.networkSettings,
                )
            )
            if (result == ChatEngineInitResult.SUCCESS) {
                settingsRepo.update {
                    generalSettings.copy(
                        loginUsername = username,
                        loginHost = host,
                    )
                }
                credentialsStore.savePassword(password)
                stateData.emit(stateData.value.copy(isLoginBusy = false))
                actionData.emit(FinishLogin)
            } else {
                stateData.emit(
                    stateData.value.copy(
                        isProfileNew = false,
                        isLoginBusy = false,
                    )
                )
                if (!isAutomatic) {
                    actionData.emit(ShowLoginError(result))
                }
            }
        }
    }

    fun openProfileSelector() {
        stateData.tryEmit(stateData.value.copy(isProfileSelectorVisible = true))
    }

    fun dismissProfileSelector() {
        stateData.tryEmit(stateData.value.copy(isProfileSelectorVisible = false))
    }
    
    fun dismissNewProfileDialog() {
        stateData.tryEmit(
            stateData.value.copy(
                isProfileNew = false,
                isPasswordWarningVisible = false,
            )
        )
    }
    
    // TODO - Remove username from file path in favor of hash or something
    fun selectProfile(path: String) {
        val selection = fileSystem.selectProfile(path) ?: run {
            actionData.tryEmit(ShowInvalidFileError)
            return
        }
        actionData.tryEmit(UpdateName(selection.username))
        stateData.tryEmit(
            stateData.value.copy(
                username = selection.username,
                isProfileSelectorVisible = false,
                filePath = selection.filePath.toString(),
                isImportRequired = selection.isImportRequired,
            )
        )
    }
    
    fun cancelProfileImport() {
        actionData.tryEmit(UpdateName(defaultUsername))
        stateData.tryEmit(
            stateData.value.copy(
                filePath = null,
                username = defaultUsername,
                isImportRequired = false,
                isRenameRequired = false,
            )
        )
    }
    
    fun confirmProfileImport() {
        val (result, outputFile) = fileSystem.importProfile(
            username = stateData.value.username,
            filePath = stateData.value.filePath.orEmpty(),
        )
        when (result) {
            ProfileImportResult.IMPORT_SUCCESSFUL -> {
                stateData.tryEmit(
                    stateData.value.copy(
                        filePath = outputFile.toString(),
                        isImportRequired = false,
                    )
                )
            }
            ProfileImportResult.PROFILE_EXISTS -> {
                stateData.tryEmit(stateData.value.copy(isRenameRequired = true))
            }
            ProfileImportResult.INVALID_FILE -> {
                cancelProfileImport()
                actionData.tryEmit(ShowInvalidFileError)
            }
        }
    }
    
    fun confirmRename(username: String) {
        if (!username.isValidFileName()) {
            actionData.tryEmit(ShowInvalidNameError)
            return
        }
        stateData.tryEmit(
            stateData.value.copy(
                username = username,
                isRenameRequired = false,
            )
        )
        confirmProfileImport()
    }

    fun toggleChatEngine() {
        val chatEngine = when (stateData.value.chatEngineType) {
            ChatEngineType.TOX -> ChatEngineType.MATRIX
            else -> ChatEngineType.TOX
        }
        settingsRepo.update { generalSettings.copy(chatEngine = chatEngine) }
        stateData.tryEmit(
            stateData.value.copy(
                chatEngineType = chatEngine,
                host = defaultHost?.takeIfNotBlank() ?: engine.hostInfo?.defaultHost.orEmpty()
            )
        )
    }
}