package nand.apps.chat.ui.logo

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.animation.core.*
import androidx.compose.animation.fadeIn
import androidx.compose.animation.fadeOut
import androidx.compose.foundation.Image
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.shape.CircleShape
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.*
import androidx.compose.ui.geometry.Size
import androidx.compose.ui.graphics.*
import androidx.compose.ui.graphics.vector.PathBuilder
import androidx.compose.ui.layout.onSizeChanged
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.unit.*
import nand.apps.chat.model.settings.theme.THEME_HUE_RANGE
import nand.apps.chat.ui.image.PlatformImage
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.toColorFilter
import nand.apps.chat.ui.util.toDp
import nand.apps.chat.ui.util.toPath
import org.jetbrains.compose.resources.painterResource
import seers.composeapp.generated.resources.*
import kotlin.math.min

private val LOGO_SIZE = 256.dp
val LOGO_COLOR = Color(0xFFFFDC53)

@Composable
fun SeersLogo(
    modifier: Modifier = Modifier,
    color: Color = LOGO_COLOR,
    rotationSpeed: Float = 0f,
    isGlowing: Boolean = false,
    onClick: (() -> Unit)? = null,
) {
    var boxSize by remember { mutableStateOf(0) }
    Box(
        contentAlignment = Alignment.Center,
        modifier = modifier.onSizeChanged {
            boxSize = min(it.width, it.height)
        },
    ) {
        val hue = LocalSettingsProvider.current.themeSettings.baseHue
        val isLightTheme = ChatTheme.colors.isLight
        val backgroundTint = remember(hue, isLightTheme) {
            when {
                hue in THEME_HUE_RANGE -> Color.hsv(
                    hue = hue,
                    saturation = if (isLightTheme) 0.4f else 0.75f,
                    value = 1f,
                )
                else -> null
            }?.toColorFilter()
        }
        val infiniteTransition = rememberInfiniteTransition()
        val ridgeRotation = rotationSpeed.takeIf { it > 0f }?.let {
            infiniteTransition.animateFloat(
                initialValue = 0f,
                targetValue = 360f,
                animationSpec = infiniteRepeatable(
                    animation = tween(10000, easing = LinearEasing),
                    repeatMode = RepeatMode.Restart,
                ),
                label = "RidgeRotate",
            )
        }
        Image(
            painter = painterResource(Res.drawable.ic_seers_base),
            contentDescription = null,
            colorFilter = backgroundTint,
            modifier = Modifier.fillMaxSize(),
        )
        Image(
            painter = painterResource(Res.drawable.ic_seers_ridges),
            contentDescription = null,
            colorFilter = backgroundTint,
            modifier = Modifier
                .fillMaxSize()
                .rotate(rotationSpeed * (ridgeRotation?.value ?: 0f)),
        )
        val colorFilter = remember(color) { color.toColorFilter() }
        PlatformImage(
            resource = Res.drawable.ic_seers_s,
            contentDescription = null,
            colorFilter = colorFilter,
            modifier = Modifier.fillMaxSize(),
        )
        if (boxSize > 0) {
            onClick?.let {
                SeersLogoClickOverlay(
                    boxSize = boxSize,
                    onClick = it,
                )
            }
            AnimatedVisibility(
                visible = isGlowing,
                enter = fadeIn(),
                exit = fadeOut(),
            ) {
                SeersLogoGlowOverlay(
                    boxSize = boxSize,
                    color = color,
                )
            }
        }
    }
}

@Composable
private fun SeersLogoClickOverlay(
    boxSize: Int,
    onClick: () -> Unit,
) {
    val size = boxSize.toDp(LocalDensity.current)
    val scale = size / LOGO_SIZE
    val logoShape = remember(scale) { LogoMaskShape(scale) }
    Box(
        modifier = Modifier
            .size(size)
            .clip(logoShape)
            .onClick(onClick = onClick),
    )
}

@Composable
private fun SeersLogoGlowOverlay(
    boxSize: Int,
    color: Color,
) {
    val infiniteTransition = rememberInfiniteTransition()
    val size = boxSize.toDp(LocalDensity.current)
    val glowAmount by infiniteTransition.animateFloat(
        initialValue = 0f,
        targetValue = 1f,
        animationSpec = infiniteRepeatable(
            animation = tween(1000, easing = FastOutSlowInEasing),
            repeatMode = RepeatMode.Reverse,
        ),
        label = "Glow",
    )
    Box(
        modifier = Modifier
            .size(size)
            .alpha(glowAmount * 0.7f)
            .background(
                brush = Brush.radialGradient(
                    (glowAmount * 0.85f) to color,
                    0.95f to color.copy(alpha = 0f),
                ),
                shape = CircleShape,
            ),
    )
}

private class LogoMaskShape(
    private val scale: Float,
) : Shape {

    override fun createOutline(size: Size, layoutDirection: LayoutDirection, density: Density): Outline {
        return getMaskOutline(scale * density.density)
    }
}

private fun getMaskOutline(scale: Float): Outline {
    val matrix = Matrix()
    matrix.scale(scale, scale)
    return Outline.Generic(OUTLINE_PATH.toPath().apply { transform(matrix) })
}

private val OUTLINE_PATH = PathBuilder().apply {
    moveTo(128.46684f, 40.641186f)
    quadToRelative(11.43765f, 0f, 11.43765f, 15.405801f)
    verticalLineToRelative(6.06895f)
    quadToRelative(7.93633f, 1.633982f, 10.27054f, 2.334204f)
    lineToRelative(3.2679f, 1.167134f)
    quadToRelative(0.23341f, 0.233504f, 2.1008f, 0.700256f)
    quadToRelative(3.50132f, -1.63395f, 6.5358f, -1.63395f)
    quadToRelative(11.67107f, 0f, 11.67107f, 15.4058f)
    verticalLineToRelative(7.469493f)
    quadToRelative(0f, 7.469492f, -2.56763f, 11.437678f)
    quadToRelative(-2.33422f, 3.968157f, -9.10344f, 3.968157f)
    quadToRelative(-10.03712f, 0f, -12.13792f, -9.803701f)
    quadToRelative(-0.46684f, -3.26793f, -6.53579f, -5.835542f)
    quadToRelative(-5.83554f, -2.567644f, -13.07161f, -2.567644f)
    quadToRelative(-10.50396f, 0f, -16.80634f, 3.968154f)
    quadToRelative(-6.30237f, 3.968154f, -6.30237f, 10.503982f)
    quadToRelative(0f, 9.570291f, 17.5066f, 13.071601f)
    lineToRelative(15.17239f, 3.26789f)
    quadToRelative(5.36869f, 1.16711f, 8.40317f, 2.1008f)
    quadToRelative(3.03448f, 0.70029f, 9.57028f, 3.73475f)
    quadToRelative(6.76922f, 2.80105f, 10.50397f, 6.53576f)
    quadToRelative(3.73474f, 3.50134f, 7.00264f, 10.50399f)
    quadToRelative(3.2679f, 6.76923f, 3.2679f, 15.40583f)
    quadToRelative(0f, 15.17239f, -10.27054f, 25.90978f)
    quadToRelative(-10.27055f, 10.73739f, -28.47742f, 14.2387f)
    verticalLineToRelative(19.37398f)
    quadToRelative(0f, 7.4695f, -2.56763f, 11.43764f)
    quadToRelative(-2.33422f, 4.20156f, -8.87002f, 4.20156f)
    quadToRelative(-7.00264f, 0f, -9.33686f, -4.20156f)
    quadToRelative(-2.33421f, -3.96814f, -2.33421f, -11.43764f)
    verticalLineToRelative(-19.37398f)
    quadToRelative(-10.03712f, -2.1008f, -18.44029f, -5.83554f)
    quadToRelative(-4.90185f, 4.20159f, -9.10343f, 4.20159f)
    quadToRelative(-11.9045f, 0f, -11.9045f, -15.63924f)
    verticalLineToRelative(-14.2387f)
    quadToRelative(0f, -15.63924f, 11.67107f, -15.63924f)
    quadToRelative(1.86737f, 0f, 3.2679f, 0.46682f)
    quadToRelative(1.63395f, 0.46688f, 2.56764f, 0.93369f)
    quadToRelative(1.16711f, 0.46685f, 2.10079f, 1.63395f)
    quadToRelative(0.93369f, 0.9337f, 1.40053f, 1.63398f)
    quadToRelative(0.70027f, 0.70023f, 1.16711f, 2.10077f)
    quadToRelative(0.46684f, 1.1671f, 0.70026f, 1.86739f)
    quadToRelative(0.23341f, 0.46681f, 0.46684f, 1.63391f)
    lineToRelative(0.23341f, 1.16714f)
    quadToRelative(1.40053f, 5.36869f, 9.33686f, 9.33685f)
    quadToRelative(8.16975f, 3.73475f, 17.74002f, 3.73475f)
    quadToRelative(11.43766f, 0f, 18.90714f, -5.13526f)
    quadToRelative(7.46949f, -5.13529f, 7.46949f, -12.60478f)
    quadToRelative(0f, -6.06895f, -4.43501f, -9.57026f)
    quadToRelative(-4.43501f, -3.73474f, -14.23871f, -5.60213f)
    lineToRelative(-17.27318f, -3.50131f)
    quadToRelative(-35.48005f, -7.23608f, -35.48005f, -36.18033f)
    quadToRelative(0f, -13.771882f, 9.33685f, -23.808987f)
    quadToRelative(9.57028f, -10.270544f, 24.50925f, -12.838188f)
    verticalLineToRelative(-6.06895f)
    quadToRelative(0f, -7.469492f, 2.33422f, -11.437647f)
    quadToRelative(2.56763f, -3.968154f, 9.33685f, -3.968154f)
    close()
}