package nand.apps.chat.ui.main

import androidx.compose.animation.*
import androidx.compose.foundation.layout.WindowInsets
import androidx.compose.foundation.layout.safeDrawing
import androidx.compose.foundation.layout.windowInsetsPadding
import androidx.compose.runtime.Composable
import androidx.compose.runtime.CompositionLocalProvider
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.ui.Modifier
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import nand.apps.chat.event.VisibleMessageTracker
import nand.apps.chat.model.event.*
import nand.apps.chat.ui.chat.message.input.MessageInputViewModel
import nand.apps.chat.ui.chat.screen.ChatScreenViewModel
import nand.apps.chat.ui.login.LoginScreen
import nand.apps.chat.ui.navigation.*
import nand.apps.chat.ui.provider.LocalEventHandler
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.provider.LocalNavRoute
import nand.apps.chat.ui.settings.SettingsNavGraph
import nand.apps.chat.ui.tutorial.TutorialNavGraph
import nand.apps.chat.ui.util.collect
import org.koin.compose.getKoin

@Composable
fun MainNavGraph(
    startRoute: NavRoute,
    modifier: Modifier = Modifier,
) {
    val navController = rememberNavController()
    val eventHandler = LocalEventHandler.current
    val visibleMessageTracker: VisibleMessageTracker by getKoin().inject()

    // Global view models
    // Declared here to prevent disposal when switching screens
    val chatScreenViewModel: ChatScreenViewModel = koinKmpViewModel()
    val messageInputViewModel: MessageInputViewModel = koinKmpViewModel()

    NavHost(
        navController = navController,
        startRoute = startRoute,
    ) { route ->
        LaunchedEffect(Unit) {
            collect(eventHandler.onEvent) { event ->
                // Navigate back to the main screen when we open a conversation
                if (event is OpenConversationEvent && route != MainRoute && MainRoute in navController.getRouteList()) {
                    navController.navigate(MainRoute, reset = true)
                }
            }
        }
        visibleMessageTracker.onNavRoute(route)
        CompositionLocalProvider(
            LocalNavController.provides(navController),
            LocalNavRoute.provides(route),
        ) {
            // Main chat window
            AnimatedVisibility(
                visible = route == MainRoute,
                enter = fadeIn(),
                exit = fadeOut(),
            ) {
                MainScreen(
                    chatScreenViewModel = chatScreenViewModel,
                    messageInputViewModel = messageInputViewModel,
                    modifier = modifier.windowInsetsPadding(WindowInsets.safeDrawing),
                )
            }

            // Profile selector and login
            AnimatedVisibility(
                visible = route == LoginRoute,
                enter = EnterTransition.None,
                exit = fadeOut(),
            ) {
                LoginScreen(
                    modifier = modifier,
                    onLoginComplete = { eventHandler += LoginCompleteEvent },
                )
            }

            // Settings pages
            AnimatedVisibility(
                visible = route.value.startsWith(SettingsRoute.value),
                enter = slideInVertically(initialOffsetY = { it }),
                exit = slideOutVertically(targetOffsetY = { it }),
            ) {
                SettingsNavGraph(
                    route = route,
                    modifier = modifier,
                )
            }
            
            // Tutorial
            AnimatedVisibility(
                visible = route.value.startsWith(TutorialRoute.value),
                enter = EnterTransition.None,
                exit = fadeOut(),
            ) {
                TutorialNavGraph(modifier)
            }
        }
    }
}