package nand.apps.chat.ui.profile.export

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.*
import nand.apps.chat.model.event.SuccessSnackbarEvent
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.res.FormattedStringResource
import nand.apps.chat.util.fifoSharedFlow
import okio.FileSystem
import okio.Path.Companion.toPath
import seers.composeapp.generated.resources.*

class ProfileExportDialogViewModel(
    private val filePaths: ChatFilePaths,
    private val fileSystem: FileSystem,
    private val selfProfileRepo: SelfProfileRepo,
    private val eventHandler: ChatEventHandler,
) : ViewModel() {
    
    val stateData = MutableStateFlow(ProfileExportDialogState())
    val actionData = fifoSharedFlow<ProfileExportDialogAction>()
    
    fun setExportType(exportType: ProfileExportType) {
        stateData.tryEmit(stateData.value.copy(exportType = exportType))
    }
    
    fun openFilePicker() {
        val isDesktop = getPlatform().isDesktop
        val exportDirectory = ChatDirectory(
            path = when {
                isDesktop -> filePaths.documentPath
                else -> filePaths.downloadPath
            },
            fileSystem = fileSystem,
        )
        val profilePath = selfProfileRepo.profilePath
        val profileName = selfProfileRepo.profileName
        val fileName = exportDirectory.resolveDuplicateFileName(
            fileName = when (stateData.value.exportType) {
                ProfileExportType.PROFILE_ONLY -> "$profileName.${profilePath.extension}"
                ProfileExportType.PROFILE_FULL -> "$profileName.$FULL_PROFILE_EXTENSION"
            }
        )
        val exportPath = exportDirectory.getFilePath(fileName)

        if (isDesktop) {
            stateData.tryEmit(stateData.value.copy(exportPath = exportPath))
        } else {
            exportProfile(exportPath.toString())
            eventHandler += SuccessSnackbarEvent(
                FormattedStringResource(Res.string.file_export_to_downloads_complete, fileName)
            )
            actionData.tryEmit(DismissDialogAction)
        }
    }
    
    fun closeFilePicker() = stateData.tryEmit(stateData.value.copy(exportPath = null))
    
    fun exportProfile(outputPath: String) {
        val profilePath = selfProfileRepo.profilePath
        stateData.tryEmit(
            stateData.value.copy(
                exportPath = null,
                isExportBusy = true,
            )
        )
        viewModelScope.launch {
            withContext(Dispatchers.IO) {
                when (stateData.value.exportType) {
                    ProfileExportType.PROFILE_ONLY -> {
                        fileSystem.copy(profilePath, outputPath.toPath())
                    }
                    ProfileExportType.PROFILE_FULL -> {
                        zipDirectory(profilePath.parent!!, outputPath.toPath())
                    }
                }
            }
            stateData.emit(stateData.value.copy(isExportBusy = false))
            actionData.emit(DismissDialogAction)
        }
    }
}