package nand.apps.chat.ui.provider

import androidx.compose.foundation.isSystemInDarkTheme
import androidx.compose.foundation.layout.BoxScope
import androidx.compose.runtime.*
import androidx.compose.ui.Modifier
import androidx.compose.ui.layout.onSizeChanged
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.unit.Density
import androidx.compose.ui.unit.IntSize
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.model.settings.DENSITY_RANGE
import nand.apps.chat.model.settings.SettingsData
import nand.apps.chat.model.settings.theme.ChatThemeSetting
import nand.apps.chat.time.DateTimeFormatter
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.share.ContentShare
import nand.apps.chat.ui.theme.*
import nand.apps.chat.util.clamp
import org.koin.compose.getKoin

@Composable
fun ChatAppSurface(
    settings: SettingsData,
    modifier: Modifier = Modifier,
    includeBackground: Boolean = true,
    theme: ChatThemeSetting = settings.themeSettings.theme,
    content: @Composable BoxScope.() -> Unit,
) {
    val chatEngine: ChatEngine by getKoin().inject()
    val eventHandler: ChatEventHandler by getKoin().inject()
    val contentShare: ContentShare by getKoin().inject()
    val dateTimeFormatter: DateTimeFormatter by getKoin().inject()
    val coroutineScope = rememberCoroutineScope()

    // Responsive design
    var surfaceSize by remember { mutableStateOf<IntSize?>(null) }

    val defaultDensity = LocalDensity.current
    val isDarkTheme = when (theme) {
        ChatThemeSetting.SYSTEM -> isSystemInDarkTheme()
        ChatThemeSetting.DARK -> true
        ChatThemeSetting.LIGHT -> false
    }
    val themeColors = remember(isDarkTheme, settings.themeSettings) {
        when {
            isDarkTheme -> generateDarkTheme(settings.themeSettings)
            else -> generateLightTheme(settings.themeSettings)
        }
    }
    ChatThemeSurface(
        colors = themeColors,
        includeBackground = includeBackground,
        modifier = modifier
            .onSizeChanged { size -> surfaceSize = size }
    ) {
        val isPortrait = surfaceSize?.run { isPortraitMode(width, height) } == true
        CompositionLocalProvider(
            LocalCoroutineScope provides coroutineScope,
            LocalDateTimeFormatProvider provides dateTimeFormatter,
            LocalChatEngineConstraints provides chatEngine.constraints,
            LocalPortraitModeProvider provides isPortrait,
            LocalSettingsProvider provides settings,
            LocalEventHandler provides eventHandler,
            LocalContentShare provides contentShare,
            LocalDensity provides Density(defaultDensity.density * settings.densityScale),
            content = { content() },
        )
    }
}

private val SettingsData.densityScale get() = generalSettings.densityScale.clamp(DENSITY_RANGE)