package nand.apps.chat.ui.settings

import androidx.compose.animation.animateColor
import androidx.compose.animation.core.*
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.rememberScrollState
import androidx.compose.foundation.text.selection.SelectionContainer
import androidx.compose.foundation.verticalScroll
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalUriHandler
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.BuildConfig.APP_VERSION
import nand.apps.chat.BuildConfig.COMMIT_HASH
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.logo.LOGO_COLOR
import nand.apps.chat.ui.logo.SeersLogo
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.toolbar.SimpleNavToolbar
import nand.apps.chat.ui.util.rememberGridBrush
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

private const val WEBSITE_URL = "https://seers.chat/"

@Composable
fun AboutScreen(
    modifier: Modifier = Modifier,
) = Column(modifier = modifier) {
    
    val chatEngine: ChatEngine by getKoin().inject()
    val uriHandler = LocalUriHandler.current
    
    SimpleNavToolbar(
        text = Res.string.about_screen_title,
    )
    Box(
        contentAlignment = Alignment.Center,
        modifier = Modifier
            .background(
                brush = rememberGridBrush(
                    size = ChatTheme.dimens.iconXLarge,
                    backgroundColor = ChatTheme.colors.background,
                    foregroundColor = ChatTheme.colors.surface,
                )
            )
            .fillMaxSize()
    ) {
        Column(
            horizontalAlignment = Alignment.CenterHorizontally,
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
            modifier = Modifier.verticalScroll(rememberScrollState()),
        ) {
            val themeHue = LocalSettingsProvider.current.themeSettings.baseHue
            var isLogoActivated by remember { mutableStateOf(false) }
            val infiniteTransition = rememberInfiniteTransition()
            val logoColor = if (isLogoActivated) {
                infiniteTransition.animateColor(
                    initialValue = Color.Red,
                    targetValue = Color.Green,
                    animationSpec = infiniteRepeatable(tween(1500, easing = LinearEasing), RepeatMode.Reverse),
                )
            } else {
                null
            }
            SeersLogo(
                color = logoColor?.value ?: LOGO_COLOR,
                isGlowing = isLogoActivated,
                modifier = Modifier.size(ChatTheme.dimens.imageThumbnail),
                onClick = if (!ChatTheme.colors.isLight && themeHue in 235f..245f) {
                    { isLogoActivated = true }
                } else {
                    null
                },
            )
            TextLine(
                text = APP_NAME,
                style = ChatTheme.typography.h4,
                color = ChatTheme.colors.onSurface,
            )
            SelectionContainer(
                modifier = Modifier.padding(bottom = ChatTheme.dimens.spacingMedium),
            ) {
                Column(horizontalAlignment = Alignment.CenterHorizontally) {
                    TextLine(
                        text = stringResource(Res.string.about_screen_app_version, APP_VERSION, COMMIT_HASH),
                        style = ChatTheme.typography.body2,
                        color = ChatTheme.colors.onBackground,
                    )
                    TextLine(
                        text = chatEngine.version,
                        style = ChatTheme.typography.body2,
                        color = ChatTheme.colors.onBackground,
                    )
                    TextLine(
                        text = stringResource(Res.string.about_screen_author),
                        style = ChatTheme.typography.body2,
                        color = ChatTheme.colors.onBackground,
                    )
                }
            }
            SimpleTextButton(
                text = stringResource(Res.string.about_screen_website),
                onClick = { uriHandler.openUri(WEBSITE_URL) },
            )
        }
    }
}