package nand.apps.chat.ui.settings

import androidx.compose.foundation.layout.ColumnScope
import androidx.compose.foundation.layout.PaddingValues
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.Close
import androidx.compose.material.icons.outlined.Search
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import androidx.compose.ui.focus.FocusRequester
import androidx.compose.ui.focus.focusRequester
import androidx.compose.ui.unit.dp
import androidx.compose.ui.util.fastFlatMap
import androidx.compose.ui.util.fastMap
import androidx.compose.ui.util.fastMapNotNull
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.IO
import kotlinx.coroutines.withContext
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.navigation.BackHandler
import nand.apps.chat.ui.settings.item.SettingsCategory
import nand.apps.chat.ui.settings.item.SettingsItem
import nand.apps.chat.ui.settings.item.SettingsScreenState
import nand.apps.chat.ui.settings.item.SettingsSeparatorItem
import nand.apps.chat.ui.settings.search.SearchSettingsItem
import nand.apps.chat.ui.settings.search.SettingsSearchMatches
import nand.apps.chat.ui.settings.search.SettingsSearchResult
import nand.apps.chat.ui.shortcut.ShortcutHandler
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.findMatchRanges
import nand.apps.chat.res.getStringBlocking
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun SettingsCategoryScreen(
    category: SettingsCategory,
    viewModel: SettingsViewModel,
    modifier: Modifier = Modifier,
    content: (@Composable ColumnScope.() -> Unit)? = null,
) {
    var isSearchActive by rememberSaveable { mutableStateOf(false) }
    var searchTerms by rememberTextFieldValue(key = isSearchActive)
    val items = remember { mutableStateListOf<SettingsItem>() }
    val state = LocalSettingsScreenProvider.current
    LaunchedEffect(category, state, isSearchActive, searchTerms.text) {
        val terms = searchTerms.text
        val filtered = mutableListOf<SettingsItem>()
        if (isSearchActive && terms.isNotBlank()) {
            withContext(Dispatchers.IO) {
                filtered.addAll(category.search(state, terms))
            }
        } else {
            filtered.addAll(category.getItems(state))
        }
        items.clear()
        items.addAll(filtered)
    }
    LaunchedEffect(category) {
        isSearchActive = false
    }
    ShortcutHandler(
        isEnabled = !isSearchActive,
        action = ShortcutAction.TOGGLE_SEARCH,
        onShortcut = { isSearchActive = true },
    )
    BackHandler(
        isEnabled = isSearchActive,
        onBack = { isSearchActive = false },
    )
    SettingsColumn(
        title = when {
            isSearchActive -> null
            else -> category.title
        },
        modifier = modifier,
        isScrollable = false,
        contentPadding = 0.dp,
        endContent = {
            if (isSearchActive) {
                val focusRequester = remember { FocusRequester() }
                LaunchedEffect(Unit) {
                    focusRequester.requestFocus()
                }
                SimpleTextField(
                    value = searchTerms,
                    onValueChange = { searchTerms = it },
                    hintText = stringResource(Res.string.settings_search),
                    modifier = Modifier
                        .weight(1f)
                        .focusRequester(focusRequester),
                    trailingIcon = {
                        IconButton(
                            icon = Icons.Outlined.Close,
                            contentDescription = stringResource(Res.string.cancel),
                            onClick = { isSearchActive = false },
                        )
                    }
                )
            } else {
                IconButton(
                    icon = Icons.Outlined.Search,
                    contentDescription = stringResource(Res.string.settings_search),
                    size = ChatTheme.dimens.iconMedium,
                    onClick = { isSearchActive = true },
                )
            }
        },
    ) {
        LazyColumn(
            contentPadding = PaddingValues(ChatTheme.dimens.spacingLarge),
        ) {
            content?.let {
                item { content() }
            }
            items(
                items = items,
            ) { item ->
                Row(
                    modifier = Modifier
                        .fillMaxWidth()
                        .animateItem()
                ) {
                    item.GetComposable(viewModel)
                }
            }
        }
    }
}

private fun SettingsCategory.search(
    state: SettingsScreenState,
    terms: String,
): List<SettingsItem> {
    val regex = try {
        Regex(Regex.escape(terms), RegexOption.IGNORE_CASE)
    } catch (ignore: Exception) {
        return emptyList()
    }
    return getItems(state).fastFlatMap { item ->
        if (item is SettingsCategory) {
            listOf(item) + item.getItems(state)
        } else {
            listOf(item)
        }
    }.fastMapNotNull { item ->
        val title = getStringBlocking(item.title)
        val desc = item.description?.let { getStringBlocking(it) }
        when {
            item is SettingsSeparatorItem -> null
            title.startsWith(terms, ignoreCase = true) -> SettingsSearchResult(
                item = item,
                score = 3,
                matches = SettingsSearchMatches(titleMatches = listOf(terms.indices)),
            )
            terms.length > 1 -> when {
                title.contains(terms, ignoreCase = true) -> SettingsSearchResult(
                    item = item,
                    score = 2,
                    matches = SettingsSearchMatches(titleMatches = title.findMatchRanges(regex)),
                )
                desc != null && desc.contains(terms, ignoreCase = true) -> SettingsSearchResult(
                    item = item,
                    score = 1,
                    matches = SettingsSearchMatches(descMatches = desc.findMatchRanges(regex)),
                )
                else -> null
            }
            else -> null
        }
    }.sortedByDescending {
        it.score
    }.fastMap { result ->
        SearchSettingsItem(
            item = result.item,
            matches = result.matches,
        )
    }
}