package nand.apps.chat.ui.settings

import androidx.compose.runtime.Composable
import androidx.compose.runtime.CompositionLocalProvider
import androidx.compose.runtime.remember
import androidx.compose.ui.Modifier
import nand.apps.chat.ui.navigation.NavRoute
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isLoggedIn
import nand.apps.chat.ui.settings.dev.DevSettingsCategory
import nand.apps.chat.ui.settings.dev.DevSettingsScreen
import nand.apps.chat.ui.settings.embed.EmbedHostsScreen
import nand.apps.chat.ui.settings.filetransfer.FileTransferAutoAcceptScreen
import nand.apps.chat.ui.settings.filetransfer.FileTransferUploadConfigScreen
import nand.apps.chat.ui.settings.item.SettingsCategory
import nand.apps.chat.ui.settings.item.SettingsScreenState
import nand.apps.chat.ui.settings.main.MainSettingsCategory
import nand.apps.chat.ui.settings.net.tox.ToxNodeSettingsScreen
import nand.apps.chat.ui.settings.shortcut.ShortcutSettingsScreen
import nand.apps.chat.ui.viewmodel.composableViewModel
import nand.apps.chat.update.AppInstaller
import org.koin.compose.getKoin

@Composable
fun SettingsNavGraph(
    route: NavRoute,
    modifier: Modifier = Modifier,
) {
    val viewModel: SettingsViewModel = composableViewModel()
    val installer: AppInstaller by getKoin().inject()
    val settings = LocalSettingsProvider.current
    val isLoggedIn = isLoggedIn()
    val state = remember(settings, isLoggedIn, installer.isSupported()) {
        SettingsScreenState(
            isLoggedIn = isLoggedIn,
            isInstallerSupported = installer.isSupported(),
            chatEngine = settings.generalSettings.chatEngine,
        )
    }
    val mainCategory = remember { MainSettingsCategory() }
    val categories = remember(state) {
        buildList {
            add(mainCategory)
            addAll(mainCategory.getItems(state).filterIsInstance<SettingsCategory>())
        }.associateBy { it.route }
    }
    CompositionLocalProvider(
        LocalSettingsScreenProvider provides state
    ) {
        categories[route]?.let { category ->
            when (category) {
                is DevSettingsCategory -> DevSettingsScreen(
                    category = category,
                    viewModel = viewModel,
                    modifier = modifier,
                )
                else -> SettingsCategoryScreen(
                    category = category,
                    viewModel = viewModel,
                    modifier = modifier,
                )
            }
        }
        when (route) {
            EmbedHostsRoute -> EmbedHostsScreen(
                viewModel = viewModel,
                modifier = modifier,
            )
            FileTransferAutoAcceptRoute -> FileTransferAutoAcceptScreen(
                viewModel = viewModel,
                modifier = modifier,
            )
            FileTransferUploadConfigRoute -> FileTransferUploadConfigScreen(
                viewModel = viewModel,
                modifier = modifier,
            )
            ToxNodeSettingsRoute -> ToxNodeSettingsScreen(
                modifier = modifier,
            )
            ShortcutSettingsRoute -> ShortcutSettingsScreen(
                viewModel = viewModel,
                modifier = modifier,
            )
            AboutScreenRoute -> AboutScreen(
                modifier = modifier,
            )
        }
    }
}