package nand.apps.chat.ui.settings.dialog

import androidx.compose.foundation.layout.*
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Modifier
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageAlignment
import nand.apps.chat.model.user.UserData
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.ui.chat.message.CompactMessageBlock
import nand.apps.chat.ui.chat.message.MessageAligner
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.provider.isLoggedIn
import nand.apps.chat.ui.settings.SettingsViewModel
import nand.apps.chat.ui.text.SimpleTextRadioButton
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.stringArrayResource
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun ChatMessageAlignmentDialog(
    viewModel: SettingsViewModel,
    onConfirm: () -> Unit,
) {
    val selfProfileRepo: SelfProfileRepo by getKoin().inject()
    val settings = LocalSettingsProvider.current.chatSettings
    var alignment by rememberSaveable(settings) { mutableStateOf(settings.messageAlignment) }
    val selfMessage = stringArrayResource(Res.array.default_status_messages).random()
    val otherMessage = stringArrayResource(Res.array.add_friend_dialog_message_defaults).random()
    SimpleDialog(
        title = Res.string.settings_chat_message_alignment_title,
        titleStyle = ChatTheme.typography.body1,
        onConfirm = {
            viewModel.update(settings.copy(messageAlignment = alignment))
            onConfirm()
        },
        onCancel = onConfirm,
    ) {
        Column(
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
            modifier = Modifier.fillMaxWidth(),
        ) {
            SampleMessageBlock(
                user = UserData.BLANK.copy(name = stringResource(Res.string.sample_username)),
                text = otherMessage.replace("%s", stringResource(Res.string.sample_username)),
                alignment = alignment,
                isSelf = false,
            )
            SampleMessageBlock(
                user = when {
                    isLoggedIn() -> selfProfileRepo.self
                    else -> UserData.BLANK.copy(name = stringResource(Res.string.your_profile))
                },
                text = selfMessage,
                alignment = alignment,
                isSelf = true,
            )
            Row(
                modifier = Modifier.fillMaxWidth(),
            ) {
                SimpleTextRadioButton(
                    text = stringResource(Res.string.settings_chat_message_alignment_left),
                    selected = alignment == ChatMessageAlignment.LEFT,
                    onClick = { alignment = ChatMessageAlignment.LEFT },
                    modifier = Modifier.weight(1f),
                )
                SimpleTextRadioButton(
                    text = stringResource(Res.string.settings_chat_message_alignment_right),
                    selected = alignment == ChatMessageAlignment.RIGHT,
                    onClick = { alignment = ChatMessageAlignment.RIGHT },
                    modifier = Modifier.weight(1f),
                )
            }
            Row(
                modifier = Modifier.fillMaxWidth(),
            ) {
                SimpleTextRadioButton(
                    text = stringResource(Res.string.settings_chat_message_alignment_self_left),
                    selected = alignment == ChatMessageAlignment.SELF_LEFT,
                    onClick = { alignment = ChatMessageAlignment.SELF_LEFT },
                    modifier = Modifier.weight(1f),
                )
                SimpleTextRadioButton(
                    text = stringResource(Res.string.settings_chat_message_alignment_self_right),
                    selected = alignment == ChatMessageAlignment.SELF_RIGHT,
                    onClick = { alignment = ChatMessageAlignment.SELF_RIGHT },
                    modifier = Modifier.weight(1f),
                )
            }
        }
    }
}

@Composable
private fun SampleMessageBlock(
    user: UserData,
    text: String,
    alignment: ChatMessageAlignment,
    isSelf: Boolean,
) = MessageAligner(
    alignment = alignment,
    isSelf = isSelf,
) {
    CompactMessageBlock(
        message = ChatMessage(
            conversationUid = user.uid,
            senderUid = user.uid,
            text = text,
        ),
        avatar = user.avatar,
        displayName = user.displayName,
    )
}