package nand.apps.chat.ui.settings.dialog

import androidx.compose.foundation.Canvas
import androidx.compose.foundation.gestures.awaitEachGesture
import androidx.compose.foundation.gestures.awaitFirstDown
import androidx.compose.foundation.layout.*
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.Refresh
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.graphics.Brush
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.StrokeCap
import androidx.compose.ui.input.pointer.pointerInput
import androidx.compose.ui.text.font.FontWeight
import nand.apps.chat.model.settings.theme.*
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.button.StyledIconButton
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.util.clamp
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import seers.composeapp.generated.resources.*

@Composable
fun ColorSettingDialog(onConfirm: () -> Unit) {
    val state = LocalSettingsProvider.current.themeSettings
    val settingsRepo: SettingsRepo by getKoin().inject()
    SimpleDialog(
        title = Res.string.settings_theme_color_title,
        titleStyle = ChatTheme.typography.body1,
        confirmLabel = null,
        cancelLabel = Res.string.confirm,
        onCancel = onConfirm,
    ) {
        Column(
            modifier = Modifier.fillMaxWidth(),
        ) {
            Row(
                horizontalArrangement = Arrangement.SpaceEvenly,
                modifier = Modifier.fillMaxWidth(),
            ) {
                SimpleTextButton(
                    text = stringResource(Res.string.settings_theme_dark),
                    backgroundColor = when {
                        !ChatTheme.colors.isLight -> ChatTheme.colors.secondary
                        else -> ChatTheme.colors.primary
                    },
                    onClick = { settingsRepo.updateTheme { copy(theme = ChatThemeSetting.DARK) } },
                )
                SimpleTextButton(
                    text = stringResource(Res.string.settings_theme_light),
                    backgroundColor = when {
                        ChatTheme.colors.isLight -> ChatTheme.colors.secondary
                        else -> ChatTheme.colors.primary
                    },
                    onClick = { settingsRepo.updateTheme { copy(theme = ChatThemeSetting.LIGHT) } },
                )
            }
            HueSlider(
                label = Res.string.settings_theme_base_label,
                hue = state.baseHue,
                defaultHue = -1f,
                onHueChanged = { settingsRepo.updateTheme { copy(baseHue = it) } },
            )
            HueSlider(
                label = Res.string.settings_theme_primary_label,
                hue = state.primaryHue,
                defaultHue = THEME_PRIMARY_HUE,
                onHueChanged = { settingsRepo.updateTheme { copy(primaryHue = it) } },
            )
            HueSlider(
                label = Res.string.settings_theme_secondary_label,
                hue = state.secondaryHue,
                defaultHue = THEME_SECONDARY_HUE,
                onHueChanged = { settingsRepo.updateTheme { copy(secondaryHue = it) } },
            )
        }
    }
}

@Composable
private fun HueSlider(
    label: StringResource,
    hue: Float,
    defaultHue: Float,
    onHueChanged: (Float) -> Unit,
) {
    val hueBrush = remember { generateHueGradient(12) }
    val strokeColor = ChatTheme.colors.onBackground2
    Row(
        verticalAlignment = Alignment.CenterVertically,
        horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
        modifier = Modifier.padding(top = ChatTheme.dimens.spacingMedium)
    ) {
        Box(
            contentAlignment = Alignment.Center,
            modifier = Modifier
                .weight(1f)
                .clip(ChatTheme.shapes.small),
        ) {
            Canvas(
                modifier = Modifier
                    .fillMaxWidth()
                    .height(ChatTheme.dimens.iconXLarge)
                    .pointerInput(Unit) {
                        awaitEachGesture {
                            val event = awaitFirstDown()
                            val newHue = ((event.position.x / size.width.toFloat()) * THEME_HUE_RANGE.endInclusive).clamp(THEME_HUE_RANGE)
                            onHueChanged(newHue)
                        }
                    }
            ) {
                drawRect(hueBrush)
                if (hue in THEME_HUE_RANGE) {
                    val x = (hue / THEME_HUE_RANGE.endInclusive) * size.width
                    drawLine(
                        color = strokeColor,
                        start = Offset(x, 0f),
                        end = Offset(x, size.height),
                        cap = StrokeCap.Round,
                        strokeWidth = 2f,
                    )
                }
            }
            TextLine(
                text = stringResource(label),
                color = Color.Black,
                style = ChatTheme.typography.body2.copy(fontWeight = FontWeight.Bold),
            )
        }
        StyledIconButton(
            icon = Icons.Outlined.Refresh,
            contentDescription = Res.string.reset,
            onClick = { onHueChanged(defaultHue) },
            modifier = Modifier.size(ChatTheme.dimens.iconXLarge),
        )
    }
}

private fun SettingsRepo.updateTheme(block: ThemeSettingsData.() -> ThemeSettingsData) {
    update { copy(themeSettings = themeSettings.block()) }
}

private fun generateHueGradient(steps: Int): Brush {
    val interval = THEME_HUE_RANGE.endInclusive / steps.toFloat()
    val colors = mutableListOf<Color>()
    for (step in 0..steps) {
        colors += Color.hsv(
            hue = (interval * step).clamp(THEME_HUE_RANGE),
            saturation = 1f,
            value = 1f,
        )
    }
    return Brush.horizontalGradient(colors)
}