package nand.apps.chat.ui.settings.net.tox

import androidx.compose.foundation.ExperimentalFoundationApi
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.lazy.rememberLazyListState
import androidx.compose.foundation.text.BasicText
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.foundation.text.selection.SelectionContainer
import androidx.compose.material.Icon
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.*
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.unit.dp
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import nand.apps.chat.model.emoji.CHAR_TO_EMOJI_CODE
import nand.apps.chat.time.formatRelativeDate
import nand.apps.chat.ui.button.IconButton
import nand.apps.chat.ui.chat.message.toStylizedMessage
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.scrollbar.VerticalScrollbar
import nand.apps.chat.ui.settings.SettingsColumn
import nand.apps.chat.ui.spinner.FollowSpinner
import nand.apps.chat.ui.text.SimpleTextField
import nand.apps.chat.ui.text.StylizedTextParameters
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.text.rememberTextFieldValue
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.tox.net.isValidIpAddress
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@OptIn(ExperimentalFoundationApi::class)
@Composable
fun ToxNodeSettingsScreen(
    modifier: Modifier,
    viewModel: ToxNodeSettingsViewModel = koinKmpViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    val settings = LocalSettingsProvider.current
    LaunchedEffect(Unit) {
        viewModel.refresh()
    }
    SettingsColumn(
        title = Res.string.settings_network_tox_nodes_title,
        isScrollable = false,
        contentPadding = 0.dp,
        modifier = modifier,
        endContent = {
            IconButton(
                icon = Icons.Outlined.Refresh,
                contentDescription = stringResource(Res.string.settings_network_tox_nodes_refresh),
                size = ChatTheme.dimens.iconMedium,
                isEnabled = !state.isLoading && state.bootstrapData != null,
                onClick = viewModel::showRefreshDialog,
                modifier = Modifier.padding(end = ChatTheme.dimens.spacingSmall),
            )
        },
    ) {
        Box(
            modifier = Modifier
                .fillMaxWidth()
                .weight(1f),
        ) {
            if (state.isLoading) {
                FollowSpinner(
                    modifier = Modifier.align(Alignment.Center)
                )
                return@Box
            }
            state.bootstrapData?.let { data ->
                val listState = rememberLazyListState()
                LazyColumn(
                    state = listState,
                    verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                    contentPadding = PaddingValues(ChatTheme.dimens.spacingLarge),
                    modifier = Modifier.fillMaxSize(),
                ) {
                    stickyHeader {
                        TextLine(
                            text = stringResource(Res.string.settings_network_tox_nodes_last_update, formatRelativeDate(data.lastRefreshMs)),
                            style = ChatTheme.typography.body1,
                            modifier = Modifier
                                .background(ChatTheme.colors.backgroundFloating)
                                .fillMaxWidth()
                        )
                    }
                    items(
                        items = data.nodes,
                        key = { it.publicKey },
                    ) { node ->
                        Row(
                            verticalAlignment = Alignment.CenterVertically,
                            modifier = Modifier
                                .fillMaxWidth()
                                .clip(ChatTheme.shapes.medium)
                                .background(ChatTheme.colors.surface)
                                .onClick { viewModel.showDetailsDialog(node) }
                                .padding(
                                    horizontal = ChatTheme.dimens.spacingMedium,
                                    vertical = ChatTheme.dimens.spacingLarge,
                                ),
                        ) {
                            if (node.location.isNotBlank()) {
                                val emoji = remember(node.location) {
                                    node.location.mapNotNull { CHAR_TO_EMOJI_CODE[it] }.joinToString(separator = "")
                                }
                                BasicText(
                                    text = emoji,
                                    style = ChatTheme.typography.h6,
                                    modifier = Modifier.padding(end = ChatTheme.dimens.spacingMedium),
                                )
                            }
                            Column(
                                modifier = Modifier.weight(1f),
                            ) {
                                TextLine(
                                    text = node.ip,
                                    style = ChatTheme.typography.body2,
                                    color = when {
                                        node.isOnline -> ChatTheme.colors.onSurface
                                        else -> ChatTheme.colors.onBackground2
                                    },
                                )
                                if (node.isOnline) {
                                    TextLine(
                                        text = when {
                                            node.isUdpOnline -> {
                                                "${stringResource(Res.string.settings_network_tox_node_udp)}: ${node.udpPort}"
                                            }
                                            node.isTcpOnline && node.tcpPorts.isNotEmpty() -> {
                                                "${stringResource(Res.string.settings_network_tox_node_tcp)}: ${node.tcpPorts.joinToString()}"
                                            }
                                            else -> ""
                                        },
                                        style = ChatTheme.typography.caption,
                                        color = ChatTheme.colors.onBackground,
                                    )
                                }
                            }
                            TextLine(
                                text = node.formattedVersion,
                                style = ChatTheme.typography.caption,
                                color = ChatTheme.colors.onBackground2,
                                modifier = Modifier.padding(end = ChatTheme.dimens.spacingMedium),
                            )
                            Icon(
                                imageVector = when {
                                    node.isOnline -> Icons.Outlined.Wifi
                                    else -> Icons.Outlined.Block
                                },
                                contentDescription = when {
                                    node.isOnline -> stringResource(Res.string.status_online)
                                    else -> stringResource(Res.string.status_offline)
                                },
                                tint = when {
                                    node.isOnline -> ChatTheme.colors.statusOnline
                                    else -> ChatTheme.colors.statusOffline
                                },
                                modifier = Modifier.size(ChatTheme.dimens.iconMedium),
                            )
                        }
                    }
                }
                VerticalScrollbar(
                    listState = listState,
                    modifier = Modifier
                        .padding(
                            top = ChatTheme.dimens.spacingLarge,
                            bottom = ChatTheme.dimens.spacingLarge,
                            end = ChatTheme.dimens.spacingXSmall,
                        )
                        .align(Alignment.CenterEnd),
                )
            }
        }
    }
    when {
        state.isRefreshDialogVisible -> {
            var urlInput by rememberTextFieldValue(settings.networkSettings.toxNodesEndpointUrl)
            SimpleDialog(
                title = Res.string.settings_network_tox_nodes_refresh,
                onConfirm = { viewModel.updateNodes(urlInput.text) },
                onCancel = viewModel::dismissRefreshDialog,
            ) {
                SimpleTextField(
                    value = urlInput,
                    onValueChange = { urlInput = it },
                    labelText = stringResource(Res.string.settings_network_tox_nodes_endpoint_url),
                    keyboardOptions = KeyboardOptions(
                        autoCorrectEnabled = false,
                        imeAction = ImeAction.Done,
                    ),
                    onSubmit = { viewModel.updateNodes(urlInput.text) },
                    modifier = Modifier.fillMaxWidth(),
                )
            }
        }
        state.detailsNode != null -> SimpleDialog(
            title = Res.string.settings_network_tox_node_details,
            confirmLabel = null,
            cancelLabel = Res.string.close,
            onCancel = viewModel::dismissDetailsDialog,
        ) {
            val node = state.detailsNode ?: return@SimpleDialog
            val colors = ChatTheme.colors
            val tcpPorts = remember(node.tcpPorts) {
                node.tcpPorts.joinToString()
            }
            val description = remember(node.motd, settings.chatSettings, colors) {
                node.motd.toStylizedMessage(
                    params = StylizedTextParameters(
                        settings = settings.chatSettings,
                        colors = colors,
                    ),
                    removeEffectTags = true,
                ).text
            }
            SelectionContainer {
                Column(
                    verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
                ) {
                    if (node.ipv4.isValidIpAddress()) {
                        NodeDetailRow(
                            label = Res.string.settings_network_tox_node_ipv4,
                            text = node.ipv4,
                        )
                    }
                    if (node.ipv6.isValidIpAddress()) {
                        NodeDetailRow(
                            label = Res.string.settings_network_tox_node_ipv6,
                            text = node.ipv6,
                        )
                    }
                    NodeDetailRow(
                        label = Res.string.settings_network_tox_node_udp,
                        text = node.udpPort.toString(),
                    )
                    if (node.tcpPorts.isNotEmpty()) {
                        NodeDetailRow(
                            label = Res.string.settings_network_tox_node_tcp,
                            text = tcpPorts,
                        )
                    }
                    if (node.formattedVersion.isNotBlank()) {
                        NodeDetailRow(
                            label = Res.string.settings_network_tox_node_version,
                            text = node.formattedVersion,
                        )
                    }
                    NodeDetailRow(
                        label = Res.string.settings_network_tox_node_uid,
                        text = node.publicKey,
                    )
                    NodeDetailRow(
                        label = Res.string.settings_network_tox_node_owner,
                        text = node.maintainer,
                    )
                    if (node.motd.isNotBlank()) {
                        Text(
                            text = description.text,
                            color = ChatTheme.colors.onBackground,
                            style = ChatTheme.typography.body2,
                        )
                    }
                }
            }
        }
    }
}

@Composable
private fun NodeDetailRow(
    label: StringResource,
    text: String,
) = Row(
    verticalAlignment = Alignment.CenterVertically,
    modifier = Modifier.fillMaxWidth(),
) {
    TextLine(
        text = stringResource(label),
        style = ChatTheme.typography.body2,
        color = ChatTheme.colors.onBackground2,
        modifier = Modifier.weight(0.2f)
    )
    TextLine(
        text = text,
        style = ChatTheme.typography.body2,
        modifier = Modifier.weight(0.8f),
    )
}