package nand.apps.chat.ui.settings.net.tox

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.repo.SettingsRepo
import nand.apps.tox.net.ToxBootstrapNode
import nand.apps.tox.net.ToxNodeManager

private val NODE_LIST_COMPARATOR = Comparator<ToxBootstrapNode> { node1, node2 ->
    val onlineComp = node2.isOnline.compareTo(node1.isOnline)
    when {
        onlineComp != 0 -> onlineComp
        else -> node2.version.compareTo(node1.version)
    }
}

class ToxNodeSettingsViewModel(
    private val nodeManager: ToxNodeManager,
    private val settingsRepo: SettingsRepo,
    private val backgroundWorker: ChatBackgroundWorker,
) : ViewModel() {

    val stateData = MutableStateFlow(ToxNodeSettingsState())
    private val state get() = stateData.value

    suspend fun refresh() {
        updateState { it.copy(isLoading = true) }
        val bootstrapData = withContext(backgroundWorker.dispatcher) {
            val data = nodeManager.parseBootstrapData()
            data.copy(
                nodes = data.nodes.sortedWith(NODE_LIST_COMPARATOR)
            )
        }
        updateState {
            it.copy(
                isLoading = false,
                bootstrapData = bootstrapData,
            )
        }
    }

    fun showRefreshDialog() = updateState {
        it.copy(isRefreshDialogVisible = true)
    }

    fun dismissRefreshDialog() = updateState {
        it.copy(isRefreshDialogVisible = false)
    }

    fun showDetailsDialog(detailsNode: ToxBootstrapNode) = updateState {
        it.copy(detailsNode = detailsNode)
    }

    fun dismissDetailsDialog() = updateState {
        it.copy(detailsNode = null)
    }

    fun updateNodes(url: String) {
        settingsRepo.update {
            copy(networkSettings = networkSettings.copy(toxNodesEndpointUrl = url))
        }
        updateState {
            it.copy(
                isRefreshDialogVisible = false,
                isLoading = true,
            )
        }
        viewModelScope.launch {
            val isSyncSuccessful = withContext(backgroundWorker.dispatcher) {
                nodeManager.syncBootstrapData(url)
            }
            if (isSyncSuccessful) {
                refresh()
            } else {
                updateState { it.copy(isLoading = false) }
            }
        }
    }

    private fun updateState(block: (ToxNodeSettingsState) -> ToxNodeSettingsState) {
        stateData.tryEmit(block(state))
    }
}