package nand.apps.chat.ui.settings.row

import androidx.compose.foundation.layout.*
import androidx.compose.material.Icon
import androidx.compose.material.Text
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.Lock
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.alpha
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.style.TextOverflow
import nand.apps.chat.ui.chat.message.addMatchStyles
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.modifier.onClick
import nand.apps.chat.ui.modifier.windowInsetPaddingStart
import nand.apps.chat.ui.provider.isLoggedIn
import nand.apps.chat.ui.settings.item.SettingsItem
import nand.apps.chat.ui.settings.search.LocalSettingsSearchProvider
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun SettingsRow(
    title: StringResource,
    description: StringResource?,
    icon: ImageVector,
    onClick: () -> Unit,
    modifier: Modifier = Modifier,
    isEnabled: Boolean = true,
    isLoginRequired: Boolean = false,
    endContent: @Composable RowScope.() -> Unit = {},
) {
    val isLocked = isLoginRequired && !isLoggedIn()
    var isLockDialogVisible by rememberSaveable { mutableStateOf(false) }
    Row(
        verticalAlignment = Alignment.CenterVertically,
        modifier = modifier
            .fillMaxWidth()
            .windowInsetPaddingStart(WindowInsets.displayCutout)
            .onClick(isEnabled) {
                if (isLocked) {
                    isLockDialogVisible = true
                } else {
                    onClick()
                }
            }
            .padding(ChatTheme.dimens.spacingMedium)
            .alpha(if (isEnabled && !isLocked) 1f else 0.5f),
    ) {
        Icon(
            imageVector = icon,
            contentDescription = stringResource(title),
            tint = ChatTheme.colors.icon,
            modifier = Modifier.size(ChatTheme.dimens.avatarSmall),
        )
        Column(
            modifier = Modifier
                .weight(1f)
                .padding(start = ChatTheme.dimens.spacingMedium),
        ) {
            val matches = LocalSettingsSearchProvider.current
            val matchStyle = SpanStyle(
                color = ChatTheme.colors.onMention,
                background = ChatTheme.colors.mention,
            )
            TextLine(
                text = stringResource(title).addMatchStyles(
                    matches = matches.titleMatches,
                    style = matchStyle,
                ),
                style = ChatTheme.typography.h6,
            )
            description?.let {
                Text(
                    text = stringResource(description).addMatchStyles(
                        matches = matches.descMatches,
                        style = matchStyle,
                    ),
                    style = ChatTheme.typography.body2,
                    color = ChatTheme.colors.onBackground,
                    maxLines = 2,
                    overflow = TextOverflow.Ellipsis,
                )
            }
        }
        if (isLocked) {
            Icon(
                imageVector = Icons.Outlined.Lock,
                contentDescription = stringResource(Res.string.settings_login_required_message),
                tint = ChatTheme.colors.icon,
                modifier = Modifier
                    .padding(end = ChatTheme.dimens.spacingMedium)
                    .size(ChatTheme.dimens.avatarSmall),
            )
        } else {
            endContent()
        }
    }
    if (isLockDialogVisible) {
        SimpleDialog(
            title = Res.string.settings_login_required_title,
            confirmLabel = null,
            cancelLabel = Res.string.close,
            onCancel = { isLockDialogVisible = false },
        ) {
            Text(
                text = stringResource(Res.string.settings_login_required_message),
                color = ChatTheme.colors.onSurface,
                style = ChatTheme.typography.body1,
            )
        }
    }
}

@Composable
@NonRestartableComposable
fun SettingsItem.Row(
    onClick: () -> Unit,
    modifier: Modifier = Modifier,
    isEnabled: Boolean = true,
    isLoginRequired: Boolean = false,
    endContent: @Composable RowScope.() -> Unit = {},
) = SettingsRow(
    title = title,
    description = description,
    icon = icon,
    onClick = onClick,
    modifier = modifier,
    isEnabled = isEnabled,
    isLoginRequired = isLoginRequired,
    endContent = endContent,
)