package nand.apps.chat.ui.spinner

import androidx.compose.animation.core.*
import androidx.compose.foundation.Canvas
import androidx.compose.foundation.background
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.size
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.geometry.CornerRadius
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.geometry.Size
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.drawscope.DrawScope
import androidx.compose.ui.graphics.drawscope.rotate
import androidx.compose.ui.graphics.drawscope.translate
import androidx.compose.ui.unit.Dp
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.rememberGridBrush

@Composable
fun FollowSpinner(
    modifier: Modifier = Modifier,
    durationMillis: Int = 2400,
    numSquares: Int = 2,
    blur: MotionBlur? = MotionBlur(),
    size: Dp = ChatTheme.dimens.iconXLarge,
    color: Color = ChatTheme.colors.primary,
    primaryBackground: Color = ChatTheme.colors.surface,
    secondaryBackground: Color = ChatTheme.colors.surface2,
) {
    val infiniteTransition = rememberInfiniteTransition()
    val position by infiniteTransition.animateFloat(
        initialValue = 0f,
        targetValue = 4f,
        animationSpec = infiniteRepeatable(tween(durationMillis, easing = LinearEasing), RepeatMode.Restart),
    )
    Box(
        contentAlignment = Alignment.Center,
        modifier = modifier,
    ) {
        Canvas(
            modifier = Modifier
                .background(
                    brush = rememberGridBrush(
                        size = size,
                        backgroundColor = primaryBackground,
                        foregroundColor = secondaryBackground,
                    ),
                    shape = ChatTheme.shapes.small,
                )
                .size(size)
        ) {
            val steps = blur?.steps ?: 1
            val blurPerStep = blur?.amountPerStep ?: 0f
            val alphaDivisor = blur?.alphaDivisor ?: 1f
            val squaresPerKeyframe = 4f / numSquares
            for (square in 0..<numSquares) {
                var alpha = 1f
                val startPosition = (position + (square * squaresPerKeyframe)) % 4
                for (step in 0..<steps) {
                    val timeStep = startPosition - (blurPerStep * step)
                    drawSquare(
                        transition = if (timeStep < 0f) timeStep + 4f else timeStep,
                        color = color,
                        alpha = alpha,
                    )
                    alpha /= alphaDivisor
                    if (alpha < 0.004f) break
                }
            }
        }
    }
}

private fun DrawScope.drawSquare(
    transition: Float,
    color: Color,
    alpha: Float,
) {
    val halfSize = this.size.height / 2
    val diameter = this.size.height / 3.5f
    val radius = diameter / 2f
    val cornerRadius = radius / 2f
    val offset = halfSize / 2f
    val frame = transition.toInt()
    val progress = transition - frame
    val positions = floatArrayOf(
        offset,
        offset + (progress * halfSize),
        offset + halfSize,
        (offset + halfSize) - (progress * halfSize),
    )
    translate(
        left = positions[(frame + 1) % 4] - halfSize,
        top = positions[frame % 4] - halfSize,
    ) {
        rotate((transition + 0.5f) * 90f) {
            drawRoundRect(
                topLeft = Offset(halfSize - radius, halfSize - radius),
                size = Size(diameter, diameter),
                cornerRadius = CornerRadius(cornerRadius, cornerRadius),
                color = color,
                alpha = alpha,
            )
        }
    }
}