package nand.apps.chat.ui.text

import androidx.compose.runtime.Composable
import androidx.compose.runtime.remember
import androidx.compose.ui.text.*
import androidx.compose.ui.text.font.FontFamily
import androidx.compose.ui.text.font.FontStyle
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.text.input.TransformedText
import androidx.compose.ui.text.style.BaselineShift
import androidx.compose.ui.text.style.TextDecoration
import androidx.compose.ui.unit.sp
import com.fleeksoft.ksoup.Ksoup
import com.fleeksoft.ksoup.nodes.Element
import com.fleeksoft.ksoup.nodes.TextNode
import nand.apps.chat.ui.theme.ChatThemeScope
import nand.apps.chat.util.HOST_CHARS_REGEX
import nand.apps.chat.util.URL_REGEX

fun TextFieldValue.takeIfValidHost(fallback: TextFieldValue) =
    if (text.isEmpty() || HOST_CHARS_REGEX.matches(text)) this else fallback

fun TransformedText.addUrlAnnotations() = copy(
    text = buildAnnotatedString {
        append(text)
        URL_REGEX.findAll(text).forEach { result ->
            val start = result.range.first
            val end = result.range.last + 1
            addLink(LinkAnnotation.Url(result.value), start, end)
        }
    },
)

fun TransformedText.applyUrlStyles(urlStyle: SpanStyle) = copy(
    text = buildAnnotatedString {
        append(text)
        text.getLinkAnnotations().forEach {
            addStyle(urlStyle, it.start, it.end)
        }
    },
)

fun AnnotatedString.getLinkAnnotations() =
    getLinkAnnotations(0, length)

fun AnnotatedString.hasUrlAnnotations(start: Int, end: Int) =
    getLinkAnnotations(start, end).isNotEmpty()

fun AnnotatedString.Builder.addLink(link: LinkAnnotation, start: Int, end: Int) {
    when (link) {
        is LinkAnnotation.Url -> addLink(
            url = link,
            start = start,
            end = end,
        )
        is LinkAnnotation.Clickable -> addLink(
            clickable = link,
            start = start,
            end = end,
        )
    }
}

@Composable
fun rememberHtmlAnnotatedString(
    htmlString: String,
    onElement: ChatThemeScope.(Element) -> SpanStyle? = { null },
) = ChatThemeScope {
    remember(htmlString) {
        buildHtmlAnnotatedString(htmlString, this, onElement)
    }
}

fun buildHtmlAnnotatedString(
    htmlString: String,
    scope: ChatThemeScope,
    onElement: ChatThemeScope.(Element) -> SpanStyle? = { null },
) = buildAnnotatedString {
    parseElement(Ksoup.parse(htmlString).body(), scope, onElement)
}

private fun AnnotatedString.Builder.parseElement(
    element: Element,
    scope: ChatThemeScope,
    onElement: ChatThemeScope.(Element) -> SpanStyle?,
) {
    element.childNodes().forEach { node ->
        when (node) {
            is TextNode -> {
                val text = node.text()
                append(text)
            }
            is Element -> when (node.tagName()) {
                "br" -> append("\n")
                else -> withStyle(scope.onElement(node) ?: scope.getDefaultHtmlStyle(node.tagName())) {
                    parseElement(node, scope, onElement)
                }
            }
        }
    }
}

private fun ChatThemeScope.getDefaultHtmlStyle(tagName: String) = when (tagName) {
    "b" -> SpanStyle(fontWeight = FontWeight.Bold)
    "i", "em" -> SpanStyle(fontStyle = FontStyle.Italic)
    "s" -> SpanStyle(textDecoration = TextDecoration.LineThrough)
    "sub" -> SpanStyle(
        baselineShift = BaselineShift.Subscript,
        fontSize = 10.sp,
    )
    "sup" -> SpanStyle(
        baselineShift = BaselineShift.Superscript,
        fontSize = 10.sp,
    )
    "code" -> SpanStyle(
        fontFamily = FontFamily.Monospace,
        background = colors.surface3,
    )
    else -> SpanStyle()
}