package nand.apps.chat.ui.text

import androidx.compose.ui.graphics.Color
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.text.input.TransformedText
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.uid.UserUid
import nand.apps.chat.model.user.UserData
import nand.apps.chat.ui.contacts.ContactsState
import nand.apps.chat.ui.theme.ChatThemeColors
import kotlin.math.max
import kotlin.math.min

const val MENTION_TAG = "mention"

fun TransformedText.addMentions(
    regex: Regex,
    contacts: ContactsState,
    colors: ChatThemeColors? = null,
): TransformedText {
    val mentions = regex.findAll(text)
    if (mentions.none()) return this

    val group = contacts.conversationContact as? ChatGroupData ?: return this

    return buildTransformedText(this) {
        var start = 0
        val selfMentionStyle = SpanStyle(
            fontWeight = FontWeight.Bold,
            color = colors?.onMention ?: Color.Unspecified,
            background = colors?.mention ?: Color.Unspecified,
        )
        val otherMentionStyle = SpanStyle(
            fontWeight = FontWeight.Bold,
            color = colors?.primary ?: Color.Unspecified,
        )
        mentions.forEach { result ->
            val uid = result.value.substring(1)
            val contact = when {
                group.uid.value == uid -> group
                else -> {
                    val peerUid = UserUid(uid)
                    contacts.friends[peerUid] ?: group.peers[peerUid]
                }
            } ?: return@forEach

            val mentionStart = result.range.first
            val mentionEnd = result.range.last + 1
            val name = when (contact) {
                is ChatGroupData -> "here"
                is UserData -> contact.displayName
                else -> ""
            }
            appendSource(start, mentionStart)
            pushStyle(
                when (uid) {
                    group.selfUid.value,
                    group.uid.value -> selfMentionStyle
                    else -> otherMentionStyle
                }
            )
            pushMentionAnnotation(uid)
            appendSource(mentionStart, mentionStart + 1)
            appendOther(name)
            pop()
            pop()
            start = mentionEnd
        }
        appendSource(start, text.length)
    }
}

fun TransformedText.addMentions(
    params: StylizedTextParameters,
): TransformedText {
    val regex = params.mentionsRegex ?: return this
    val contactsState = params.contactsState ?: return this
    return addMentions(
        regex = regex,
        contacts = contactsState,
        colors = params.colors,
    )
}

// TODO - Fix cursor position when backspacing
fun TextFieldValue.checkMentions(
    original: TextFieldValue,
    originalTransformed: TransformedText,
): TextFieldValue {
    if (text == original.text || !original.text.contains('@')) return this

    val mapping = originalTransformed.offsetMapping
    val start = mapping.originalToTransformed(original.selection.min)
    val end = mapping.originalToTransformed(original.selection.max)
    val mapOffset = if (original.selection.min > 0 && mapping.originalToTransformed(original.selection.min - 1) == start) {
        -1
    } else {
        0
    }
    val mentions = originalTransformed.text.getStringAnnotations(
        tag = MENTION_TAG,
        start = start + mapOffset,
        end = end + mapOffset,
    ).sortedBy { it.start }
    return if (mentions.isNotEmpty()) {
        val minIndex = min(original.selection.min, mapping.transformedToOriginal(mentions.first().start))
        val maxIndex = max(original.selection.max, mapping.transformedToOriginal(mentions.last().end))
        val before = original.text.substring(0, minIndex)
        val after = original.text.substring(maxIndex, original.text.length)
        val addedText = text.substring(
            startIndex = original.selection.min.coerceAtMost(selection.min),
            endIndex = selection.min,
        )
        copy(
            text = before + addedText + after,
            selection = selection,
        )
    } else {
        this
    }
}