package nand.apps.chat.ui.text

import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.ExperimentalTextApi
import androidx.compose.ui.text.TextRange
import androidx.compose.ui.text.buildAnnotatedString
import androidx.compose.ui.text.input.TextFieldValue
import nand.apps.chat.util.REGEX_DIGITS
import nand.apps.chat.util.fastForEach
import okio.utf8Size

fun TextFieldValue.insert(text: String) = TextFieldValue(
    text = this.text.substring(0..<selection.start) + text + this.text.substring(selection.end..<this.text.length),
    selection = TextRange(selection.end + text.length)
)

fun TextFieldValue.insertNewline() = insert("\n")

fun TextFieldValue.replace(regex: Regex, replacement: String) =
    copy(text = text.replace(regex, replacement))

fun TextFieldValue.remove(regex: Regex) =
    copy(text = text.replace(regex, ""))

fun TextFieldValue.uppercase() =
    copy(text = text.uppercase())

fun TextFieldValue.takeIfUnderLength(length: Int, fallback: TextFieldValue): TextFieldValue {
    val maxPossibleLength = text.length * 3
    return when {
        maxPossibleLength > length && text.utf8Size() > length -> fallback
        else -> this
    }
}

fun TextFieldValue.takeIfNumber(
    fallback: TextFieldValue,
    maxDigits: Int = Int.MAX_VALUE,
) = if (text.isEmpty() || REGEX_DIGITS.matches(text) && text.length <= maxDigits) this else fallback

@OptIn(ExperimentalTextApi::class)
fun AnnotatedString.Builder.copyStyles(
    src: AnnotatedString,
    indexMap: Map<Int, Int>,
) {
    // Needs more copy-pasting
    src.spanStyles.fastForEach { range ->
        addStyle(
            style = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.paragraphStyles.fastForEach { range ->
        addStyle(
            style = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.getTtsAnnotations(0, src.length).fastForEach { range ->
        addTtsAnnotation(
            ttsAnnotation = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.getLinkAnnotations().fastForEach { range ->
        addLink(
            link = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
    src.getStringAnnotations(0, src.length).fastForEach { range ->
        addStringAnnotation(
            tag = range.tag,
            annotation = range.item,
            start = indexMap[range.start] ?: indexError(range.start),
            end = indexMap[range.end] ?: indexError(range.end),
        )
    }
}

@OptIn(ExperimentalTextApi::class)
fun AnnotatedString.removeStyles() = buildAnnotatedString {
    append(text)
    getTtsAnnotations(0, length).fastForEach { range ->
        addTtsAnnotation(
            ttsAnnotation = range.item,
            start = range.start,
            end = range.end,
        )
    }
    getLinkAnnotations().fastForEach { range ->
        addLink(
            link = range.item,
            start = range.start,
            end = range.end,
        )
    }
    getStringAnnotations(0, length).fastForEach { range ->
        if (range.tag == UNICODE_FONT_TAG) return@fastForEach
        addStringAnnotation(
            tag = range.tag,
            annotation = range.item,
            start = range.start,
            end = range.end,
        )
    }
}

private fun indexError(index: Int): Nothing {
    error("Missing index: $index")
}