package nand.apps.chat.ui.tutorial

import androidx.compose.foundation.background
import androidx.compose.foundation.layout.*
import androidx.compose.material.SnackbarHostState
import androidx.compose.material.Text
import androidx.compose.runtime.*
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.style.TextAlign
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.ui.button.SimpleTextButton
import nand.apps.chat.ui.login.showLoginError
import nand.apps.chat.ui.modifier.blockInput
import nand.apps.chat.ui.navigation.MainRoute
import nand.apps.chat.ui.provider.LocalNavController
import nand.apps.chat.ui.provider.LocalSettingsProvider
import nand.apps.chat.ui.settings.dialog.NetworkProxySettingDialog
import nand.apps.chat.ui.snackbar.SimpleSnackbarHost
import nand.apps.chat.ui.spinner.FollowSpinner
import nand.apps.chat.ui.text.SimpleTextCheckbox
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.LaunchedCollect
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

@Composable
fun TutorialNetworkSettingsScreen(
    viewModel: TutorialViewModel,
) {
    val navController = LocalNavController.current
    val settings = LocalSettingsProvider.current.networkSettings
    val state by viewModel.stateData.collectAsState()
    val isUdpEnabled = rememberSaveable { mutableStateOf(settings.isUdpEnabled) }
    val isIpv6Enabled = rememberSaveable { mutableStateOf(settings.isIPv6Enabled) }
    val isLanEnabled = rememberSaveable { mutableStateOf(settings.isLanEnabled) }
    var isProxyDialogVisible by rememberSaveable { mutableStateOf(false) }
    val errorSnackbar = remember { SnackbarHostState() }
    Box(
        modifier = Modifier.fillMaxSize(),
    ) {
        TutorialStepColumn(
            title = stringResource(Res.string.tutorial_network_settings_title),
            modifier = Modifier.align(Alignment.Center),
        ) {
            LaunchedCollect(viewModel.loginResult) { result ->
                if (result == ChatEngineInitResult.SUCCESS) {
                    viewModel.completeTutorial()
                    navController.navigate(MainRoute, reset = true)
                } else {
                    errorSnackbar.showLoginError(result)
                }
            }
            Text(
                text = stringResource(Res.string.tutorial_network_settings_desc),
                color = ChatTheme.colors.onBackground,
                style = ChatTheme.typography.body2,
                textAlign = TextAlign.Center,
                modifier = Modifier.padding(
                    vertical = ChatTheme.dimens.spacingMedium,
                    horizontal = ChatTheme.dimens.spacingXLarge,
                ),
            )
            SimpleTextCheckbox(
                text = Res.string.settings_network_udp_enabled_title,
                checkState = isUdpEnabled,
                isEnabled = !state.isLoggingIn,
            )
            SimpleTextCheckbox(
                text = Res.string.settings_network_ipv6_enabled_title,
                checkState = isIpv6Enabled,
                isEnabled = !state.isLoggingIn,
            )
            SimpleTextCheckbox(
                text = Res.string.settings_network_lan_enabled_title,
                checkState = isLanEnabled,
                isEnabled = !state.isLoggingIn,
            )
            SimpleTextButton(
                text = stringResource(Res.string.settings_network_proxy_title),
                onClick = { isProxyDialogVisible = true },
                isEnabled = !state.isLoggingIn,
            )
            SimpleTextButton(
                text = stringResource(Res.string.tutorial_complete),
                onClick = {
                    viewModel.updateNetworkSettings(
                        isUdpEnabled = isUdpEnabled.value,
                        isIpv6Enabled = isIpv6Enabled.value,
                        isLanEnabled = isLanEnabled.value,
                    )
                    viewModel.performLogin()
                },
                isEnabled = !state.isLoggingIn,
            )
        }
        SimpleSnackbarHost(
            state = errorSnackbar,
            backgroundColor = ChatTheme.colors.error,
            modifier = Modifier
                .align(Alignment.BottomCenter)
                .fillMaxWidth()
                .padding(ChatTheme.dimens.spacingLarge)
        )
        if (state.isLoggingIn) {
            Box(
                contentAlignment = Alignment.Center,
                modifier = Modifier
                    .background(ChatTheme.colors.backgroundFloating)
                    .matchParentSize()
                    .blockInput()
            ) {
                FollowSpinner()
            }
        }
    }
    if (isProxyDialogVisible) {
        NetworkProxySettingDialog(
            onUpdateProxy = { viewModel.updateProxySettings(it) },
            onDismiss = { isProxyDialogVisible = false },
        )
    }
}