package nand.apps.chat.ui.update

import androidx.compose.foundation.layout.*
import androidx.compose.material.Icon
import androidx.compose.material.LinearProgressIndicator
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.outlined.Error
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.model.update.AppReleaseInfo
import nand.apps.chat.ui.dialog.SimpleDialog
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.viewmodel.composableViewModel
import org.jetbrains.compose.resources.stringResource
import org.koin.core.parameter.parametersOf
import seers.composeapp.generated.resources.*

@Composable
fun AppUpdateDialog(
    release: AppReleaseInfo,
    autoDownload: Boolean,
    onCancel: () -> Unit,
    viewModel: AppUpdateDialogViewModel = composableViewModel(
        parameters = { parametersOf(AppUpdateDialogState(release = release)) }
    )
) {
    val state by viewModel.stateData.collectAsState()
    LaunchedEffect(Unit) {
        if (autoDownload) viewModel.download()
    }
    SimpleDialog(
        title = Res.string.app_updater_dialog_title,
        confirmLabel = when {
            state.packagePath != null -> Res.string.app_updater_install
            else -> Res.string.download
        },
        isConfirmEnabled = !state.isDownloading,
        onConfirm = viewModel::onConfirm,
        onCancel = onCancel,
    ) {
        Column(
            horizontalAlignment = Alignment.CenterHorizontally,
            verticalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingMedium),
            modifier = Modifier
                .fillMaxWidth()
                .padding(bottom = ChatTheme.dimens.spacingMedium),
        ) {
            when {
                state.isDownloading -> {
                    TextLine(
                        text = stringResource(Res.string.app_updater_downloading_package, release.download.fileName),
                        style = ChatTheme.typography.body1,
                    )
                    Row(
                        verticalAlignment = Alignment.CenterVertically,
                        horizontalArrangement = Arrangement.spacedBy(ChatTheme.dimens.spacingSmall),
                    ) {
                        LinearProgressIndicator(
                            progress = state.downloadProgress / 100f,
                            modifier = Modifier.fillMaxWidth(),
                        )
                        TextLine(
                            text = "${state.downloadProgress}%",
                            style = ChatTheme.typography.body1,
                        )
                    }
                }
                state.isDownloadFailure || state.isInstallFailure -> {
                    Icon(
                        imageVector = Icons.Outlined.Error,
                        contentDescription = null,
                        modifier = Modifier.size(ChatTheme.dimens.iconMedium),
                    )
                    TextLine(
                        text = stringResource(
                            when {
                                state.isDownloadFailure -> Res.string.app_updater_download_failed
                                else -> Res.string.app_updater_install_failed
                            }
                        ),
                        style = ChatTheme.typography.body1,
                    )
                }
                state.packagePath != null -> TextLine(
                    text = stringResource(Res.string.app_updater_download_complete),
                    style = ChatTheme.typography.body1,
                )
                else -> TextLine(
                    text = stringResource(Res.string.app_updater_new_version_available, APP_NAME, release.version.numberString),
                    style = ChatTheme.typography.body1,
                )
            }
        }
    }
}