package nand.apps.chat.ui.util

import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.graphics.BlendMode
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.Path
import androidx.compose.ui.graphics.vector.*
import androidx.compose.ui.unit.Dp

fun ImageVector.copy(
    name: String = this.name,
    defaultWidth: Dp = this.defaultWidth,
    defaultHeight: Dp = this.defaultHeight,
    viewportWidth: Float = this.viewportWidth,
    viewportHeight: Float = this.viewportHeight,
    tintColor: Color = this.tintColor,
    tintBlendMode: BlendMode = this.tintBlendMode,
    autoMirror: Boolean = this.autoMirror,
) = ImageVector.Builder(
    name = name,
    defaultWidth = defaultWidth,
    defaultHeight = defaultHeight,
    viewportWidth = viewportWidth,
    viewportHeight = viewportHeight,
    tintColor = tintColor,
    tintBlendMode = tintBlendMode,
    autoMirror = autoMirror,
).run {
    root.forEach { node ->
        when (node) {
            is VectorPath -> addPath(
                pathData = node.pathData,
                pathFillType = node.pathFillType,
                name = node.name,
                fill = node.fill,
                fillAlpha = node.fillAlpha,
                stroke = node.stroke,
                strokeAlpha = node.strokeAlpha,
                strokeLineWidth = node.strokeLineWidth,
                strokeLineCap = node.strokeLineCap,
                strokeLineJoin = node.strokeLineJoin,
                strokeLineMiter = node.strokeLineMiter,
                trimPathStart = node.trimPathStart,
                trimPathEnd = node.trimPathEnd,
                trimPathOffset = node.trimPathOffset,
            )
            is VectorGroup -> addGroup(
                name = node.name,
                rotate = node.rotation,
                pivotX = node.pivotX,
                pivotY = node.pivotY,
                scaleX = node.scaleX,
                scaleY = node.scaleY,
                translationX = node.translationX,
                translationY = node.translationY,
                clipPathData = node.clipPathData,
            )
        }
    }
    build()
}

fun PathBuilder.toPath() = Path().let { path ->
    var pos = Offset(0f, 0f)
    nodes.forEach { node ->
        when (node) {
            is PathNode.RelativeMoveTo -> {
                path.relativeMoveTo(node.dx, node.dy)
                pos += Offset(node.dx, node.dy)
            }
            is PathNode.MoveTo -> {
                path.moveTo(node.x, node.y)
                pos = Offset(node.x, node.y)
            }
            is PathNode.RelativeLineTo -> {
                path.relativeLineTo(node.dx, node.dy)
                pos += Offset(node.dx, node.dy)
            }
            is PathNode.LineTo -> {
                path.lineTo(node.x, node.y)
                pos = Offset(node.x, node.y)
            }
            is PathNode.RelativeHorizontalTo -> {
                path.relativeLineTo(node.dx, 0f)
                pos += Offset(node.dx, 0f)
            }
            is PathNode.HorizontalTo -> {
                path.lineTo(node.x, pos.y)
                pos = Offset(node.x, pos.y)
            }
            is PathNode.RelativeVerticalTo -> {
                path.relativeLineTo(0f, node.dy)
                pos += Offset(0f, node.dy)
            }
            is PathNode.VerticalTo -> {
                path.lineTo(pos.x, node.y)
                pos = Offset(pos.x, node.y)
            }
            is PathNode.RelativeCurveTo -> {
                path.relativeCubicTo(node.dx1, node.dy1, node.dx2, node.dy2, node.dx3, node.dy3)
                pos += Offset(node.dx3, node.dy3)
            }
            is PathNode.CurveTo -> {
                path.cubicTo(node.x1, node.y1, node.x2, node.y2, node.x3, node.y3)
                pos = Offset(node.x3, node.y3)
            }
            is PathNode.RelativeQuadTo -> {
                path.relativeQuadraticTo(node.dx1, node.dy1, node.dx2, node.dy2)
                pos += Offset(node.dx2, node.dy2)
            }
            is PathNode.QuadTo -> {
                path.quadraticTo(node.x1, node.y1, node.x2, node.y2)
                pos = Offset(node.x2, node.y2)
            }
            is PathNode.RelativeArcTo -> {
                error("Unsupported operation RelativeArcTo")
            }
            is PathNode.ArcTo -> {
                error("Unsupported operation ArcTo")
            }
            is PathNode.ReflectiveCurveTo -> {
                error("Unsupported operation ReflectiveCurveTo")
            }
            is PathNode.ReflectiveQuadTo -> {
                error("Unsupported operation ReflectiveQuadTo")
            }
            is PathNode.RelativeReflectiveCurveTo -> {
                error("Unsupported operation RelativeReflectiveCurveTo")
            }
            is PathNode.RelativeReflectiveQuadTo -> {
                error("Unsupported operation RelativeReflectiveQuadTo")
            }
            is PathNode.Close -> {
                path.close()
            }
        }
    }
    path
}