package nand.apps.chat.update

import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.request.*
import io.ktor.http.*
import io.ktor.utils.io.*
import kotlinx.coroutines.*
import kotlinx.io.readByteArray
import kotlinx.serialization.json.jsonObject
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.ChatBackgroundWorker
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.event.AppUpdateReadyEvent
import nand.apps.chat.model.update.AppReleaseDownload
import nand.apps.chat.model.update.AppReleaseInfo
import nand.apps.chat.model.user.ChatClientVersion
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.time.MINUTE_MILLIS
import nand.apps.chat.time.currentTimeMillis
import nand.apps.chat.util.*
import okio.Path
import kotlin.math.roundToInt

private const val URL_LATEST_RELEASE = "https://gitlab.com/api/v4/projects/nand-apps%2Fseers-lodge/releases/permalink/latest"
private const val DOWNLOAD_PACKET_SIZE = 65536L
private const val CHECK_INTERVAL = 30 * MINUTE_MILLIS

class AppUpdaterImpl(
    private val client: HttpClient,
    private val installer: AppInstaller,
    private val fileSystem: ChatFileSystem,
    private val settingsRepo: SettingsRepo,
    private val eventHandler: ChatEventHandler,
    private val backgroundWorker: ChatBackgroundWorker,
) : AppUpdater {

    private var lastCheckTime = 0L
    private val settings get() = settingsRepo.settings.appUpdateSettings
    private val logger = getLogger("AppUpdater")

    override suspend fun checkLoop() {
        while (installer.isSupported()) {
            val curTime = currentTimeMillis()
            if (settings.isAutoUpdateEnabled && curTime - lastCheckTime > settings.updateCheckInterval) {
                lastCheckTime = curTime
                getLatestReleaseInfo()?.let { release ->
                    if (release.version > LOCAL_CLIENT_VERSION) {
                        eventHandler += AppUpdateReadyEvent(release)
                    }
                }
            }
            delay(CHECK_INTERVAL)
        }
    }

    override suspend fun download(
        request: AppReleaseDownload,
        onProgress: (Int) -> Unit,
    ): Path? {
        val downloadDir = fileSystem.getTempDirectory("install")
        downloadDir.removeFile(request.fileName)
        return try {
            withContext(backgroundWorker.dispatcher) {
                val response = client.get(request.url)
                val totalBytes = response.headers["Content-Length"]?.toLongOrNull() ?: error("Failed to read content length")
                var writtenBytes = 0L
                var lastProgress = 0
                check(response.status.isSuccess())
                val channel: ByteReadChannel = response.body()
                while (!channel.isClosedForRead) {
                    val packet = channel.readRemaining(DOWNLOAD_PACKET_SIZE)
                    val bytes = packet.readByteArray()
                    if (!downloadDir.appendToFile(request.fileName, bytes)) {
                        error("Failed to write data to ${request.fileName}")
                    }
                    writtenBytes += bytes.size
                    val progress = ((writtenBytes.toDouble() / totalBytes) * 100).roundToInt()
                    if (progress != lastProgress) {
                        lastProgress = progress
                        onProgress(progress)
                    }
                }
            }
            downloadDir.getFilePath(request.fileName)
        } catch (e: Exception) {
            logger.error("Failed to download ${request.url}", e)
            null
        }
    }

    override suspend fun getLatestReleaseInfo(): AppReleaseInfo? {
        val releaseInfo = withContext(backgroundWorker.dispatcher) {
            client.getJsonObject(URL_LATEST_RELEASE)
        } ?: run {
            logger.warn("Failed to query for latest app version")
            return null
        }
        val version = releaseInfo["name"].getString() ?: run {
            logger.error("Release metadata is missing \"name\"")
            return null
        }
        val parts = version.split('.')
        if (parts.size != 3) {
            logger.error("Release name isn't formatted properly: $version")
            return null
        }
        val links = releaseInfo.find("assets", "links")?.jsonArraySafe() ?: run {
            logger.error("Release links not found")
            return null
        }
        val fileExtension = installer.getPackageFileExtension()
        val download = links.find { link ->
            link.find("name")?.getString()?.endsWith(".$fileExtension") == true
        }?.jsonObject ?: run {
            logger.error("Failed to find download for release $version")
            return null
        }
        val fileName = download["name"].getString() ?: run {
            logger.error("Failed to get download file name")
            return null
        }
        val downloadUrl = download["direct_asset_url"].getString() ?: run {
            logger.error("Failed to get download URL")
            return null
        }
        return AppReleaseInfo(
            version = ChatClientVersion(
                name = APP_NAME,
                major = parts[0].toIntOrZero(),
                minor = parts[1].toIntOrZero(),
                patch = parts[2].toIntOrZero(),
            ),
            download = AppReleaseDownload(
                fileName = fileName,
                url = downloadUrl,
            ),
        )
    }
}