package nand.apps.chat.di

import io.ktor.client.*
import nand.apps.chat.audio.*
import nand.apps.chat.audio.id3.Id3TagReader
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.mock.MockChatEngineEncryption
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.engine.mock.getMockKoinModule
import nand.apps.chat.io.*
import nand.apps.chat.net.packet.ChatPacketConverter
import nand.apps.chat.net.packet.ChatPacketRepo
import nand.apps.chat.repo.NotificationRepo
import nand.apps.chat.repo.*
import nand.apps.chat.emoji.EmojiRepo
import nand.apps.chat.emoji.TestEmojiRepo
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.engine.DefaultChatEngineCallback
import nand.apps.chat.engine.mock.MockChatEngineImpl
import nand.apps.chat.engine.tox.TestToxNodeManager
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.event.ChatEventHandlerImpl
import nand.apps.chat.event.VisibleMessageTracker
import nand.apps.chat.net.*
import nand.apps.chat.net.packet.ChatPacketRepoImpl
import nand.apps.chat.net.remote.ChatRemoteServer
import nand.apps.chat.platform.AppStateProvider
import nand.apps.chat.platform.TestAppStateProvider
import nand.apps.chat.res.StringResourceProvider
import nand.apps.chat.res.TestStringResourceProvider
import nand.apps.chat.search.ChatSearch
import nand.apps.chat.time.DateTimeFormatter
import nand.apps.chat.ui.chat.message.input.MessageInputViewModel
import nand.apps.chat.ui.contacts.ContactListViewModel
import nand.apps.chat.ui.dialog.audio.AudioRecorderDialogViewModel
import nand.apps.chat.ui.embed.audio.AudioEmbedViewModel
import nand.apps.chat.ui.emoji.EmojiPopupViewModel
import nand.apps.chat.ui.file.sender.ChatFileSenderViewModel
import nand.apps.chat.ui.settings.net.tox.ToxNodeSettingsViewModel
import nand.apps.chat.ui.update.AppUpdateDialogViewModel
import nand.apps.chat.update.AppInstaller
import nand.apps.chat.update.AppUpdater
import nand.apps.chat.update.TestAppInstaller
import nand.apps.chat.update.TestAppUpdater
import nand.apps.tox.net.ToxNodeManager
import okio.FileSystem
import okio.fakefilesystem.FakeFileSystem
import org.koin.core.Koin
import org.koin.core.KoinApplication
import org.koin.core.context.startKoin
import org.koin.core.module.dsl.singleOf
import org.koin.core.module.dsl.factoryOf
import org.koin.dsl.KoinAppDeclaration
import org.koin.dsl.module

fun initKoinTest(
    appDeclaration: KoinAppDeclaration = {},
): KoinApplication {
    val registeredModules = module {
        // Global
        single<Koin> { getKoin() }
        single<FileSystem> { FakeFileSystem() }
        single<HttpClient> { createTestHttpClient() }
        single<AppStateProvider> { TestAppStateProvider() }
        single<ChatFilePaths> { TestFilePaths() }
        single<ChatBackgroundWorker> { TestBackgroundWorker() }
        single<AudioFileWriterFactory> { TestAudioFileWriterFactory(get()) }
        singleOf(::ChatFileSystem)
        singleOf(::ChatPacketConverter)
        singleOf(::DateTimeFormatter)
        
        // Core chat engine
        single<ChatEngine> { MockChatEngineImpl(get()) }
        single<ChatEngineEncryption> { MockChatEngineEncryption() }
        single<ChatEngineConstraints> {
            val engine: ChatEngine by getKoin().inject()
            engine.constraints
        }
        single<ChatEngineCallback> { DefaultChatEngineCallback(get()) }
        singleOf(::ChatRemoteServer)
        
        // Repos
        single<SettingsRepo> { SettingsRepoImpl(get(), get()) }
        single<FileTransferRepo> { FileTransferRepoImpl(get()) }
        single<ChatAvatarRepo> { ChatAvatarRepoImpl(get()) }
        single<SelfProfileRepo> { SelfProfileRepoImpl(get()) }
        single<CallRepo> { CallRepoImpl(get()) }
        single<VoicemailRepo> { VoicemailRepoImpl(get()) }
        single<ChatMessageRepo> { ChatMessageRepoImpl(get()) }
        single<ContactRepo> { ContactRepoImpl(get()) }
        single<ChatPacketRepo> { ChatPacketRepoImpl(get()) }
        single<NotificationRepo> { NotificationRepoImpl(get()) }
        single<EmojiRepo> { TestEmojiRepo() }

        // View models
        factoryOf(::MessageInputViewModel)
        factoryOf(::ContactListViewModel)
        factoryOf(::ChatFileSenderViewModel)
        factoryOf(::AudioEmbedViewModel)
        factoryOf(::AudioRecorderDialogViewModel)
        factoryOf(::AppUpdateDialogViewModel)
        factoryOf(::EmojiPopupViewModel)
        factoryOf(::ToxNodeSettingsViewModel)

        // Other
        single<ChatEventHandler> { ChatEventHandlerImpl(get()) }
        singleOf(::VisibleMessageTracker)
        singleOf(::ChatSearch)
        singleOf(::Id3TagReader)
        singleOf(::AudioMetadataReader)
        single<AudioPlayer> { TestAudioPlayer() }
        single<AudioRecorder> { TestAudioRecorder() }
        single<FileUploader> { TestFileUploader() }
        single<FileDownloader> { TestFileDownloader() }
        single<AppInstaller> { TestAppInstaller() }
        single<AppUpdater> { TestAppUpdater(get(), get()) }
        single<ToxNodeManager> { TestToxNodeManager() }
        single<StringResourceProvider> { TestStringResourceProvider() }
        singleOf(::TempFileCache)
    }
    return startKoin {
        modules(registeredModules)
        modules(getMockKoinModule())
        appDeclaration()
    }
}