package nand.apps.chat.repo

import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.test.runTest
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.mock.MockChatEngine
import nand.apps.chat.event.VisibleMessageTracker
import nand.apps.chat.model.file.ChatFileTransfer
import nand.apps.chat.model.file.ChatFileTransferDirection
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.message.UnreadMessageCounts
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.uid.FileUid
import nand.apps.chat.platform.AppStateProvider
import nand.apps.chat.platform.TestAppStateProvider
import nand.apps.chat.ui.navigation.MainRoute
import nand.apps.chat.util.*
import org.koin.core.context.stopKoin
import kotlin.test.*

class ChatMessageRepoTests {

    private lateinit var messageRepo: ChatMessageRepo
    private lateinit var contactRepo: ContactRepo
    private lateinit var selfProfileRepo: SelfProfileRepo
    private lateinit var chatAvatarRepo: ChatAvatarRepo
    private lateinit var visibleMessageTracker: VisibleMessageTracker
    private lateinit var appStateProvider: TestAppStateProvider

    private val repos by lazy { listOf(contactRepo, messageRepo, chatAvatarRepo, selfProfileRepo) }
    private lateinit var newMessageChannel: Channel<ChatMessage>
    private lateinit var messagesChannel: Channel<ContactUid>
    private val self by lazy { selfProfileRepo.self }

    @BeforeTest
    fun setup() {
        val koin = initKoinTest().koin
        val chatEngine = koin.get<ChatEngine>() as MockChatEngine
        messageRepo = koin.get()
        contactRepo = koin.get()
        selfProfileRepo = koin.get()
        chatAvatarRepo = koin.get()
        visibleMessageTracker = koin.get()
        appStateProvider = koin.get<AppStateProvider>() as TestAppStateProvider
        runTest {
            chatEngine.setSelf(TEST_SELF)
            chatEngine.setContacts(
                listOf(
                    GROUP_BOOBS,
                    GROUP_NOOBS,
                    FRIEND_CHRIS,
                    FRIEND_LARRY,
                    FRIEND_MGC,
                    FRIEND_ROBBY,
                    TEST_SELF,
                )
            )
            repos.forEach {
                it.onEngineInit(TEST_PROFILE)
                it.onEngineStart(this)
            }
        }
    }

    @AfterTest
    fun teardown() {
        repos.forEach { it.onEngineStop() }
        stopKoin()
    }

    @Test
    fun `Received message from friend in group`() = repoTest {
        // Add message
        val message = ChatMessage(
            conversationUid = GROUP_BOOBS.uid,
            senderUid = FRIEND_LARRY.uid,
            text = "Hi",
        )
        addMessage(message)
        assertEquals(
            expected = listOf(message.copy(isUnread = true)),
            actual = getMessages(GROUP_BOOBS.uid),
        )
        newMessageChannel.assertValue(message.copy(isUnread = true))
        assertEquals(
            expected = UnreadMessageCounts(groupMessages = 1),
            actual = getUnreadMessageCounts(GROUP_BOOBS.uid),
        )

        // Open conversation
        visibleMessageTracker.onConversationChanged(GROUP_BOOBS.uid)
        assertEquals(
            expected = listOf(message),
            actual = getMessages(GROUP_BOOBS.uid),
        )

        // Message edited
        val updated = message.copy(text = "Hello")
        updateMessage(
            message = updated,
            localOnly = true,
        )
        messagesChannel.assertValue(GROUP_BOOBS.uid)
        assertEquals(
            expected = listOf(updated),
            actual = getMessages(GROUP_BOOBS.uid),
        )
    }

    @Test
    fun `Add reaction to message`() = repoTest {
        // Send message
        val message = ChatMessage(
            conversationUid = FRIEND_MGC.uid,
            senderUid = self.uid,
            text = "Hi",
        )
        sendMessage(message)
        assertEquals(
            expected = listOf(message),
            actual = getMessages(FRIEND_MGC.uid),
        )
        newMessageChannel.assertValue(message)

        // Add reaction
        val updated = message.copy(reactions = mapOf(self.uid to setOf("\uD83D\uDE05")))
        addReaction(
            message = message,
            reaction = "\uD83D\uDE05",
        )
        assertEquals(
            expected = updated,
            actual = getMessage(FRIEND_MGC.uid, message.uid),
        )
        messagesChannel.assertValue(FRIEND_MGC.uid)
    }

    @Test
    fun `Remove message`() = repoTest {
        // Send message
        val message = ChatMessage(
            conversationUid = GROUP_NOOBS.uid,
            senderUid = self.uid,
            text = "Hi",
        )
        sendMessage(message)
        assertEquals(
            expected = listOf(message),
            actual = getMessages(GROUP_NOOBS.uid),
        )
        newMessageChannel.assertValue(message)

        // Remove message
        removeMessage(
            message = message,
            localOnly = true,
        )
        assertEquals(
            expected = emptyList(),
            actual = getMessages(GROUP_NOOBS.uid),
        )
        messagesChannel.assertValue(GROUP_NOOBS.uid)
    }

    @Test
    fun `Add file transfer message`() = repoTest {
        // Send message
        val fileTransfer = ChatFileTransfer(
            fileUid = FileUid("Test"),
            friendUid = FRIEND_CHRIS.uid,
            name = "File.png",
            path = "file.png",
            size = 1_000_000L,
            direction = ChatFileTransferDirection.INCOMING,
        )
        visibleMessageTracker.onConversationChanged(FRIEND_CHRIS.uid)
        addFileTransferMessage(
            transfer = fileTransfer,
            isAutoAccepted = false,
        )
        val fileMessage = getMessages(FRIEND_CHRIS.uid).first()
        val message = ChatMessage(
            uid = fileMessage.uid,
            timestamp = fileMessage.timestamp,
            conversationUid = FRIEND_CHRIS.uid,
            senderUid = FRIEND_CHRIS.uid,
            text = "user_sent_file_message",
            type = ChatMessageType.ACTION,
            fileTransferUid = fileTransfer.uid,
        )
        assertEquals(
            expected = listOf(message),
            actual = getMessages(FRIEND_CHRIS.uid),
        )
        newMessageChannel.assertValue(message)
    }

    private fun repoTest(block: ChatMessageRepo.() -> Unit) = runTest {
        newMessageChannel = collect(messageRepo.onNewMessage)
        messagesChannel = collect(messageRepo.onMessagesChanged)
        visibleMessageTracker.onNavRoute(MainRoute)
        messageRepo.block()
    }
}