package nand.apps.chat.repo

import kotlinx.coroutines.test.runTest
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.model.settings.GeneralSettingsData
import nand.apps.chat.model.settings.SettingsData
import nand.apps.chat.model.settings.theme.ChatThemeSetting
import nand.apps.chat.model.settings.theme.ThemeSettingsData
import nand.apps.chat.model.settings.update.AppUpdateSettingsData
import okio.FileSystem
import okio.Path.Companion.toPath
import okio.fakefilesystem.FakeFileSystem
import org.koin.core.context.stopKoin
import kotlin.test.*

class SettingsRepoTests {

    private lateinit var fileSystem: FakeFileSystem
    private lateinit var settingsRepo: SettingsRepo
    private val settingsPath = "config/settings.json".toPath()

    @BeforeTest
    fun setup() {
        val koin = initKoinTest().koin
        settingsRepo = koin.get()
        fileSystem = koin.get<FileSystem>() as FakeFileSystem
        fileSystem.emulateUnix()
    }

    @AfterTest
    fun teardown() {
        stopKoin()
    }

    @Test
    fun `Settings saved to file exists and has correct version`() = runTest {
        settingsRepo.save()
        assertTrue(fileSystem.exists(settingsPath), "Settings file doesn't exist")

        val rawJson = fileSystem.read(settingsPath) { readByteArray().decodeToString() }
        assertEquals(
            expected = "{\"version\":1}",
            actual = rawJson,
        )
    }

    @Test
    fun `Load settings from file`() = runTest {
        writeSettingsFile("{\"version\":0,\"generalSettings\":{\"loginUsername\":\"Jimbo\"}}")
        settingsRepo.load()
        assertEquals(
            expected = SettingsData(
                version = SETTINGS_VERSION,
                generalSettings = GeneralSettingsData(loginUsername = "Jimbo"),
            ),
            actual = settingsRepo.settings,
        )
    }

    @Test
    fun `Upgrade old theme settings version`() = runTest {
        writeSettingsFile("""
            {
              "generalSettings": {
                "appTheme": "LIGHT",
                "appThemeHue": 123.45
              }
            }
        """.trimIndent())
        settingsRepo.load()
        assertEquals(
            expected = SettingsData(
                version = SETTINGS_VERSION,
                generalSettings = GeneralSettingsData(
                    appTheme = ChatThemeSetting.LIGHT,
                    appThemeHue = 123.45f,
                ),
                themeSettings = ThemeSettingsData(
                    theme = ChatThemeSetting.LIGHT,
                    baseHue = 123.45f,
                ),
            ),
            actual = settingsRepo.settings,
        )
    }

    @Test
    fun `Upgrade old app update settings version`() = runTest {
        writeSettingsFile("""
            {
              "appUpdateSettingsData": {
                "isAutoUpdateEnabled": true,
                "updateCheckInterval": 123456
              }
            }
        """.trimIndent())
        settingsRepo.load()
        val expectedUpdateSettings = AppUpdateSettingsData(
            isAutoUpdateEnabled = true,
            updateCheckInterval = 123456L,
        )
        assertEquals(
            expected = SettingsData(
                version = SETTINGS_VERSION,
                appUpdateSettings = expectedUpdateSettings,
                appUpdateSettingsData = expectedUpdateSettings,
            ),
            actual = settingsRepo.settings,
        )
    }

    private fun writeSettingsFile(rawJson: String) {
        fileSystem.createDirectories(settingsPath.parent!!)
        fileSystem.write(settingsPath) { write(rawJson.encodeToByteArray()) }
    }
}