package nand.apps.chat.search

import kotlinx.coroutines.runBlocking
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.repo.SelfProfileRepo
import org.koin.core.context.stopKoin
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test
import kotlin.test.assertEquals

class ChatSearchTests {

    private lateinit var chatSearch: ChatSearch
    private lateinit var selfProfileRepo: SelfProfileRepo
    private lateinit var contactRepo: ContactRepo
    private lateinit var chatAvatarRepo: ChatAvatarRepo

    private val self by lazy { contactRepo.self }
    private val friend by lazy { contactRepo.getFriends().first() }

    @BeforeTest
    fun setup() {
        initKoinTest().koin.apply {
            chatSearch = get()
            selfProfileRepo = get()
            contactRepo = get()
            chatAvatarRepo = get()
        }
        runBlocking {
            val profile = ChatProfileData("", "")
            chatAvatarRepo.onEngineInit(profile)
            selfProfileRepo.onEngineInit(profile)
            contactRepo.onEngineInit(profile)
        }
    }

    @AfterTest
    fun teardown() {
        stopKoin()
        selfProfileRepo.onEngineStop()
        contactRepo.onEngineStop()
    }

    @Test
    fun `Direct message search`() {
        val messages = listOf(
            createSelfMessage("This is a message"),
            createFriendMessage("This is another message"),
            createFriendMessage("It sure is"),
        )
        val filtered = chatSearch.filterMessages(
            messages = messages,
            terms = "message",
            params = ChatSearchParams(),
        )
        filtered.assertSize(2)
        assertEquals(
            expected = messages[0].copy(textMatches = listOf(10..16)),
            actual = filtered[0],
        )
        assertEquals(
            expected = messages[1].copy(textMatches = listOf(16..22)),
            actual = filtered[1]
        )
    }

    @Test
    fun `Message username search`() {
        val messages = listOf(
            createSelfMessage("Something to say"),
            createFriendMessage("Great cool awesome"),
            createSelfMessage("Yep"),
        )
        val filtered = chatSearch.filterMessages(
            messages = messages,
            terms = "NAND",
            params = ChatSearchParams(matchCase = true),
        )
        filtered.assertSize(2)
        assertEquals(
            expected = messages[0].copy(nameMatches = listOf(0..3)),
            actual = filtered[0],
        )
        assertEquals(
            expected = messages[2].copy(nameMatches = listOf(0..3)),
            actual = filtered[1]
        )
    }

    @Test
    fun `Message search excluding usernames`() {
        val messages = listOf(
            createSelfMessage("im saying words"),
            createFriendMessage("ur an idiot nand"),
            createSelfMessage("no!!"),
        )
        val filtered = chatSearch.filterMessages(
            messages = messages,
            terms = "nand",
            params = ChatSearchParams(matchNames = false),
        )
        filtered.assertSize(1)
        assertEquals(
            expected = messages[1].copy(textMatches = listOf(12..15)),
            actual = filtered[0],
        )
    }

    @Test
    fun `Message regex search`() {
        val messages = listOf(
            createSelfMessage("This is my message"),
            createFriendMessage("What is?"),
            createSelfMessage("This message"),
        )
        val filtered = chatSearch.filterMessages(
            messages = messages,
            terms = "[\\w]+\\?",
            params = ChatSearchParams(matchRegex = true),
        )
        filtered.assertSize(1)
        assertEquals(
            expected = messages[1].copy(textMatches = listOf(5..7)),
            actual = filtered[0],
        )
    }

    private fun createSelfMessage(text: String) = ChatMessage(
        conversationUid = friend.uid,
        senderUid = self.uid,
        text = text,
    )

    private fun createFriendMessage(text: String) = ChatMessage(
        conversationUid = friend.uid,
        senderUid = friend.uid,
        text = text,
    )

    private fun List<ChatMessage>.assertSize(size: Int) = assertEquals(
        expected = size,
        actual = this.size,
        message = "Search results size mismatch",
    )
}