package nand.apps.chat.ui.chat.message.input

import androidx.compose.ui.input.key.Key
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.TextRange
import androidx.compose.ui.text.input.TextFieldValue
import androidx.compose.ui.text.input.TransformedText
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.*
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.model.event.*
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.settings.shortcut.ShortcutAction
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.model.user.UserData
import nand.apps.chat.net.FileUploader
import nand.apps.chat.net.TestFileUploader
import nand.apps.chat.text.TestChatColors
import nand.apps.chat.text.assertEquals
import nand.apps.chat.ui.chat.message.toStylizedInputText
import nand.apps.chat.ui.contacts.ContactsState
import nand.apps.chat.ui.shortcut.ChatKeyEvent
import nand.apps.chat.ui.text.StylizedTextParameters
import nand.apps.chat.ui.text.toTransformedText
import nand.apps.chat.ui.viewmodel.ViewModelTestScope
import nand.apps.chat.util.get
import okio.Path.Companion.toPath
import kotlin.test.*

private typealias MessageInputViewModelTestScope = ViewModelTestScope<MessageInputViewModel, MessageInputState, MessageInputAction>

@OptIn(ExperimentalCoroutinesApi::class)
class MessageInputViewModelTests {

    private lateinit var scope: MessageInputViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Submit message to friend`() = viewModelTest {
        // Open conversation for random friend
        val friend = contactRepo.getFriends().random()
        openConversation(friend)

        // Submit message and check for event + reset text
        submitMessage("Test")
    }

    @Test
    fun `Send message keyboard events`() = viewModelTest {
        val group = contactRepo.getGroups().random()
        openConversation(group)

        // Enter message
        val message = "Hi this is a message"
        enterText(message)
        pressKey(Key.Enter)
        assertSentMessage(
            conversationUid = group.uid,
            text = message,
        )

        // Add newline
        val newMessage = "Hi this \nis a message"
        enterText(message, 8)
        pressKey(Key.Enter, isShiftPressed = true)
        assertUpdateTextAction(newMessage, 9)
        pressKey(Key.Enter)
        assertSentMessage(
            conversationUid = group.uid,
            text = newMessage,
        )

        // Check message history cycle
        pressKey(Key.DirectionUp)
        assertUpdateTextAction(newMessage, 21)
        pressKey(Key.DirectionUp)
        assertUpdateTextAction(message, 20)
        pressKey(Key.DirectionDown)
        assertUpdateTextAction(newMessage, 21)
        pressKey(Key.DirectionDown)
        assertUpdateTextAction()
    }

    @Test
    fun `Edit message in friend chat`() = viewModelTest {
        // Open friend conversation
        val conversation = contactRepo.getFriends().random()
        var state = openConversation(conversation)

        // Submit message to edit
        submitMessage("Salve")
        assertSentMessage(
            conversationUid = conversation.uid,
            text = "Salve",
        )

        // Prompt to edit message
        val message = messageRepo.getMessages(conversation.uid).first()
        pushEvent(PromptEditMessageEvent(message))
        assertState(state.copy(editMessage = message))
        assertUpdateTextAction("Salve", 5)

        // Change message to "Hello"
        submitMessage("Hello")
        assertEvent(
            UpdateMessageEvent(
                message = message.copy(
                    text = "Hello",
                    isEdited = true,
                ),
                localOnly = false,
            )
        )
        assertState(state.copy(editMessage = null))
    }

    @Test
    fun `React to message in group chat`() = viewModelTest {
        // Open group conversation
        val conversation = contactRepo.getGroups().random()
        val state = openConversation(conversation)

        // Submit message
        submitMessage("Test")
        assertSentMessage(
            conversationUid = conversation.uid,
            text = "Test",
        )

        // Prompt to react to message
        val message = messageRepo.getMessages(conversation.uid).first()
        pushEvent(PromptReactMessageEvent(message))
        assertState(
            state.copy(
                reactMessage = message,
                isEmojiPopupVisible = true,
            )
        )

        // Add reaction
        val emoji = "\uD83E\uDD76"
        viewModel.addReaction(emoji)
        assertEvent(
            ReactMessageEvent(
                message = message,
                reaction = emoji,
            )
        )
        assertState(
            state.copy(
                reactMessage = null,
                isEmojiPopupVisible = false,
            )
        )
    }

    @Test
    fun `Add and remove mention`() = viewModelTest {
        // Open group conversation
        val conversation = contactRepo.getGroups().random()
        val conversationUid = conversation.uid.value
        val originalMessage = "@$conversationUid"
        val updatedMessage = "@${conversationUid.substring(0, conversationUid.length - 1)}"
        val styleParams = StylizedTextParameters(
            settings = settingsRepo.settings.chatSettings,
            colors = TestChatColors,
            tagPreviewStyle = null,
            mentionsRegex = constraints.mentionRegex,
            contactsState = ContactsState(
                conversationUid = conversation.uid,
                friends = contactRepo.getFriends().associateBy { it.uid },
                groups = contactRepo.getGroups().associateBy { it.uid },
            ),
        )

        // Open conversation and initialize text style
        openConversation(conversation)
        viewModel.onTextStyleChanged(
            textStyle = null,
            colors = TestChatColors,
        )
        assertUpdateTextAction(
            stylizedText = styleParams.apply(""),
        )

        // Enter mention with message
        val stylizedMention = styleParams.apply(originalMessage)
        enterText(originalMessage, assertAction = false)
        assertUpdateTextAction(
            text = originalMessage,
            selectionIndex = originalMessage.length,
            stylizedText = stylizedMention,
        )
        assertEquals(stylizedMention.text.text, "@here")

        // Backspace mention and ensure it's completely removed
        enterText(updatedMessage, assertAction = false)
        assertUpdateTextAction(
            stylizedText = styleParams.apply(""),
        )
    }

    @Test
    fun `Mention popup states`() = viewModelTest {
        // Trigger message popup by entering an "@"
        val state = MessageInputState(mentionIndex = 9)
        enterText("Testing @")
        assertState(state)

        // Ensure key is passed to mention state
        val mentionKeys = listOf(Key.DirectionUp, Key.DirectionDown, Key.Enter, Key.NumPadEnter)
        mentionKeys.forEach { key ->
            pressKey(key)
            assertMentionKey(key)
        }

        // Dismiss the popup by removing the "@"
        enterText("Testing ")
        assertState(
            state.copy(
                mentionIndex = -1,
                mentionKeyEvent = null,
            )
        )
    }

    @Test
    fun `Shortcut actions`() = viewModelTest {
        // Open group conversation
        val state = openConversation(contactRepo.getGroups().random())

        // Open emoji popup
        viewModel.onShortcutAction(ShortcutAction.TOGGLE_EMOJI_POPUP)
        assertState(state.copy(isEmojiPopupVisible = true))

        // Dismiss emoji popup
        viewModel.onShortcutAction(ShortcutAction.TOGGLE_EMOJI_POPUP)
        assertState(state.copy(isEmojiPopupVisible = false))

        // Prompt to send file
        viewModel.onShortcutAction(ShortcutAction.SEND_FILE)
        assertState(state.copy(isGroupUploadDialogVisible = true))
    }

    @Test
    fun `Send a file to a group`() = viewModelTest {
        // Must be in a conversation first
        val filePath = "/some/path"
        viewModel.onSendFile(filePath)
        assertNoStateChange()

        val conversation = contactRepo.getGroups().random()
        val state = openConversation(conversation)

        // Must be a valid file path
        viewModel.onSendFile("crab")
        assertNoStateChange()

        // File must exist
        viewModel.onSendFile(filePath)
        assertNoStateChange()

        // Create file path and send (error)
        fileSystem.createDirectories("some".toPath())
        fileSystem.write(filePath.toPath()){}
        viewModel.onSendFile(filePath)
        assertState(state.copy(isBusy = true))
        assertState(state.copy(isBusy = false, isUploadError = true))

        // Re-attempt but succeed
        val responseUrl = "https://seers.chat/"
        val expectedInput = "Here's a link:\n$responseUrl"
        fileUploader.mockResponse(filePath, responseUrl)
        enterText("Here's a link:")
        viewModel.onSendFile(filePath)
        assertState(state.copy(isBusy = true))
        assertState(
            state.copy(
                isBusy = false,
                isUploadError = false,
            )
        )
        assertUpdateTextAction(expectedInput, 34)
    }

    private fun viewModelTest(
        block: MessageInputViewModelTestScope.() -> Unit,
    ) = runTest {
        val koin = initKoinTest().koin
        val viewModel: MessageInputViewModel = koin.get()
        scope = ViewModelTestScope(koin)
        scope.initTestScope(
            scope = this,
            viewModel = viewModel,
            stateData = viewModel.stateData,
            actionData = viewModel.actionData,
        )
        scope.assertState(MessageInputState())
        scope.block()
    }
}

private val MessageInputViewModelTestScope.fileUploader get() = koin.get<FileUploader>() as TestFileUploader

private val MessageInputViewModelTestScope.constraints get() = koin.get<ChatEngineConstraints>()

private fun MessageInputViewModelTestScope.openConversation(conversation: ChatContact): MessageInputState {
    val state = MessageInputState(
        conversationUid = conversation.uid,
        conversationContact = when (conversation) {
            is UserData,
            is ChatGroupData -> conversation

            else -> null
        },
        conversationTitle = when (conversation) {
            is UserData -> conversation.displayName
            is ChatGroupData -> conversation.title
            else -> null
        },
        isFileTransferEnabled = conversation is UserData,
    )
    pushEvent(OpenConversationEvent(conversation.uid))
    assertState(state)
    return state
}

private fun MessageInputViewModelTestScope.enterText(
    text: String,
    selectionIndex: Int = text.length,
    assertAction: Boolean = true,
) {
    viewModel.onTextFieldChanged(TextFieldValue(text, TextRange(selectionIndex)))
    if (assertAction) assertUpdateTextAction(text, selectionIndex)
}

private fun MessageInputViewModelTestScope.submitMessage(text: String) {
    enterText(text)
    viewModel.submitMessage()
}

private fun MessageInputViewModelTestScope.assertSentMessage(
    conversationUid: ContactUid,
    text: String,
) {
    assertEvent(
        SendMessageEvent(
            conversationUid = conversationUid,
            text = text,
        )
    )
    assertUpdateTextAction()
}

private fun MessageInputViewModelTestScope.assertUpdateTextAction(
    text: String = "",
    selectionIndex: Int = 0,
    stylizedText: TransformedText = text.toTransformedText(),
) {
    val textField = TextFieldValue(text, TextRange(selectionIndex))
    val nextAction = actionChannel.get()
    assertIs<UpdateTextAction>(nextAction)
    assertEquals(textField, nextAction.textField)
    assertEquals(stylizedText, nextAction.stylizedText)
}

private fun MessageInputViewModelTestScope.pressKey(
    key: Key,
    isShiftPressed: Boolean = false,
    isCtrlPressed: Boolean = false,
    isAltPressed: Boolean = false,
) = viewModel.onKeyDown(
    ChatKeyEvent(
        key = key,
        isShiftPressed = isShiftPressed,
        isCtrlPressed = isCtrlPressed,
        isAltPressed = isAltPressed,
        timestamp = -1L,
    )
)

private fun MessageInputViewModelTestScope.assertMentionKey(key: Key) = assertState(
    viewModel.stateData.value.copy(
        mentionKeyEvent = ChatKeyEvent(key, timestamp = -1L),
    )
)

private fun StylizedTextParameters.apply(text: String) =
    AnnotatedString(text).toStylizedInputText(this)