package nand.apps.chat.ui.contacts

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.*
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.engine.mock.MockChatEngine
import nand.apps.chat.model.contact.ChatContact
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.uid.FriendRequestUid
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.util.*
import nand.apps.chat.ui.viewmodel.ViewModelTestScope
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test

private typealias ContactListViewModelTestScope = ViewModelTestScope<ContactListViewModel, ContactListState, Unit>

@OptIn(ExperimentalCoroutinesApi::class)
class ContactListViewModelTests {

    private lateinit var scope: ContactListViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Changing friend statuses sorts contacts`() = viewModelTest {
        assertInitialState()

        // MGC comes online -> move to top of the list
        contactRepo.onFriendConnectionProtocol(FRIEND_MGC.uid, ConnectionProtocol.TCP)
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_MGC.copy(protocol = ConnectionProtocol.TCP),
            FRIEND_CHRIS,
            FRIEND_LARRY,
            FRIEND_ROBBY,
        )

        // Larry comes online -> move to top of the list above MGC
        contactRepo.onFriendConnectionProtocol(FRIEND_LARRY.uid, ConnectionProtocol.UDP)
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_LARRY.copy(protocol = ConnectionProtocol.UDP),
            FRIEND_MGC.copy(protocol = ConnectionProtocol.TCP),
            FRIEND_CHRIS,
            FRIEND_ROBBY,
        )

        // Larry goes into away mode -> sort order does not change
        contactRepo.onFriendStatus(FRIEND_LARRY.uid, UserStatus.AWAY)
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_LARRY.copy(protocol = ConnectionProtocol.UDP, status = UserStatus.AWAY),
            FRIEND_MGC.copy(protocol = ConnectionProtocol.TCP),
            FRIEND_CHRIS,
            FRIEND_ROBBY,
        )
    }

    @Test
    fun `Change friend aliases sorts contacts`() = viewModelTest {
        assertInitialState()

        // MGC -> "CGM" -> move to the top of the list
        contactRepo.setFriendAlias(FRIEND_MGC.uid, "CGM")
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_MGC.copy(alias = "CGM"),
            FRIEND_CHRIS,
            FRIEND_LARRY,
            FRIEND_ROBBY,
        )

        // Chris -> "zitz" -> move to the bottom of the list
        contactRepo.setFriendAlias(FRIEND_CHRIS.uid, "zitz")
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_MGC.copy(alias = "CGM"),
            FRIEND_LARRY,
            FRIEND_ROBBY,
            FRIEND_CHRIS.copy(alias = "zitz"),
        )
    }

    @Test
    fun `Remove friend + group sorts contacts`() = viewModelTest {
        assertInitialState()
        contactRepo.removeFriend(FRIEND_LARRY.uid)
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_CHRIS,
            FRIEND_MGC,
            FRIEND_ROBBY,
        )
        contactRepo.removeGroup(GROUP_BOOBS.uid, null)
        assertContactsList(
            GROUP_NOOBS,
            FRIEND_CHRIS,
            FRIEND_MGC,
            FRIEND_ROBBY,
        )
    }

    @Test
    fun `Friend request moves to the top of the list`() = viewModelTest {
        assertInitialState()

        val friendRequest1 = FriendRequest(
            uid = FriendRequestUid("FRIEND_REQUEST_1"),
            message = "Hey there buddy",
        )
        val friendRequest2 = FriendRequest(
            uid = FriendRequestUid("FRIEND_REQUEST_2"),
            message = "Aha!",
        )

        // First friend request is at the top
        contactRepo.onFriendRequest(friendRequest1)
        assertContactsList(
            friendRequest1,
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_CHRIS,
            FRIEND_LARRY,
            FRIEND_MGC,
            FRIEND_ROBBY,
        )

        // Second friend request is under the first one (temporal sort order)
        contactRepo.onFriendRequest(friendRequest2)
        assertContactsList(
            friendRequest1,
            friendRequest2,
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_CHRIS,
            FRIEND_LARRY,
            FRIEND_MGC,
            FRIEND_ROBBY,
        )
    }

    @Test
    fun `Search filters contacts by name`() = viewModelTest {
        assertInitialState()

        viewModel.search("C")
        assertContactsList(FRIEND_CHRIS)

        viewModel.search("The")
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
        )

        viewModel.search("ob")
        assertContactsList(
            GROUP_BOOBS,
            GROUP_NOOBS,
            FRIEND_ROBBY,
        )
    }

    @Test
    fun `Open and dismiss dialogs`() = viewModelTest {
        val state = assertInitialState()
        ContactDialogType.entries.forEach { dialogType ->
            viewModel.openDialog(dialogType)
            assertState(state.copy(dialog = dialogType))
        }
        viewModel.dismissDialog()
        assertState(state.copy(dialog = null))
    }

    private fun viewModelTest(block: ContactListViewModelTestScope.() -> Unit) = runTest {
        val koin = initKoinTest().koin
        val viewModel: ContactListViewModel = koin.get()
        val chatEngine = koin.get<ChatEngine>() as MockChatEngine
        chatEngine.setContacts(listOf(GROUP_NOOBS, GROUP_BOOBS, FRIEND_CHRIS, FRIEND_LARRY, FRIEND_ROBBY, FRIEND_MGC))
        scope = ContactListViewModelTestScope(koin)
        scope.initTestScope(
            scope = this,
            viewModel = viewModel,
            stateData = viewModel.stateData,
        )
        scope.block()
    }
}

private fun ContactListViewModelTestScope.assertContactsList(
    vararg contacts: ChatContact,
) = assertState(
    ContactListState(
        self = selfProfileRepo.self,
        contacts = contacts.toList(),
    )
)

private fun ContactListViewModelTestScope.assertInitialState() = assertContactsList(
    GROUP_BOOBS,
    GROUP_NOOBS,
    FRIEND_CHRIS,
    FRIEND_LARRY,
    FRIEND_MGC,
    FRIEND_ROBBY,
)