package nand.apps.chat.ui.dialog

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.UnconfinedTestDispatcher
import kotlinx.coroutines.test.resetMain
import kotlinx.coroutines.test.runTest
import kotlinx.coroutines.test.setMain
import nand.apps.audio.AudioFrame
import nand.apps.chat.audio.AudioRecorder
import nand.apps.chat.audio.TestAudioRecorder
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.ui.dialog.audio.AudioRecorderDialogState
import nand.apps.chat.ui.dialog.audio.AudioRecorderDialogViewModel
import nand.apps.chat.ui.viewmodel.ViewModelTestScope
import okio.Path.Companion.toPath
import org.koin.core.parameter.parametersOf
import kotlin.test.*

private const val BIT_RATE = 64000
private const val FILE_PATH = "audio/test.opus"

private typealias AudioRecorderViewModelTestScope = ViewModelTestScope<AudioRecorderDialogViewModel, AudioRecorderDialogState, Unit>

@OptIn(ExperimentalCoroutinesApi::class)
class AudioRecorderDialogViewModelTests {

    private lateinit var scope: AudioRecorderViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Start and stop audio recording`() = viewModelTest { initialState ->
        // Record a few seconds of audio
        viewModel.toggleRecording()
        var state = assertState(
            initialState.copy(
                isActive = true,
                isSharedRecorderActive = false,
                recordDurationSecs = 0.0,
            )
        )
        assertTrue(recorder.isActive, "Audio recorder is not active")
        val numFrames = 3
        val numChannels = 2
        val samplesPerFrame = 48
        for (i in 0..<numFrames) {
            val value = i.toShort()
            recorder.onFrame.tryEmit(
                AudioFrame(
                    samples = ShortArray(samplesPerFrame) { value },
                    channels = numChannels,
                    sampleRate = samplesPerFrame / numChannels,
                )
            )
        }
        state = assertState(state.copy(recordDurationSecs = 1.0))
        state = assertState(state.copy(recordDurationSecs = 2.0))
        state = assertState(state.copy(recordDurationSecs = 3.0))

        // Stop recording
        viewModel.toggleRecording()
        assertTrue(!recorder.isActive, "Audio recorder is still active")
        state = assertState(state.copy(isActive = false))
        state = assertState(state.copy(isPathValid = true))
        assertNoStateChange()

        // Check file for bytes we wrote earlier
        fileSystem.read(FILE_PATH.toPath()) {
            // TestAudioFileWriter writes header info first
            val sampleRate = readInt()
            val channels = readInt()
            val bitRate = readInt()
            assertEquals(
                expected = samplesPerFrame / numChannels,
                actual = sampleRate,
            )
            assertEquals(
                expected = numChannels,
                actual = channels,
            )
            assertEquals(
                expected = BIT_RATE,
                actual = bitRate,
            )
            for (i in 0..<numFrames) {
                val value = i.toShort()
                for (s in 0..<samplesPerFrame) {
                    val sample = readShort()
                    assertEquals(
                        expected = value,
                        actual = sample,
                    )
                }
            }
        }
    }

    @Test
    fun `Start recording and delete file before finishing`() = viewModelTest { initialState ->
        // Start recording
        viewModel.toggleRecording()
        val state = assertState(
            initialState.copy(
                isActive = true,
                isSharedRecorderActive = false,
                recordDurationSecs = 0.0,
            )
        )
        assertTrue(recorder.isActive, "Audio recorder is not active")
        recorder.onFrame.tryEmit(
            AudioFrame(
                samples = shortArrayOf(),
                channels = 1,
                sampleRate = 1,
            )
        )
        assertTrue(fileSystem.exists(FILE_PATH.toPath()), "Audio file does not exist")

        // Delete audio
        viewModel.deleteTempFile()
        assertTrue(!fileSystem.exists(FILE_PATH.toPath()), "Audio file still exists after deletion")
    }

    private fun viewModelTest(block: AudioRecorderViewModelTestScope.(AudioRecorderDialogState) -> Unit) = runTest {
        val koin = initKoinTest().koin
        val initialState = AudioRecorderDialogState(
            filePath = FILE_PATH,
            bitRate = BIT_RATE,
            metadata = null,
            shouldEncrypt = false,
        )
        scope = AudioRecorderViewModelTestScope(koin)
        val viewModel: AudioRecorderDialogViewModel = koin.get(parameters = { parametersOf(initialState, scope.delayScope) })
        scope.initTestScope(
            scope = this,
            viewModel = viewModel,
            stateData = viewModel.stateData,
        )
        with(scope) {
            assertState(initialState)
            block(initialState)
        }
    }
}

private val AudioRecorderViewModelTestScope.recorder get() = koin.get<AudioRecorder>() as TestAudioRecorder