package nand.apps.chat.ui.embed.audio

import io.ktor.http.*
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.*
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.model.audio.AudioPlaybackState
import nand.apps.chat.ui.viewmodel.ViewModelTestScope
import nand.apps.chat.util.*
import org.koin.core.parameter.parametersOf
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test

private typealias AudioEmbedViewModelTestScope = ViewModelTestScope<AudioEmbedViewModel, AudioEmbedState, AudioEmbedAction>

@OptIn(ExperimentalCoroutinesApi::class)
class AudioEmbedViewModelTests {

    private lateinit var scope: AudioEmbedViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Audio playback and seeking`() = viewModelTest(MP3_SAMPLE_V2) { initialState ->
        var seekTo = 20.0
        onDelay {
            runTest {
                viewModel.seekTo(seekTo)
                seekTo += 20.0
            }
        }
        viewModel.togglePlayback()
        var state = assertState(initialState.copy(playbackState = AudioPlaybackState.LOADING))
        state = assertState(
            state.copy(
                playbackState = AudioPlaybackState.PLAYING,
                playablePath = "cache/.tmp/audio/${MP3_SAMPLE_V2.sha256()}.mp3",
                metadata = MP3_SAMPLE_V2_METADATA,
            )
        )
        state = assertState(state.copy(duration = 60.0))
        state = assertState(state.copy(timestamp = 20.0))
        state = assertState(state.copy(timestamp = 40.0))
        state = assertState(state.copy(timestamp = 60.0))
        state = assertState(state.copy(playbackState = AudioPlaybackState.STOPPED))
        assertState(state.copy(timestamp = 0.0))
        assertNoStateChange()
    }

    @Test
    fun `Audio playback and stop`() = viewModelTest(MP3_SAMPLE_V1) { initialState ->
        onDelay {
            runTest {
                val currentState = viewModel.stateData.value
                if (currentState.timestamp < 25) {
                    viewModel.seekTo(25.0)
                } else {
                    viewModel.stopPlayback()
                }
            }
        }
        viewModel.togglePlayback()
        var state = assertState(initialState.copy(playbackState = AudioPlaybackState.LOADING))
        state = assertState(
            state.copy(
                playbackState = AudioPlaybackState.PLAYING,
                playablePath = "cache/.tmp/audio/${MP3_SAMPLE_V1.sha256()}.mp3",
                metadata = MP3_SAMPLE_V1_METADATA,
            )
        )
        state = assertState(state.copy(duration = 60.0))
        state = assertState(state.copy(timestamp = 25.0))
        state = assertState(state.copy(playbackState = AudioPlaybackState.STOPPED))
        assertState(state.copy(timestamp = 0.0))
        assertNoStateChange()
    }

    @Test
    fun `Audio deletion dialog`() = viewModelTest { state ->
        viewModel.showDeleteDialog()
        assertState(state.copy(isDeleteDialogVisible = true))
        viewModel.dismissDeleteDialog()
        assertState(state.copy(isDeleteDialogVisible = false))
        viewModel.showDeleteDialog()
        assertState(state.copy(isDeleteDialogVisible = true))
        viewModel.confirmDeletion()
        assertState(state.copy(isDeleteDialogVisible = false, playablePath = null))
    }

    private fun viewModelTest(
        filePath: String = MP3_SAMPLE_V1,
        block: AudioEmbedViewModelTestScope.(AudioEmbedState) -> Unit,
    ) = runTest {
        val koin = initKoinTest().koin
        val initialState = AudioEmbedState(
            uri = filePath,
            type = ContentType.Audio.MPEG,
            reuseCache = false,
        )
        scope = AudioEmbedViewModelTestScope(koin)
        val viewModel: AudioEmbedViewModel = koin.get(parameters = { parametersOf(initialState, scope.delayScope) })
        scope.initTestScope(
            scope = this,
            viewModel = viewModel,
            stateData = viewModel.stateData,
            actionData = viewModel.actionData,
        )
        with(scope) {
            fileSystem.writeTestResource(filePath)
            assertState(initialState)
            block(initialState)
        }
    }
}