package nand.apps.chat.ui.emoji

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.UnconfinedTestDispatcher
import kotlinx.coroutines.test.resetMain
import kotlinx.coroutines.test.runTest
import kotlinx.coroutines.test.setMain
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.emoji.EmojiRepo
import nand.apps.chat.emoji.TestEmojiRepo
import nand.apps.chat.model.emoji.EmojiData
import nand.apps.chat.model.emoji.EmojiGroup
import nand.apps.chat.ui.viewmodel.ViewModelTestScope
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test

private typealias EmojiPopupViewModelTestScope = ViewModelTestScope<EmojiPopupViewModel, EmojiPopupState, EmojiPopupAction>

@OptIn(ExperimentalCoroutinesApi::class)
class EmojiPopupViewModelTests {

    private lateinit var scope: EmojiPopupViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Emoji popup search and refresh`() = viewModelTest {
        // Initial refresh
        viewModel.refresh()
        var state = assertState(
            EmojiPopupState(
                emojis = EMOJIS[EmojiGroup.SMILEYS].orEmpty(),
                groups = EMOJIS.keys.toList(),
                isLoading = false,
            )
        )

        // Change group to food
        viewModel.setGroup(EmojiGroup.FOOD)
        state = assertState(state.copy(currentGroup = EmojiGroup.FOOD))
        state = assertState(state.copy(emojis = EMOJIS[EmojiGroup.FOOD].orEmpty()))

        // Toggle search on
        viewModel.toggleSearch()
        state = assertState(state.copy(isSearchVisible = true))

        // Search for all food emojis with "melon"
        viewModel.search("melon")
        state = assertState(state.copy(searchTerms = "melon"))
        state = assertState(state.copy(emojis = EMOJIS[EmojiGroup.FOOD]?.filter { it.description.contains("melon") }.orEmpty()))

        // Toggle search off
        viewModel.toggleSearch()
        state = assertState(
            state.copy(
                isSearchVisible = false,
                searchTerms = "",
            )
        )
        assertState(state.copy(emojis = EMOJIS[EmojiGroup.FOOD].orEmpty()))
    }

    private fun viewModelTest(block: EmojiPopupViewModelTestScope.() -> Unit) = runTest {
        val koin = initKoinTest().koin
        val emojiRepo = koin.get<EmojiRepo>() as TestEmojiRepo
        emojiRepo.groups = EMOJIS
        val viewModel: EmojiPopupViewModel = koin.get()
        scope = EmojiPopupViewModelTestScope(koin)
        scope.initTestScope(
            scope = this,
            viewModel = viewModel,
            stateData = viewModel.stateData,
            actionData = viewModel.actionData,
        )
        with(scope) {
            assertState(EmojiPopupState())
            block()
        }
    }
}

private val EMOJIS = mapOf(
    EmojiGroup.SMILEYS to listOf(
        EmojiData("\uD83D\uDE00", "grinning face"),
        EmojiData("\uD83D\uDE03", "grinning face with big eyes"),
        EmojiData("\uD83D\uDE04", "grinning face with smiling eyes"),
    ),
    EmojiGroup.FOOD to listOf(
        EmojiData("\uD83C\uDF47", "grapes"),
        EmojiData("\uD83C\uDF48", "melon"),
        EmojiData("\uD83C\uDF49", "watermelon"),
    )
)