package nand.apps.chat.ui.file.sender

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.test.*
import nand.apps.chat.camera.CameraLauncher
import nand.apps.chat.camera.TestCameraLauncher
import nand.apps.chat.di.initKoinTest
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.ChatFileBrowser
import nand.apps.chat.io.TestFileBrowser
import nand.apps.chat.io.nameWithoutExtension
import nand.apps.chat.model.event.ChatEvent
import nand.apps.chat.model.event.TakePhotoEvent
import nand.apps.chat.repo.ChatAvatarRepo
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.util.TEST_PROFILE
import nand.apps.chat.util.assertValue
import nand.apps.chat.util.collect
import nand.apps.chat.util.get
import okio.FileSystem
import okio.Path.Companion.toPath
import okio.fakefilesystem.FakeFileSystem
import org.koin.core.context.stopKoin
import org.koin.dsl.module
import kotlin.test.*

@OptIn(ExperimentalCoroutinesApi::class)
class ChatFileSenderViewModelTests {

    private lateinit var scope: ChatFileSenderViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Self name change updates state`() = viewModelTest { initialState ->
        selfProfileRepo.setCurrentDisplayName("Doggie")
        assertState(initialState.copy(selfName = "Doggie"))
    }

    @Test
    fun `Attach file workflow using file picker`() = viewModelTest { initialState ->
        // Check that we use the "pictures" directory as fallback
        viewModel.attachFile()
        assertState(
            initialState.copy(
                isFilePickerVisible = true,
                defaultFilePickerDirectory = "pictures",
            )
        )

        // Change to "send_directory" and check again
        settingsRepo.update { chatSettings.copy(defaultSendDirectory = "send_directory") }
        viewModel.attachFile()
        assertState(
            initialState.copy(
                isFilePickerVisible = true,
                defaultFilePickerDirectory = "send_directory",
            )
        )

        // Simulate file selection
        viewModel.confirmFilePickerSelection("other_directory/test.png".toPath())
        assertEquals(
            expected = "other_directory",
            actual = settingsRepo.settings.chatSettings.defaultSendDirectory,
        )
        assertAction(ConfirmFileAction("other_directory/test.png"))
    }

    @Test
    fun `Attach file workflow using file browser`() = viewModelTest(isBrowserSupported = true) {
        // Immediately skips to confirm file
        viewModel.attachFile()
        assertEquals(
            expected = "test",
            actual = settingsRepo.settings.chatSettings.defaultSendDirectory,
        )
        assertAction(ConfirmFileAction("test/file.txt"))
    }

    @Test
    fun `Select file using gallery`() = viewModelTest(isGallerySupported = true) { initialState ->
        viewModel.launchGallery()
        assertState(
            initialState.copy(
                cropImagePath = "test/image.png",
                outputImagePath = "test/image-cropped.jpg",
            )
        )

        // Confirm image crop - image doesn't exist
        viewModel.cropFile()
        assertAction(DismissAction)

        // Create image and retry
        fileSystem.createDirectories("test".toPath())
        fileSystem.write("test/image-cropped.jpg".toPath()) { writeInt(1) }
        viewModel.cropFile()
        assertAction(ConfirmFileAction("test/image-cropped.jpg"))
    }

    @Test
    fun `Take photo using camera`() = viewModelTest(isCameraSupported = true) { initialState ->
        viewModel.launchCamera()
        val takePhoto = eventChannel.get()
        assertIs<TakePhotoEvent>(takePhoto, "TakePhoto event not received")
        assertTrue(takePhoto.path.nameWithoutExtension.toLongOrNull() != null, "Photo name isn't using UNIX timestamp")
        assertState(
            initialState.copy(
                cropImagePath = takePhoto.path.toString(),
                outputImagePath = "pictures/.tmp/${takePhoto.path.nameWithoutExtension}-cropped.jpg",
            )
        )
    }

    @Test
    fun `Record audio and confirm file`() = viewModelTest { initialState ->
        // Launch recorder and verify expected file name format
        viewModel.launchAudioRecorder()
        val state = stateChannel.get() ?: error("State is null")
        val fileNameRegex = Regex("cache/\\.tmp/audio/\\d{4}-\\d{2}-\\d{2} \\d{1,2}_\\d{2}_\\d{2} (AM|PM).opus")
        assertTrue(
            actual = state.isRecordAudioDialogVisible,
            message = "Audio recorder not visible",
        )
        assertTrue(
            actual = state.audioFilePath.matches(fileNameRegex),
            message = "Incorrect audio file path: ${state.audioFilePath}",
        )

        // Confirm selection
        viewModel.onAudioRecorded()
        assertState(state.copy(isRecordAudioDialogVisible = false))
        assertAction(ConfirmFileAction(state.audioFilePath))
    }

    private fun viewModelTest(
        isCameraSupported: Boolean = false,
        isBrowserSupported: Boolean = false,
        isGallerySupported: Boolean = false,
        block: ChatFileSenderViewModelTestScope.(ChatFileSenderState) -> Unit,
    ) = runTest {
        scope = ChatFileSenderViewModelTestScope(
            isCameraSupported = isCameraSupported,
            isBrowserSupported = isBrowserSupported,
            isGallerySupported = isGallerySupported,
        )
        scope.initTestScope(this)
        with(scope) {
            val initialState = assertState(
                ChatFileSenderState(
                    selfName = selfProfileRepo.self.name,
                    isCameraSupported = isCameraSupported,
                    isGallerySupported = isGallerySupported,
                )
            )
            block(initialState)
        }
    }
}

private class ChatFileSenderViewModelTestScope(
    isCameraSupported: Boolean,
    isBrowserSupported: Boolean,
    isGallerySupported: Boolean,
) {

    private val koin by lazy {
        initKoinTest {
            val modules = module {
                single<CameraLauncher> { TestCameraLauncher(isCameraSupported, get()) }
                single<ChatFileBrowser> { TestFileBrowser(isBrowserSupported, isGallerySupported) }
            }
            modules(modules)
        }.koin
    }
    val contactRepo: ContactRepo by koin.inject()
    val selfProfileRepo: SelfProfileRepo by koin.inject()
    val chatAvatarRepo: ChatAvatarRepo by koin.inject()
    val settingsRepo: SettingsRepo by koin.inject()
    val fileSystem: FakeFileSystem by lazy { koin.get<FileSystem>() as FakeFileSystem }
    val eventHandler: ChatEventHandler by koin.inject()
    val viewModel: ChatFileSenderViewModel by koin.inject()

    private val repos by lazy {
        listOf(contactRepo, chatAvatarRepo, selfProfileRepo)
    }

    lateinit var coroutineScope: TestScope
    lateinit var stateChannel: Channel<ChatFileSenderState>
    lateinit var actionChannel: Channel<ChatFileSenderAction>
    lateinit var eventChannel: Channel<ChatEvent>

    init {
        fileSystem.emulateUnix()
        runTest {
            repos.forEach {
                it.onEngineInit(TEST_PROFILE)
                it.onEngineStart(this)
            }
        }
    }

    fun initTestScope(scope: TestScope) {
        this.coroutineScope = scope
        this.stateChannel = scope.collect(viewModel.stateData)
        this.actionChannel = scope.collect(viewModel.actionData)
        this.eventChannel = scope.collect(eventHandler.onEvent)
    }

    fun destroy() {
        repos.forEach { it.onEngineStop() }
        stopKoin()
    }

    fun assertState(expected: ChatFileSenderState): ChatFileSenderState {
        stateChannel.assertValue(expected)
        assertStateValid(expected)
        return expected
    }

    fun assertAction(expected: ChatFileSenderAction) {
        actionChannel.assertValue(expected)
    }

    private fun assertStateValid(state: ChatFileSenderState) {
        if (state.isFilePickerVisible) {
            assertTrue(
                actual = state.defaultFilePickerDirectory.isNotBlank(),
                message = "File picker directory is blank while file picker is visible",
            )
        }
        if (state.cropImagePath.isNotBlank()) {
            assertTrue(
                actual = state.outputImagePath.isNotBlank(),
                message = "Cropped image output path is blank",
            )
        }
        if (state.isRecordAudioDialogVisible) {
            assertTrue(
                actual = state.audioFilePath.isNotBlank(),
                message = "Audio path is blank while audio recorder is visible",
            )
        }
    }
}