package nand.apps.chat.ui.viewmodel

import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.launch
import kotlinx.coroutines.test.TestScope
import kotlinx.coroutines.test.advanceUntilIdle
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.io.TestCoroutineDelayScope
import nand.apps.chat.model.event.ChatEvent
import nand.apps.chat.repo.*
import nand.apps.chat.util.TEST_PROFILE
import nand.apps.chat.util.assertValue
import nand.apps.chat.util.collect
import okio.FileSystem
import okio.fakefilesystem.FakeFileSystem
import org.koin.core.Koin
import org.koin.core.context.stopKoin

class ViewModelTestScope<V, S, A>(val koin: Koin) {

    private var _viewModel: V? = null
    val viewModel by lazy { _viewModel ?: error("Must call initTestScope() first") }
    val delayScope = TestCoroutineDelayScope()

    val contactRepo: ContactRepo by koin.inject()
    val messageRepo: ChatMessageRepo by koin.inject()
    val callRepo: CallRepo by koin.inject()
    val fileTransferRepo: FileTransferRepo by koin.inject()
    val settingsRepo: SettingsRepo by koin.inject()
    val selfProfileRepo: SelfProfileRepo by koin.inject()
    val chatAvatarRepo: ChatAvatarRepo by koin.inject()
    val eventHandler: ChatEventHandler by koin.inject()
    val engineCallback: ChatEngineCallback by koin.inject()
    val fileSystem: FakeFileSystem = koin.inject<FileSystem>().value as FakeFileSystem

    lateinit var coroutineScope: TestScope
    lateinit var stateChannel: Channel<S>
    lateinit var actionChannel: Channel<A>
    lateinit var eventChannel: Channel<ChatEvent>

    @OptIn(ExperimentalCoroutinesApi::class)
    fun initTestScope(
        scope: TestScope,
        viewModel: V,
        stateData: Flow<S>,
        actionData: Flow<A>? = null,
    ) {
        fileSystem.emulateUnix()
        this.coroutineScope = scope
        scope.launch {
            engineCallback.onEngineInit(TEST_PROFILE)
            engineCallback.onEngineStart(scope)
        }
        scope.advanceUntilIdle()
        _viewModel = viewModel
        this.stateChannel = scope.collect(stateData)
        actionData?.let { this.actionChannel = scope.collect(it) }
        this.eventChannel = scope.collect(eventHandler.onEvent)
    }

    fun destroy() {
        engineCallback.onEngineStop()
        stopKoin()
    }

    fun assertState(expected: S): S {
        stateChannel.assertValue(expected)
        return expected
    }

    fun assertAction(expected: A) {
        actionChannel.assertValue(expected)
    }

    fun assertEvent(expected: ChatEvent) {
        eventChannel.assertValue(expected)
    }

    fun assertNoStateChange() = stateChannel.assertValue(null)

    fun assertNoAction() = actionChannel.assertValue(null)

    fun assertNoEvent() = eventChannel.assertValue(null)

    fun pushEvent(event: ChatEvent) {
        eventHandler += event
        assertEvent(event)
    }

    fun onDelay(block: suspend () -> Unit) {
        delayScope.block = block
    }
}