import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.runBlocking
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.di.initDesktopKoin
import nand.apps.chat.engine.ChatEngine
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.engine.ChatEngineType
import nand.apps.chat.model.group.ChatGroupInvite
import nand.apps.chat.model.net.NetworkProxyType
import nand.apps.chat.model.settings.net.ChatRemoteSettingsData
import nand.apps.chat.model.settings.net.DEFAULT_REMOTE_SERVER_HOST
import nand.apps.chat.model.settings.net.DEFAULT_REMOTE_SERVER_PORT
import nand.apps.chat.model.settings.net.NetworkProxySettingsData
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.model.user.FriendRequest
import nand.apps.chat.profile.getDefaultProfilePath
import nand.apps.chat.profile.getPrimaryProfile
import nand.apps.chat.repo.ContactRepo
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.util.collect
import okio.Path
import org.jetbrains.skiko.MainUIDispatcher
import kotlin.system.exitProcess

class HeadlessApp {

    private val scope = CoroutineScope(MainUIDispatcher)
    private val koin = initDesktopKoin(scope)
    private val engine: ChatEngine by koin.inject()
    private val contactRepo: ContactRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val settingsRepo: SettingsRepo by koin.inject()
    private val fileSystem: ChatFileSystem by koin.inject()

    private val profileDirectory = fileSystem.getProfileDirectory()
    private val settings get() = settingsRepo.settings

    init {
        runBlocking { settingsRepo.load() }
    }

    fun startLoginFlow() {
        if (settings.generalSettings.isTutorialComplete) {
            promptUsername()
        } else {
            tutorialFlow()
        }
    }

    private fun tutorialFlow() {
        println("Welcome to $APP_NAME")
        println("This is the CLI version of the app intended for running chat bots")
        println()
        if (promptConfirmation("Would you like to setup a proxy?")) {
            promptProxyType()
        } else {
            promptConnectionType()
        }
    }

    private fun promptProxyType() {
        println("Select proxy type [HTTP/SOCKS]:")
        val proxyType = when (val type = readlnOrNull()?.uppercase()) {
            "HTTP", "HTTPS" -> NetworkProxyType.HTTP
            "SOCKS", "SOCKS5" -> NetworkProxyType.SOCKS
            else -> {
                printErr("Invalid proxy type $type")
                promptProxyType()
                return
            }
        }
        promptProxyHost(proxyType)
    }

    private fun promptProxyHost(type: NetworkProxyType) {
        println("Enter proxy host:")
        val host = readlnOrNull() ?: error("Proxy host is required")
        val port = readlnOrNull()?.toIntOrNull()?.takeIf { it > 0 } ?: error("Proxy port is required")
        if (!promptConfirmation("Using $type proxy for $host with port $port. Is this correct?")) {
            promptProxyType()
            return
        }
        setProxySettings(
            NetworkProxySettingsData(
                type = type,
                host = host,
                port = port,
            )
        )
        println("Successfully setup network proxy.")
        promptConnectionType()
    }

    private fun setProxySettings(settings: NetworkProxySettingsData?) {
        settingsRepo.update { copy(networkSettings = networkSettings.copy(proxy = settings)) }
        saveSettings()
    }

    private fun promptConnectionType() {
        val isUdpEnabled = promptConfirmation("Would you like to enable direct connections to peers via UDP?")
        settingsRepo.update {
            copy(
                generalSettings = generalSettings.copy(isTutorialComplete = true),
                networkSettings = networkSettings.copy(isUdpEnabled = isUdpEnabled)
            )
        }
        saveSettings()
        promptRemote()
    }

    private fun promptRemote() {
        println("Enter host for remote connections [$DEFAULT_REMOTE_SERVER_HOST]:")
        val host = readlnOrNull() ?: DEFAULT_REMOTE_SERVER_HOST

        println("Enter port for remote connections [$DEFAULT_REMOTE_SERVER_PORT]:")
        val port = readlnOrNull()?.toIntOrNull() ?: DEFAULT_REMOTE_SERVER_PORT

        updateRemoteSettings {
            it.copy(
                isEnabled = true,
                host = host,
                port = port,
            )
        }
        promptUsername()
    }

    private fun promptUsername() {
        println("Enter profile username:")
        val username = readlnOrNull() ?: error("Username is required")
        checkValidUsername(username)
    }

    private fun checkValidUsername(username: String) {
        val dir = profileDirectory.getSubDirectory(username)
        val profile = dir.getPrimaryProfile() ?: dir.getDefaultProfilePath(ChatEngineType.TOX)
        if (profile.name in dir) {
            promptPassword(username)
        } else {
            if (promptConfirmation("$username does not exist. Create new profile?")) {
                promptPassword(username)
            } else {
                promptUsername()
            }
        }
    }

    private fun promptPassword(username: String) {
        println("Enter profile password:")
        val password = readlnOrNull().orEmpty()
        attemptLogin(username, password)
    }

    private fun attemptLogin(username: String, password: String) {
        val profileData = ChatProfileData(
            filePath = getProfile(username).toString(),
            username = username,
            password = password,
            networkSettings = settings.networkSettings,
        )
        val result = runBlocking { engine.initialize(profileData) }
        when (result) {
            ChatEngineInitResult.SUCCESS -> {
                println("Successfully logged into $username")
                println("Your chat address: ${selfProfileRepo.getSelfAddress()}")
                engine.start(scope)
                scope.collect(contactRepo.onFriendRequest) { request -> onFriendRequest(request) }
                scope.collect(contactRepo.onGroupInvite) { invite -> onGroupInvite(invite) }
            }
            ChatEngineInitResult.INVALID_CREDENTIALS -> {
                printErr("Incorrect password for \"$username\"")
                promptPassword(username)
            }
            ChatEngineInitResult.INVALID_PROFILE -> {
                printErr("Invalid profile \"$username\"")
                promptUsername()
            }
            ChatEngineInitResult.PROXY_BAD_HOST,
            ChatEngineInitResult.PROXY_BAD_PORT,
            ChatEngineInitResult.PROXY_NOT_FOUND -> {
                printErr("Invalid proxy setup ($result)")
                when {
                    promptConfirmation("Edit proxy settings?") -> promptProxyType()
                    promptConfirmation("Remove proxy?") -> {
                        setProxySettings(null)
                        attemptLogin(username, password)
                    }
                    else -> exitProcess(1)
                }
            }
            ChatEngineInitResult.INTERNAL_ERROR -> {
                if (promptConfirmation("Failed to login due to unexpected error. Retry?")) {
                    promptUsername()
                } else {
                    exitProcess(1)
                }
            }
        }
    }

    // TODO - Timeout readLine() after 5 minutes
    private fun onFriendRequest(request: FriendRequest) {
        println("Received friend request from ${request.uid.value}")
        println("Message: ${request.message}")
        if (promptConfirmation("Add this friend?")) {
            contactRepo.acceptFriendRequest(request)
        } else {
            contactRepo.declineFriendRequest(request)
        }
    }

    // TODO - Timeout readLine() after 5 minutes
    private fun onGroupInvite(invite: ChatGroupInvite) {
        val friend = contactRepo.getFriend(invite.friendUid) ?: return
        println("Received group invite from ${friend.displayName}")
        println("Group key: ${invite.cookie}")
        if (promptConfirmation("Join this group?")) {
            contactRepo.acceptGroupInvite(invite)
        } else {
            contactRepo.removeGroupInvite(invite)
        }
    }

    private fun getProfile(username: String): Path {
        val dir = profileDirectory.getSubDirectory(username)
        return dir.getPrimaryProfile() ?: dir.getDefaultProfilePath(ChatEngineType.TOX)
    }

    private fun printErr(text: String) = System.err.println(text)

    private fun promptConfirmation(text: String): Boolean {
        println("$text [y/n]")
        val answer = readlnOrNull()?.lowercase()
        return answer == "y" || answer == "yes"
    }

    private fun saveSettings() = runBlocking { settingsRepo.save() }

    private fun updateRemoteSettings(block: (ChatRemoteSettingsData) -> ChatRemoteSettingsData) {
        settingsRepo.update {
            copy(networkSettings = networkSettings.copy(remoteSettings = block(networkSettings.remoteSettings)))
        }
        saveSettings()
    }
}

fun main() {
    val app = HeadlessApp()
    app.startLoginFlow()
}