/**
 * A chatbot that listens for "/wiki" commands from other users and responds with a Wikipedia URL
 *
 * Setup:
 * 1) Launch Seers Lodge and create/login to the bot account you want to use
 * 2) Invite the bot to the groups you want to serve
 * 3) Turn on the "Enable Remote Access" network setting
 * 4) Run this demo and enter the bot's login credentials
 * 5) Your bot should now respond to "/wiki" searches in chat (i.e. "/wiki banana boat food")
 */

package demo

import io.ktor.http.*
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import kotlinx.serialization.json.jsonObject
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.message.ChatMessageType
import nand.apps.chat.model.settings.net.ChatRemoteSettingsData
import nand.apps.chat.model.settings.net.DEFAULT_REMOTE_SERVER_HOST
import nand.apps.chat.model.settings.net.DEFAULT_REMOTE_SERVER_PORT
import nand.apps.chat.model.user.ChatProfileData
import nand.apps.chat.net.remote.ChatRemoteClient
import nand.apps.chat.net.remote.command.HandshakeCommand
import nand.apps.chat.net.remote.command.SendMessageCommand
import nand.apps.chat.net.remote.event.*
import nand.apps.chat.util.*

// Host and port that the app is listening on
private const val HOST = DEFAULT_REMOTE_SERVER_HOST
private const val PORT = DEFAULT_REMOTE_SERVER_PORT

// Set your preferred language here
private const val WIKIPEDIA_LANGUAGE = "en"
private const val WIKIPEDIA_URL_HOST = "https://$WIKIPEDIA_LANGUAGE.wikipedia.org"

private val httpClient by lazy { createDefaultHttpClient() }
private val logger = getLogger("ChatRemoteClientDemo")

fun main() {
    println("Please enter the profile username to connect to:")
    val userName = readlnOrNull() ?: error("Username is required")

    println("Please enter profile password:")
    val password = readlnOrNull().orEmpty()

    runBlocking {
        val profile = ChatProfileData(
            username = userName,
            password = password,
            filePath = "",
        )
        val client = ChatRemoteClient()
        client.init(
            profile = profile,
            coroutineScope = this,
            settings = ChatRemoteSettingsData(
                host = HOST,
                port = PORT,
            ),
        )
        launch {
            client.onEvent.collect { event ->
                client.onEvent(event)
            }
        }
        try {
            client.sendCommand(HandshakeCommand())
        } catch (e: Exception) {
            logger.error("Failed to send handshake command - aborting")
            throw e
        }
    }
}

private suspend fun ChatRemoteClient.onEvent(event: ChatRemoteEvent) {
    when (event) {
        is CommandResponseEvent,
        is ContactListEvent,
        is FriendRequestEvent,
        is GroupInviteEvent -> Unit
        is NewMessageEvent -> {
            if (event.message.startsWith("/wiki ")) {
                try {
                    searchWiki(event)
                } catch (e: Exception) {
                    logger.error("Failed to perform wiki search (${event.message})", e)
                }
            }
        }
    }
}

/**
 * Performs a Wikipedia search using "/wiki" commands sent by other users and sends a message
 * containing a URL for the first result
 */
private suspend fun ChatRemoteClient.searchWiki(event: NewMessageEvent) {
    val query = event.message.substringAfter("/wiki ").trim().encodeURLQueryComponent()
    val url = "$WIKIPEDIA_URL_HOST/w/api.php?action=query&format=json&list=search&srsearch=$query"
    val results = httpClient.getJsonObject(url)?.find("query", "search").jsonArraySafe()
    val primaryResult = results.firstOrNull()?.jsonObject ?: return
    val title = primaryResult["title"].getString() ?: return
    sendCommand(
        SendMessageCommand(
            conversationUid = event.conversationUid,
            contactUid = event.conversationUid,
            message = "$WIKIPEDIA_URL_HOST/wiki/${title.replace(" ", "_").encodeURLQueryComponent()}",
            type = ChatMessageType.NORMAL,
        )
    )
}