package nand.apps.chat.audio

import kotlinx.coroutines.CoroutineScope
import nand.apps.audio.AudioFrame
import nand.apps.audio.AudioHandle
import nand.apps.chat.BuildConfig.HEADLESS_APP
import nand.apps.chat.audio.file.DesktopAudioFileHandle
import nand.apps.chat.audio.file.DesktopAudioFileStreamer
import nand.apps.chat.audio.resampler.DesktopAudioResampler
import nand.apps.chat.log.getLogger
import okio.Path
import java.util.*
import javax.sound.sampled.*
import kotlin.math.roundToLong

// Wait 3 seconds before pausing the audio line
private const val SLEEP_TIMEOUT = 3000L

private const val DEFAULT_FRAMES_PER_SECOND = 50

// Barebones audio player used for both sound effects and VoIP
class DesktopAudioPlayer(
    coroutineScope: CoroutineScope,
    private val soundEffectCache: SoundEffectCache,
) : DesktopAudioLine(), AudioPlayer {

    private val fps = DEFAULT_FRAMES_PER_SECOND
    private val bufSizeFrames = sampleRate / fps
    private val bufSizeBytes = bufSizeFrames * channels * (bitsPerSample / 8)
    private val bufFrac = DoubleArray(bufSizeFrames)

    private val queued = ArrayList<DesktopAudioHandle>()
    private val removed = ArrayList<DesktopAudioHandle>()
    private var globalGroupId = 0L
    private val fileStreamer = DesktopAudioFileStreamer(this, coroutineScope)
    private val resampler = DesktopAudioResampler(this, coroutineScope)

    private var isSleeping = true
    private var sleepStartTime = 0L
    private val sleepDelay = 1000L / fps

    private val logger = getLogger("AudioPlayer")

    init {
        if (!HEADLESS_APP) {
            isRunning = true
            val t = Thread(this, "AudioPlayer")
            t.isDaemon = true
            t.priority = Thread.MAX_PRIORITY
            t.start()
        }
    }

    private fun shutdown() {
        isRunning = false
    }

    override fun queueSound(name: String): AudioHandle? {
        return soundEffectCache.getSoundEffectPath(name)?.let { path ->
            queueFile(path)
        }
    }

    override fun queueFile(path: Path, startTime: Double): AudioHandle? {
        return fileStreamer.streamFile(path, startTime) ?: run {
            logger.warn("Attempted to queue unsupported audio file: $path")
            null
        }
    }

    override fun queueFrame(frame: AudioFrame, last: AudioHandle?): AudioHandle? {
        return queue(frame)?.also { handle ->
            if (last is DesktopAudioHandle) {
                last.next = handle
                handle.prev = last
                handle.groupId = last.groupId
                handle.time = handle.time.coerceAtLeast(last.time + (last.frame.duration * 1000).roundToLong())
            } else {
                handle.groupId = globalGroupId++
            }
            if (handle.needsResample()) {
                resampler.submit(handle)
            }
        }
    }

    override fun stopSound(handle: AudioHandle) {
        when (handle) {
            is DesktopAudioHandle -> remove(handle)
            is DesktopAudioFileHandle -> handle.close()
        }
    }

    internal fun queueHandle(handle: DesktopAudioHandle, delay: Int = 0): Boolean {
        if (!isRunning) return false

        synchronized(queued) {
            handle.time = curTime() + delay
            queued.add(handle)
            isSleeping = false
        }

        return true
    }

    private fun createHandle(wav: AudioFrame): DesktopAudioHandle {
        return DesktopAudioHandle(wav.setChannelCount(channels))
    }

    private fun queue(wav: AudioFrame, delay: Int = 0): DesktopAudioHandle? {
        val handle = createHandle(wav)
        return if (handle.needsResample()) {
            handle
        } else {
            val isQueued = queueHandle(handle, delay)
            if (isQueued) handle else null
        }
    }

    private fun remove(handle: DesktopAudioHandle) {
        synchronized(queued) {
            queued.remove(handle)
            removed.add(handle)
        }
    }

    private fun removeAll(handles: Collection<DesktopAudioHandle>) {
        synchronized(queued) {
            queued.removeAll(handles.toSet())
            removed.addAll(handles)
        }
    }

    private fun clear() {
        synchronized(queued) {
            removed.addAll(queued)
            queued.clear()
        }
    }

    private fun DesktopAudioHandle.needsResample() = frame.sampleRate != sampleRate

    // Relative to system time in milliseconds
    private fun curTime() = System.nanoTime() / 1000000L

    override fun run() {
        try {
            for (i in 0..<bufSizeFrames) {
                bufFrac[i] = i.toDouble() / bufSizeFrames
            }
            val speakerInfo = DataLine.Info(SourceDataLine::class.java, audioFormat)
            val speaker = AudioSystem.getLine(speakerInfo) as SourceDataLine
            speaker.open(audioFormat, bufSizeBytes)
            speaker.start()

            val data = ByteArray(bufSizeBytes)
            val wav = Array(channels) { DoubleArray(bufSizeFrames) }
            val groups = mutableSetOf<Long>()
            var isSpeakerStarted = true
            while (isRunning) {

                // Speaker pause/resume
                when {
                    isSleeping -> {
                        if (isSpeakerStarted) {
                            speaker.write(data, 0, bufSizeBytes)
                            if (curTime() - sleepStartTime > SLEEP_TIMEOUT) {
                                speaker.drain()
                                speaker.stop()
                                isSpeakerStarted = false
                            }
                        } else {
                            Thread.sleep(sleepDelay)
                        }
                        continue
                    }
                    !isSpeakerStarted -> {
                        speaker.start()
                        isSpeakerStarted = true
                    }
                }

                val curTime = curTime()

                synchronized(queued) {

                    val ready = ArrayList(queued)
                    groups.clear()
                    
                    // Queue cleanup
                    var i = 0
                    var j = 0
                    while (i < queued.size) {
                        val s = queued[i]
                        when {
                            s.time > curTime -> ready.removeAt(j--)
                            s.isFinished -> queued.removeAt(i--)
                        }
                        i++
                        j++
                    }
                    for (h in removed) {
                        ready.add(h)
                        h.setVolumeCalculator(null)
                        h.volume = 0.0
                    }
                    
                    if (queued.isEmpty() && removed.isEmpty()) {
                        // Nothing to play so just sleep
                        data.fill(0.toByte())
                        isSleeping = true
                        sleepStartTime = curTime
                    } else {
                        for (c in 0..<channels) wav[c].fill(0.0)

                        // Audio mixing
                        for (handle in ready) {
                            if (handle.prev?.isFinished == false) {
                                // Do not play the next sample in this group before the previous is finished
                                continue
                            }
                            var s = handle.getNextPlayable() ?: continue
                            if (s.groupId != -1L) {
                                // Prevent overlap of samples in the same group
                                if (s.groupId in groups) continue else groups += s.groupId
                            }
                            val dstVol = s.volume
                            val volChange = dstVol != s.srcVol
                            for (f in 0..<bufSizeFrames) {
                                if (s.isFinished) s = s.getNextPlayable() ?: break
                                var vol = s.srcVol
                                if (volChange) {
                                    // Volume smoothing
                                    vol = s.srcVol * (1 - bufFrac[f]) + (dstVol * bufFrac[f])
                                }
                                for (c in 0..<channels) {
                                    val amp = s.frame.samples[s.pos * 2 + c] / SAMPLE_RANGE
                                    wav[c][f] += amp * vol
                                }
                                s.increment()
                            }
                            s.srcVol = dstVol
                        }
                        convertSamplesToBytes(wav, data)
                        for (h in removed) h.finish()
                        removed.clear()
                    }
                }
                
                speaker.write(data, 0, bufSizeBytes)
            }

            speaker.drain()
            speaker.close()
        } catch (e: Exception) {
            logger.error("Something went horribly wrong", e)
        }
    }
    
    private fun convertSamplesToBytes(samples: Array<DoubleArray>, out: ByteArray) {
        var d = 0
        for (f in 0..<bufSizeFrames) {
            for (c in 0..<channels) {
                var s = samples[c][f]
                when {
                    s > 1 -> s = 1.0
                    s < -1 -> s = -1.0
                }
                if (s < 0) s += 2.0
                val sh = (s * Short.MAX_VALUE).toInt()
                var b2 = sh shr 8
                var b1 = sh and 0xFF
                if (b1 > 127) b1 -= 256
                if (b2 > 127) b2 -= 256
                out[d++] = b1.toByte()
                out[d++] = b2.toByte()
            }
        }
    }
}