package nand.apps.chat.audio.file

import io.ktor.http.*
import kotlinx.coroutines.*
import nand.apps.audio.AudioFileReader
import nand.apps.chat.audio.*
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.file.OPUS
import nand.apps.chat.model.file.contentTypeFromFileName
import nand.apps.mpeg.MpegFileReader
import nand.apps.opus.OpusFileReader
import okio.Path

private const val BUFFER_SIZE = 960 * 2
private const val HANDLE_BUFFER_CAPACITY = 5

class DesktopAudioFileStreamer(
    private val audioPlayer: DesktopAudioPlayer,
    private val coroutineScope: CoroutineScope,
) {

    private val outChannels = audioPlayer.channels
    private val logger = getLogger("DesktopAudioFileStreamer")

    fun streamFile(path: Path, startTime: Double): DesktopAudioFileHandle? {
        val file = getFileReader(path)?.takeIf { it.isOpen } ?: return null
        if (startTime > 0) file.seekTo(startTime)
        val job = coroutineScope.launch(Dispatchers.IO) {
            val buffer = ShortArray(BUFFER_SIZE)
            val handleBuffer = mutableListOf<DesktopAudioHandle>()
            var lastHandle: DesktopAudioHandle? = null
            while (file.isOpen) {
                val frame = file.readToFrame(buffer)?.setChannelCount(
                    channels = outChannels,
                ) ?: run {
                    file.close()
                    return@launch
                }
                val handle = withContext(Dispatchers.Main) {
                    audioPlayer.queueFrame(
                        frame = frame,
                        last = lastHandle,
                    ) as? DesktopAudioHandle ?: run {
                        logger.error("Error queueing audio frame")
                        null
                    }
                } ?: return@launch
                lastHandle = handle
                handleBuffer += handle
                handleBuffer.delayUntilVacant(HANDLE_BUFFER_CAPACITY)
            }
        }
        return DesktopAudioFileHandle(
            file = file,
            job = job,
        )
    }

    private fun getFileReader(path: Path): AudioFileReader? {
        val pathString = path.toString()
        return when (contentTypeFromFileName(path.name)) {
            ContentType.Audio.MPEG -> MpegFileReader(pathString)
            ContentType.Audio.OPUS -> OpusFileReader(pathString)
            else -> null
        }
    }

    private suspend fun MutableList<DesktopAudioHandle>.delayUntilVacant(
        capacity: Int = 1,
        delayMs: Long = 20L,
    ) {
        assert(capacity > 0)
        while (size >= capacity) {
            removeIf { it.isFinished }
            if (size >= capacity) delay(delayMs)
        }
    }
}