package nand.apps.chat.audio.resampler

import kotlinx.coroutines.*
import nand.apps.audio.AudioFrame
import nand.apps.chat.audio.DesktopAudioHandle
import nand.apps.chat.audio.DesktopAudioPlayer
import nand.apps.soxr.SoxrNative
import kotlin.math.roundToLong

private const val MAX_DELAYS = 4

class DesktopAudioResampler(
    private val audioPlayer: DesktopAudioPlayer,
    private val coroutineScope: CoroutineScope,
) {
    private val sampleRate = audioPlayer.sampleRate
    private val queued = mutableSetOf<Long>()

    fun submit(handle: DesktopAudioHandle) {
        if (queued.add(handle.groupId)) {
            coroutineScope.launch(Dispatchers.IO) {
                process(handle)
            }
        }
    }

    private suspend fun process(firstHandle: DesktopAudioHandle) {
        val groupId = firstHandle.groupId
        val frameMs = (firstHandle.duration * 1000).roundToLong()
        var node = firstHandle
        var prevFrame = node.prev?.frame
        var numDelays = 0
        var buffer = ShortArray(node.numSamples * 3)
        while (numDelays < MAX_DELAYS) {
            val next = node.next
            if (next != null) {
                val requiredSize = prevFrame.numSamples + node.numSamples + next.numSamples
                if (requiredSize > buffer.size) {
                    buffer = ShortArray(requiredSize)
                }
                val startIndex = prevFrame.put(buffer, 0)
                val endIndex = node.put(buffer, startIndex)
                next.put(buffer, endIndex)
                val resampled = try {
                    SoxrNative.resample(
                        samples = buffer,
                        channels = node.frame.channels,
                        inputRate = node.frame.sampleRate,
                        outputRate = sampleRate,
                    )
                } catch (ignore: Exception) {
                    null
                }
                prevFrame = node.frame
                node.frame = resampled?.let { samples ->
                    val ratio = samples.size.toDouble() / buffer.size
                    val trStartIndex = (startIndex * ratio).toInt()
                    val trEndIndex = (endIndex * ratio).toInt()
                    AudioFrame(
                        samples = samples.sliceArray(trStartIndex..<trEndIndex),
                        channels = node.frame.channels,
                        sampleRate = sampleRate,
                    )
                } ?: node.frame
                withContext(Dispatchers.Main) {
                    audioPlayer.queueHandle(node)
                }
                node = next
                numDelays = 0
            } else {
                delay(frameMs)
                numDelays++
            }
        }
        withContext(Dispatchers.Main) {
            queued -= groupId
            node.finish()
        }
    }

    private val DesktopAudioHandle?.numSamples get() = this?.frame?.samples?.size ?: 0

    private val AudioFrame?.numSamples get() = this?.samples?.size ?: 0

    private fun DesktopAudioHandle?.put(buffer: ShortArray, index: Int): Int {
        return this?.frame?.put(buffer, index) ?: 0
    }

    private fun AudioFrame?.put(buffer: ShortArray, index: Int): Int {
        this ?: return 0
        samples.copyInto(buffer, index)
        return index + numSamples
    }
}