package nand.apps.chat.ui.widget

import androidx.compose.foundation.background
import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.shape.CircleShape
import androidx.compose.material.Icon
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Close
import androidx.compose.runtime.*
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.Density
import androidx.compose.ui.unit.DpSize
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.compose.ui.window.DialogWindow
import androidx.compose.ui.window.WindowPosition
import androidx.compose.ui.window.rememberDialogState
import com.hoc081098.kmp.viewmodel.koin.compose.koinKmpViewModel
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.ui.logo.LOGO_COLOR
import nand.apps.chat.ui.logo.SeersLogo
import nand.apps.chat.ui.provider.LocalLoginState
import nand.apps.chat.ui.settings.notifications.toAlignment
import nand.apps.chat.ui.text.TextLine
import nand.apps.chat.ui.theme.ChatTheme
import nand.apps.chat.ui.util.LaunchedCollect
import org.jetbrains.compose.resources.stringResource
import seers.composeapp.generated.resources.*

private val WIDGET_BASE_SIZE = 36.dp
private val LOGO_COLOR_DISCONNECTED = Color(0xFFBAA13D)

@Composable
fun ChatWidget(
    viewModel: ChatWidgetViewModel = koinKmpViewModel(),
) {
    val state by viewModel.stateData.collectAsState()
    val settings = state.settings
    val size = WIDGET_BASE_SIZE * settings.size.scale
    val dialogState = rememberDialogState(
        width = size,
        height = size,
        position = WindowPosition(settings.placement.toAlignment()),
    )

    // Update widget state when settings are changed
    LaunchedCollect(viewModel.actionData) { action ->
        when (action) {
            is UpdateDialogAction -> {
                val sizeDp = WIDGET_BASE_SIZE * action.size.scale
                dialogState.size = DpSize(sizeDp, sizeDp)
                dialogState.position = WindowPosition(action.placement.toAlignment())
            }
        }
    }

    val defaultDensity = LocalDensity.current

    if (!settings.isEnabled) return
    
    DialogWindow(
        state = dialogState,
        title = stringResource(Res.string.chat_widget_title, APP_NAME),
        transparent = true,
        undecorated = true,
        resizable = false,
        focusable = false,
        alwaysOnTop = true,
        onCloseRequest = {},
    ) {
        // Flash the widget icon
        val isHighlighted = state.unreadCount > 0 && settings.isFlashEnabled
        ChatTheme {
            CompositionLocalProvider(
                LocalDensity provides Density(defaultDensity.density * settings.size.scale)
            ) {
                Box(
                    modifier = Modifier
                        .clip(ChatTheme.shapes.medium)
                        .background(ChatTheme.colors.backgroundFloating)
                        .clickable { viewModel.toggleAppWindow() }
                        .fillMaxSize(),
                ) {
                    val isLoggedIn = LocalLoginState.current
                    Box(
                        modifier = Modifier
                            .fillMaxSize()
                            .padding(5.dp)
                    ) {
                        val badgeText = when {
                            state.unreadCount > 0 -> state.unreadCount.toString()
                            state.unreadGroupMessages > 0 && settings.isGroupUnreadEnabled -> "٭"
                            else -> ""
                        }
                        SeersLogo(
                            color = when {
                                isLoggedIn && !state.isOnline && !isHighlighted -> LOGO_COLOR_DISCONNECTED
                                else -> LOGO_COLOR
                            },
                            isGlowing = isLoggedIn && isHighlighted,
                            modifier = Modifier.fillMaxSize(),
                        )
                        if (isLoggedIn && state.isOnline && badgeText.isNotEmpty()) {
                            TextLine(
                                text = badgeText,
                                color = Color.White,
                                style = ChatTheme.typography.overline.copy(lineHeight = 0.sp),
                                textAlign = TextAlign.Center,
                                modifier = Modifier
                                    .clip(CircleShape)
                                    .background(Color.Red)
                                    .padding(start = (1.5).dp)
                                    .align(Alignment.BottomEnd),
                            )
                        }
                    }
                    if (isLoggedIn && !state.isOnline) {
                        Icon(
                            imageVector = Icons.Default.Close,
                            contentDescription = null,
                            tint = Color.Red,
                            modifier = Modifier
                                .size(16.dp)
                                .align(Alignment.BottomEnd),
                        )
                    }
                }
            }
        }
    }
}