package nand.apps.chat.ui.widget

import com.hoc081098.kmp.viewmodel.ViewModel
import kotlinx.coroutines.flow.MutableStateFlow
import nand.apps.chat.event.ChatEventHandler
import nand.apps.chat.model.event.RequestAppFocusEvent
import nand.apps.chat.model.message.UnreadMessageCounts
import nand.apps.chat.model.settings.notifications.ChatWidgetSettingsData
import nand.apps.chat.model.uid.ContactUid
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.SelfProfileRepo
import nand.apps.chat.repo.SettingsRepo
import nand.apps.chat.ui.util.collect
import nand.apps.chat.util.fifoSharedFlow

class ChatWidgetViewModel(
    settingsRepo: SettingsRepo,
    messageRepo: ChatMessageRepo,
    selfProfileRepo: SelfProfileRepo,
    private val eventHandler: ChatEventHandler,
) : ViewModel() {

    val stateData = MutableStateFlow(ChatWidgetState())
    val actionData = fifoSharedFlow<ChatWidgetAction>()
    private val state get() = stateData.value

    init {
        collect(settingsRepo.observer) { settings ->
            onSettingsChanged(settings.notificationSettings.widgetSettings)
        }
        collect(messageRepo.unreadObserver) { unreadStates ->
            onUnreadStatesChanged(unreadStates)
        }
        collect(selfProfileRepo.selfObserver) { self ->
            updateState { it.copy(isOnline = self.isOnline) }
        }
    }

    fun toggleAppWindow() {
        eventHandler += RequestAppFocusEvent(toggle = true)
    }

    private fun onSettingsChanged(settings: ChatWidgetSettingsData) {
        val oldSettings = state.settings
        if (settings != oldSettings) {
            updateState { it.copy(settings = settings) }
            if (settings.isEnabled && (!oldSettings.isEnabled || settings.size != oldSettings.size || settings.placement != oldSettings.placement)) {
                actionData.tryEmit(UpdateDialogAction(size = settings.size, placement = settings.placement))
            }
        }
    }

    private fun onUnreadStatesChanged(unreadStates: Map<ContactUid, UnreadMessageCounts>) {
        updateState { state ->
            state.copy(
                unreadDirectMessages = unreadStates.values.sumOf { it.directMessages },
                unreadGroupMessages = unreadStates.values.sumOf { it.groupMessages },
            )
        }
    }

    private fun updateState(block: (ChatWidgetState) -> ChatWidgetState) {
        stateData.tryEmit(block(state))
    }
}