package nand.apps.chat.update

import nand.apps.chat.BuildConfig.APP_DIR_NAME
import nand.apps.chat.BuildConfig.APP_NAME
import nand.apps.chat.io.ChatFileSystem
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.update.AppReleaseInfo
import nand.apps.chat.platform.ChatPlatform
import nand.apps.chat.platform.getPlatform
import nand.apps.chat.res.StringResourceProvider
import okio.Path
import org.jetbrains.compose.resources.ExperimentalResourceApi
import seers.composeapp.generated.resources.*
import java.io.File
import kotlin.system.exitProcess

class DesktopAppInstaller(
    private val fileSystem: ChatFileSystem,
    private val stringProvider: StringResourceProvider,
) : AppInstaller {

    private val platform = getPlatform()
    private val logger = getLogger("DesktopAppInstaller")

    override fun isSupported() = platform.isDesktop && platform != ChatPlatform.DESKTOP_OTHER

    override fun getPackageFileExtension() = when (platform) {
        ChatPlatform.LINUX -> "deb"
        ChatPlatform.MAC -> "dmg"
        ChatPlatform.WINDOWS -> "msi"
        else -> error("Unsupported platform!")
    }

    @OptIn(ExperimentalResourceApi::class)
    override suspend fun install(release: AppReleaseInfo, packagePath: Path) {
        val scriptFileName = when (platform) {
            ChatPlatform.LINUX -> "update-linux.sh"
            ChatPlatform.MAC -> "update-mac.sh"
            ChatPlatform.WINDOWS -> "update-windows.bat"
            else -> return
        }
        val exePath = when (platform) {
            ChatPlatform.LINUX -> "/opt/$APP_DIR_NAME/bin/$APP_NAME"
            ChatPlatform.MAC -> "/Applications/$APP_NAME.app"
            ChatPlatform.WINDOWS -> "C:\\Program Files\\$APP_NAME\\$APP_NAME.exe"
            else -> return
        }
        val vars = mapOf(
            "APP_NAME" to APP_NAME,
            "PACKAGE_PATH" to packagePath.toString(),
            "EXECUTABLE_PATH" to exePath,
            "INSTALLING" to stringProvider.getString(Res.string.app_updater_notification_installing, release.version),
            "INSTALL_SUCCESS" to stringProvider.getString(Res.string.app_updater_notification_install_finished, release.version),
            "INSTALL_FAILED" to stringProvider.getString(Res.string.app_updater_notification_install_failed, release.version),
        )
        var script = Res.readBytes("files/scripts/$scriptFileName").decodeToString()
        vars.forEach { (key, value) ->
            script = script.replace("\$$key", value)
        }
        val installDir = fileSystem.getTempDirectory("install")
        if (installDir.writeToFile(scriptFileName, script.encodeToByteArray())) {
            try {
                val scriptPath = installDir.getFilePath(scriptFileName).toString()
                File(scriptPath).setExecutable(true)
                ProcessBuilder().run {
                    if (platform == ChatPlatform.WINDOWS) {
                        command(scriptPath)
                    } else {
                        command("nohup", scriptPath)
                    }
                    redirectErrorStream(true)
                    start()
                }
                exitProcess(0)
            } catch (e: Exception) {
                logger.error("Failed to install $packagePath", e)
            }
        }
    }
}