package nand.apps.chat.ui.widget

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.*
import nand.apps.chat.di.initDesktopKoinTest
import nand.apps.chat.model.event.RequestAppFocusEvent
import nand.apps.chat.model.event.SetSelfStatusEvent
import nand.apps.chat.model.settings.SettingsData
import nand.apps.chat.model.settings.notifications.ChatTraySettingsData
import nand.apps.chat.model.settings.notifications.NotificationSettingsData
import nand.apps.chat.model.user.UserStatus
import nand.apps.chat.ui.viewmodel.ViewModelTestScope
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test

private typealias ChatTrayIconViewModelTestScope = ViewModelTestScope<ChatTrayIconViewModel, ChatTrayIconState, Unit>

@OptIn(ExperimentalCoroutinesApi::class)
class ChatTrayIconViewModelTests {

    private lateinit var scope: ChatTrayIconViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Settings change updates state`() = viewModelTest {
        val settings = ChatTraySettingsData(isEnabled = true,)
        updateSettings(settings)
        assertState(ChatTrayIconState(settings = settings))
    }

    @Test
    fun `Self user change updates state`() = viewModelTest {
        selfProfileRepo.setSelfStatus(UserStatus.AWAY)
        assertState(ChatTrayIconState(status = UserStatus.AWAY))
        selfProfileRepo.setSelfStatus(UserStatus.DEFAULT)
        assertState(ChatTrayIconState(status = UserStatus.DEFAULT))
        viewModel.setSelfStatus(UserStatus.BUSY)
        assertEvent(SetSelfStatusEvent(UserStatus.BUSY))
        assertState(ChatTrayIconState(status = UserStatus.BUSY))
    }

    @Test
    fun `Toggle app window fires event`() = viewModelTest {
        viewModel.toggleAppWindow()
        assertEvent(RequestAppFocusEvent(toggle = true, isTrayIconEvent = true))
    }

    private fun viewModelTest(
        block: ChatTrayIconViewModelTestScope.() -> Unit,
    ) = runTest {
        val koin = initDesktopKoinTest().koin
        val viewModel: ChatTrayIconViewModel = koin.get()
        scope = ViewModelTestScope(koin)
        scope.initTestScope(
            scope = this,
            viewModel = viewModel,
            stateData = viewModel.stateData,
        )
        with(scope) {
            assertState(ChatTrayIconState())
            block()
        }
    }
}

private fun ChatTrayIconViewModelTestScope.updateSettings(settings: ChatTraySettingsData) {
    settingsRepo.settings = SettingsData(notificationSettings = NotificationSettingsData(traySettings = settings))
}