package nand.apps.chat.ui.widget

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.ExperimentalCoroutinesApi
import kotlinx.coroutines.test.*
import nand.apps.chat.di.initDesktopKoinTest
import nand.apps.chat.model.event.RequestAppFocusEvent
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.net.ConnectionProtocol
import nand.apps.chat.model.settings.SettingsData
import nand.apps.chat.model.settings.notifications.ChatWidgetPlacement
import nand.apps.chat.model.settings.notifications.ChatWidgetSettingsData
import nand.apps.chat.model.settings.notifications.ChatWidgetSize
import nand.apps.chat.model.settings.notifications.NotificationSettingsData
import nand.apps.chat.ui.viewmodel.ViewModelTestScope
import kotlin.test.AfterTest
import kotlin.test.BeforeTest
import kotlin.test.Test

private typealias ChatWidgetViewModelTestScope = ViewModelTestScope<ChatWidgetViewModel, ChatWidgetState, ChatWidgetAction>

@OptIn(ExperimentalCoroutinesApi::class)
class ChatWidgetViewModelTests {

    private lateinit var scope: ChatWidgetViewModelTestScope

    @BeforeTest
    fun setup() {
        Dispatchers.setMain(UnconfinedTestDispatcher())
    }

    @AfterTest
    fun teardown() {
        if (this::scope.isInitialized) scope.destroy()
        Dispatchers.resetMain()
    }

    @Test
    fun `Settings change fires dialog action`() = viewModelTest {
        val settings = ChatWidgetSettingsData(
            isEnabled = true,
            isFlashEnabled = true,
            size = ChatWidgetSize.SMALL,
            placement = ChatWidgetPlacement.BOTTOM_LEFT,
        )
        val action = UpdateDialogAction(size = ChatWidgetSize.SMALL, placement = ChatWidgetPlacement.BOTTOM_LEFT)

        // Change size and placement
        updateSettings(settings)
        assertAction(action)

        // Disable widget = no action
        updateSettings(settings.copy(isEnabled = false))
        assertNoAction()

        // Enable widget = action
        updateSettings(settings.copy(isEnabled = true))
        assertAction(action)

        // Change unrelated settings = no action
        updateSettings(settings.copy(isGroupUnreadNumberShown = true))
        assertNoAction()
    }

    @Test
    fun `Toggling app window fires event`() = viewModelTest {
        viewModel.toggleAppWindow()
        assertEvent(RequestAppFocusEvent(toggle = true))
    }

    @Test
    fun `Self user updates state`() = viewModelTest {
        selfProfileRepo.onSelfConnectionProtocol(ConnectionProtocol.UDP)
        assertState(ChatWidgetState(isOnline = true))
        selfProfileRepo.onSelfConnectionProtocol(ConnectionProtocol.NONE)
        assertState(ChatWidgetState(isOnline = false))
    }

    @Test
    fun `Message unread counts update state`() = viewModelTest {
        val friend = contactRepo.getFriends().random()
        val group = contactRepo.getGroups().random()
        messageRepo.addMessage(
            ChatMessage(
                conversationUid = friend.uid,
                senderUid = friend.uid,
                text = "Hello",
            )
        )
        assertState(ChatWidgetState(unreadDirectMessages = 1))
        messageRepo.addMessage(
            ChatMessage(
                conversationUid = group.uid,
                senderUid = friend.uid,
                text = "Hello",
            )
        )
        assertState(ChatWidgetState(unreadDirectMessages = 1, unreadGroupMessages = 1))
    }

    private fun viewModelTest(block: ChatWidgetViewModelTestScope.() -> Unit) = runTest {
        val koin = initDesktopKoinTest().koin
        val viewModel: ChatWidgetViewModel = koin.get()
        scope = ViewModelTestScope(koin)
        scope.initTestScope(
            scope = this,
            viewModel = viewModel,
            stateData = viewModel.stateData,
            actionData = viewModel.actionData,
        )
        scope.assertState(ChatWidgetState())
        scope.block()
    }
}

private fun ChatWidgetViewModelTestScope.updateSettings(settings: ChatWidgetSettingsData) {
    settingsRepo.settings = SettingsData(notificationSettings = NotificationSettingsData(widgetSettings = settings))
}