package nand.apps.chat.audio.id3

import com.fleeksoft.charset.Charsets
import com.fleeksoft.charset.decodeToString
import nand.apps.chat.io.getFileMetadata
import nand.apps.chat.util.takeIfNotBlank
import nand.apps.chat.util.toIntLittleEndian
import okio.BufferedSource
import okio.FileSystem
import okio.IOException
import okio.Path.Companion.toPath
import kotlin.String

private const val ID3_MAGIC_SIZE = 3L

private const val V1_SIZE = 128
private const val V1_BLOCK_SIZE = 30L
private val V1_MAGIC = "TAG".encodeToByteArray()

private val V2_MAGIC = "ID3".encodeToByteArray()
private val UTF_BOM_BE = byteArrayOf(0xFE.toByte(), 0xFF.toByte())
private const val V2_COMM_HEADER_LENGTH = 7L
private const val V2_TAG_ART = "APIC"
private const val V2_TAG_PRIVATE = "PRIV"
private const val V2_TAG_TITLE = "TIT2"
private const val V2_TAG_ARTIST = "TPE1"

private const val BYTE_0 = 0.toByte()
private const val BYTE_1 = 1.toByte()

class Id3TagReader(
    private val fileSystem: FileSystem,
) {

    @Throws(IOException::class)
    fun readTags(path: String): Id3TagData? {
        val ioPath = path.toPath()
        val fileSize = fileSystem.getFileMetadata(ioPath)?.size ?: 0
        if (fileSize < V1_SIZE) {
            // If the file is less than 128 bytes then it's probably invalid
            return null
        }
        fileSystem.read(ioPath) {
            val v2Magic = readByteArray(ID3_MAGIC_SIZE)
            if (v2Magic.contentEquals(V2_MAGIC)) {
                return readV2Tags()
            } else {
                skip(fileSize - ID3_MAGIC_SIZE - V1_SIZE)
                val v1Magic = readByteArray(ID3_MAGIC_SIZE)
                if (v1Magic.contentEquals(V1_MAGIC)) {
                    return readV1Tags()
                }
            }
            return null
        }
    }

    private fun BufferedSource.readV2Tags(): Id3TagData {
        val utf16Le = Charsets.forName("UTF-16LE")
        val utf16Be = Charsets.forName("UTF-16BE")
        val major = readByte()
        val minor = readByte()
        skip(1) // Flags
        var size = 0L
        for (index in 0..<4) {
            val byte = readByte()
            size = size or ((byte.toLong() and 0x7f) shl ((3 - index) * 7))
        }
        var numRead = 4L
        val tags = mutableMapOf<String, String?>()
        while (numRead < size) {
            val frameIdBytes = readByteArray(4)
            if (frameIdBytes[0] == BYTE_0) break
            val frameId = frameIdBytes.decodeToString()
            val frameSize = readByteArray(4).toIntLittleEndian()
            var remaining = frameSize.toLong()
            skip(2) // Frame flags
            when (frameId) {
                V2_TAG_ART, // TODO - Read image
                V2_TAG_PRIVATE -> skip(remaining)
                else -> {
                    val isUnicode = readByte() == BYTE_1
                    remaining--
                    if (frameId == "COMM") {
                        skip(V2_COMM_HEADER_LENGTH)
                        remaining -= V2_COMM_HEADER_LENGTH
                    }
                    tags[frameId] = if (isUnicode) {
                        val bom = readByteArray(2)
                        val isBigEndian = bom.contentEquals(UTF_BOM_BE)
                        remaining -= 2
                        if (remaining > 0) {
                            readByteArray(remaining).decodeToString(if (isBigEndian) utf16Be else utf16Le)
                        } else {
                            null
                        }
                    } else {
                        readAsciiString(remaining)
                    }
                }
            }

            numRead += remaining + 10
        }
        return Id3TagData(
            title = tags[V2_TAG_TITLE],
            artist = tags[V2_TAG_ARTIST],
            version = "ID3v2.$major.$minor",
        )
    }

    private fun BufferedSource.readV1Tags(): Id3TagData {
        val title = readAsciiString(V1_BLOCK_SIZE)
        val artist = readAsciiString(V1_BLOCK_SIZE)
        return Id3TagData(
            title = title,
            artist = artist,
            version = "ID3v1",
        )
    }

    private fun BufferedSource.readAsciiString(numBytes: Long): String? {
        val data = readByteArray(numBytes)
        var length = data.indexOf(BYTE_0)
        if (length == -1) length = data.size
        return data.decodeToString(Charsets.ISO_8859_1, 0, length).takeIfNotBlank()
    }
}