package nand.apps.chat.util

import kotlinx.coroutines.ExperimentalForInheritanceCoroutinesApi
import kotlinx.coroutines.flow.FlowCollector
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow

@OptIn(ExperimentalForInheritanceCoroutinesApi::class)
class MutableStateFlowMap<K, V> : MutableMap<K, V>, StateFlow<Map<K, V>> {

    private val impl = mutableMapOf<K, V>()
    private val stateFlow = MutableStateFlow(emptyMap<K, V>())

    override val entries get() = impl.entries
    override val size get() = impl.size
    override val keys get() = impl.keys
    override val values get() = impl.values
    override val value get() = stateFlow.value
    override val replayCache get() = stateFlow.replayCache

    override fun clear() {
        if (isNotEmpty()) {
            impl.clear()
            emitState()
        }
    }

    override fun isEmpty() = impl.isEmpty()

    override fun remove(key: K): V? {
        return impl.remove(key).also { value ->
            if (value != null) emitState()
        }
    }

    override fun put(key: K, value: V): V? {
        return impl.put(key, value).also { previous ->
            if (value != previous) emitState()
        }
    }

    override fun putAll(from: Map<out K, V>) {
        impl.putAll(from)
        emitState()
    }

    override fun get(key: K) = impl[key]

    override fun containsValue(value: V) = impl.containsValue(value)

    override fun containsKey(key: K) = impl.containsKey(key)

    override suspend fun collect(collector: FlowCollector<Map<K, V>>) = stateFlow.collect(collector)

    private fun emitState() {
        stateFlow.tryEmit(impl.toMap())
    }
}