package nand.apps.chat.util

import okio.ByteString.Companion.toByteString
import kotlin.math.pow
import kotlin.math.roundToInt
import kotlin.uuid.ExperimentalUuidApi
import kotlin.uuid.Uuid

private const val THOUSAND = 1000
private const val TEN_THOUSAND = 10_000
private const val MILLION = 1_000_000
private const val TEN_MILLION = 10_000_000
private const val BILLION = 1_000_000_000
private const val TEN_BILLION = 10_000_000_000

val REGEX_DIGITS = Regex("\\d+")
val REGEX_NON_DIGITS = Regex("\\D+")

const val CHARACTER_RANGE_HEX = "0-9A-Fa-f"

@OptIn(ExperimentalUuidApi::class)
fun randomUid() = Uuid.random().toString()

fun String.findMatchRanges(regex: Regex) = regex.findAll(this).toList().map { it.range }

fun String.takeIfNotBlank() = takeIf { it.isNotBlank() }

fun String.trimToLength(length: Int) =
    if (this.length > length) substring(0, length) else this

fun String.sha256() = encodeToByteArray().toByteString().sha256().hex().uppercase()

fun String.sha256Bytes() = encodeToByteArray().toByteString().sha256().toByteArray()

fun String?.toIntOrZero() = this?.toIntOrNull() ?: 0

fun String?.toLongOrZero() = this?.toLongOrNull() ?: 0L

@OptIn(ExperimentalStdlibApi::class)
fun String.decodeHex(): ByteArray? {
    return try {
        hexToByteArray()
    } catch (ignore: Exception) {
        null
    }
}

fun Double.roundToDecimal(decimalPlaces: Int): Float {
    val exp = 10.0.pow(decimalPlaces)
    return (this * exp).roundToInt() / exp.toFloat()
}

fun Double.roundToDecimalString(decimalPlaces: Int) =
    roundToDecimal(decimalPlaces).toString().trimEnd('.', '0')

fun formatFileSize(sizeBytes: Long, kbSize: Int = 1000): String {
    val kb = sizeBytes / kbSize.toDouble()
    val mb = kb / kbSize.toDouble()
    val gb = mb / kbSize.toDouble()
    return when {
        sizeBytes < kbSize -> "$sizeBytes B"
        kb < kbSize -> "${kb.roundToInt()} KB"
        mb < kbSize -> "${mb.roundToDecimalString(if (mb >= 100) 1 else 2)} MB"
        else -> "${gb.roundToDecimalString(if (gb >= 100) 1 else 2)} GB"
    }
}

fun formatViewCount(count: Long) = when {
    count < THOUSAND -> count.toString()
    count < MILLION -> (count / THOUSAND.toDouble()).roundToDecimalString(if (count >= TEN_THOUSAND) 0 else 1) + "K"
    count < BILLION -> (count / MILLION.toDouble()).roundToDecimalString(if (count >= TEN_MILLION) 0 else 1) + "M"
    else -> (count / BILLION.toDouble()).roundToDecimalString(if (count >= TEN_BILLION) 0 else 1) + "B"
}

fun formatPlural(word: String, quantity: Int) =
    if (quantity == 1) word else "${word}s"

fun formatPlural(word: String, quantity: Long) =
    if (quantity == 1L) word else "${word}s"

fun Int.toDurationString(): String {
    var seconds = this
    var minutes = seconds / 60
    val hours = minutes / 60
    seconds -= minutes * 60
    minutes -= hours * 60
    return buildString {
        if (hours > 0) {
            append(hours)
            append(if (minutes < 10) ":0$minutes" else ":$minutes")
        } else {
            append(minutes)
        }
        append(if (seconds < 10) ":0$seconds" else ":$seconds")
    }
}