package nand.apps.chat.engine.matrix

import io.ktor.http.*
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import nand.apps.audio.AudioFrame
import nand.apps.chat.engine.ChatEngineCallback
import nand.apps.chat.engine.matrix.dev.MatrixDevCommandHandler
import nand.apps.chat.engine.matrix.profile.MatrixProfileAuthenticator
import nand.apps.chat.engine.matrix.profile.MatrixProfileData
import nand.apps.chat.engine.matrix.profile.MatrixProfileRepo
import nand.apps.chat.log.getLogger
import nand.apps.chat.matrix.BuildConfig.TRIXNITY_VERSION
import nand.apps.chat.model.call.CallControl
import nand.apps.chat.model.engine.ChatEngineConstraints
import nand.apps.chat.model.engine.ChatEngineHostInfo
import nand.apps.chat.model.engine.ChatEngineInitResult
import nand.apps.chat.model.file.ChatFileControl
import nand.apps.chat.model.file.ChatFileControlResponse
import nand.apps.chat.model.file.ChatFileType
import nand.apps.chat.model.group.*
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.uid.*
import nand.apps.chat.model.user.*
import nand.apps.chat.platform.getPlatformInfo
import nand.apps.chat.repo.SelfProfileRepo
import net.folivo.trixnity.client.*
import net.folivo.trixnity.client.media.InMemoryMediaStore
import net.folivo.trixnity.client.room.message.text
import net.folivo.trixnity.clientserverapi.model.authentication.IdentifierType.User
import net.folivo.trixnity.core.model.RoomId
import net.folivo.trixnity.utils.toByteArray
import okio.ByteString.Companion.toByteString
import org.koin.core.Koin

private const val DEFAULT_HOST = "matrix.org"
private const val AVATAR_SIZE = 128L

class MatrixChatEngineImpl(koin: Koin) : MatrixChatEngine {

    private var client: MatrixClient? = null
    private var coroutineScope: CoroutineScope? = null
    private var isStarted = false
    private var self: UserData? = null

    private val repo: MatrixChatEngineRepo by koin.inject()
    private val profileRepo: MatrixProfileRepo by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()
    private val engineCallback: ChatEngineCallback by koin.inject()
    private val mediaStore = InMemoryMediaStore()
    private val repositoriesModule by lazy { getMatrixRepoKoinModule() }

    private val logger = getLogger("MatrixChatEngine")

    override val version = "Trixnity v$TRIXNITY_VERSION"

    override val hostInfo get() = ChatEngineHostInfo(
        host = client?.baseUrl?.host,
        defaultHost = DEFAULT_HOST,
    )

    override val constraints = ChatEngineConstraints(
        isPasswordRequired = true,
        isHostRequired = true,
        maxNameLength = 256, // Pulled from Synapse
        maxStatusMessageLength = 0, // Not supported
        maxMessageLength = 65536, // Is there one?
        maxGroupNameLength = 256,
        addressLength = -1, // Variable length
        uidLength = -1, // Variable length
        uidCharacterRange = "0-9A-Za-z@.",
    )

    override val encryption by koin.inject<MatrixChatEngineEncryption>()
    override val authenticator by koin.inject<MatrixProfileAuthenticator>()
    override val commandHandler by koin.inject<MatrixDevCommandHandler>()

    override suspend fun initialize(profile: ChatProfileData): ChatEngineInitResult {
        client?.stopSync(wait = true)

        // Load device ID
        profileRepo.init(profile)
        val matrixProfile = profileRepo.load()

        // Login to Matrix
        val host = profile.host ?: DEFAULT_HOST
        client = MatrixClient.fromStore(
            repositoriesModule = repositoriesModule,
            mediaStore = mediaStore,
        ).getOrNull() ?: MatrixClient.login(
            baseUrl = Url("https://$host/"),
            identifier = User(profile.username),
            password = profile.password,
            deviceId = matrixProfile?.deviceId,
            initialDeviceDisplayName = "${selfProfileRepo.clientVersion} (${getPlatformInfo()})",
            repositoriesModule = repositoriesModule,
            mediaStore = mediaStore,
        ).getOrNull()

        // Initialize self data
        self = client?.let {
            UserData(
                uid = it.userId.toUserUid(),
                name = profile.username,
            )
        }

        // Save newly generated device ID
        client?.deviceId?.takeIf { it != matrixProfile?.deviceId }?.let {
            profileRepo.save(MatrixProfileData(deviceId = it))
        }

        return client?.let {
            engineCallback.onEngineInit(profile)
            ChatEngineInitResult.SUCCESS
        } ?: ChatEngineInitResult.INVALID_CREDENTIALS
    }

    override fun start(coroutineScope: CoroutineScope) {
        val client = client ?: run {
            logger.warn("Attempted to start engine without valid client")
            return
        }

        if (isStarted) {
            logger.warn("Attempted to start engine that's already active")
            return
        }

        this.coroutineScope = coroutineScope

        coroutineScope.launch {
            client.startSync()
            isStarted = true
            repo.init(client, coroutineScope)
            engineCallback.onEngineStart(coroutineScope)
        }
    }

    override fun stop() {
        if (isStarted) {
            engineCallback.onEngineStop()
            launch { client ->
                client.logout()
                client.stopSync()
            }
            repo.shutdown()
        }
        client = null
        self = null
        isStarted = false
    }

    override fun getSelfUserData(): UserData {
        return self ?: UserData.BLANK
    }

    override fun setSelfUserData(name: String, statusMessage: String): Boolean {
        return launch { it.setDisplayName(name) }
    }

    override fun setSelfStatus(status: UserStatus) {
        // TODO - Implement
    }

    override fun getSelfAddress(): String {
        return client?.userId?.full.orEmpty()
    }

    override fun resetSelfAddress() {
        // Not supported
    }

    override suspend fun getFriendList(): List<UserData> {
        // Not supported
        return emptyList()
    }

    override suspend fun getFriendRequestList(): List<FriendRequest> {
        // Not supported
        return emptyList()
    }

    override fun sendFriendRequest(address: String, message: String): UserUid? {
        // Not supported
        return null
    }

    override fun acceptFriendRequest(requestUid: FriendRequestUid): UserUid? {
        // Not supported
        return null
    }

    override fun declineFriendRequest(requestUid: FriendRequestUid) {
        // Not supported
    }

    override fun removeFriend(friendUid: UserUid): Boolean {
        // Not supported
        return false
    }

    override fun setFriendAlias(friendUid: UserUid, alias: String?): Boolean {
        // Not supported
        return false
    }

    override fun sendMessage(message: ChatMessage): ChatMessage? {
        return runClient { client ->
            val transactionId = client.room.sendMessage(RoomId(message.conversationUid.value)) {
                text(message.text)
            }
            message.copy(uid = MessageUid(transactionId))
        }
    }

    override fun editMessage(old: ChatMessage, new: ChatMessage): Boolean {
        // TODO - Implement
        return false
    }

    override fun setMessageReactions(message: ChatMessage, reactions: Set<String>): Boolean {
        // TODO - Implement
        return false
    }

    override fun removeMessage(message: ChatMessage): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendPacket(uid: ContactUid, packet: ByteArray): Boolean {
        // Not supported
        return false
    }

    override fun sendPeerPacket(groupUid: GroupUid, peerUid: UserUid, packet: ByteArray): Boolean {
        // Not supported
        return false
    }

    override suspend fun getGroupList(): List<ChatGroupData> {
        return repo.getGroupList()
    }

    override fun getGroupPeerList(groupUid: GroupUid): Map<UserUid, UserData> {
        return repo.getGroup(groupUid)?.peers.orEmpty()
    }

    override fun createGroup(title: String, type: ChatGroupType, visibility: ChatGroupVisibility): ChatGroupData? {
        // TODO - Implement
        return null
    }

    override fun setGroupTitle(groupUid: GroupUid, title: String): Boolean {
        // TODO - Implement
        return false
    }

    override fun setGroupDescription(groupUid: GroupUid, description: String): Boolean {
        // Not supported
        return false
    }

    override fun setGroupDescriptionLocked(groupUid: GroupUid, isLocked: Boolean): Boolean {
        // Not supported
        return false
    }

    override fun setGroupPassword(groupUid: GroupUid, password: String?): Boolean {
        // Not supported
        return false
    }

    override fun setGroupNickname(groupUid: GroupUid, name: String): Boolean {
        // Not supported
        return false
    }

    override fun setGroupAvatar(groupUid: GroupUid, avatarHash: String): Boolean {
        // TODO - Implement
        return false
    }

    override fun setGroupAudioEnabled(groupUid: GroupUid, isEnabled: Boolean): Boolean {
        // TODO - Implement
        return false
    }

    override fun setGroupPeerLimit(groupUid: GroupUid, peerLimit: Int): Boolean {
        // Not supported
        return false
    }

    override fun setGroupTalkState(groupUid: GroupUid, talkState: ChatGroupTalkState): Boolean {
        // Not supported
        return false
    }

    override fun setGroupVisibility(groupUid: GroupUid, visibility: ChatGroupVisibility): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendGroupInvite(friendUid: UserUid, groupUid: GroupUid): Boolean {
        // TODO - Implement
        return false
    }

    override fun acceptGroupInvite(invite: ChatGroupInvite): ChatGroupData? {
        // TODO - Implement
        return null
    }

    override fun setGroupRole(groupUid: GroupUid, peerUid: UserUid, role: ChatGroupRole): Boolean {
        // TODO - Implement
        return false
    }

    override fun kickGroupPeer(groupUid: GroupUid, peerUid: UserUid): Boolean {
        // TODO - Implement
        return false
    }

    override fun setGroupPeerBlocked(groupUid: GroupUid, peerUid: UserUid, block: Boolean): Boolean {
        // TODO - Implement
        return false
    }

    override fun joinGroup(groupUid: GroupUid, selfNickname: String, password: String?): Boolean {
        // TODO - Implement
        return false
    }

    override fun removeGroup(groupUid: GroupUid, partingMessage: String?): Boolean {
        return launch { it.room.forgetRoom(groupUid.toMatrixId(it.baseUrl.host)) }
    }

    override fun setContactMuted(uid: ContactUid, isMuted: Boolean): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendFile(
        friendUid: UserUid,
        kind: ChatFileType,
        fileSize: Long,
        fileName: String,
        fileUid: FileUid
    ): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendFileChunk(friendUid: UserUid, fileUid: FileUid, position: Long, data: ByteArray): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendFileControl(
        friendUid: UserUid,
        fileUid: FileUid,
        control: ChatFileControl
    ): ChatFileControlResponse {
        // TODO - Implement
        return ChatFileControlResponse.DENIED
    }

    override fun startCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean {
        // TODO - Implement
        return false
    }

    override fun answerCall(uid: ContactUid, isAudioEnabled: Boolean, isVideoEnabled: Boolean): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendCallControl(uid: ContactUid, control: CallControl): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendAudioFrame(uid: ContactUid, frame: AudioFrame): Boolean {
        // TODO - Implement
        return false
    }

    override fun sendAudioBitRate(uid: ContactUid, bitRate: Int): Boolean {
        // TODO - Implement
        return false
    }

    override fun generateHash(data: ByteArray): ByteArray? {
        return data.toByteString().sha256().toByteArray()
    }

    override fun onDisplayNameUpdate(name: String) {
        self = self?.copy(name = name)
    }

    override suspend fun getAvatarData(url: String): ByteArray? {
        return client?.media?.getThumbnail(
            uri = url,
            width = AVATAR_SIZE,
            height = AVATAR_SIZE,
        )?.getOrNull()?.toByteArray()
    }

    private fun launch(block: suspend CoroutineScope.(MatrixClient) -> Unit): Boolean {
        val client = client ?: return false
        return coroutineScope?.takeIf { isStarted }?.launch { block(client) } != null
    }

    private fun <T> runClient(block: suspend CoroutineScope.(MatrixClient) -> T): T? = runBlocking {
        client?.takeIf { isStarted }?.let { block(it) }
    }
}