package nand.apps.chat.engine.matrix

import kotlinx.coroutines.CoroutineScope
import nand.apps.chat.engine.matrix.group.MatrixRoomEventListener
import nand.apps.chat.model.uid.GroupUid
import nand.apps.chat.repo.SelfProfileRepo
import net.folivo.trixnity.client.*
import net.folivo.trixnity.client.store.Room
import org.koin.core.Koin

class MatrixChatEngineRepo(private val koin: Koin) {

    private val chatEngine: MatrixChatEngine by koin.inject()
    private val selfProfileRepo: SelfProfileRepo by koin.inject()

    private val roomListeners = mutableMapOf<GroupUid, MatrixRoomEventListener>()

    private lateinit var client: MatrixClient
    private lateinit var coroutineScheduler: MatrixCoroutineScheduler
    private lateinit var coroutineScope: CoroutineScope

    fun init(
        client: MatrixClient,
        coroutineScope: CoroutineScope,
    ) {
        this.client = client
        this.coroutineScope = coroutineScope
        this.coroutineScheduler = MatrixCoroutineScheduler(coroutineScope)

        coroutineScheduler.collect(client.room.getAll().flattenValues()) { rooms ->
            rooms.forEach(::onRoomSync)
        }
        coroutineScheduler.collect(client.avatarUrl) { avatarUrl ->
            selfProfileRepo.setCurrentAvatarUri(avatarUrl)
        }
        coroutineScheduler.collect(client.displayName) { displayName ->
            displayName?.let {
                chatEngine.onDisplayNameUpdate(displayName)
                selfProfileRepo.setCurrentDisplayName(displayName)
            }
        }
    }

    fun shutdown() {
        roomListeners.values.forEach { it.shutdown() }
        roomListeners.clear()
        coroutineScheduler.shutdown()
    }

    fun getGroupList() = roomListeners.values.map { it.group }

    fun getGroup(groupUid: GroupUid) = roomListeners[groupUid]?.group

    private fun onRoomSync(room: Room) {
        val groupUid = room.roomId.toGroupUid()
        val handler = roomListeners[groupUid] ?: MatrixRoomEventListener(
            client = client,
            coroutineScope = coroutineScope,
            koin = koin,
        ).also {
            roomListeners[groupUid] = it
        }
        handler.onSync(room)
    }
}