package nand.apps.chat.engine.matrix.group

import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.first
import nand.apps.chat.engine.matrix.MatrixCoroutineScheduler
import nand.apps.chat.engine.matrix.toGroupUid
import nand.apps.chat.engine.matrix.toUserUid
import nand.apps.chat.log.getLogger
import nand.apps.chat.model.group.ChatGroupData
import nand.apps.chat.model.group.ChatGroupFeature
import nand.apps.chat.model.message.ChatMessage
import nand.apps.chat.model.uid.MessageUid
import nand.apps.chat.model.user.UserData
import nand.apps.chat.repo.ChatMessageRepo
import nand.apps.chat.repo.ContactRepo
import net.folivo.trixnity.client.MatrixClient
import net.folivo.trixnity.client.flattenValues
import net.folivo.trixnity.client.room
import net.folivo.trixnity.client.store.Room
import net.folivo.trixnity.client.store.RoomUser
import net.folivo.trixnity.client.user
import net.folivo.trixnity.core.model.EventId
import net.folivo.trixnity.core.model.RoomId
import net.folivo.trixnity.core.model.events.ClientEvent
import net.folivo.trixnity.core.model.events.m.RelatesTo
import net.folivo.trixnity.core.model.events.m.room.RoomMessageEventContent
import org.koin.core.Koin

class MatrixRoomEventListener(
    private val client: MatrixClient,
    koin: Koin,
    coroutineScope: CoroutineScope,
) {
    private val contactRepo: ContactRepo by koin.inject()
    private val messageRepo: ChatMessageRepo by koin.inject()

    private val coroutineScheduler = MatrixCoroutineScheduler(coroutineScope)
    private val processedEvents = mutableSetOf<EventId>()
    private val deferredEvents = mutableListOf<ClientEvent.RoomEvent<*>>()
    private var isInitializing = true

    private val logger = getLogger("MatrixGroupEventListener")

    lateinit var group: ChatGroupData
        private set

    fun shutdown() {
        coroutineScheduler.shutdown()
    }

    fun onSync(room: Room) {
        val roomId = room.roomId
        val groupUid = roomId.toGroupUid()
        val groupName = room.name?.explicitName.orEmpty()
        val groupAvatarUri = room.avatarUrl

        if (!this::group.isInitialized) {
            group = ChatGroupData(
                uid = groupUid,
                selfUid = client.userId.toUserUid(),
                title = groupName,
                avatarUri = groupAvatarUri,
                isOnline = true,
                features = setOf(ChatGroupFeature.EDIT_TITLE),
            ).also {
                contactRepo.onGroupAdded(it)
            }
            fetchLatestEvents(roomId)
            coroutineScheduler.collect(client.user.getAll(roomId).flattenValues()) { users ->
                updatePeers(users)
            }
            coroutineScheduler.collect(client.room.getLastTimelineEvent(roomId)) { timelineEvents ->
                timelineEvents?.first()?.let { onRoomEvent(it.event) }
            }
        } else {
            // TODO - What else is updated here?
            group = group.copy(
                title = groupName,
                avatarUri = groupAvatarUri,
            )
            if (group.title != groupName) {
                contactRepo.onGroupTitle(groupUid, groupName)
            }
            if (group.avatarUri != groupAvatarUri) {
                contactRepo.onGroupAvatar(groupUid, groupAvatarUri)
            }
        }
    }

    private fun updatePeers(users: List<RoomUser>) {
        group = group.copy(
            peers = users.associate { user ->
                val uid = user.userId.toUserUid()
                uid to UserData(
                    uid = uid,
                    name = user.name,
                )
            },
        )
        contactRepo.onGroupUsersChanged(group.uid)
    }

    // One-time fetch when the group is first synced
    // TODO - Probably a better way to do this by exclusively using the flow functions
    private fun fetchLatestEvents(roomId: RoomId) {
        coroutineScheduler.launch {
            val events = client.api.room.getEvents(
                roomId = roomId,
                from = "",
                limit = 20,
            ).getOrNull()?.chunk.orEmpty()
            isInitializing = false
            events.forEach(::onRoomEvent)
            deferredEvents.forEach(::onRoomEvent)
            deferredEvents.clear()
        }
    }

    private fun onRoomEvent(event: ClientEvent.RoomEvent<*>) {
        logger.debug("onRoomEvent($event)")
        if (event.id in processedEvents) return
        processedEvents += event.id

        if (isInitializing) {
            deferredEvents += event
            return
        }

        when (val content = event.content) {
            is RoomMessageEventContent -> onMessageEvent(
                event = event,
                content = content,
            )
            else -> {}
        }
    }

    private fun onMessageEvent(
        event: ClientEvent.RoomEvent<*>,
        content: RoomMessageEventContent,
    ) {
        val messageUid: MessageUid
        val text: String
        val isEdited: Boolean
        when (val relatesTo = content.relatesTo) {
            is RelatesTo.Replace -> {
                messageUid = MessageUid(relatesTo.eventId.full)
                text = (relatesTo.newContent as? RoomMessageEventContent)?.body ?: content.body
                isEdited = true
            }
            else -> {
                messageUid = MessageUid(event.id.full)
                text = content.body
                isEdited = false
            }
        }
        val groupUid = event.roomId.toGroupUid()
        val senderUid = event.sender.toUserUid()
        val selfUid = client.userId.toUserUid()
        val existing = messageRepo.getMessage(
            conversationUid = groupUid,
            messageUid = messageUid,
        ) ?: event.unsigned?.transactionId?.let {
            messageRepo.getMessage(
                conversationUid = groupUid,
                messageUid = MessageUid(it),
            )
        }
        existing?.let { message ->
            if (message.senderUid != senderUid) {
                logger.warn("Received message update with unexpected sender (expected: ${message.senderUid}, actual: $senderUid)")
                return
            }
            messageRepo.updateMessage(
                message = message.copy(
                    uid = messageUid,
                    text = text,
                    timestamp = event.originTimestamp,
                    isDelivered = senderUid == selfUid,
                    isEdited = isEdited,
                ),
                localOnly = true,
                oldUid = message.uid,
            )
        } ?: run {
            val message = ChatMessage(
                conversationUid = groupUid,
                senderUid = senderUid,
                uid = messageUid,
                text = text,
                timestamp = event.originTimestamp,
            )
            messageRepo.addMessage(message)
        }
    }
}